! Written by I. Misztal, 1998
! modification 05/31/2001 by S. Tsuruta
! last modified 01/11/2005
! Genomic Selection Options added by I. Aguilar 09/23/09
module BLUPF90
    use kinds; use sparsem; use sparseop; use pcg;
    use fdistri_mod ! ajout ofilangi Juillet 2013 pour calculer la pvalue
    implicit none

    integer,parameter::   maxcorr=50   ! maximum number of correlated effects; used for g 50
    !        Types of effects
    integer,parameter::effcross=0,& !effects can be cross-classified
        effcov=1     !or covariables

    !        Types of random effects
    integer,  parameter ::  g_fixed=1,&       ! fixed effect
        g_diag=2, &       ! diagonal
        g_A=3, &          ! additive animal
        g_A_UPG=4, &      ! additive animal with unknown parent groups
        g_A_UPG_INB=5, &  ! additive animal with unknown parent groups and inbreeding
        g_As=6,&          ! additive sire
        g_PD =7,&         ! parental dominance
        g_usr=8,&         ! user defined
        g_usr_inv=9,&     ! user defined and inverted
        g_A_UnSym=10,&    ! additive animal unsymetric
        g_last=11         ! last type

    !! =========================================================================================
    !! Type Param_blupf90
    !!
    !! Type contenant les parametres de l'analyse. Cette structure de données remplace le fichier parametre en entrée de blupf90
    !!
    !! =========================================================================================
    type Param_blupf90
        integer             :: ntrait=0,&                    !number of traits
                               neff=0,&                      !number of effects
                               miss=0                      !value of missing trait/effect


        integer,allocatable :: pos_y(:)        !positions of observations
        integer ::             pos_weight      ! position of weight of records; zero if none


        integer,allocatable :: pos_eff(:,:),&  !positions of effects for each trait
                               nlev(:),&       !number of levels
                               effecttype(:),& !type of effects
                               nestedcov(:,:),&!position of nesting effect for each trait  if the effect is nested covariable
                               randomtype(:),& ! status of each effect, as above
                               randomnumb(:),&   ! number of consecutive correlated effects
                               offset(:)
        real (rh), allocatable :: r(:,:),&      !residual (co)variance matrix
                                  gcov(:,:,:)      ! The random (co)variance matrix for each trait

        !! DEFAULT VALUE => mettre -1

        real               :: conv_crit = -1
        integer            :: maxrounds = -1
        character (20)     :: solv_method = "PCG"
        real               :: r_factor = -1
        integer            :: blksize = -1
        !  logical            :: residual = .false.
        !  logical            :: stdresidual = .false.
        integer            :: mis = 0
        logical            :: lchk_lev = .true.

    end type  Param_blupf90

    !!
    !! =========================================================================================
    !! Type Solution_blupf90
    !!
    !! Type contenant les sorties du programme blupf90. Les sorties matricielles A,B,C et D sont celle décrite dans la specification de a.ricard
    !!
    !! =========================================================================================
    type Solution_blupf90

        real , dimension(:,:),pointer :: A
        real , dimension(:,:),pointer :: B
        real , dimension(:,:),pointer :: C
        ! D is a sparse matrix...seulement les individus phenotype et genotypé ont une valeur
        real , dimension(:,:),pointer :: D
        integer ,dimension(:),pointer :: Dani

        integer :: nlevtot = 0! nombre total d'effet
        integer :: nhaplo  = 0! nombre d'haplotype
        integer :: nani    = 0! nombre d'animaux pour les effets polygenique

        ! perfo correspond à nani (voir lecture.f90)
        real :: test=0.d0
        real :: pvalue=0.d0
        real , dimension(:),pointer :: sol  ! Solution pour chaque niveau d'haplotype (dim 1 dans le cas de 1 unique SNP)
        real :: solmu=0.d0
        real , dimension(:),pointer :: std  ! STD Solution pour chaque niveau d'haplotype (dim 1 dans le cas de 1 unique SNP)
        real :: stdmu=0.d0

    end type Solution_blupf90

    !! =========================================================================================

    ! Execution du blup
    public :: exec_blupf90
    ! Liberation de la memoire de la structure des parametres du type(blup paramblupf90)
    public :: release_paramblupf90
    ! Liberation de la memoire des tableaux allouer pour la solution des parametres
    public :: release_solutionblupf90


contains


    ! OFI , adaptation du programme blupf90 pour eviter les entrees/sorties
    !
    !
    !
    subroutine exec_blupf90(params,sizendata,ndata,ncol,data,ndata_ped,data_ped,nguser,guser,solution,writeok)
        type(Param_blupf90), intent(in) :: params
        integer            ,intent(in)  :: sizendata
        integer            ,intent(in)  :: ndata
        integer            ,intent(in)  :: ncol
        real, dimension(sizendata,ncol),intent(in) :: data
        integer            ,intent(in)   :: ndata_ped
        integer, dimension(ndata_ped,3),intent(in)  :: data_ped
        integer            ,intent(in)              :: nguser
        real, dimension(nguser,3),intent(in)  :: guser
        type(Solution_blupf90) , intent(out)  :: solution
        logical                ,intent(in)    :: writeok

        real,allocatable :: y(:),&                      ! observation value
            indata(:)                   ! one line of input data

        real (rh) ::        weight_y                    ! weight for records

        type (sparse_hashm)::xx
        type (sparse_ija):: xx_ija
        !ANNE
        type (sparse_hashm)::hh
        type (sparse_ija)::hh_ija
        integer :: annef,annei,annej,at,idata
        !FIN ANNE
        real (rh), allocatable:: rinv(:,:),xy(:),sol(:),g(:,:,:)        !storage for equations
        real (rh), parameter :: set_eig=1d-5
        real (rh), allocatable :: weight_cov(:,:)
        integer,allocatable:: address(:,:),&    ! start and address of each effect
            nlevsum(:)
        integer :: neq,io,&                     ! number of equations and io-status
            data_len,&                   ! length of data record to read
            i,j,k,l,nrec,&          ! extra variables
            maxrounds=5000,&             ! unit number of each pedigree file
            blksize=1
        real (rh) ::  val,sumxx,sum2
        real :: second,conv_crit=1e-12,r_factor=1.4,x(3)
        character (20):: xc(3),solv_method='PCG'
        !logical :: residual=.false.,stdresidual=.false.,&
        logical :: first_snp,lzerosol=.true.,lprior=.false.,lchk_lev=.true.

        ! call read_parameters
        ! ANNE read_parameters=voir dans libs/blups1
        ! change defaults if optional parameters present and check positive definitness
        ! call defaults_and_checks()

        if ( params%conv_crit /= -1.0 ) then
            conv_crit = params%conv_crit
            if ( writeok ) print*,'* convergence criterion (default=1e-12):',conv_crit
        endif
        if ( params%conv_crit /= -1.0 ) then
            maxrounds = params%maxrounds
            if ( writeok ) print*,'* maximum number of iterations (default=5000):',maxrounds
        endif
        if ( params%solv_method /= "" ) then
            solv_method=params%solv_method
            if ( writeok ) print*,'* solving method (default=PCG):',solv_method
        endif
        if ( params%r_factor /= -1 ) then
            r_factor = params%r_factor
            if ( writeok ) print*,'* relaxation factor for SOR (default=1.4):',r_factor
        endif
        if ( params%blksize /= -1 ) then
            blksize = params%blksize
            if ( writeok ) print*,'* block size for block diagonal matrix (default=1):',blksize
        endif

        !    if ( params%lchk_lev /= -1 ) then
        !       if (params%lchk_lev==0) lchk_lev=.false.
        !       if (params%lchk_lev==1) lchk_lev=.true.
        !       print*,'* check levels (default=.true.):',lchk_lev
        !    endif

        !if (stdresidual) solv_method='FSPAK'

        allocate (g(params%neff,maxcorr,maxcorr))
        g = params%gcov

        call pos_def(params%r,'Warning !!! R not positive definite: fixed',min_eig=set_eig)
        do i=1,params%neff
            if (params%randomnumb(i).ne.0) then
                j=params%randomnumb(i)*params%ntrait
                call pos_def(g(i,:j,:j),'Warning !!! G not positive definite: fixed',min_eig=set_eig)
            endif
        enddo

        allocate (rinv(params%ntrait,params%ntrait))

         !call defaults_and_checks_GS()
         !call print_parameters('BLUPF90 1.45')
        neq=params%ntrait*sum(params%nlev)
        data_len=max(params%pos_weight,maxval(params%pos_y),maxval(params%pos_eff),maxval(params%nestedcov))
        if ( writeok ) then
         print*,'Data record length = ',data_len
         print*,'# equations = ',neq
        end if
        allocate (xy(neq),sol(neq),address(params%neff,params%ntrait),&
            weight_cov(params%neff,params%ntrait),y(params%ntrait),nlevsum(params%neff))

        call setup_sum(params,nlevsum)
 
        call setup_g(params,g,set_eig)                    ! invert G matrices

        ! open(io_d,file=datafile)         !data file


        call zerom(xx,neq,100000); xy=0; nrec=0
        sol=0

        !ANNE

        ! On suppose maintenant que l effet d interet est l effet neff-1
        ! On suppose que l effet animal est l effet neff
        ! il reste donc neff-2 effets fixes

        annef=params%neff-1
        at=1
        if ( writeok ) then
         print *,'numero de leffet considere  ',annef
         print *,'numero du caractere   ',at
        end if

        !  annef=neff
        call zerom(hh,params%nlev(annef),100000)

        if ( writeok ) then
         print *,'Nombre de niveau  ',params%nlev(annef)
        end if

        sumxx=0;sum2=0
        !FIN ANNE
        allocate (indata(data_len))
        do idata=1,ndata
            indata = data (idata,:)
            nrec=nrec+1
            call decode_record(params,indata,weight_y,y)
            !   call find_addresses
            call find_addresses0(params,indata,weight_cov,address)
            call find_rinv(params,rinv,y)
            do i=1,params%neff
                do j=1,params%neff
                    do k=1,params%ntrait
                        do l=1,params%ntrait
                            val=weight_cov(i,k)*weight_cov(j,l)*weight_y*rinv(k,l)
                            !                 if (val /= 0) &
                            call addm(val,address(i,k),address(j,l),xx)
                            sumxx=sumxx+val
                            sum2=sum2+weight_y*rinv(k,l)
                        enddo
                    enddo
                enddo

                do k=1,params%ntrait
                    do l=1,params%ntrait
                        xy(address(i,k))=xy(address(i,k))+rinv(k,l)*y(l)*weight_cov(i,k) &
                            *weight_y
                    enddo
                enddo
            enddo
        enddo
        deallocate (indata)

        if ( writeok ) then
         print*,'read ',nrec,' records in ',second(),' s,  ',xx%filled,' nonzeroes'
        end if

        !
        !     Random effects' contributions
        first_snp=.true.
        do i=1,params%neff
            ! print *,i,params%randomtype(i)
            select case (params%randomtype(i))
                case (g_fixed)
                continue                ! fixed effect, do nothing
            case (g_diag)
                call add_g_diag(params,g,i,xx)
            case (g_A, g_As, g_A_UPG,g_A_UPG_INB)
                call add_g_add(params,g,ndata_ped,data_ped,params%randomtype(i),i,xx,writeok)
            !        if (whichH==3 .and. first_snp) then
            !           print *,"add_genomic n'est pas porté car le module genomic n'est pas implémenté... OFI"
            !           stop
            !           !call add_genomic(i,unit_eff)
            !           first_snp = .false.
            !        endif
            case (g_PD)
                ! call add_g_domin(params,i,xx)
                write (0,*) "pas implementé..."
                stop
            case (g_usr)
                call add_g_usr(params,g,nguser,guser,i,'',xx)
            case (g_usr_inv)
                call add_g_usr(params,g,nguser,guser,i,'inv',xx)
            case default
                print*,'unimplemented random type',params%randomtype(i)
        endselect
    enddo
    !stop

    if ( writeok ) then
     print*,'finished peds in ',second(),' s,  ',xx%filled,' nonzeroes'
    end if

    !
    if (neq <= 14) then
        if ( writeok ) print*,'left hand side'
        call printm(xx)
        if ( writeok ) print  '( '' right hand side:'',10(/10f8.2))',xy
    endif

    !if (lprior) then
    !   call read_solutions(prior_sol_file)
    !endif
    if(solv_method == 'FSPAK') then
        ! Comment the line above and uncomments the lines below only if
        ! solutions by FSPAK desired

        !ANNE
        ! changement de xx ici

        !pour recuperer les termes de linverse pour calculer un etest F

        !**********************************************************************************************************
        ! finalement il suuffit de rajouter 0 et da appeler juste la combinaison d effet donc modifier aussi pour cov
        !***********************************************************************************************************


        do i=1,params%nlev(annef)
            j=address1(params,annef,i,at)
            do l=i+1,params%nlev(annef)
                k=address1(params,annef,l,at)
                val=0.D0
                if (k.ge.j) then
                    call addm(val,j,k,xx)
                else
                    call addm(val,k,j,xx)
                endif
            enddo
            ! Ajout pour la matrice SNP*effet fixe
            do annei=1,params%neff-2
                do l=1,params%nlev(annei)
                    k=address1(params,annei,l,at)
                    val=0.D0
                    if (k.ge.j) then
                        call addm(val,j,k,xx)
                    else
                        call addm(val,k,j,xx)
                    endif
                enddo
            enddo
        enddo
        ! Ajout pour la matrice SNP*animal
        do l=1,params%nlev(params%neff)
            k=address1(params,params%neff,l,at)
            do i=1,params%nlev(annef)
                j=address1(params,annef,i,at)
                val=getm(j,k,xx)
                if (val.ne.0) then
                    do annei=1,params%nlev(annef)
                        annej=address1(params,annef,annei,at)
                        val=0.D0
                        if (k.ge.annej) then
                            call addm(val,annej,k,xx)
                        else
                            call addm(val,k,annej,xx)
                        endif
                    enddo
                    go to 521
                endif
            enddo
521     continue
        enddo


        !FIN ANNE
        xx_ija=xx

        call fspak90('factorize',xx_ija)
        call fspak90('solve',xx_ija,rs8=xy,sol8=sol)


    elseif(solv_method == 'SOR') then
        call default_iter(conv=conv_crit,maxround=maxrounds,relax=r_factor,zerosol=lzerosol) !each parameter optional
        call solve_iterm(xx,xy,sol)
    elseif(solv_method == 'PCG') then
        call default_iter(conv=conv_crit,maxround=maxrounds,zerosol=lzerosol) !each parameter optional
        call solve_pcg(xx,xy,sol,blksize)
        print *,'solutions de pcg ',sol(1)
    else
        print*,'Wrong solv_method ',solv_method
        stop
    endif

    if (neq <=16) print  '( '' solution:'' ,10(/10f8.2))',sol

    call store_solutions(params,ndata,annef,at,hh,hh_ija,xx,xx_ija,sol,solution)
    !if(residual) call write_adjusted_data
    !if(stdresidual) call write_studentized_residuals
    deallocate (xy,sol,address,weight_cov,y,nlevsum,rinv,g)
    call reset(xx)
    call reset(xx_ija)
    call reset(hh)
    call reset(hh_ija)

end subroutine exec_blupf90
    
subroutine decode_record(params,indata,weight_y,y)
    type(Param_blupf90), intent(in)  :: params
    real , dimension(:) , intent(in) :: indata
    real , dimension(:) , intent(inout) :: y
    real (rh) , intent(inout)           :: weight_y

    integer :: i
    ! decodes data record into y's and weight
    if (params%pos_weight /= 0) then
        weight_y = indata(params%pos_weight)
    else
        weight_y=1
    endif
    do i=1,params%ntrait
        y(i)=indata(params%pos_y(i))
    enddo
end subroutine


subroutine find_addresses0(params,indata,weight_cov,address)
    type(Param_blupf90), intent(in)    :: params
    real , dimension(:) , intent(in) :: indata
    real (rh) , dimension(:,:) ,intent(inout) :: weight_cov
    ! integer , dimension(:,:,:) ,intent(in) :: address1
    integer , dimension(:,:) ,intent(inout) :: address

    ! modified 09/04/2007
    integer :: i,j,lev

    do i=1,params%neff
        do j=1,params%ntrait
            select case (params%effecttype(i))
                case (effcross)
                    if(params%pos_eff(i,j) > 0) then
                        lev=int(indata(params%pos_eff(i,j)))
                        if(params%lchk_lev .and. lev > params%nlev(i) .and. params%nlev(i)/=0) then
                            print '(/4(a,i0)/)',' at Effect: ',i,' trait ',j,' level ',lev,' > # animals: ',params%nlev(i)
                            stop
                        endif
                    else
                        lev=0
                    endif
                    weight_cov(i,j)=1.0
                case  (effcov)
                    if(params%pos_eff(i,j) > 0) then
                        weight_cov(i,j)=indata(params%pos_eff(i,j))
                    else
                        weight_cov(i,j)=0
                    endif
                    if(params%nestedcov(i,j) == 0) then
                        lev=1
                    elseif (params%nestedcov(i,j) > 0) then
                        lev=int(indata(params%nestedcov(i,j)))
                        if(params%lchk_lev .and. lev > params%nlev(i) .and. params%nlev(i)/=0) then
                            print '(/4(a,i0)/)',' at Effect: ',i,' trait ',j,' level ',lev,' > # animals: ',params%nlev(i)
                            stop
                        endif
                    else
                        print*,'wrong description of nested covariable'
                        stop
                    endif
                case default
                    print*,'unimplemented effect ',i
                    stop
            end select

            if (lev /= 0) then
                address(i,j)=address1(params,i,lev,j)
            else
                address(i,j)=1    !dummy address
                weight_cov(i,j)=0.0
            endif
        enddo
    enddo
end subroutine


subroutine find_addresses(params,indata,weight_cov,address)
    type(Param_blupf90), intent(in) :: params
    real  , dimension(:) ,intent(in) :: indata
    real (rh) , dimension(:,:) ,intent(inout) :: weight_cov
    integer , dimension(:,:) ,intent(inout) :: address

    ! modified 10/25/2002
    integer :: i,j,lev

    do i=1,params%neff
        do j=1,params%ntrait

            select case (params%effecttype(i))
                case (effcross)
                    if(params%pos_eff(i,j) >0) then
                        lev=int(indata(params%pos_eff(i,j)))
                    else
                        lev=0
                    endif
                    weight_cov(i,j)=1.0
                case  (effcov)
                    if(params%pos_eff(i,j) >0) then
                        weight_cov(i,j)=indata(params%pos_eff(i,j))
                    else
                        weight_cov(i,j)=0
                    endif
                    if (params%nestedcov(i,j) == 0) then
                        lev=1
                    elseif (params%nestedcov(i,j) > 0) then
                        lev=int(indata(params%nestedcov(i,j)))
                    else
                        print*,'wrong description of nested covariable'
                        stop
                    endif
                case default
                    print*,'unimplemented effect ',i
                    stop
            end select

            if (lev /= 0) then
                address(i,j)=address1(params,i,lev,j)
            else
                address(i,j)=1    !dummy address
                weight_cov(i,j)=0.0
            endif
    
        enddo
    enddo
end subroutine

 
function address1(params,e,l,t)
    ! returns address for level l of effect e and trait t
    type(Param_blupf90), intent(in) :: params
    integer :: e,l,t, address1,i

    address1= params%offset(e)+(l-1)*params%ntrait+t
end function

 
subroutine find_rinv(params,rinv,y)
    type(Param_blupf90), intent(in) :: params
    real (rh), dimension(params%ntrait,params%ntrait),intent(inout) :: rinv      !inverse residual (co)variance matrix (tableau temporaire)
    real , dimension(:),intent(inout) :: y
    ! calculates inv(Q R Q), where Q is an identity matrix zeroed for
    ! elements corresponding to y(i)=mis
    interface
        subroutine ginv1(a,n,nmax,tol,rank)
            use kinds
            integer::n,nmax,rank
            real (rh)::a(nmax,nmax),tol
        end subroutine
    end interface

    integer :: i,irank
    !
    rinv=params%r
    do i=1,params%ntrait
        if (y(i) == params%miss) then
            rinv(i,:)=0; rinv(:,i)=0
        endif
    enddo
    call ginv1(rinv,params%ntrait,params%ntrait,real(1d-8,rh),irank)

end subroutine
  
subroutine setup_g(params,g,set_eig)
    type(Param_blupf90), intent(in) :: params
    real (rh),dimension(params%neff,params%ntrait,params%ntrait) , intent(inout):: g
    real (rh)          , intent(in) :: set_eig

    ! inverts g matrices
    interface
        subroutine ginv1(a,n,nmax,tol,rank)
            use kinds
            integer::n,nmax,rank
            real (rh)::a(nmax,nmax),tol
        end subroutine
    end interface
 
    integer :: i,j,rank
  
    do i=1,params%neff
        if (params%randomnumb(i).ne.0) then
            j=params%randomnumb(i)*params%ntrait
            call pos_def(g(i,:j,:j),'G not positive definite: fixed',min_eig=set_eig)
            !call printmat(g(i,:j,:j),'G','(9g12.5)')
            call ginv1(g(i,:,:),params%randomnumb(i)*params%ntrait,maxcorr,real(1d-12,rh),rank)
        endif
    enddo
end subroutine

subroutine add_g_diag(params,g,eff,xx)
    type(Param_blupf90), intent(in) :: params
    real (rh),dimension(params%neff,params%ntrait,params%ntrait) , intent(inout):: g
    type (sparse_hashm)   , intent(inout) :: xx
    ! adds diagonal (IID) contributions to MME
    integer :: eff, i,j,k,l,m,t1,t2
    do i=1,params%nlev(eff)
        do j=0,params%randomnumb(eff)-1
            do k=0,params%randomnumb(eff)-1
                do t1=1,params%ntrait
                    do t2=1,params%ntrait
                        m=address1(params,eff+j,i,t1); l=address1(params,eff+k,i,t2)
                        !                 xx(m,l)=xx(m,l)+g(eff,t1+j*ntrait,t2+k*ntrait)
                        call addm(g(eff,t1+j*params%ntrait,t2+k*params%ntrait),m,l,xx)
                    enddo
                enddo
            enddo
        enddo
    enddo
end subroutine

! inputs doit contenir l'ensemble des enregistrements.....
subroutine add_g_add(params,g,ninputs,inputs,type,eff,xx,writeok)
    type(Param_blupf90)           , intent(in) :: params
    real (rh),dimension(params%neff,params%ntrait,params%ntrait) , intent(inout):: g
    integer                       , intent(in) :: ninputs
    integer,    dimension(ninputs,4) , intent(in) :: inputs
    type (sparse_hashm)           , intent(inout) :: xx
    logical                       ,intent(in)     :: writeok
    ! generates contributions for additive sire or animal effects
    integer :: type,eff,i,i1,j,t1,t2,k,l,m,n,io,p(4),ped_len,nped,ii
    real (rh) ::w(3),res(4),mendel,val
    !
    select case (type)
        case (g_A)
            w=(/1., -.5, -.5/)
            res=(/2., 4/3., 1., 0./)
            ped_len=3
        case (g_A_UPG)
            w=(/1., -.5, -.5/)
            res=(/2., 4/3., 1., 0./)
            ped_len=4
        case (g_A_UPG_INB)
            w=(/1., -.5, -.5/)
            ped_len=4
        case (g_As)
            w=(/1., -.5, -.25/)
            res=(/16/11., 4/3., 16/15., 1./)
            ped_len=3
    end select

    nped=0
    do ii=1,ninputs
        ! read(unit_eff,*,iostat=io)p(1:ped_len)
        p(1:ped_len) = inputs(ii,:ped_len)
        nped=nped+1
        select case (type)
            case (g_A)
                i=1; if(p(2)==0) i=i+1; if (p(3) == 0) i=i+1
                mendel=res(i)
            case (g_A_UPG)
                mendel=res(p(4))
            case (g_A_UPG_INB)
                mendel=p(4)/1000.
            case (g_As)
                i=1; if(p(2)==0) i=i+2; if (p(3) == 0) i=i+1
                mendel=res(i)
        end select

        do i=0,params%randomnumb(eff) - 1
            do i1=1,3
                if (PARAMS%lchk_lev .and. p(i1) > params%nlev(eff+i) .and. params%nlev(eff+i)/=0) then
                    print '(/3(a,i0)/)',' at Effect: ',eff+i,' level ',p(i1),' > # animals: ',params%nlev(eff+i)
                    stop
                endif
            enddo
            do j=0,params%randomnumb(eff) - 1
                do t1=1,params%ntrait
                    do t2=1,params%ntrait
                        if (params%pos_eff(eff+i,t1)/=0 .and. params%pos_eff(eff+j,t2)/=0) then
                            do k=1,3
                                do l=1,3
                                    if (p(k) /=0 .and. p(l) /=0) then
                                        m=address1(params,eff+i,p(k),t1)
                                        n=address1(params,eff+j,p(l),t2)
                                        val=g(eff,t1+i*params%ntrait,t2+j*params%ntrait)&
                                            *w(k)*w(l)*mendel
                                        call addm(val,m,n,xx)
                                    endif
                                enddo
                            enddo
                        endif
                    enddo
                enddo
            enddo
        enddo
    enddo
    if ( writeok ) print*,' read ',nped,' additive pedigrees'
end subroutine add_g_add

!! inputs(,1) -> row, inputs(,2) -> row, inputs(,3) -> x
subroutine add_g_usr(params,g,ninputs,inputs,eff,text,xx)
    type(Param_blupf90), intent(in)     :: params
    real (rh),dimension(params%neff,params%ntrait,params%ntrait) , intent(inout):: g
    integer                     , intent(in) :: ninputs
    real , dimension(ninputs,3) , intent(in) :: inputs
    type (sparse_hashm)   , intent(inout) :: xx

    ! generates contributions user defined matrix; if text is 'inv',
    !    this matrix is inverted
    !
    integer :: eff,i,j,k,m,n,row,col,t1,t2,type=0,maxrow=0,maxcol=0,maxdiag=0,ii
    integer,parameter::upper=1,lower=2
    real (rh) ::val,x,ident(params%nlev(eff)),sol(params%nlev(eff))
    type (sparse_hashm)::a_usr
    type (sparse_ija)::a_usr_ija
    character (*)::text
    !
    call init(a_usr)
    call zerom(a_usr,params%nlev(eff))
    n=0

    do ii=1,ninputs
        row = nint(inputs(ii,1))
        col = nint(inputs(ii,2))
        x = inputs(ii,3)
        ! read(unit_eff,*,iostat=io)row,col,x
        ! if (io /= 0) exit
        n=n+1
    
        ! Only upper or lower half stiored matrices are accepted
        if (row /=col) then
            if (type == 0) then
                if (row < col) type = upper
                if (row > col) type = lower
            endif

            if ((type == upper .and. row > col) .or. &
                (type == lower .and. row < col)) then
                print*,'Mixed upper and lower diagonals in g_usr_inv:',row,col,x
                stop
            endif
        endif
        
        call addm(x,row,col,a_usr) 
        if (row /=col) call addm(x,col,row,a_usr) 
        maxrow=max(maxrow,row)
        maxcol=max(maxcol,col)
        if (row==col) maxdiag=max(maxdiag,row)
    enddo
     
    print*,'g_usr_inv: read ',n,' elements'
    if (n == 0) then
        print*,'User defined file for effect ',eff,' empty'
        stop
    endif
     
    !df_random(eff)=params%nlev(eff)

    print*,'largest row, column, diagonal:',maxrow,maxcol,maxdiag

    !print*,'a_usr'
    !call printm(a_usr)
  
    if (text(1:3)=='inv') then ! invert a_usr
        a_usr_ija=a_usr
        call reset(a_usr)
        call init(a_usr)
        call zerom(a_usr,params%nlev(eff))
        print*,'a_usr_ija'
        !call printm(a_usr_ija)
        do i=1,params%nlev(eff)
            ident=0; ident(i)=1
            call fspak90('solve',a_usr_ija,ident,sol)
            do j=1,params%nlev(eff)
                if (sol(j) /= 0) call addm(sol(j),i,j,a_usr)
            enddo
        enddo
        !       call fspak90('stat',a_usr_ija)
        call reset(a_usr_ija)
    else
         !ainv(eff)=a_usr
    endif
  
    if (params%nlev(eff) <= 20) then
        print*,'user defined matrix'
        call printm(a_usr)
    endif
     
  
    do i=0,params%randomnumb(eff) - 1 !loops for correlated effects
        do j=0,params%randomnumb(eff) - 1
            do t1=1,params%ntrait
                do t2=1,params%ntrait
                    if (params%pos_eff(eff+i,t1)/=0 .and. params%pos_eff(eff+j,t2)/=0) then
                        do k=1,a_usr%nel
                            row=a_usr%x(1,k)
                            if (row /=0) then
                                col=a_usr%x(2,k)
                                x=a_usr%x(3,k)
                                m=address1(params,eff+i,row,t1)
                                n=address1(params,eff+j,col,t2)
                                val=g(eff,t1+i*params%ntrait,t2+j*params%ntrait)*x
                                call addm(val,m,n,xx)
                                !		           if (m /=n) call addm(val,n,m,xx)
                                if (row /= col) call addm(val,n,m,xx)
                            endif
                        enddo
                    endif
                enddo
            enddo
        enddo
    enddo
    call reset(a_usr)
end subroutine


!subroutine add_g_domin(params,g,ninputs,inputs,eff,xx)
!    type(Param_blupf90), intent(in)     :: params
!    real (rh),dimension(params%neff,params%ntrait,params%ntrait) , intent(inout):: g
!    integer                     , intent(in) :: ninputs
!    real , dimension(ninputs,3) , intent(in) :: inputs
!    type (sparse_hashm)   , intent(inout) :: xx
!    ! adds parental dominance matrix using files created by RENDOMN
!    integer :: eff,i,i1,j,t1,t2,k,l,m,n,io,p(10),max_class,d,nped
!    real (rh) ::w(9)=(/1.0,-.5,-.5,-.5,-.5,.25,.25,.25,.25/),r(0:255)
!    logical,save::first=.true.
!    real (rh) ::  val
!
!    ! compute residuals for different classes of missing s-d combinations
!    if (first) then
!        call creatres(r)
!        first=.false.
!    endif
!
!    max_class=0
!
!    !  contributions to diagonals for classes with parental classes
!
!    nped=0
!    do
!        ! read(unit_eff,*,iostat=io)p(1:10)
!        ! if (io /= 0) exit
!        nped=nped+1
!        max_class=max(max_class,p(1))
!        do i=0,params%randomnumb(eff) - 1
!            do i1=1,9
!                if(params%lchk_lev .and. p(i1) > params%nlev(eff+i) .and. params%nlev(eff+i)/=0) then
!                    print '(/3(a,i0)/)',' at Effect: ',eff+i,' level ',p(i1),' > # levels: ',params%nlev(eff+i)
!                    stop
!                endif
!            enddo
!            do j=0,params%randomnumb(eff) - 1
!                do t1=1,params%ntrait
!                    do t2=1,params%ntrait
!                        do k=1,9
!                            do l=1,9
!                                if (p(k) /=0 .and. p(l) /=0) then
!                                    m=address1(params,eff+i,p(k),t1)
!                                    n=address1(params,eff+j,p(l),t2)
!                                    val=g(eff,t1+i*params%ntrait,t2+j*params%ntrait)&
!                                        *r(p(10))*w(k)*w(l)
!                                    call addm(val,m,n,xx)
!                                endif
!                            enddo
!                        enddo
!                    enddo
!                enddo
!            enddo
!        enddo
!    enddo
!
!    !  contributions to diagonals for classes without parental classes
!
!    do d=max_class+1,params%nlev(eff)
!        do i=0,params%randomnumb(eff) - 1
!            do j=0,params%randomnumb(eff) - 1
!                do t1=1,params%ntrait
!                    do t2=1,params%ntrait
!                        m=address1(params,eff+i,d,t1)
!                        n=address1(params,eff+j,d,t2)
!                        val=g(eff,t1+i*params%ntrait,t2+j*params%ntrait)
!                        call addm(val,m,n,xx)
!                    enddo
!                enddo
!            enddo
!        enddo
!    enddo
!
!    print*,' read ',nped,' dominance pedigrees'
!
!end subroutine


subroutine store_solutions(params,ndata,annef,at,hh,hh_ija,xx,xx_ija,sol,solution)
    type(Param_blupf90)   , intent(in)    :: params
    integer               , intent(in)    :: ndata,annef,at
    type (sparse_hashm)   , intent(inout) :: hh
    type (sparse_ija)     , intent(inout) :: hh_ija
    type (sparse_hashm)   , intent(inout) :: xx
    type (sparse_ija)     , intent(inout) :: xx_ija
    real(rh),dimension(:),intent(in)      :: sol
    type(Solution_blupf90) , intent(inout)  :: solution
    ! store solutions in file 'solutions'

    integer l,t,j,effet,niveau,i,k,ccount
    !ANNE
    real (rh), allocatable:: hb(:),hd(:)
    real (rh) :: hc,val
    integer :: annej,annei
    character(50) ::      fmt_sol_se=        '(2i4,i10,2f16.8)'               ! format for solutions file (sol and se)
    real (r8) :: a,b,tmp
    external :: INCOB
    allocate (hb(params%nlev(annef)),hd(params%nlev(annef)))

    solution%nlevtot = sum(params%nlev(:params%neff-2))
    solution%nhaplo = params%nlev(annef)
    solution%nani = ndata ! params%nlev(params%neff)

    !print *,"Nlevtot=",solution%nlevtot
    !print *,"Nhaplo=",solution%nhaplo
    !print *,"Nani=",solution%nani

    allocate (solution%A(solution%nhaplo,solution%nlevtot))
    allocate (solution%B(solution%nhaplo,solution%nhaplo))
    allocate (solution%C(solution%nhaplo,solution%nlevtot))
    allocate (solution%D(solution%nhaplo,ndata))
    allocate (solution%Dani(ndata))
    allocate (solution%sol(solution%nhaplo))
    allocate (solution%std(solution%nhaplo))

    !  io_s=8765
    !  open(io_s,file='solu_'//datafile)
    !   open(io_s,file='solu_TEST')
    !   open(io_t,file='test_TEST')
    !  open(io_mat,file='mat_'//datafile)

    !FIN ANNE

    ! if(sol_se == 'se') then
    if( .true. ) then

        !  write(io_s,*) 'trait/effect level  solution          s.e.'

        if (params%solv_method /= 'FSPAK') xx_ija=xx
        !ANNE
        !print *,'avant '
        do i=1,params%nlev(annef)
            do j=i,params%nlev(annef)
                k=address1(params,annef,i,at)
                l=address1(params,annef,j,at)
                val=getm(k,l,xx_ija)
            !         print *,'matrice   ',i,j,val
            enddo
        enddo

        ! Impression des la matrice SNP*effet fixe avant inversion

        do i=1,params%nlev(annef)
            annej=0
            do annei=1,params%neff-2
                do j=1,params%nlev(annei)
                    annej=annej+1
                    k=address1(params,annef,i,at)
                    l=address1(params,annei,j,at)
                    if (k.le.l) then
                        val=getm(k,l,xx_ija)
                    else
                        val=getm(l,k,xx_ija)
                    endif
                    ! write (io_mat,fmt=*) "0  ",i,annej,val    !!! Ecriture de la matrice A...
                    solution%A(i,annej) = val
                enddo
            enddo
        enddo



        !FIN ANNE

        call fspak90('inverse',xx_ija)

        !ANNE
        !print *,'inverse '
        ! do i=1,params%nlev(annef)
        !      do j=i,params%nlev(annef)
        !         k=address1(params,annef,i,at)
        !         l=address1(params,annef,j,at)
        !         val=getm(k,l,xx_ija)
        !         print *,'matrice  ',i,j,val
        !enddo
        !enddo
        !FIN ANNE
        !ANNE
        !    Calcul du test

            ! Effet

        !print *,'effet utilise  store',annef
        !print *,'trait utilise  store ',at
        ! trait
        solution%B=0.d0
        !regression simple
        if (params%nlev(annef).eq.1) then
            j=address1(params,annef,1,at)
            hb(1)=sol(j)
            k=address1(params,annef,1,at)
            val=getm(j,k,xx_ija)
            hc=hb(1)*hb(1)/val
            !  write (io_t,fmt=*) "1   ",annef,hc,params%nlev(annef)
            !  print *,'test  ',hc, params%nlev(annef)

            solution%test=hc
            tmp=real(solution%nani)/(real(solution%nani)+solution%test)
            a = solution%nani/2.d0
            b = 1.d0/2.d0
            solution%pvalue = BETAI(a,b,tmp)
            !call INCOB(a,b,tmp,bix)
            !print *,bix
            !solution%pvalue = bix
           ! print *,"betai==",a,b,tmp,hc,hb(1),val
           ! print *,"pvalue=",solution%pvalue," test=",solution%test," tmp=",tmp
            !solution%pvalue=betai(solution%nani/2.d0,1.d0/2.d0,tmp)

            ! Ecriture inverse SNP*SNP
            ! write (io_mat,fmt=*) "1  ",params%nlev(annef),params%nlev(annef),val
            solution%B(1,1) = val

        else

            !plusieurs genotypes ou haplotypes

            do i=2,params%nlev(annef)
                j=address1(params,annef,i-1,at)
                k=address1(params,annef,i,at)
                hb(i-1)=sol(j)-sol(k)

            enddo

            do i=1,params%nlev(annef)-1
                do j=i,params%nlev(annef)-1
                    k=address1(params,annef,i,at)
                    l=address1(params,annef,j,at)
                    val=getm(k,l,xx_ija)
                    ! Ecriture inverse SNP*SNP
                    ! write (io_mat,fmt=*) "1  ",i,j,val
                    solution%B(i,j)=val
                    call addm(val,i,j,hh)
                    k=address1(params,annef,i+1,at)
                    l=address1(params,annef,j,at)
                    if (k.le.l) then
                        val=-getm(k,l,xx_ija)
                    else
                        val=-getm(l,k,xx_ija)
                    endif
                    call addm(val,i,j,hh)
                    k=address1(params,annef,i,at)
                    l=address1(params,annef,j+1,at)
                    val=-getm(k,l,xx_ija)
                    call addm(val,i,j,hh)
                    k=address1(params,annef,i+1,at)
                    l=address1(params,annef,j+1,at)
                    val=getm(k,l,xx_ija)
                    ! Ecriture inverse SNP*SNP
                    if ((i.eq.params%nlev(annef)-1).and.(i.eq.j)) then
                        !write (io_mat,fmt=*) "1  ",i+1,j+1,val  !!! Ecriture de la matrice B...
                        solution%B(i+1,j+1)=val
                    endif

                    call addm(val,i,j,hh)

                enddo
            enddo
        endif

        ! Impression des la matrice SNP*effet fixe

        do i=1,params%nlev(annef)
            annej=0
            do annei=1,params%neff-2
                do j=1,params%nlev(annei)
                    annej=annej+1
                    k=address1(params,annef,i,at)
                    l=address1(params,annei,j,at)
                    if (k.le.l) then
                        val=getm(k,l,xx_ija)
                    else
                        val=getm(l,k,xx_ija)
                    endif
                    ! write (io_mat,fmt=*) "2  ",i,annej,val
            !        print *,i,annej,val
                    solution%C(i,annej)=val
                enddo
            enddo
        enddo

        ! Impression des la matrice SNP*animal
        solution%D=0.d0
        solution%Dani=0
        do i=1,params%nlev(annef)
            ccount=0
            do j=1,params%nlev(params%neff)
                k=address1(params,annef,i,at)
                l=address1(params,params%neff,j,at)
                if (k.le.l) then
                    val=getm(k,l,xx_ija)
                else
                    val=getm(l,k,xx_ija)
                endif
                if (val.ne.0) then
                    !write (io_mat,fmt=*) "3  ",i,j,val
                    ccount=ccount+1     ! Attention valeur par ordre croissant d'animaux qui ont une perf et un genotype
                    solution%D(i,ccount)=val
                    solution%Dani(ccount)=j
                endif
            enddo
        enddo

        if (params%nlev(annef).gt.1) then
            !    print *,'avant ija'
            hh_ija=hh
            !    print *,'avant inverse'
            call fspak90('inverse',hh_ija)
            hd=mult(hb,hh_ija)
            hc=0.D0
            do i=1,params%nlev(annef)-1
                hc=hc+hd(i)*hb(i)
            enddo

            hc=hc/dfloat(params%nlev(annef)-1)
            !    write (io_t,fmt=*) "1   ",annef,hc,params%nlev(annef)
            !    print *,'test  ',hc, params%nlev(annef)
            solution%test=hc
            tmp=real(solution%nani)/(real(solution%nani)+solution%test)
            a = solution%nani/2.d0
            b = 1.d0/2.d0
            !print *,"betai==",a,b,tmp
            solution%pvalue = BETAI(a,b,tmp)

            !print *,"pvalue=",solution%pvalue
            !solution%pvalue=betai(solution%nani/2.d0,1.d0/2.d0,tmp)
        endif


        !FIN ANNE

        !Moyenne general (1er effet)
        j=address1(params,1,1,1)
        solution%solmu=sol(j)
        solution%stdmu=sqrt(getm(j,j,xx_ija))

        !Snp/haplotype sol (Avant dernier effet)
        do l=1,params%nlev(params%neff-1)
            j=address1(params,params%neff-1,l,1)
            solution%sol(l)=sol(j)
            solution%std(l)=sqrt(getm(j,j,xx_ija))
        end do


    !     do e=1,params%neff-1  ! ANNE ecriture des effets sauf effet animal
    !       do l=1,params%nlev(e)
    !
    !          do t=1,params%ntrait
    !             j=address1(params,e,l,t)
    !	     write(io_s,fmt=fmt_sol_se)t,e,l,sol(j),sqrt(getm(j,j,xx_ija))
    !
    !      !     write(*,fmt=fmt_sol_se)t,e,l,sol(j),sqrt(getm(j,j,xx_ija))
    !          enddo
    !       enddo
    !    enddo
    !    print*,'solutions and s.e. stored in file: "solutions"'
    else
    !    write(io_s,'(''trait/effect level  solution'')')
    !    do e=1,params%neff
    !       do l=1,params%nlev(e)
    !          do t=1,params%ntrait
    !             write(io_s,fmt=fmt_sol_se)t,e,l,sol(address1(params,e,l,t))
    !          enddo
    !       enddo
    !    enddo
    !    print*,'solutions stored in file: "solutions"'
    endif

!close(io_s)

end subroutine

!  subroutine read_solutions(afile)
!  ! read solutions in file 'solutions'
!  integer e,l,t,j,i,io
!  real(rh) s
!  character*(*) :: afile
!  open(io_s,file=afile)
!
!  read(io_s,*)
!  do i=1,neq
!     read(io_s,fmt=fmt_sol_se,iostat=io)t,e,l,s
!     if(io/=0)  exit
!     sol(address1(e,l,t)) = s
!  enddo
!  print*,'solutions readed from file: "',afile,'"'
!  close(io_s)
!
!  end subroutine
  
!  subroutine defaults_and_checks(params)
!   type(Param_blupf90), intent(in) :: params
!  ! check optional parameters and check for positive definitness
!  integer::i
!
!   if ( params%conv_crit /= -1.0 ) then
!      conv_crit = params%conv_crit
!      print*,'* convergence criterion (default=1e-12):',conv_crit
!   endif
!   if ( params%conv_crit /= -1.0 ) then
!      maxrounds = params%maxrounds
!      print*,'* maximum number of iterations (default=5000):',maxrounds
!   endif
!  if ( params%solv_method /= "" ) then
!      solv_method="PCG"
!      print*,'* solving method (default=PCG):',solv_method
!   endif
!   if ( params%r_factor /= -1 ) then
!      r_factor = params%r_factor
!      print*,'* relaxation factor for SOR (default=1.4):',r_factor
!   endif
!   if ( params%blksize /= -1 ) then
!      blksize = params%blksize
!      print*,'* block size for block diagonal matrix (default=1):',blksize
!   endif
!
!   residual=params%residual
!   if (residual) &
!     print*,'* yhat and residual will be shown in yhat_residual'
!
!   stdresidual=params%stdresidual
!   if (stdresidual) &
!      print*,'* yhat and student residual will be shown in yhat_student_residual'
!
!    if ( params%mis /= -1 ) then
!       mis = params%mis
!       print*,'* missing observation (default=0):',mis
!    endif
!    if ( params%lchk_lev /= -1 ) then
!       if (params%lchk_lev==0) lchk_lev=.false.
!       if (params%lchk_lev==1) lchk_lev=.true.
!       print*,'* check levels (default=.true.):',lchk_lev
!    endif
!
!  if (stdresidual) solv_method='FSPAK'
!
!  call pos_def(r,'Warning !!! R not positive definite: fixed',min_eig=set_eig)
!  do i=1,params%neff
!   if (params%randomnumb(i).ne.0) then
!       j=params%randomnumb(i)*params%ntrait
!       call pos_def(params%g(i,:j,:j),'Warning !!! G not positive definite: fixed',min_eig=set_eig)
!   endif
!  enddo
!  end subroutine

!  subroutine write_adjusted_data
!
!  ! write yhat and residual
!  integer::i,j,k
!  character (40)::form
!  real :: y2(ntrait)
!  form='(100f14.6)'
!  rewind io_d
!  open(io_s+1,file='yhat_residual')
!
!  do j=1,nrec
!    read(io_d,*,iostat=io)indata
!    call decode_record
!    call find_addresses
!    y2=y
!    do k=1,ntrait
!       if (y(k) /= mis) then
!          do i=1,neff
!            y(k)=y(k)- weight_cov(i,k)*sol(address(i,k))
!          enddo
!       endif
!    enddo
!    if (j<10) print*, y2,y
!    write(io_s+1,form)y2-y,y
!  enddo
!
!  print*,'wrote yhat and residual in file "yhat_residual"'
!  end subroutine write_adjusted_data
!
!
!  subroutine write_studentized_residuals
!  ! compute studentized residual by e/sqrt(mse*(1-Hii)
!  integer::i,j,k,l,m,ii,jj,t,e
!  character (40)::form
!  real :: y2(ntrait),nn(ntrait),df(ntrait)
!  real(r8) :: ee(ntrait),seff(neff,ntrait),mse(ntrait)
!  real(r4) :: h(ntrait)
!  real(rh) ::val
!  !
!
!  !if (ntrait >1) stop 'studentized residual only for ntrait==1'
!
!  form='(100f14.6)'
!  rewind io_d
!  open(io_s+1,file='yhat_student_residual')
!  ee=0
!  nn=0
!  do j=1,nrec
!    read(io_d,*,iostat=io)indata
!    call decode_record
!    call find_addresses
!    y2=y
!    do k=1,ntrait
!       if (y(k) /= mis) then ! mis
!          nn(k)=nn(k)+1
!          do i=1,neff
!            !y(k)=y(k)- weight_cov(i,k)*weight_y*sol(address(i,k))
!            y(k)=y(k)- weight_cov(i,k)*sol(address(i,k))
!          enddo
!       endif
!    enddo
!    y=y*sqrt(weight_y)
!
!
!    where(y2/=mis)
!       ee=ee+y**2
!    endwhere
!  enddo
!
!  df=0
!  do t=1, ntrait
!     do e=1,neff
!        do l=1,nlev(e)
!           if (sol(address1(e,l,t))/=0.d0) then
!              df(t)=df(t)+1
!           endif
!        enddo
!     enddo
!  enddo
!
!  if(sol_se /= 'se') call fspak90('inverse',xx_ija,rank=i)
!  !xrank = i
!  xx=xx_ija
!  print '(a)','wrote yhat and student residual in file "yhat_student_residual"'
!  print '(a)','sum(e**2) '
!  print '(10(f15.2,x))',ee
!  print '(a)','nrec '
!  print '(10i10)',int(nn)
!  print '(a)','parameters sol /=0'
!  print '(10i10)',int(df)
!  print '(a)','error df'
!  print '(10i10)',int(nn-df)
!  print '(a)','MSE '
!  print '(10f10.3)',ee/(nn-df)
!
!  mse=ee/(nn-df)
!
!  if (neq<10) call printm(xx_ija)
!  rewind io_d
!  do i=1,nrec
!     read(io_d,*,iostat=io)indata
!     call decode_record
!     call find_addresses
!
!     ! get residual
!     y2=y
!     do k=1,ntrait
!        if (y(k) /= mis) then ! mis
!           do j=1,neff
!            y(k)=y(k)- weight_cov(j,k)*sol(address(j,k))
!           enddo
!        endif
!     enddo
!
!
!     !  get diagonal of Hat matrix
!     seff=0
!     h=0
!     do k=1,ntrait
!        if (y(k) == mis) cycle
!        do l=1,neff
!           ii = address(l,k)
!           do m=1,neff
!              jj = address(m,k)
!              if (ii>jj) then
!                 !val=getm(jj,ii,xx_ija)
!                 val=getm(jj,ii,xx)
!              else
!                 !val=getm(ii,jj,xx_ija)
!                 val=getm(ii,jj,xx)
!              endif
!              seff(m,k)=seff(m,k) + weight_cov(l,k)*weight_y*val*rinv(k,k)
!           enddo
!        enddo
!        do m=1,neff
!           h(k) = h(k) + seff(m,k)*weight_cov(m,k)
!        enddo
!     enddo
!
!     h=min(h,real(1,kind(h)))
!
!     ee=0
!     where (h/=1.d0)
!        ee=y/(sqrt(mse*(1.d0-h))/sqrt(weight_y))
!     endwhere
!
!     if (i<10) print*, y2,y,ee,sqrt(mse*(1.d0-h))/sqrt(weight_y),h
!     write(io_s+1,form)ee
!     !write(999,'(100f30.20)') h
!  enddo  ! nrec
!  end subroutine write_studentized_residuals

subroutine setup_sum(params,nlevsum)
    type(Param_blupf90), intent(in)     :: params
    integer  ,dimension(:) , intent(inout) :: nlevsum

    integer  :: i

    do i=1, params%neff
        nlevsum(i)=sum(params%nlev(1:i-1))
    enddo
end subroutine
 
!  subroutine add_genomic(i,unit)
!  integer :: i,unit
!  !
!  call setup_genomic(i,unit)
!  if (sameWeights) then
!     !print*,'wGimA22i',wGimA22i
!     call add_g_usr_gs(xx,i,nlevsum,randomnumb(i),ntrait,GimA22i,wGimA22i)
!  else
!     !print*,'wGi,wA22i',wGi,wA22i
!     call add_g_usr_gs(xx,i,nlevsum,randomnumb(i),ntrait,Gi,wGi)
!     call add_g_usr_gs(xx,i,nlevsum,randomnumb(i),ntrait,A22i,wA22i)
!  endif
!  end subroutine


subroutine release_paramblupf90(paramblupf90)
    type(Param_blupf90), intent(inout)         :: paramblupf90

    deallocate (paramblupf90%pos_y)
    deallocate (paramblupf90%r)
    deallocate (paramblupf90%pos_eff)
    deallocate (paramblupf90%nlev)
    deallocate (paramblupf90%effecttype)
    deallocate (paramblupf90%nestedcov)
    deallocate (paramblupf90%randomnumb)
    deallocate (paramblupf90%randomtype)
    deallocate (paramblupf90%gcov)
    deallocate (paramblupf90%offset)

end subroutine release_paramblupf90


subroutine release_solutionblupf90(solblupf90)
    type(Solution_blupf90), intent(inout)         :: solblupf90

    deallocate (solblupf90%A,solblupf90%B,solblupf90%C,solblupf90%D,solblupf90%Dani)
    deallocate (solblupf90%sol,solblupf90%std)
    solblupf90%nlevtot = 0
    solblupf90%nhaplo = 0
    solblupf90%nani = 0

end subroutine release_solutionblupf90

end module blupf90
 

