module ilapack
    use kinds

    implicit none

!    interface DGESV
!         module procedure DGESV_lapack
!    end interface
!
!    interface DSYEV
!         module procedure DSYEV_lapack
!    end interface

  public :: DGESV_lapack,DSYEV_lapack

contains
    SUBROUTINE DGESV_lapack(x,x_o)
        real(r8) ,intent(in)  :: x(:,:)
        real(r8) ,intent(out) :: x_o(:,:)

        INTEGER :: N,NRHS,INFO,LDA,LDB
        DOUBLE PRECISION , DIMENSION(:,:) , allocatable :: A,B
        DOUBLE PRECISION , DIMENSION(:) , allocatable :: IPIV

        EXTERNAL :: DGESV

        N = SIZE(x,1)
        NRHS = N
        LDA = size(x,1)
        LDB = size(x_o,1)

        allocate(A(size(x,1),size(x,2)),&
            B(size(x_o,1),size(x_o,2)),&
            IPIV(N))
        A = x
#ifdef LAPACK
        call DGESV(N,NRHS,A,LDA,IPIV,B,LDB,INFO)
#else
        write (0,*) "Please, You should use MKL INTEL or LAPACK Librairie (DGESV)."
        stop
#endif
        x_o = B
        IF (INFO /= 0) write (0,*) "Error DGESV INFO=",INFO

        deallocate (A,B,IPIV)

    END SUBROUTINE DGESV_lapack

    SUBROUTINE DSYEV_lapack(d, v, JOBZ)
        real(r8) , intent(inout) :: d(:,:)
        real(r8) , intent(out)   :: v(:)
        character(len=1), intent(in) :: JOBZ

        CHARACTER(len=1) :: UPLO='U'
        INTEGER          :: N,LDA,LWORK,INFO

        DOUBLE PRECISION , ALLOCATABLE, DIMENSION(:,:)  :: A
        DOUBLE PRECISION , ALLOCATABLE, DIMENSION(:)    :: WORK,W

        external :: DSYEV

        N = size(d,1)
        LDA = N
        LWORK = N * 5
        allocate (A(N,N))
        A = d
        allocate (W(N),WORK(LWORK))

#ifdef LAPACK
        call DSYEV(JOBZ,UPLO,N,A,LDA,W,WORK,LWORK,INFO)
#else
        write (0,*) "Please, You should use MKL INTEL or LAPACK Librairie (DSYEV)."
        stop
#endif
        v = W
        d = A

        deallocate (A,W,WORK)

    END SUBROUTINE DSYEV_lapack

end module ilapack
