module io
    use const
    use cli
    implicit none

    public :: load_genotype
    public :: set_map

    contains

        subroutine load_genotype(genotypeFile,npos,id,genotype,nu)
        character(len=len_long_cli)               ,intent(in) :: genotypeFile
        integer                                   ,intent(in) :: npos
        integer , dimension(:)     , pointer,intent(inout)    :: id
        integer , dimension(:,:)     , pointer,intent(inout)  :: genotype
        integer                               ,intent(inout)  :: nu

        !Local
        integer :: ios,i,unit_val=1234,izero,nup,idread
        character(len=2)  :: lexeme
        character(len=10) :: nposstring
        character(len=npos)                                   :: bufsnp

        !Lecture des genotypes/haplotypes status
        open(unit=unit_val,file=genotypeFile,iostat=ios,action="read")

        if ( ios /= 0 ) then
            write (0,*) "Can not open file ",trim(genotypeFile)
            stop
        end if

       write (nposstring,fmt="(I7)") npos

        nup = 0
        do while (ios == 0)
            read (unit_val,*,iostat=ios) lexeme
            if ( ios == 0) nup = nup+1
        end do

        !
        !  LECTURE DES ID/GENOTYPES
        !
        allocate (id(nup))
        allocate (genotype(nup,npos))

        nu=0
        ios=0
        izero = ichar('0')
        genotype=-1
        id=-1
        rewind(unit_val)
        do while ( ios == 0 )
            idread = -1
            read (unit_val,'(I'//trim(sizeid)//',1x,(a'//trim(nposstring)//'))',iostat=ios) idread,bufsnp
            if ( ios == 0 .and. idread > 0 ) then
                nu = nu+1
                id(nu)=idread
                genotype(nu,:) = (/ (ichar(bufsnp(i:i))-izero,i=1,npos) /)
            end if
            if (.not. all( genotype(nu,:) >=0 .and. genotype(nu,:) <=5 )) then
             write (0,*) "Error reading ID=",idread,". Bad construction of genotype file. Check otpions [npos=",&
              npos,"] and [idlength=",trim(sizeid),"]."
             do i=1,npos
              if (genotype(nu,i) <0 .or. genotype(nu,i) >5) write (0,*) "SNP ",i,genotype(nu,i)
             end do
              stop
            end if
        end do

        close(unit_val)

        if ( nu <= 0 ) then
            write (0,*) "Can not read genotype file :[",trim(genotypeFile),"]"
            stop
        end if

    end subroutine load_genotype

    ! mapFile   : the file
    ! startpos  : array of dimension nchr with started position
    ! endpos    : array of dimension nchr with end position
    ! nchr      : number of chromosome
    ! npostotal : nombre total de marqueurs dans le fichiers de génotype
    !
    ! initialize data structure startpos,endpos,nchr
    subroutine set_map(mapFile,chr,startpos,endpos,nmark,nchr)
      character(len=len_long_cli)               ,intent(in)  :: mapFile
      character(len=20) ,dimension(:),pointer  ,intent(out)  :: chr
      integer , dimension(:)           , pointer,intent(out) :: startpos
      integer , dimension(:)           , pointer,intent(out) :: endpos
      integer , dimension(:)           , pointer,intent(out) :: nmark
      integer                                   ,intent(out) :: nchr

      integer , parameter :: unit_val = 814
      integer , parameter :: MAX_CHR = 200

      integer :: ios

      open(unit=unit_val,file=mapFile,iostat=ios,action="read")

      if ( ios /= 0 ) then
            write (0,*) "Can not open file ",trim(mapFile)
            stop
      end if

      allocate (chr(MAX_CHR),startpos(MAX_CHR),endpos(MAX_CHR),nmark(MAX_CHR))
      nchr=0
      startpos=999999
      nmark=0
      endpos=0

      do while (ios == 0)
          nchr=nchr+1
          read (unit_val,*,iostat=ios) chr(nchr),startpos(nchr),endpos(nchr)
          if ( ios == 0 ) then
            if ( startpos(nchr)<=0 .or. endpos(nchr)<=0 ) then
              write (0,*) "Can not defined position <=0 in the map File CHR=",chr(nchr)
              stop
            end if
            nmark(nchr) = endpos(nchr) - startpos(nchr) + 1
          end if
      end do
      nchr=nchr-1

      close(unit_val)

    end subroutine set_map
end module io
