!   Olivier Filangi aT Rennes dot INRA dot FR
!
!   Test du module blupf90
!
!   Creation Juillet 2013
!
program PROG_TESTMODBLUPF90
    use blupf90
    implicit none
    real, dimension(347,8)              :: data
    integer, dimension(:,:),allocatable :: data_ped

    real, dimension(1,15)               :: Aref
    real, dimension(1,1)                :: Bref
    real, dimension(1,15)               :: Cref
    real, dimension(1,347)              :: Dref
    real, dimension(347)                :: Daniref
    integer                             :: nlevtotref,nhaploref,naniref
    real                                :: testref,pvalueref,solmuref,stdmuref
    real, dimension(1)                  :: solref,stdref

    write (*,*) " ** Comparatif des solutions generées avec gfortran ** "

    allocate(data_ped(9481,3))
    call set_data
    ! vrai solution dans solution_mat pour comparer
    call set_solution
    call test_subroutine_blupf90
    deallocate (data_ped)

contains

    subroutine test_subroutine_blupf90

        integer                             :: ndata,ncol,ndatap,nguser,i,j,counterror

        real,dimension(:,:),allocatable     :: guser

        type(Param_blupf90)                 :: params_i
        type(Solution_blupf90)              :: sol

        ndata = 347   !nombre d'enregistrement
        ncol = 8      !nombre de colonne enregistrée
        ndatap=9481   ! nombre d'animaux dans la genealogie

        !Initialisation du fichier parametre
        params_i%ntrait=1     ! number of trait
        params_i%neff=5       ! number of effect
        allocate (params_i%pos_y(params_i%ntrait))  ! position of trait
        params_i%pos_y(1) = 8
        params_i%pos_weight = 0
        allocate (params_i%pos_eff(params_i%neff,params_i%ntrait))
        allocate (params_i%nlev(params_i%neff))
        allocate (params_i%effecttype(params_i%neff))
        allocate (params_i%nestedcov(params_i%neff,params_i%ntrait))

        params_i%pos_eff(1,1)=2
        params_i%pos_eff(2,1)=3
        params_i%pos_eff(3,1)=4
        params_i%pos_eff(4,1)=6
        params_i%pos_eff(5,1)=1

        params_i%nlev(1)=1
        params_i%nlev(2)=3
        params_i%nlev(3)=11
        params_i%nlev(4)=1 ! covariate
        params_i%nlev(5)=9481

        params_i%effecttype(1)=effcross
        params_i%effecttype(2)=effcross
        params_i%effecttype(3)=effcross
        params_i%effecttype(4)=effcov ! covariate
        params_i%effecttype(5)=effcross

        params_i%nestedcov = 0

        !random residual
        allocate (params_i%r(params_i%ntrait,params_i%ntrait))
        params_i%r(1,1)=20.831895

        allocate (params_i%randomnumb(params_i%neff))
        allocate (params_i%randomtype(params_i%neff))
        params_i%randomtype=g_fixed
        params_i%randomnumb=0
        !L'effet 5
        params_i%randomnumb(5)=1
        params_i%randomtype(5)=g_A !add_animal

        allocate (params_i%gcov(params_i%neff,maxcorr,maxcorr)) ! mieux vaut garder maxcor comme size max..
        params_i%gcov(5,1,1)=12.234605

        params_i%solv_method="FSPAK"

        allocate (params_i%offset(params_i%neff))
        do i=1,params_i%neff
            params_i%offset(i)=sum(params_i%nlev(1:i-1))*params_i%ntrait
        enddo

        nguser=0
        allocate (guser(nguser,3))
        guser=0.d0

        call exec_blupf90(params_i,ndata,ncol,data,ndatap,data_ped,nguser,guser,sol,.false.)

        counterror=0
        if ( count(Aref == sol%A) /= size(Aref) ) then
            write (*,*) " ======= Error with result sol%A ========= "
            write (*,*) "number of diff ",count(Aref /= sol%A), "/",size(Aref)
            counterror=counterror+1
        else
            write (*,*) "mat A ok!"
        end if

        if ( count(Bref == sol%B) /= size(Bref) ) then
            write (*,*) " ======= Error with result sol%B ========= "
            write (*,*) "number of diff ",count(Bref /= sol%B), "/",size(Bref)
            counterror=counterror+1
        else
            write (*,*) "mat B ok!"
        end if

        if ( count(Cref == sol%C) /= size(Cref) ) then
            write (*,*) " ======= Error with result sol%C ========= "
            write (*,*) "number of diff ",count(Cref /= sol%C), "/",size(Cref)
            counterror=counterror+1
        else
            write (*,*) "mat C ok!"
        end if

        if ( count(Dref == sol%D) /= size(Dref) ) then
            write (*,*) " ======= Error with result sol%D ========= "
            write (*,*) "number of diff ",count(Dref /= sol%D), "/",size(Dref)
            counterror=counterror+1
            do i=1,size(Dref,1)
                do j=1,size(Dref,2)
                    if ( Dref(i,j) /= sol%D(i,j) ) then
                        print *,"indice:",i,j," ref=",Dref(i,j)," val=",sol%D(i,j)
                    end if
                end do
            end do
        else
            write (*,*) "mat D ok!"
        end if

        if ( count(Daniref == sol%Dani) /= size(Daniref) ) then
            write (*,*) " ======= Error with result sol%Dani ========= "
            write (*,*) "number of diff ",count(Daniref /= sol%Dani), "/",size(Daniref)
            counterror=counterror+1
        else
            write (*,*) "mat Dani ok!"
        end if

        if( nlevtotref /= sol%nlevtot) then
            print *,"** DIFF ** nlevtot ref=",nlevtotref," val=",sol%nlevtot
            counterror=counterror+1
        end if

        if( nhaploref /= sol%nhaplo) then
            print *,"** DIFF ** nhaplo ref=",nhaploref," val=",sol%nhaplo
            counterror=counterror+1
        end if

        if( naniref /= sol%nani) then
            print *,"** DIFF ** nani ref=",naniref," val=",sol%nani
            counterror=counterror+1
        end if

        if( testref /= sol%test) then
            print *,"** DIFF ** test ref=",testref," val=",sol%test
            counterror=counterror+1
        end if

        if( pvalueref /= sol%pvalue) then
            print *,"** DIFF ** pvalue ref=",pvalueref," val=",sol%pvalue
            counterror=counterror+1
        end if

        if( solmuref /= sol%solmu) then
            print *,"** DIFF ** solmu ref=",solmuref," val=",sol%solmu
            counterror=counterror+1
        end if

        if( stdmuref /= sol%stdmu) then
            print *,"** DIFF ** stdmu ref=",stdmuref," val=",sol%stdmu
            counterror=counterror+1
        end if

        if( count(solref == sol%sol) /= size(solref)) then
            print *,"** DIFF ** sol ref=",solref," val=",sol%sol
            counterror=counterror+1
        end if

        if( count(stdref == sol%std) /= size(stdref) ) then
            print *,"** DIFF ** std ref=",stdref," val=",sol%std
            counterror=counterror+1
        end if

        write (*,*) "Nombre d'erreur =",counterror

        deallocate (guser)

        call release_paramblupf90(params_i)
        call release_solutionblupf90(sol)

    end subroutine test_subroutine_blupf90



    subroutine set_data
        data(1,1:8)=(/1397.000000,1.000000,1.000000,1.000000,3.000000,2.000000,0.066581,163.000000/)
        data(2,1:8)=(/1401.000000,1.000000,2.000000,9.000000,3.000000,2.000000,0.883816,172.000000/)
        data(3,1:8)=(/1403.000000,1.000000,1.000000,6.000000,3.000000,3.000000,0.374420,188.000000/)
        data(4,1:8)=(/1405.000000,1.000000,1.000000,5.000000,3.000000,2.000000,0.245958,167.000000/)
        data(5,1:8)=(/1406.000000,1.000000,3.000000,1.000000,3.000000,2.000000,0.668208,175.000000/)
        data(6,1:8)=(/1407.000000,1.000000,1.000000,7.000000,3.000000,3.000000,0.042387,172.000000/)
        data(7,1:8)=(/1409.000000,1.000000,3.000000,9.000000,3.000000,3.000000,0.421606,178.000000/)
        data(8,1:8)=(/1410.000000,1.000000,3.000000,11.000000,3.000000,3.000000,-0.351421,178.000000/)
        data(9,1:8)=(/6351.000000,1.000000,3.000000,8.000000,2.000000,3.000000,0.078155,176.000000/)
        data(10,1:8)=(/6378.000000,1.000000,1.000000,2.000000,4.000000,3.000000,1.122801,175.000000/)
        data(11,1:8)=(/6384.000000,1.000000,3.000000,1.000000,5.000000,3.000000,0.746619,172.500000/)
        data(12,1:8)=(/6386.000000,1.000000,2.000000,1.000000,5.000000,3.000000,0.503704,172.000000/)
        data(13,1:8)=(/6406.000000,1.000000,2.000000,7.000000,5.000000,3.000000,0.371814,171.000000/)
        data(14,1:8)=(/6407.000000,1.000000,3.000000,11.000000,5.000000,3.000000,0.584915,172.000000/)
        data(15,1:8)=(/6408.000000,1.000000,1.000000,8.000000,5.000000,2.000000,0.210421,173.000000/)
        data(16,1:8)=(/6409.000000,1.000000,1.000000,9.000000,1.000000,2.000000,0.791107,174.000000/)
        data(17,1:8)=(/6410.000000,1.000000,1.000000,11.000000,1.000000,3.000000,0.054818,169.000000/)
        data(18,1:8)=(/6412.000000,1.000000,3.000000,5.000000,5.000000,3.000000,0.786483,170.000000/)
        data(19,1:8)=(/6413.000000,1.000000,1.000000,5.000000,5.000000,3.000000,-0.149962,181.000000/)
        data(20,1:8)=(/6414.000000,1.000000,3.000000,9.000000,4.000000,2.000000,1.222180,185.000000/)
        data(21,1:8)=(/6415.000000,1.000000,1.000000,11.000000,4.000000,3.000000,0.263898,182.000000/)
        data(22,1:8)=(/6419.000000,1.000000,1.000000,11.000000,4.000000,3.000000,-0.108748,172.000000/)
        data(23,1:8)=(/6421.000000,1.000000,3.000000,3.000000,5.000000,2.000000,0.043748,172.000000/)
        data(24,1:8)=(/6423.000000,1.000000,1.000000,1.000000,5.000000,3.000000,0.453585,170.000000/)
        data(25,1:8)=(/6426.000000,1.000000,2.000000,3.000000,5.000000,3.000000,0.498323,186.000000/)
        data(26,1:8)=(/6429.000000,1.000000,1.000000,2.000000,4.000000,3.000000,0.424398,179.000000/)
        data(27,1:8)=(/6430.000000,1.000000,3.000000,5.000000,4.000000,3.000000,0.715130,179.000000/)
        data(28,1:8)=(/6431.000000,1.000000,1.000000,7.000000,4.000000,3.000000,0.155678,180.000000/)
        data(29,1:8)=(/6434.000000,1.000000,3.000000,5.000000,5.000000,2.000000,-0.105523,181.000000/)
        data(30,1:8)=(/6435.000000,1.000000,1.000000,11.000000,1.000000,3.000000,-0.251090,175.000000/)
        data(31,1:8)=(/6437.000000,1.000000,1.000000,3.000000,4.000000,3.000000,0.355884,184.000000/)
        data(32,1:8)=(/6439.000000,1.000000,3.000000,9.000000,5.000000,3.000000,0.147528,176.000000/)
        data(33,1:8)=(/6440.000000,1.000000,3.000000,5.000000,5.000000,2.000000,0.199173,174.000000/)
        data(34,1:8)=(/6441.000000,1.000000,3.000000,1.000000,5.000000,3.000000,0.822875,180.000000/)
        data(35,1:8)=(/6442.000000,1.000000,3.000000,3.000000,4.000000,1.000000,0.439354,175.000000/)
        data(36,1:8)=(/6443.000000,1.000000,3.000000,5.000000,4.000000,2.000000,1.040546,174.000000/)
        data(37,1:8)=(/6445.000000,1.000000,1.000000,9.000000,2.000000,3.000000,0.300882,182.000000/)
        data(38,1:8)=(/6446.000000,1.000000,3.000000,11.000000,5.000000,2.000000,0.658089,169.000000/)
        data(39,1:8)=(/6447.000000,1.000000,1.000000,5.000000,1.000000,2.000000,0.945390,180.000000/)
        data(40,1:8)=(/6448.000000,1.000000,1.000000,7.000000,3.000000,1.000000,0.333409,177.000000/)
        data(41,1:8)=(/6449.000000,1.000000,3.000000,1.000000,4.000000,2.000000,0.323285,163.000000/)
        data(42,1:8)=(/6450.000000,1.000000,3.000000,9.000000,4.000000,3.000000,-0.089720,178.000000/)
        data(43,1:8)=(/6452.000000,1.000000,2.000000,5.000000,5.000000,3.000000,0.520142,167.000000/)
        data(44,1:8)=(/7559.000000,1.000000,1.000000,6.000000,4.000000,2.000000,0.374420,173.000000/)
        data(45,1:8)=(/7596.000000,1.000000,2.000000,11.000000,5.000000,2.000000,0.516938,159.000000/)
        data(46,1:8)=(/7617.000000,1.000000,1.000000,3.000000,1.000000,2.000000,0.286165,184.000000/)
        data(47,1:8)=(/7628.000000,1.000000,3.000000,1.000000,5.000000,3.000000,0.361843,166.000000/)
        data(48,1:8)=(/7635.000000,1.000000,2.000000,11.000000,5.000000,3.000000,-0.004454,170.000000/)
        data(49,1:8)=(/7637.000000,1.000000,1.000000,4.000000,5.000000,3.000000,0.297720,177.000000/)
        data(50,1:8)=(/7640.000000,1.000000,3.000000,11.000000,5.000000,3.000000,-0.190451,168.000000/)
        data(51,1:8)=(/7655.000000,1.000000,1.000000,9.000000,5.000000,2.000000,-0.108381,167.000000/)
        data(52,1:8)=(/7656.000000,1.000000,3.000000,7.000000,1.000000,2.000000,0.030263,176.000000/)
        data(53,1:8)=(/7657.000000,1.000000,1.000000,11.000000,5.000000,1.000000,0.428574,176.000000/)
        data(54,1:8)=(/7658.000000,1.000000,3.000000,7.000000,5.000000,2.000000,-0.215322,179.000000/)
        data(55,1:8)=(/7662.000000,1.000000,2.000000,7.000000,5.000000,2.000000,-0.061193,177.000000/)
        data(56,1:8)=(/7663.000000,1.000000,1.000000,4.000000,5.000000,1.000000,0.612756,171.000000/)
        data(57,1:8)=(/7665.000000,1.000000,3.000000,1.000000,5.000000,3.000000,0.688764,167.500000/)
        data(58,1:8)=(/7666.000000,1.000000,3.000000,1.000000,5.000000,2.000000,0.475788,179.000000/)
        data(59,1:8)=(/7668.000000,1.000000,3.000000,5.000000,5.000000,3.000000,1.030504,173.000000/)
        data(60,1:8)=(/7672.000000,1.000000,1.000000,7.000000,4.000000,3.000000,0.420167,166.000000/)
        data(61,1:8)=(/7675.000000,1.000000,1.000000,5.000000,5.000000,3.000000,0.160106,179.000000/)
        data(62,1:8)=(/7677.000000,1.000000,1.000000,3.000000,5.000000,2.000000,0.190949,174.000000/)
        data(63,1:8)=(/7679.000000,1.000000,3.000000,1.000000,5.000000,2.000000,0.462931,174.000000/)
        data(64,1:8)=(/7684.000000,1.000000,1.000000,3.000000,5.000000,3.000000,0.437364,177.000000/)
        data(65,1:8)=(/7687.000000,1.000000,1.000000,5.000000,4.000000,3.000000,0.621617,176.000000/)
        data(66,1:8)=(/7688.000000,1.000000,3.000000,7.000000,5.000000,2.000000,0.148055,180.000000/)
        data(67,1:8)=(/7690.000000,1.000000,1.000000,2.000000,5.000000,3.000000,0.710350,176.000000/)
        data(68,1:8)=(/7691.000000,1.000000,1.000000,3.000000,5.000000,3.000000,0.176784,178.000000/)
        data(69,1:8)=(/7693.000000,1.000000,1.000000,11.000000,4.000000,2.000000,-0.485460,179.000000/)
        data(70,1:8)=(/7698.000000,1.000000,3.000000,5.000000,5.000000,3.000000,0.361967,168.000000/)
        data(71,1:8)=(/7702.000000,1.000000,3.000000,7.000000,2.000000,3.000000,0.434779,183.000000/)
        data(72,1:8)=(/7704.000000,1.000000,3.000000,9.000000,5.000000,3.000000,0.569849,173.000000/)
        data(73,1:8)=(/7707.000000,1.000000,3.000000,8.000000,5.000000,3.000000,0.009522,181.000000/)
        data(74,1:8)=(/7711.000000,1.000000,3.000000,7.000000,1.000000,1.000000,0.543021,176.000000/)
        data(75,1:8)=(/7714.000000,1.000000,3.000000,11.000000,5.000000,3.000000,-0.132645,174.000000/)
        data(76,1:8)=(/7716.000000,1.000000,3.000000,11.000000,5.000000,3.000000,0.227713,167.000000/)
        data(77,1:8)=(/7717.000000,1.000000,2.000000,7.000000,5.000000,2.000000,0.292697,193.000000/)
        data(78,1:8)=(/7719.000000,1.000000,2.000000,7.000000,5.000000,3.000000,-0.344658,172.000000/)
        data(79,1:8)=(/7721.000000,1.000000,3.000000,11.000000,5.000000,3.000000,0.084933,165.000000/)
        data(80,1:8)=(/7724.000000,1.000000,1.000000,5.000000,5.000000,3.000000,0.737968,176.000000/)
        data(81,1:8)=(/7725.000000,1.000000,1.000000,11.000000,5.000000,3.000000,0.544817,174.000000/)
        data(82,1:8)=(/7726.000000,1.000000,3.000000,4.000000,5.000000,3.000000,0.230388,172.000000/)
        data(83,1:8)=(/7727.000000,1.000000,2.000000,11.000000,5.000000,3.000000,0.127874,167.000000/)
        data(84,1:8)=(/7728.000000,1.000000,1.000000,1.000000,5.000000,2.000000,0.433275,176.000000/)
        data(85,1:8)=(/7730.000000,1.000000,1.000000,4.000000,5.000000,3.000000,0.274077,165.000000/)
        data(86,1:8)=(/7732.000000,1.000000,3.000000,1.000000,5.000000,2.000000,-0.097903,176.000000/)
        data(87,1:8)=(/7734.000000,1.000000,1.000000,9.000000,5.000000,2.000000,0.153206,177.000000/)
        data(88,1:8)=(/7735.000000,1.000000,1.000000,3.000000,5.000000,3.000000,0.292874,187.000000/)
        data(89,1:8)=(/7736.000000,1.000000,2.000000,5.000000,5.000000,2.000000,0.359294,162.000000/)
        data(90,1:8)=(/7737.000000,1.000000,1.000000,9.000000,5.000000,3.000000,0.718388,174.000000/)
        data(91,1:8)=(/7930.000000,1.000000,3.000000,5.000000,1.000000,2.000000,0.332268,182.000000/)
        data(92,1:8)=(/7938.000000,1.000000,1.000000,9.000000,4.000000,3.000000,0.237441,179.000000/)
        data(93,1:8)=(/7939.000000,1.000000,3.000000,7.000000,5.000000,3.000000,0.071799,170.000000/)
        data(94,1:8)=(/7943.000000,1.000000,1.000000,7.000000,5.000000,3.000000,0.200448,185.000000/)
        data(95,1:8)=(/7947.000000,1.000000,1.000000,3.000000,5.000000,2.000000,0.307893,177.000000/)
        data(96,1:8)=(/7948.000000,1.000000,1.000000,11.000000,5.000000,2.000000,0.662992,179.000000/)
        data(97,1:8)=(/7949.000000,1.000000,3.000000,3.000000,5.000000,3.000000,0.247212,172.000000/)
        data(98,1:8)=(/7950.000000,1.000000,3.000000,11.000000,5.000000,2.000000,0.290343,168.000000/)
        data(99,1:8)=(/7991.000000,1.000000,1.000000,1.000000,5.000000,2.000000,0.259796,179.000000/)
        data(100,1:8)=(/7992.000000,1.000000,2.000000,1.000000,5.000000,2.000000,0.458524,180.000000/)
        data(101,1:8)=(/8200.000000,1.000000,2.000000,2.000000,5.000000,3.000000,0.408410,178.000000/)
        data(102,1:8)=(/8205.000000,1.000000,2.000000,10.000000,5.000000,3.000000,0.288295,178.000000/)
        data(103,1:8)=(/8211.000000,1.000000,2.000000,10.000000,5.000000,3.000000,0.288295,176.000000/)
        data(104,1:8)=(/8244.000000,1.000000,2.000000,10.000000,5.000000,3.000000,0.288295,179.000000/)
        data(105,1:8)=(/8275.000000,1.000000,3.000000,3.000000,3.000000,2.000000,0.577312,165.000000/)
        data(106,1:8)=(/8286.000000,1.000000,2.000000,10.000000,5.000000,3.000000,-0.368290,171.000000/)
        data(107,1:8)=(/8295.000000,1.000000,3.000000,7.000000,5.000000,1.000000,0.681254,184.000000/)
        data(108,1:8)=(/8307.000000,1.000000,2.000000,7.000000,5.000000,2.000000,-0.050713,176.000000/)
        data(109,1:8)=(/8311.000000,1.000000,1.000000,9.000000,5.000000,2.000000,0.154084,174.000000/)
        data(110,1:8)=(/8312.000000,1.000000,3.000000,8.000000,1.000000,3.000000,0.414717,180.000000/)
        data(111,1:8)=(/8313.000000,1.000000,1.000000,2.000000,5.000000,3.000000,0.479985,173.000000/)
        data(112,1:8)=(/8314.000000,1.000000,2.000000,10.000000,5.000000,3.000000,0.579170,169.000000/)
        data(113,1:8)=(/8324.000000,1.000000,3.000000,5.000000,2.000000,3.000000,0.679337,183.000000/)
        data(114,1:8)=(/8328.000000,1.000000,3.000000,11.000000,5.000000,3.000000,0.798815,168.000000/)
        data(115,1:8)=(/8330.000000,1.000000,1.000000,3.000000,5.000000,2.000000,0.145467,183.000000/)
        data(116,1:8)=(/8331.000000,1.000000,3.000000,1.000000,5.000000,2.000000,0.775786,177.000000/)
        data(117,1:8)=(/8334.000000,1.000000,3.000000,9.000000,5.000000,3.000000,0.034484,181.000000/)
        data(118,1:8)=(/8338.000000,1.000000,1.000000,11.000000,5.000000,1.000000,0.110527,166.000000/)
        data(119,1:8)=(/8342.000000,1.000000,1.000000,8.000000,5.000000,3.000000,0.574715,171.000000/)
        data(120,1:8)=(/8343.000000,1.000000,3.000000,11.000000,1.000000,2.000000,-0.115993,171.000000/)
        data(121,1:8)=(/8347.000000,1.000000,1.000000,6.000000,4.000000,3.000000,0.451317,186.000000/)
        data(122,1:8)=(/8351.000000,1.000000,3.000000,7.000000,5.000000,2.000000,0.378164,183.000000/)
        data(123,1:8)=(/8355.000000,1.000000,1.000000,3.000000,5.000000,3.000000,0.460472,179.000000/)
        data(124,1:8)=(/8360.000000,1.000000,1.000000,2.000000,1.000000,3.000000,0.828165,180.000000/)
        data(125,1:8)=(/8361.000000,1.000000,1.000000,7.000000,5.000000,3.000000,0.253967,177.000000/)
        data(126,1:8)=(/8362.000000,1.000000,3.000000,2.000000,4.000000,2.000000,-0.043965,168.000000/)
        data(127,1:8)=(/8364.000000,1.000000,1.000000,9.000000,5.000000,3.000000,0.481405,178.000000/)
        data(128,1:8)=(/8365.000000,1.000000,3.000000,4.000000,5.000000,3.000000,0.197914,188.000000/)
        data(129,1:8)=(/8366.000000,1.000000,1.000000,1.000000,4.000000,3.000000,0.411115,166.000000/)
        data(130,1:8)=(/8367.000000,1.000000,1.000000,3.000000,5.000000,2.000000,0.306162,173.000000/)
        data(131,1:8)=(/8368.000000,1.000000,1.000000,5.000000,5.000000,2.000000,0.107469,155.000000/)
        data(132,1:8)=(/8369.000000,1.000000,1.000000,3.000000,5.000000,3.000000,0.023980,172.000000/)
        data(133,1:8)=(/8374.000000,1.000000,1.000000,2.000000,5.000000,3.000000,0.489955,167.000000/)
        data(134,1:8)=(/8375.000000,1.000000,1.000000,2.000000,4.000000,2.000000,0.637736,186.000000/)
        data(135,1:8)=(/8376.000000,1.000000,3.000000,2.000000,4.000000,3.000000,-0.110623,178.000000/)
        data(136,1:8)=(/8377.000000,1.000000,3.000000,4.000000,5.000000,3.000000,0.190622,180.000000/)
        data(137,1:8)=(/8379.000000,1.000000,1.000000,7.000000,5.000000,2.000000,0.566728,172.000000/)
        data(138,1:8)=(/8380.000000,1.000000,1.000000,7.000000,4.000000,3.000000,0.604016,183.000000/)
        data(139,1:8)=(/8383.000000,1.000000,3.000000,5.000000,5.000000,2.000000,-0.054655,167.000000/)
        data(140,1:8)=(/8384.000000,1.000000,3.000000,1.000000,5.000000,2.000000,-0.211028,182.000000/)
        data(141,1:8)=(/8385.000000,1.000000,3.000000,9.000000,4.000000,3.000000,0.955376,188.000000/)
        data(142,1:8)=(/8386.000000,1.000000,2.000000,11.000000,5.000000,2.000000,0.165862,167.000000/)
        data(143,1:8)=(/8387.000000,1.000000,3.000000,11.000000,5.000000,3.000000,-0.286207,182.000000/)
        data(144,1:8)=(/8388.000000,1.000000,1.000000,2.000000,5.000000,2.000000,0.795110,187.000000/)
        data(145,1:8)=(/8389.000000,1.000000,3.000000,1.000000,5.000000,3.000000,1.153343,172.000000/)
        data(146,1:8)=(/8392.000000,1.000000,3.000000,2.000000,5.000000,2.000000,0.006420,172.000000/)
        data(147,1:8)=(/8393.000000,1.000000,3.000000,11.000000,1.000000,3.000000,0.432858,166.000000/)
        data(148,1:8)=(/8394.000000,1.000000,3.000000,2.000000,4.000000,3.000000,0.484129,188.000000/)
        data(149,1:8)=(/8395.000000,1.000000,1.000000,2.000000,5.000000,3.000000,0.408410,197.000000/)
        data(150,1:8)=(/8396.000000,1.000000,1.000000,3.000000,4.000000,2.000000,0.676786,175.000000/)
        data(151,1:8)=(/8398.000000,1.000000,3.000000,1.000000,5.000000,3.000000,0.297593,172.000000/)
        data(152,1:8)=(/8399.000000,1.000000,1.000000,4.000000,5.000000,3.000000,1.003342,173.000000/)
        data(153,1:8)=(/8403.000000,1.000000,3.000000,9.000000,5.000000,2.000000,0.387729,172.000000/)
        data(154,1:8)=(/8404.000000,1.000000,1.000000,7.000000,4.000000,3.000000,-0.027977,183.000000/)
        data(155,1:8)=(/8405.000000,1.000000,1.000000,3.000000,5.000000,2.000000,0.243787,174.000000/)
        data(156,1:8)=(/8406.000000,1.000000,1.000000,3.000000,5.000000,3.000000,0.530046,165.000000/)
        data(157,1:8)=(/8407.000000,1.000000,3.000000,11.000000,4.000000,3.000000,0.097275,175.000000/)
        data(158,1:8)=(/8408.000000,1.000000,1.000000,3.000000,4.000000,3.000000,0.413688,183.000000/)
        data(159,1:8)=(/8409.000000,1.000000,2.000000,7.000000,5.000000,3.000000,-0.014122,172.000000/)
        data(160,1:8)=(/8413.000000,1.000000,3.000000,1.000000,5.000000,3.000000,0.276156,174.000000/)
        data(161,1:8)=(/8415.000000,1.000000,3.000000,4.000000,4.000000,2.000000,0.730742,179.000000/)
        data(162,1:8)=(/8416.000000,1.000000,3.000000,2.000000,1.000000,3.000000,0.408410,176.000000/)
        data(163,1:8)=(/8417.000000,1.000000,3.000000,4.000000,5.000000,2.000000,0.689483,172.000000/)
        data(164,1:8)=(/8418.000000,1.000000,3.000000,4.000000,1.000000,3.000000,0.662269,182.000000/)
        data(165,1:8)=(/8419.000000,1.000000,1.000000,3.000000,5.000000,3.000000,0.482644,172.000000/)
        data(166,1:8)=(/8421.000000,1.000000,1.000000,9.000000,5.000000,3.000000,0.629916,173.000000/)
        data(167,1:8)=(/8422.000000,1.000000,3.000000,2.000000,5.000000,3.000000,0.034381,178.000000/)
        data(168,1:8)=(/8423.000000,1.000000,1.000000,3.000000,4.000000,2.000000,0.001313,174.000000/)
        data(169,1:8)=(/8424.000000,1.000000,3.000000,4.000000,5.000000,2.000000,0.281394,176.000000/)
        data(170,1:8)=(/8426.000000,1.000000,1.000000,3.000000,4.000000,2.000000,0.315656,185.000000/)
        data(171,1:8)=(/8427.000000,1.000000,1.000000,7.000000,5.000000,2.000000,0.644154,173.000000/)
        data(172,1:8)=(/8428.000000,1.000000,1.000000,1.000000,5.000000,3.000000,0.055931,174.000000/)
        data(173,1:8)=(/8429.000000,1.000000,1.000000,11.000000,5.000000,3.000000,-0.087838,169.000000/)
        data(174,1:8)=(/8430.000000,1.000000,3.000000,4.000000,5.000000,3.000000,0.907756,168.000000/)
        data(175,1:8)=(/8431.000000,1.000000,1.000000,3.000000,5.000000,3.000000,0.373901,176.000000/)
        data(176,1:8)=(/8434.000000,1.000000,3.000000,3.000000,5.000000,3.000000,0.008306,177.000000/)
        data(177,1:8)=(/8435.000000,1.000000,2.000000,1.000000,5.000000,2.000000,0.130575,171.000000/)
        data(178,1:8)=(/8515.000000,1.000000,3.000000,7.000000,1.000000,3.000000,0.851920,188.000000/)
        data(179,1:8)=(/8516.000000,1.000000,3.000000,1.000000,1.000000,3.000000,0.440008,183.000000/)
        data(180,1:8)=(/8946.000000,1.000000,2.000000,10.000000,5.000000,1.000000,0.288295,177.000000/)
        data(181,1:8)=(/9025.000000,1.000000,1.000000,11.000000,3.000000,3.000000,-0.251215,156.000000/)
        data(182,1:8)=(/9045.000000,1.000000,1.000000,1.000000,4.000000,3.000000,0.029404,169.000000/)
        data(183,1:8)=(/9056.000000,1.000000,1.000000,5.000000,1.000000,2.000000,0.009811,166.000000/)
        data(184,1:8)=(/9075.000000,1.000000,1.000000,6.000000,5.000000,3.000000,0.428159,177.000000/)
        data(185,1:8)=(/9080.000000,1.000000,3.000000,4.000000,5.000000,3.000000,-0.286640,189.000000/)
        data(186,1:8)=(/9081.000000,1.000000,3.000000,9.000000,5.000000,3.000000,0.970786,175.000000/)
        data(187,1:8)=(/9086.000000,1.000000,3.000000,1.000000,5.000000,2.000000,0.492987,176.000000/)
        data(188,1:8)=(/9090.000000,1.000000,1.000000,11.000000,5.000000,3.000000,0.430427,175.000000/)
        data(189,1:8)=(/9101.000000,1.000000,3.000000,4.000000,5.000000,2.000000,0.457534,170.000000/)
        data(190,1:8)=(/9103.000000,1.000000,3.000000,1.000000,1.000000,3.000000,0.966394,190.000000/)
        data(191,1:8)=(/9105.000000,1.000000,3.000000,7.000000,1.000000,3.000000,0.377840,177.000000/)
        data(192,1:8)=(/9115.000000,1.000000,1.000000,1.000000,4.000000,3.000000,0.940186,175.000000/)
        data(193,1:8)=(/9119.000000,1.000000,1.000000,9.000000,4.000000,3.000000,0.356751,176.000000/)
        data(194,1:8)=(/9120.000000,1.000000,3.000000,9.000000,5.000000,2.000000,1.024141,176.000000/)
        data(195,1:8)=(/9122.000000,1.000000,3.000000,3.000000,5.000000,2.000000,0.966628,177.000000/)
        data(196,1:8)=(/9123.000000,1.000000,3.000000,8.000000,4.000000,3.000000,0.410663,189.000000/)
        data(197,1:8)=(/9128.000000,1.000000,1.000000,1.000000,4.000000,2.000000,0.183755,170.000000/)
        data(198,1:8)=(/9130.000000,1.000000,1.000000,5.000000,5.000000,3.000000,1.025567,174.000000/)
        data(199,1:8)=(/9131.000000,1.000000,2.000000,11.000000,5.000000,3.000000,-0.015856,165.000000/)
        data(200,1:8)=(/9134.000000,1.000000,3.000000,5.000000,5.000000,3.000000,0.709148,172.000000/)
        data(201,1:8)=(/9141.000000,1.000000,1.000000,11.000000,4.000000,3.000000,0.382457,173.000000/)
        data(202,1:8)=(/9143.000000,1.000000,1.000000,5.000000,5.000000,3.000000,0.674281,174.000000/)
        data(203,1:8)=(/9145.000000,1.000000,1.000000,7.000000,3.000000,3.000000,0.029209,180.000000/)
        data(204,1:8)=(/9151.000000,1.000000,3.000000,3.000000,5.000000,3.000000,0.090852,188.000000/)
        data(205,1:8)=(/9152.000000,1.000000,3.000000,1.000000,5.000000,3.000000,0.949088,169.000000/)
        data(206,1:8)=(/9153.000000,1.000000,3.000000,11.000000,5.000000,3.000000,0.589452,165.000000/)
        data(207,1:8)=(/9155.000000,1.000000,1.000000,8.000000,4.000000,3.000000,1.036786,173.000000/)
        data(208,1:8)=(/9157.000000,1.000000,1.000000,3.000000,5.000000,3.000000,0.901318,172.000000/)
        data(209,1:8)=(/9158.000000,1.000000,1.000000,2.000000,5.000000,3.000000,0.094023,182.000000/)
        data(210,1:8)=(/9159.000000,1.000000,2.000000,7.000000,5.000000,3.000000,0.274215,172.000000/)
        data(211,1:8)=(/9160.000000,1.000000,1.000000,2.000000,5.000000,3.000000,0.725313,180.000000/)
        data(212,1:8)=(/9161.000000,1.000000,1.000000,2.000000,2.000000,2.000000,0.637081,186.000000/)
        data(213,1:8)=(/9162.000000,1.000000,1.000000,2.000000,4.000000,3.000000,0.397003,177.000000/)
        data(214,1:8)=(/9163.000000,1.000000,3.000000,5.000000,4.000000,3.000000,0.469609,176.000000/)
        data(215,1:8)=(/9164.000000,1.000000,3.000000,7.000000,5.000000,3.000000,0.090280,171.000000/)
        data(216,1:8)=(/9165.000000,1.000000,3.000000,9.000000,5.000000,2.000000,0.405092,177.000000/)
        data(217,1:8)=(/9167.000000,1.000000,1.000000,5.000000,4.000000,3.000000,0.628648,163.000000/)
        data(218,1:8)=(/9172.000000,1.000000,3.000000,7.000000,5.000000,3.000000,0.668811,172.000000/)
        data(219,1:8)=(/9177.000000,1.000000,1.000000,9.000000,5.000000,3.000000,0.700409,176.000000/)
        data(220,1:8)=(/9178.000000,1.000000,1.000000,11.000000,5.000000,2.000000,0.218347,172.000000/)
        data(221,1:8)=(/9183.000000,1.000000,1.000000,1.000000,5.000000,3.000000,0.309996,188.000000/)
        data(222,1:8)=(/9189.000000,1.000000,1.000000,3.000000,5.000000,2.000000,0.633070,173.000000/)
        data(223,1:8)=(/9195.000000,1.000000,1.000000,5.000000,5.000000,2.000000,-0.210929,174.000000/)
        data(224,1:8)=(/9196.000000,1.000000,3.000000,5.000000,4.000000,3.000000,0.657753,185.500000/)
        data(225,1:8)=(/9197.000000,1.000000,3.000000,7.000000,5.000000,2.000000,-0.070454,185.000000/)
        data(226,1:8)=(/9201.000000,1.000000,3.000000,5.000000,3.000000,3.000000,0.255248,171.000000/)
        data(227,1:8)=(/9202.000000,1.000000,3.000000,2.000000,5.000000,3.000000,-0.137084,178.000000/)
        data(228,1:8)=(/9206.000000,1.000000,1.000000,9.000000,4.000000,3.000000,0.650747,183.000000/)
        data(229,1:8)=(/9213.000000,1.000000,1.000000,3.000000,5.000000,3.000000,-0.091724,180.000000/)
        data(230,1:8)=(/9217.000000,1.000000,3.000000,5.000000,4.000000,1.000000,0.498665,176.000000/)
        data(231,1:8)=(/9218.000000,1.000000,2.000000,11.000000,5.000000,1.000000,0.737457,166.000000/)
        data(232,1:8)=(/9225.000000,1.000000,2.000000,1.000000,5.000000,3.000000,0.252357,184.000000/)
        data(233,1:8)=(/9228.000000,1.000000,3.000000,1.000000,4.000000,3.000000,0.540521,184.000000/)
        data(234,1:8)=(/9229.000000,1.000000,3.000000,9.000000,4.000000,2.000000,-0.018533,175.000000/)
        data(235,1:8)=(/9230.000000,1.000000,1.000000,7.000000,4.000000,2.000000,0.187097,174.000000/)
        data(236,1:8)=(/9234.000000,1.000000,1.000000,7.000000,5.000000,2.000000,0.075037,177.000000/)
        data(237,1:8)=(/9235.000000,1.000000,3.000000,11.000000,5.000000,3.000000,0.553737,172.000000/)
        data(238,1:8)=(/9236.000000,1.000000,3.000000,11.000000,5.000000,3.000000,0.568292,179.000000/)
        data(239,1:8)=(/9237.000000,1.000000,3.000000,9.000000,5.000000,3.000000,0.145775,177.000000/)
        data(240,1:8)=(/9240.000000,1.000000,1.000000,1.000000,1.000000,3.000000,0.754491,173.000000/)
        data(241,1:8)=(/9244.000000,1.000000,3.000000,11.000000,5.000000,3.000000,0.532110,180.000000/)
        data(242,1:8)=(/9245.000000,1.000000,3.000000,5.000000,5.000000,3.000000,0.391727,176.000000/)
        data(243,1:8)=(/9246.000000,1.000000,1.000000,5.000000,5.000000,3.000000,1.086497,176.000000/)
        data(244,1:8)=(/9249.000000,1.000000,1.000000,7.000000,5.000000,2.000000,0.315401,172.000000/)
        data(245,1:8)=(/9250.000000,1.000000,1.000000,5.000000,5.000000,2.000000,0.440863,168.000000/)
        data(246,1:8)=(/9252.000000,1.000000,3.000000,7.000000,5.000000,3.000000,0.038321,177.000000/)
        data(247,1:8)=(/9253.000000,1.000000,2.000000,1.000000,4.000000,3.000000,0.508006,182.000000/)
        data(248,1:8)=(/9254.000000,1.000000,1.000000,9.000000,5.000000,3.000000,0.361435,176.000000/)
        data(249,1:8)=(/9255.000000,1.000000,1.000000,11.000000,5.000000,3.000000,0.484922,171.000000/)
        data(250,1:8)=(/9256.000000,1.000000,1.000000,5.000000,1.000000,3.000000,0.777836,171.000000/)
        data(251,1:8)=(/9258.000000,1.000000,1.000000,6.000000,1.000000,2.000000,-0.070813,177.000000/)
        data(252,1:8)=(/9259.000000,1.000000,3.000000,11.000000,5.000000,2.000000,0.212801,179.000000/)
        data(253,1:8)=(/9261.000000,1.000000,1.000000,1.000000,1.000000,2.000000,0.204954,181.000000/)
        data(254,1:8)=(/9264.000000,1.000000,3.000000,9.000000,5.000000,2.000000,0.009046,183.000000/)
        data(255,1:8)=(/9265.000000,1.000000,1.000000,3.000000,5.000000,3.000000,-0.003875,174.000000/)
        data(256,1:8)=(/9266.000000,1.000000,2.000000,9.000000,4.000000,3.000000,0.033372,181.000000/)
        data(257,1:8)=(/9267.000000,1.000000,3.000000,1.000000,5.000000,3.000000,-0.094506,174.000000/)
        data(258,1:8)=(/9269.000000,1.000000,1.000000,11.000000,5.000000,3.000000,0.528851,161.000000/)
        data(259,1:8)=(/9270.000000,1.000000,1.000000,7.000000,5.000000,3.000000,0.336001,179.000000/)
        data(260,1:8)=(/9272.000000,1.000000,1.000000,1.000000,5.000000,3.000000,-0.008680,177.000000/)
        data(261,1:8)=(/9274.000000,1.000000,3.000000,11.000000,4.000000,2.000000,0.749607,171.000000/)
        data(262,1:8)=(/9276.000000,1.000000,1.000000,7.000000,5.000000,3.000000,0.358254,176.000000/)
        data(263,1:8)=(/9277.000000,1.000000,3.000000,3.000000,4.000000,2.000000,0.828291,176.000000/)
        data(264,1:8)=(/9282.000000,1.000000,2.000000,10.000000,5.000000,3.000000,0.654006,182.000000/)
        data(265,1:8)=(/9283.000000,1.000000,1.000000,1.000000,1.000000,3.000000,0.571723,170.000000/)
        data(266,1:8)=(/9284.000000,1.000000,1.000000,11.000000,5.000000,1.000000,0.302314,180.000000/)
        data(267,1:8)=(/9287.000000,1.000000,3.000000,5.000000,5.000000,3.000000,0.493760,169.000000/)
        data(268,1:8)=(/9288.000000,1.000000,1.000000,5.000000,5.000000,3.000000,0.583475,180.000000/)
        data(269,1:8)=(/9290.000000,1.000000,1.000000,9.000000,5.000000,2.000000,1.091067,178.000000/)
        data(270,1:8)=(/9291.000000,1.000000,3.000000,1.000000,5.000000,2.000000,0.756132,174.000000/)
        data(271,1:8)=(/9292.000000,1.000000,1.000000,4.000000,5.000000,3.000000,-0.033310,178.000000/)
        data(272,1:8)=(/9294.000000,1.000000,3.000000,11.000000,5.000000,3.000000,0.098966,172.000000/)
        data(273,1:8)=(/9297.000000,1.000000,3.000000,7.000000,4.000000,3.000000,0.963653,178.000000/)
        data(274,1:8)=(/9300.000000,1.000000,1.000000,5.000000,5.000000,3.000000,0.637981,176.000000/)
        data(275,1:8)=(/9301.000000,1.000000,1.000000,1.000000,1.000000,3.000000,0.817641,182.000000/)
        data(276,1:8)=(/9302.000000,1.000000,1.000000,3.000000,5.000000,2.000000,0.426321,170.000000/)
        data(277,1:8)=(/9303.000000,1.000000,1.000000,7.000000,5.000000,3.000000,0.059103,171.000000/)
        data(278,1:8)=(/9304.000000,1.000000,1.000000,3.000000,5.000000,2.000000,0.480723,185.000000/)
        data(279,1:8)=(/9305.000000,1.000000,1.000000,7.000000,5.000000,3.000000,0.086665,181.000000/)
        data(280,1:8)=(/9306.000000,1.000000,3.000000,11.000000,5.000000,3.000000,0.391897,174.000000/)
        data(281,1:8)=(/9307.000000,1.000000,1.000000,5.000000,5.000000,3.000000,0.663855,175.000000/)
        data(282,1:8)=(/9309.000000,1.000000,2.000000,3.000000,1.000000,3.000000,0.710177,178.000000/)
        data(283,1:8)=(/9311.000000,1.000000,1.000000,5.000000,1.000000,2.000000,0.643648,174.000000/)
        data(284,1:8)=(/9314.000000,1.000000,1.000000,11.000000,5.000000,3.000000,0.061087,172.500000/)
        data(285,1:8)=(/9315.000000,1.000000,3.000000,3.000000,5.000000,2.000000,0.380479,170.000000/)
        data(286,1:8)=(/9316.000000,1.000000,1.000000,2.000000,5.000000,3.000000,0.547978,176.000000/)
        data(287,1:8)=(/9319.000000,1.000000,3.000000,11.000000,4.000000,3.000000,0.320638,180.000000/)
        data(288,1:8)=(/9320.000000,1.000000,3.000000,7.000000,5.000000,2.000000,0.515052,179.000000/)
        data(289,1:8)=(/9321.000000,1.000000,1.000000,1.000000,5.000000,3.000000,1.160615,176.000000/)
        data(290,1:8)=(/9322.000000,1.000000,1.000000,1.000000,5.000000,3.000000,0.351034,167.000000/)
        data(291,1:8)=(/9323.000000,1.000000,2.000000,11.000000,5.000000,3.000000,0.226164,165.000000/)
        data(292,1:8)=(/9324.000000,1.000000,1.000000,5.000000,5.000000,3.000000,1.413630,171.000000/)
        data(293,1:8)=(/9326.000000,1.000000,3.000000,7.000000,5.000000,3.000000,0.418262,182.000000/)
        data(294,1:8)=(/9327.000000,1.000000,3.000000,1.000000,5.000000,3.000000,0.473848,176.000000/)
        data(295,1:8)=(/9328.000000,1.000000,1.000000,9.000000,1.000000,3.000000,0.127006,176.000000/)
        data(296,1:8)=(/9329.000000,1.000000,3.000000,1.000000,5.000000,2.000000,0.596957,169.000000/)
        data(297,1:8)=(/9330.000000,1.000000,1.000000,11.000000,5.000000,3.000000,0.183162,168.000000/)
        data(298,1:8)=(/9333.000000,1.000000,3.000000,1.000000,5.000000,2.000000,0.370570,190.000000/)
        data(299,1:8)=(/9334.000000,1.000000,3.000000,3.000000,4.000000,3.000000,0.275476,181.000000/)
        data(300,1:8)=(/9335.000000,1.000000,3.000000,3.000000,4.000000,3.000000,0.093913,179.000000/)
        data(301,1:8)=(/9338.000000,1.000000,1.000000,1.000000,5.000000,2.000000,-0.231785,179.000000/)
        data(302,1:8)=(/9339.000000,1.000000,1.000000,3.000000,5.000000,2.000000,0.728998,178.000000/)
        data(303,1:8)=(/9340.000000,1.000000,3.000000,3.000000,5.000000,3.000000,0.420182,171.000000/)
        data(304,1:8)=(/9341.000000,1.000000,3.000000,4.000000,5.000000,3.000000,0.995219,173.000000/)
        data(305,1:8)=(/9342.000000,1.000000,1.000000,5.000000,5.000000,3.000000,0.722572,175.000000/)
        data(306,1:8)=(/9343.000000,1.000000,1.000000,3.000000,4.000000,3.000000,-0.114983,179.000000/)
        data(307,1:8)=(/9344.000000,1.000000,1.000000,6.000000,4.000000,3.000000,0.678449,172.000000/)
        data(308,1:8)=(/9345.000000,1.000000,1.000000,9.000000,4.000000,2.000000,0.415085,177.000000/)
        data(309,1:8)=(/9346.000000,1.000000,3.000000,1.000000,5.000000,2.000000,0.463610,170.000000/)
        data(310,1:8)=(/9347.000000,1.000000,1.000000,4.000000,5.000000,2.000000,0.697227,172.000000/)
        data(311,1:8)=(/9348.000000,1.000000,3.000000,3.000000,5.000000,2.000000,0.550332,178.000000/)
        data(312,1:8)=(/9350.000000,1.000000,3.000000,1.000000,5.000000,3.000000,1.173297,179.500000/)
        data(313,1:8)=(/9351.000000,1.000000,3.000000,9.000000,4.000000,2.000000,0.007082,183.000000/)
        data(314,1:8)=(/9352.000000,1.000000,1.000000,4.000000,2.000000,2.000000,0.607405,171.000000/)
        data(315,1:8)=(/9353.000000,1.000000,3.000000,7.000000,5.000000,3.000000,0.314316,172.000000/)
        data(316,1:8)=(/9354.000000,1.000000,3.000000,11.000000,5.000000,2.000000,0.260802,166.000000/)
        data(317,1:8)=(/9356.000000,1.000000,3.000000,11.000000,5.000000,3.000000,0.640528,173.000000/)
        data(318,1:8)=(/9357.000000,1.000000,3.000000,3.000000,5.000000,2.000000,0.226942,171.000000/)
        data(319,1:8)=(/9358.000000,1.000000,2.000000,11.000000,5.000000,3.000000,0.173498,172.000000/)
        data(320,1:8)=(/9359.000000,1.000000,3.000000,4.000000,5.000000,2.000000,0.081905,177.000000/)
        data(321,1:8)=(/9360.000000,1.000000,1.000000,3.000000,4.000000,3.000000,0.135633,189.000000/)
        data(322,1:8)=(/9361.000000,1.000000,1.000000,3.000000,5.000000,2.000000,0.602094,168.000000/)
        data(323,1:8)=(/9362.000000,1.000000,1.000000,11.000000,5.000000,3.000000,0.767378,167.000000/)
        data(324,1:8)=(/9364.000000,1.000000,1.000000,11.000000,4.000000,3.000000,0.422519,172.000000/)
        data(325,1:8)=(/9368.000000,1.000000,2.000000,5.000000,5.000000,2.000000,0.745980,168.000000/)
        data(326,1:8)=(/9369.000000,1.000000,3.000000,2.000000,5.000000,3.000000,0.408410,183.000000/)
        data(327,1:8)=(/9370.000000,1.000000,1.000000,9.000000,4.000000,3.000000,0.596118,177.000000/)
        data(328,1:8)=(/9371.000000,1.000000,3.000000,3.000000,4.000000,3.000000,0.262405,197.000000/)
        data(329,1:8)=(/9372.000000,1.000000,3.000000,8.000000,4.000000,3.000000,0.529594,173.000000/)
        data(330,1:8)=(/9374.000000,1.000000,3.000000,7.000000,4.000000,3.000000,0.432082,175.000000/)
        data(331,1:8)=(/9376.000000,1.000000,1.000000,9.000000,5.000000,3.000000,0.615743,169.000000/)
        data(332,1:8)=(/9377.000000,1.000000,1.000000,9.000000,4.000000,3.000000,0.213343,173.000000/)
        data(333,1:8)=(/9378.000000,1.000000,3.000000,4.000000,5.000000,3.000000,0.446167,176.000000/)
        data(334,1:8)=(/9381.000000,1.000000,3.000000,2.000000,5.000000,3.000000,0.045052,179.000000/)
        data(335,1:8)=(/9382.000000,1.000000,1.000000,1.000000,5.000000,2.000000,0.411921,172.500000/)
        data(336,1:8)=(/9383.000000,1.000000,1.000000,11.000000,4.000000,2.000000,0.391276,173.000000/)
        data(337,1:8)=(/9384.000000,1.000000,3.000000,3.000000,5.000000,2.000000,0.220702,173.000000/)
        data(338,1:8)=(/9385.000000,1.000000,1.000000,6.000000,4.000000,3.000000,0.384986,180.000000/)
        data(339,1:8)=(/9386.000000,1.000000,3.000000,3.000000,4.000000,3.000000,0.566977,188.000000/)
        data(340,1:8)=(/9387.000000,1.000000,3.000000,4.000000,5.000000,3.000000,0.379029,180.000000/)
        data(341,1:8)=(/9388.000000,1.000000,3.000000,5.000000,4.000000,3.000000,0.263225,171.000000/)
        data(342,1:8)=(/9390.000000,1.000000,3.000000,3.000000,4.000000,2.000000,0.344008,189.000000/)
        data(343,1:8)=(/9392.000000,1.000000,1.000000,1.000000,5.000000,2.000000,0.486409,172.000000/)
        data(344,1:8)=(/9393.000000,1.000000,1.000000,1.000000,5.000000,3.000000,0.304582,178.000000/)
        data(345,1:8)=(/9395.000000,1.000000,3.000000,4.000000,5.000000,3.000000,0.523742,181.000000/)
        data(346,1:8)=(/9475.000000,1.000000,1.000000,5.000000,3.000000,3.000000,0.526573,176.000000/)
        data(347,1:8)=(/9480.000000,1.000000,3.000000,1.000000,3.000000,2.000000,0.463610,172.500000/)



        data_ped(1,1:3)=(/1,0,0/)
        data_ped(2,1:3)=(/2,0,0/)
        data_ped(3,1:3)=(/3,0,0/)
        data_ped(4,1:3)=(/4,0,0/)
        data_ped(5,1:3)=(/5,0,0/)
        data_ped(6,1:3)=(/6,0,0/)
        data_ped(7,1:3)=(/7,0,0/)
        data_ped(8,1:3)=(/8,0,0/)
        data_ped(9,1:3)=(/9,0,0/)
        data_ped(10,1:3)=(/10,0,0/)
        data_ped(11,1:3)=(/11,0,0/)
        data_ped(12,1:3)=(/12,0,0/)
        data_ped(13,1:3)=(/13,0,0/)
        data_ped(14,1:3)=(/14,0,0/)
        data_ped(15,1:3)=(/15,0,0/)
        data_ped(16,1:3)=(/16,0,0/)
        data_ped(17,1:3)=(/17,0,0/)
        data_ped(18,1:3)=(/18,0,0/)
        data_ped(19,1:3)=(/19,0,0/)
        data_ped(20,1:3)=(/20,0,0/)
        data_ped(21,1:3)=(/21,0,0/)
        data_ped(22,1:3)=(/22,0,0/)
        data_ped(23,1:3)=(/23,0,0/)
        data_ped(24,1:3)=(/24,0,0/)
        data_ped(25,1:3)=(/25,0,0/)
        data_ped(26,1:3)=(/26,0,0/)
        data_ped(27,1:3)=(/27,0,0/)
        data_ped(28,1:3)=(/28,0,0/)
        data_ped(29,1:3)=(/29,0,0/)
        data_ped(30,1:3)=(/30,0,0/)
        data_ped(31,1:3)=(/31,0,0/)
        data_ped(32,1:3)=(/32,0,0/)
        data_ped(33,1:3)=(/33,0,0/)
        data_ped(34,1:3)=(/34,0,0/)
        data_ped(35,1:3)=(/35,0,0/)
        data_ped(36,1:3)=(/36,0,0/)
        data_ped(37,1:3)=(/37,0,0/)
        data_ped(38,1:3)=(/38,0,0/)
        data_ped(39,1:3)=(/39,0,0/)
        data_ped(40,1:3)=(/40,0,0/)
        data_ped(41,1:3)=(/41,0,0/)
        data_ped(42,1:3)=(/42,0,0/)
        data_ped(43,1:3)=(/43,0,0/)
        data_ped(44,1:3)=(/44,0,0/)
        data_ped(45,1:3)=(/45,0,0/)
        data_ped(46,1:3)=(/46,0,0/)
        data_ped(47,1:3)=(/47,0,0/)
        data_ped(48,1:3)=(/48,0,0/)
        data_ped(49,1:3)=(/49,0,0/)
        data_ped(50,1:3)=(/50,0,0/)
        data_ped(51,1:3)=(/51,0,0/)
        data_ped(52,1:3)=(/52,0,0/)
        data_ped(53,1:3)=(/53,0,0/)
        data_ped(54,1:3)=(/54,0,0/)
        data_ped(55,1:3)=(/55,0,0/)
        data_ped(56,1:3)=(/56,0,0/)
        data_ped(57,1:3)=(/57,0,0/)
        data_ped(58,1:3)=(/58,0,0/)
        data_ped(59,1:3)=(/59,0,0/)
        data_ped(60,1:3)=(/60,0,0/)
        data_ped(61,1:3)=(/61,0,0/)
        data_ped(62,1:3)=(/62,0,0/)
        data_ped(63,1:3)=(/63,0,0/)
        data_ped(64,1:3)=(/64,0,0/)
        data_ped(65,1:3)=(/65,0,0/)
        data_ped(66,1:3)=(/66,0,0/)
        data_ped(67,1:3)=(/67,0,0/)
        data_ped(68,1:3)=(/68,0,0/)
        data_ped(69,1:3)=(/69,0,0/)
        data_ped(70,1:3)=(/70,0,0/)
        data_ped(71,1:3)=(/71,0,0/)
        data_ped(72,1:3)=(/72,0,0/)
        data_ped(73,1:3)=(/73,0,0/)
        data_ped(74,1:3)=(/74,0,0/)
        data_ped(75,1:3)=(/75,0,0/)
        data_ped(76,1:3)=(/76,0,0/)
        data_ped(77,1:3)=(/77,0,0/)
        data_ped(78,1:3)=(/78,0,0/)
        data_ped(79,1:3)=(/79,0,0/)
        data_ped(80,1:3)=(/80,0,0/)
        data_ped(81,1:3)=(/81,0,0/)
        data_ped(82,1:3)=(/82,0,0/)
        data_ped(83,1:3)=(/83,0,0/)
        data_ped(84,1:3)=(/84,0,0/)
        data_ped(85,1:3)=(/85,0,0/)
        data_ped(86,1:3)=(/86,0,0/)
        data_ped(87,1:3)=(/87,0,0/)
        data_ped(88,1:3)=(/88,0,0/)
        data_ped(89,1:3)=(/89,0,0/)
        data_ped(90,1:3)=(/90,0,0/)
        data_ped(91,1:3)=(/91,0,0/)
        data_ped(92,1:3)=(/92,0,0/)
        data_ped(93,1:3)=(/93,0,0/)
        data_ped(94,1:3)=(/94,0,0/)
        data_ped(95,1:3)=(/95,0,0/)
        data_ped(96,1:3)=(/96,0,0/)
        data_ped(97,1:3)=(/97,0,0/)
        data_ped(98,1:3)=(/98,0,0/)
        data_ped(99,1:3)=(/99,0,0/)
        data_ped(100,1:3)=(/100,0,0/)
        data_ped(101,1:3)=(/101,0,0/)
        data_ped(102,1:3)=(/102,0,0/)
        data_ped(103,1:3)=(/103,0,0/)
        data_ped(104,1:3)=(/104,0,0/)
        data_ped(105,1:3)=(/105,0,0/)
        data_ped(106,1:3)=(/106,0,0/)
        data_ped(107,1:3)=(/107,0,0/)
        data_ped(108,1:3)=(/108,0,0/)
        data_ped(109,1:3)=(/109,0,0/)
        data_ped(110,1:3)=(/110,0,0/)
        data_ped(111,1:3)=(/111,0,0/)
        data_ped(112,1:3)=(/112,0,0/)
        data_ped(113,1:3)=(/113,0,0/)
        data_ped(114,1:3)=(/114,0,0/)
        data_ped(115,1:3)=(/115,0,0/)
        data_ped(116,1:3)=(/116,0,0/)
        data_ped(117,1:3)=(/117,0,0/)
        data_ped(118,1:3)=(/118,0,0/)
        data_ped(119,1:3)=(/119,0,0/)
        data_ped(120,1:3)=(/120,0,0/)
        data_ped(121,1:3)=(/121,0,0/)
        data_ped(122,1:3)=(/122,0,0/)
        data_ped(123,1:3)=(/123,0,0/)
        data_ped(124,1:3)=(/124,0,0/)
        data_ped(125,1:3)=(/125,0,0/)
        data_ped(126,1:3)=(/126,0,0/)
        data_ped(127,1:3)=(/127,0,0/)
        data_ped(128,1:3)=(/128,0,0/)
        data_ped(129,1:3)=(/129,0,0/)
        data_ped(130,1:3)=(/130,0,0/)
        data_ped(131,1:3)=(/131,0,0/)
        data_ped(132,1:3)=(/132,0,0/)
        data_ped(133,1:3)=(/133,0,0/)
        data_ped(134,1:3)=(/134,0,0/)
        data_ped(135,1:3)=(/135,0,0/)
        data_ped(136,1:3)=(/136,0,0/)
        data_ped(137,1:3)=(/137,0,0/)
        data_ped(138,1:3)=(/138,0,0/)
        data_ped(139,1:3)=(/139,0,0/)
        data_ped(140,1:3)=(/140,0,0/)
        data_ped(141,1:3)=(/141,0,0/)
        data_ped(142,1:3)=(/142,0,0/)
        data_ped(143,1:3)=(/143,0,0/)
        data_ped(144,1:3)=(/144,0,0/)
        data_ped(145,1:3)=(/145,0,0/)
        data_ped(146,1:3)=(/146,0,0/)
        data_ped(147,1:3)=(/147,0,0/)
        data_ped(148,1:3)=(/148,0,0/)
        data_ped(149,1:3)=(/149,0,0/)
        data_ped(150,1:3)=(/150,0,0/)
        data_ped(151,1:3)=(/151,0,0/)
        data_ped(152,1:3)=(/152,0,0/)
        data_ped(153,1:3)=(/153,0,0/)
        data_ped(154,1:3)=(/154,0,0/)
        data_ped(155,1:3)=(/155,0,0/)
        data_ped(156,1:3)=(/156,0,0/)
        data_ped(157,1:3)=(/157,0,0/)
        data_ped(158,1:3)=(/158,0,0/)
        data_ped(159,1:3)=(/159,0,0/)
        data_ped(160,1:3)=(/160,0,0/)
        data_ped(161,1:3)=(/161,0,0/)
        data_ped(162,1:3)=(/162,0,0/)
        data_ped(163,1:3)=(/163,0,0/)
        data_ped(164,1:3)=(/164,0,0/)
        data_ped(165,1:3)=(/165,0,0/)
        data_ped(166,1:3)=(/166,0,0/)
        data_ped(167,1:3)=(/167,0,0/)
        data_ped(168,1:3)=(/168,0,0/)
        data_ped(169,1:3)=(/169,0,0/)
        data_ped(170,1:3)=(/170,0,0/)
        data_ped(171,1:3)=(/171,0,0/)
        data_ped(172,1:3)=(/172,0,0/)
        data_ped(173,1:3)=(/173,0,0/)
        data_ped(174,1:3)=(/174,0,0/)
        data_ped(175,1:3)=(/175,0,0/)
        data_ped(176,1:3)=(/176,0,0/)
        data_ped(177,1:3)=(/177,0,0/)
        data_ped(178,1:3)=(/178,0,0/)
        data_ped(179,1:3)=(/179,0,0/)
        data_ped(180,1:3)=(/180,0,0/)
        data_ped(181,1:3)=(/181,0,0/)
        data_ped(182,1:3)=(/182,0,0/)
        data_ped(183,1:3)=(/183,0,0/)
        data_ped(184,1:3)=(/184,0,0/)
        data_ped(185,1:3)=(/185,0,0/)
        data_ped(186,1:3)=(/186,0,0/)
        data_ped(187,1:3)=(/187,0,0/)
        data_ped(188,1:3)=(/188,0,0/)
        data_ped(189,1:3)=(/189,0,0/)
        data_ped(190,1:3)=(/190,0,0/)
        data_ped(191,1:3)=(/191,0,0/)
        data_ped(192,1:3)=(/192,0,0/)
        data_ped(193,1:3)=(/193,0,0/)
        data_ped(194,1:3)=(/194,0,0/)
        data_ped(195,1:3)=(/195,0,0/)
        data_ped(196,1:3)=(/196,0,0/)
        data_ped(197,1:3)=(/197,0,0/)
        data_ped(198,1:3)=(/198,0,0/)
        data_ped(199,1:3)=(/199,0,0/)
        data_ped(200,1:3)=(/200,0,0/)
        data_ped(201,1:3)=(/201,0,0/)
        data_ped(202,1:3)=(/202,0,0/)
        data_ped(203,1:3)=(/203,0,0/)
        data_ped(204,1:3)=(/204,0,0/)
        data_ped(205,1:3)=(/205,0,0/)
        data_ped(206,1:3)=(/206,0,0/)
        data_ped(207,1:3)=(/207,0,0/)
        data_ped(208,1:3)=(/208,0,0/)
        data_ped(209,1:3)=(/209,0,0/)
        data_ped(210,1:3)=(/210,0,0/)
        data_ped(211,1:3)=(/211,0,0/)
        data_ped(212,1:3)=(/212,0,0/)
        data_ped(213,1:3)=(/213,0,0/)
        data_ped(214,1:3)=(/214,0,0/)
        data_ped(215,1:3)=(/215,0,0/)
        data_ped(216,1:3)=(/216,0,0/)
        data_ped(217,1:3)=(/217,0,0/)
        data_ped(218,1:3)=(/218,0,0/)
        data_ped(219,1:3)=(/219,0,0/)
        data_ped(220,1:3)=(/220,0,0/)
        data_ped(221,1:3)=(/221,0,0/)
        data_ped(222,1:3)=(/222,0,0/)
        data_ped(223,1:3)=(/223,0,0/)
        data_ped(224,1:3)=(/224,0,0/)
        data_ped(225,1:3)=(/225,0,0/)
        data_ped(226,1:3)=(/226,0,0/)
        data_ped(227,1:3)=(/227,0,0/)
        data_ped(228,1:3)=(/228,0,0/)
        data_ped(229,1:3)=(/229,0,0/)
        data_ped(230,1:3)=(/230,0,0/)
        data_ped(231,1:3)=(/231,0,0/)
        data_ped(232,1:3)=(/232,0,0/)
        data_ped(233,1:3)=(/233,0,0/)
        data_ped(234,1:3)=(/234,0,0/)
        data_ped(235,1:3)=(/235,0,0/)
        data_ped(236,1:3)=(/236,0,0/)
        data_ped(237,1:3)=(/237,0,0/)
        data_ped(238,1:3)=(/238,0,0/)
        data_ped(239,1:3)=(/239,0,0/)
        data_ped(240,1:3)=(/240,0,0/)
        data_ped(241,1:3)=(/241,0,0/)
        data_ped(242,1:3)=(/242,0,0/)
        data_ped(243,1:3)=(/243,0,0/)
        data_ped(244,1:3)=(/244,0,0/)
        data_ped(245,1:3)=(/245,0,0/)
        data_ped(246,1:3)=(/246,0,0/)
        data_ped(247,1:3)=(/247,0,0/)
        data_ped(248,1:3)=(/248,0,0/)
        data_ped(249,1:3)=(/249,0,0/)
        data_ped(250,1:3)=(/250,0,0/)
        data_ped(251,1:3)=(/251,0,0/)
        data_ped(252,1:3)=(/252,0,0/)
        data_ped(253,1:3)=(/253,0,0/)
        data_ped(254,1:3)=(/254,0,0/)
        data_ped(255,1:3)=(/255,0,0/)
        data_ped(256,1:3)=(/256,0,0/)
        data_ped(257,1:3)=(/257,0,0/)
        data_ped(258,1:3)=(/258,0,0/)
        data_ped(259,1:3)=(/259,0,0/)
        data_ped(260,1:3)=(/260,0,0/)
        data_ped(261,1:3)=(/261,0,0/)
        data_ped(262,1:3)=(/262,0,0/)
        data_ped(263,1:3)=(/263,0,0/)
        data_ped(264,1:3)=(/264,0,0/)
        data_ped(265,1:3)=(/265,0,0/)
        data_ped(266,1:3)=(/266,0,0/)
        data_ped(267,1:3)=(/267,0,0/)
        data_ped(268,1:3)=(/268,0,0/)
        data_ped(269,1:3)=(/269,0,0/)
        data_ped(270,1:3)=(/270,0,0/)
        data_ped(271,1:3)=(/271,0,0/)
        data_ped(272,1:3)=(/272,0,0/)
        data_ped(273,1:3)=(/273,0,0/)
        data_ped(274,1:3)=(/274,0,0/)
        data_ped(275,1:3)=(/275,0,0/)
        data_ped(276,1:3)=(/276,0,0/)
        data_ped(277,1:3)=(/277,0,0/)
        data_ped(278,1:3)=(/278,0,0/)
        data_ped(279,1:3)=(/279,0,0/)
        data_ped(280,1:3)=(/280,0,0/)
        data_ped(281,1:3)=(/281,0,0/)
        data_ped(282,1:3)=(/282,0,0/)
        data_ped(283,1:3)=(/283,0,0/)
        data_ped(284,1:3)=(/284,0,0/)
        data_ped(285,1:3)=(/285,0,0/)
        data_ped(286,1:3)=(/286,0,0/)
        data_ped(287,1:3)=(/287,0,0/)
        data_ped(288,1:3)=(/288,0,0/)
        data_ped(289,1:3)=(/289,0,0/)
        data_ped(290,1:3)=(/290,0,0/)
        data_ped(291,1:3)=(/291,0,0/)
        data_ped(292,1:3)=(/292,0,0/)
        data_ped(293,1:3)=(/293,0,0/)
        data_ped(294,1:3)=(/294,0,0/)
        data_ped(295,1:3)=(/295,0,0/)
        data_ped(296,1:3)=(/296,0,0/)
        data_ped(297,1:3)=(/297,0,0/)
        data_ped(298,1:3)=(/298,0,0/)
        data_ped(299,1:3)=(/299,0,0/)
        data_ped(300,1:3)=(/300,0,0/)
        data_ped(301,1:3)=(/301,0,0/)
        data_ped(302,1:3)=(/302,0,0/)
        data_ped(303,1:3)=(/303,0,0/)
        data_ped(304,1:3)=(/304,0,0/)
        data_ped(305,1:3)=(/305,0,0/)
        data_ped(306,1:3)=(/306,0,0/)
        data_ped(307,1:3)=(/307,0,0/)
        data_ped(308,1:3)=(/308,0,0/)
        data_ped(309,1:3)=(/309,0,0/)
        data_ped(310,1:3)=(/310,0,0/)
        data_ped(311,1:3)=(/311,0,0/)
        data_ped(312,1:3)=(/312,0,0/)
        data_ped(313,1:3)=(/313,0,0/)
        data_ped(314,1:3)=(/314,0,0/)
        data_ped(315,1:3)=(/315,0,0/)
        data_ped(316,1:3)=(/316,0,0/)
        data_ped(317,1:3)=(/317,0,0/)
        data_ped(318,1:3)=(/318,0,0/)
        data_ped(319,1:3)=(/319,0,0/)
        data_ped(320,1:3)=(/320,0,0/)
        data_ped(321,1:3)=(/321,0,0/)
        data_ped(322,1:3)=(/322,0,0/)
        data_ped(323,1:3)=(/323,0,0/)
        data_ped(324,1:3)=(/324,0,0/)
        data_ped(325,1:3)=(/325,0,0/)
        data_ped(326,1:3)=(/326,0,0/)
        data_ped(327,1:3)=(/327,0,0/)
        data_ped(328,1:3)=(/328,0,0/)
        data_ped(329,1:3)=(/329,0,0/)
        data_ped(330,1:3)=(/330,0,0/)
        data_ped(331,1:3)=(/331,0,0/)
        data_ped(332,1:3)=(/332,0,0/)
        data_ped(333,1:3)=(/333,0,0/)
        data_ped(334,1:3)=(/334,0,0/)
        data_ped(335,1:3)=(/335,0,0/)
        data_ped(336,1:3)=(/336,0,0/)
        data_ped(337,1:3)=(/337,0,0/)
        data_ped(338,1:3)=(/338,0,0/)
        data_ped(339,1:3)=(/339,0,0/)
        data_ped(340,1:3)=(/340,0,0/)
        data_ped(341,1:3)=(/341,0,0/)
        data_ped(342,1:3)=(/342,0,0/)
        data_ped(343,1:3)=(/343,0,0/)
        data_ped(344,1:3)=(/344,0,0/)
        data_ped(345,1:3)=(/345,0,0/)
        data_ped(346,1:3)=(/346,0,0/)
        data_ped(347,1:3)=(/347,0,0/)
        data_ped(348,1:3)=(/348,0,0/)
        data_ped(349,1:3)=(/349,0,0/)
        data_ped(350,1:3)=(/350,0,0/)
        data_ped(351,1:3)=(/351,0,0/)
        data_ped(352,1:3)=(/352,0,0/)
        data_ped(353,1:3)=(/353,0,0/)
        data_ped(354,1:3)=(/354,0,0/)
        data_ped(355,1:3)=(/355,0,0/)
        data_ped(356,1:3)=(/356,0,0/)
        data_ped(357,1:3)=(/357,0,0/)
        data_ped(358,1:3)=(/358,0,0/)
        data_ped(359,1:3)=(/359,0,0/)
        data_ped(360,1:3)=(/360,0,0/)
        data_ped(361,1:3)=(/361,0,0/)
        data_ped(362,1:3)=(/362,0,0/)
        data_ped(363,1:3)=(/363,0,0/)
        data_ped(364,1:3)=(/364,0,0/)
        data_ped(365,1:3)=(/365,0,0/)
        data_ped(366,1:3)=(/366,0,0/)
        data_ped(367,1:3)=(/367,0,0/)
        data_ped(368,1:3)=(/368,0,0/)
        data_ped(369,1:3)=(/369,0,0/)
        data_ped(370,1:3)=(/370,0,0/)
        data_ped(371,1:3)=(/371,0,0/)
        data_ped(372,1:3)=(/372,0,0/)
        data_ped(373,1:3)=(/373,0,0/)
        data_ped(374,1:3)=(/374,0,0/)
        data_ped(375,1:3)=(/375,0,0/)
        data_ped(376,1:3)=(/376,0,0/)
        data_ped(377,1:3)=(/377,0,0/)
        data_ped(378,1:3)=(/378,0,0/)
        data_ped(379,1:3)=(/379,0,0/)
        data_ped(380,1:3)=(/380,0,0/)
        data_ped(381,1:3)=(/381,0,0/)
        data_ped(382,1:3)=(/382,0,0/)
        data_ped(383,1:3)=(/383,0,0/)
        data_ped(384,1:3)=(/384,0,0/)
        data_ped(385,1:3)=(/385,0,0/)
        data_ped(386,1:3)=(/386,0,0/)
        data_ped(387,1:3)=(/387,0,0/)
        data_ped(388,1:3)=(/388,0,0/)
        data_ped(389,1:3)=(/389,0,0/)
        data_ped(390,1:3)=(/390,0,0/)
        data_ped(391,1:3)=(/391,0,0/)
        data_ped(392,1:3)=(/392,0,0/)
        data_ped(393,1:3)=(/393,0,0/)
        data_ped(394,1:3)=(/394,0,0/)
        data_ped(395,1:3)=(/395,0,0/)
        data_ped(396,1:3)=(/396,0,0/)
        data_ped(397,1:3)=(/397,0,0/)
        data_ped(398,1:3)=(/398,0,0/)
        data_ped(399,1:3)=(/399,0,0/)
        data_ped(400,1:3)=(/400,0,0/)
        data_ped(401,1:3)=(/401,0,0/)
        data_ped(402,1:3)=(/402,0,0/)
        data_ped(403,1:3)=(/403,0,0/)
        data_ped(404,1:3)=(/404,0,0/)
        data_ped(405,1:3)=(/405,0,0/)
        data_ped(406,1:3)=(/406,0,0/)
        data_ped(407,1:3)=(/407,0,0/)
        data_ped(408,1:3)=(/408,0,0/)
        data_ped(409,1:3)=(/409,0,0/)
        data_ped(410,1:3)=(/410,0,0/)
        data_ped(411,1:3)=(/411,0,0/)
        data_ped(412,1:3)=(/412,0,0/)
        data_ped(413,1:3)=(/413,0,0/)
        data_ped(414,1:3)=(/414,0,0/)
        data_ped(415,1:3)=(/415,0,0/)
        data_ped(416,1:3)=(/416,0,0/)
        data_ped(417,1:3)=(/417,0,0/)
        data_ped(418,1:3)=(/418,0,0/)
        data_ped(419,1:3)=(/419,0,0/)
        data_ped(420,1:3)=(/420,0,0/)
        data_ped(421,1:3)=(/421,0,0/)
        data_ped(422,1:3)=(/422,0,0/)
        data_ped(423,1:3)=(/423,0,0/)
        data_ped(424,1:3)=(/424,0,0/)
        data_ped(425,1:3)=(/425,0,0/)
        data_ped(426,1:3)=(/426,0,0/)
        data_ped(427,1:3)=(/427,0,0/)
        data_ped(428,1:3)=(/428,0,0/)
        data_ped(429,1:3)=(/429,0,0/)
        data_ped(430,1:3)=(/430,0,0/)
        data_ped(431,1:3)=(/431,0,0/)
        data_ped(432,1:3)=(/432,0,0/)
        data_ped(433,1:3)=(/433,0,0/)
        data_ped(434,1:3)=(/434,0,0/)
        data_ped(435,1:3)=(/435,0,0/)
        data_ped(436,1:3)=(/436,0,0/)
        data_ped(437,1:3)=(/437,0,0/)
        data_ped(438,1:3)=(/438,0,0/)
        data_ped(439,1:3)=(/439,0,0/)
        data_ped(440,1:3)=(/440,0,0/)
        data_ped(441,1:3)=(/441,0,0/)
        data_ped(442,1:3)=(/442,0,0/)
        data_ped(443,1:3)=(/443,0,0/)
        data_ped(444,1:3)=(/444,0,0/)
        data_ped(445,1:3)=(/445,0,0/)
        data_ped(446,1:3)=(/446,0,0/)
        data_ped(447,1:3)=(/447,0,0/)
        data_ped(448,1:3)=(/448,0,0/)
        data_ped(449,1:3)=(/449,0,0/)
        data_ped(450,1:3)=(/450,0,0/)
        data_ped(451,1:3)=(/451,0,0/)
        data_ped(452,1:3)=(/452,0,0/)
        data_ped(453,1:3)=(/453,0,0/)
        data_ped(454,1:3)=(/454,0,0/)
        data_ped(455,1:3)=(/455,0,0/)
        data_ped(456,1:3)=(/456,0,0/)
        data_ped(457,1:3)=(/457,0,0/)
        data_ped(458,1:3)=(/458,0,0/)
        data_ped(459,1:3)=(/459,0,0/)
        data_ped(460,1:3)=(/460,0,0/)
        data_ped(461,1:3)=(/461,0,0/)
        data_ped(462,1:3)=(/462,0,0/)
        data_ped(463,1:3)=(/463,0,0/)
        data_ped(464,1:3)=(/464,0,0/)
        data_ped(465,1:3)=(/465,0,0/)
        data_ped(466,1:3)=(/466,0,0/)
        data_ped(467,1:3)=(/467,0,0/)
        data_ped(468,1:3)=(/468,0,0/)
        data_ped(469,1:3)=(/469,0,0/)
        data_ped(470,1:3)=(/470,0,0/)
        data_ped(471,1:3)=(/471,0,0/)
        data_ped(472,1:3)=(/472,0,0/)
        data_ped(473,1:3)=(/473,0,0/)
        data_ped(474,1:3)=(/474,0,0/)
        data_ped(475,1:3)=(/475,0,0/)
        data_ped(476,1:3)=(/476,0,0/)
        data_ped(477,1:3)=(/477,0,0/)
        data_ped(478,1:3)=(/478,0,0/)
        data_ped(479,1:3)=(/479,0,0/)
        data_ped(480,1:3)=(/480,0,0/)
        data_ped(481,1:3)=(/481,0,0/)
        data_ped(482,1:3)=(/482,0,0/)
        data_ped(483,1:3)=(/483,0,0/)
        data_ped(484,1:3)=(/484,0,0/)
        data_ped(485,1:3)=(/485,0,0/)
        data_ped(486,1:3)=(/486,0,0/)
        data_ped(487,1:3)=(/487,0,0/)
        data_ped(488,1:3)=(/488,0,0/)
        data_ped(489,1:3)=(/489,0,0/)
        data_ped(490,1:3)=(/490,0,0/)
        data_ped(491,1:3)=(/491,0,0/)
        data_ped(492,1:3)=(/492,0,0/)
        data_ped(493,1:3)=(/493,0,0/)
        data_ped(494,1:3)=(/494,0,0/)
        data_ped(495,1:3)=(/495,0,0/)
        data_ped(496,1:3)=(/496,0,0/)
        data_ped(497,1:3)=(/497,0,0/)
        data_ped(498,1:3)=(/498,0,0/)
        data_ped(499,1:3)=(/499,0,0/)
        data_ped(500,1:3)=(/500,0,0/)
        data_ped(501,1:3)=(/501,0,0/)
        data_ped(502,1:3)=(/502,0,0/)
        data_ped(503,1:3)=(/503,0,0/)
        data_ped(504,1:3)=(/504,0,0/)
        data_ped(505,1:3)=(/505,0,0/)
        data_ped(506,1:3)=(/506,0,0/)
        data_ped(507,1:3)=(/507,0,0/)
        data_ped(508,1:3)=(/508,0,0/)
        data_ped(509,1:3)=(/509,0,0/)
        data_ped(510,1:3)=(/510,0,0/)
        data_ped(511,1:3)=(/511,0,0/)
        data_ped(512,1:3)=(/512,0,0/)
        data_ped(513,1:3)=(/513,0,0/)
        data_ped(514,1:3)=(/514,0,0/)
        data_ped(515,1:3)=(/515,0,0/)
        data_ped(516,1:3)=(/516,0,0/)
        data_ped(517,1:3)=(/517,0,0/)
        data_ped(518,1:3)=(/518,0,0/)
        data_ped(519,1:3)=(/519,0,0/)
        data_ped(520,1:3)=(/520,0,0/)
        data_ped(521,1:3)=(/521,0,0/)
        data_ped(522,1:3)=(/522,0,0/)
        data_ped(523,1:3)=(/523,0,0/)
        data_ped(524,1:3)=(/524,0,0/)
        data_ped(525,1:3)=(/525,0,0/)
        data_ped(526,1:3)=(/526,0,0/)
        data_ped(527,1:3)=(/527,0,0/)
        data_ped(528,1:3)=(/528,0,0/)
        data_ped(529,1:3)=(/529,0,0/)
        data_ped(530,1:3)=(/530,0,0/)
        data_ped(531,1:3)=(/531,0,0/)
        data_ped(532,1:3)=(/532,0,0/)
        data_ped(533,1:3)=(/533,0,0/)
        data_ped(534,1:3)=(/534,0,0/)
        data_ped(535,1:3)=(/535,0,0/)
        data_ped(536,1:3)=(/536,0,0/)
        data_ped(537,1:3)=(/537,0,0/)
        data_ped(538,1:3)=(/538,0,0/)
        data_ped(539,1:3)=(/539,0,0/)
        data_ped(540,1:3)=(/540,0,0/)
        data_ped(541,1:3)=(/541,0,0/)
        data_ped(542,1:3)=(/542,0,0/)
        data_ped(543,1:3)=(/543,0,0/)
        data_ped(544,1:3)=(/544,0,0/)
        data_ped(545,1:3)=(/545,0,0/)
        data_ped(546,1:3)=(/546,0,0/)
        data_ped(547,1:3)=(/547,0,0/)
        data_ped(548,1:3)=(/548,0,0/)
        data_ped(549,1:3)=(/549,0,0/)
        data_ped(550,1:3)=(/550,0,0/)
        data_ped(551,1:3)=(/551,0,0/)
        data_ped(552,1:3)=(/552,0,0/)
        data_ped(553,1:3)=(/553,0,0/)
        data_ped(554,1:3)=(/554,0,0/)
        data_ped(555,1:3)=(/555,0,0/)
        data_ped(556,1:3)=(/556,0,0/)
        data_ped(557,1:3)=(/557,0,0/)
        data_ped(558,1:3)=(/558,0,0/)
        data_ped(559,1:3)=(/559,0,0/)
        data_ped(560,1:3)=(/560,0,0/)
        data_ped(561,1:3)=(/561,0,0/)
        data_ped(562,1:3)=(/562,0,0/)
        data_ped(563,1:3)=(/563,0,0/)
        data_ped(564,1:3)=(/564,0,0/)
        data_ped(565,1:3)=(/565,0,0/)
        data_ped(566,1:3)=(/566,0,0/)
        data_ped(567,1:3)=(/567,0,0/)
        data_ped(568,1:3)=(/568,0,0/)
        data_ped(569,1:3)=(/569,0,0/)
        data_ped(570,1:3)=(/570,0,0/)
        data_ped(571,1:3)=(/571,0,0/)
        data_ped(572,1:3)=(/572,0,0/)
        data_ped(573,1:3)=(/573,0,0/)
        data_ped(574,1:3)=(/574,0,0/)
        data_ped(575,1:3)=(/575,0,0/)
        data_ped(576,1:3)=(/576,0,0/)
        data_ped(577,1:3)=(/577,0,0/)
        data_ped(578,1:3)=(/578,0,0/)
        data_ped(579,1:3)=(/579,0,0/)
        data_ped(580,1:3)=(/580,0,0/)
        data_ped(581,1:3)=(/581,0,0/)
        data_ped(582,1:3)=(/582,0,0/)
        data_ped(583,1:3)=(/583,0,0/)
        data_ped(584,1:3)=(/584,0,0/)
        data_ped(585,1:3)=(/585,0,0/)
        data_ped(586,1:3)=(/586,0,0/)
        data_ped(587,1:3)=(/587,0,0/)
        data_ped(588,1:3)=(/588,0,0/)
        data_ped(589,1:3)=(/589,0,0/)
        data_ped(590,1:3)=(/590,0,0/)
        data_ped(591,1:3)=(/591,0,0/)
        data_ped(592,1:3)=(/592,0,0/)
        data_ped(593,1:3)=(/593,0,0/)
        data_ped(594,1:3)=(/594,0,0/)
        data_ped(595,1:3)=(/595,0,0/)
        data_ped(596,1:3)=(/596,0,0/)
        data_ped(597,1:3)=(/597,0,0/)
        data_ped(598,1:3)=(/598,0,0/)
        data_ped(599,1:3)=(/599,0,0/)
        data_ped(600,1:3)=(/600,0,0/)
        data_ped(601,1:3)=(/601,0,0/)
        data_ped(602,1:3)=(/602,0,0/)
        data_ped(603,1:3)=(/603,0,0/)
        data_ped(604,1:3)=(/604,0,0/)
        data_ped(605,1:3)=(/605,0,0/)
        data_ped(606,1:3)=(/606,0,0/)
        data_ped(607,1:3)=(/607,0,0/)
        data_ped(608,1:3)=(/608,0,0/)
        data_ped(609,1:3)=(/609,0,0/)
        data_ped(610,1:3)=(/610,0,0/)
        data_ped(611,1:3)=(/611,0,0/)
        data_ped(612,1:3)=(/612,0,0/)
        data_ped(613,1:3)=(/613,0,0/)
        data_ped(614,1:3)=(/614,0,0/)
        data_ped(615,1:3)=(/615,0,0/)
        data_ped(616,1:3)=(/616,0,0/)
        data_ped(617,1:3)=(/617,0,0/)
        data_ped(618,1:3)=(/618,0,0/)
        data_ped(619,1:3)=(/619,0,0/)
        data_ped(620,1:3)=(/620,0,0/)
        data_ped(621,1:3)=(/621,0,0/)
        data_ped(622,1:3)=(/622,0,0/)
        data_ped(623,1:3)=(/623,0,0/)
        data_ped(624,1:3)=(/624,0,0/)
        data_ped(625,1:3)=(/625,0,0/)
        data_ped(626,1:3)=(/626,0,0/)
        data_ped(627,1:3)=(/627,0,0/)
        data_ped(628,1:3)=(/628,0,0/)
        data_ped(629,1:3)=(/629,0,0/)
        data_ped(630,1:3)=(/630,0,0/)
        data_ped(631,1:3)=(/631,0,0/)
        data_ped(632,1:3)=(/632,0,0/)
        data_ped(633,1:3)=(/633,0,0/)
        data_ped(634,1:3)=(/634,0,0/)
        data_ped(635,1:3)=(/635,0,0/)
        data_ped(636,1:3)=(/636,0,0/)
        data_ped(637,1:3)=(/637,0,0/)
        data_ped(638,1:3)=(/638,0,0/)
        data_ped(639,1:3)=(/639,0,0/)
        data_ped(640,1:3)=(/640,0,0/)
        data_ped(641,1:3)=(/641,0,0/)
        data_ped(642,1:3)=(/642,0,0/)
        data_ped(643,1:3)=(/643,0,0/)
        data_ped(644,1:3)=(/644,0,0/)
        data_ped(645,1:3)=(/645,0,0/)
        data_ped(646,1:3)=(/646,0,0/)
        data_ped(647,1:3)=(/647,0,0/)
        data_ped(648,1:3)=(/648,0,0/)
        data_ped(649,1:3)=(/649,0,0/)
        data_ped(650,1:3)=(/650,0,0/)
        data_ped(651,1:3)=(/651,0,0/)
        data_ped(652,1:3)=(/652,0,0/)
        data_ped(653,1:3)=(/653,0,0/)
        data_ped(654,1:3)=(/654,0,0/)
        data_ped(655,1:3)=(/655,0,0/)
        data_ped(656,1:3)=(/656,0,0/)
        data_ped(657,1:3)=(/657,0,0/)
        data_ped(658,1:3)=(/658,0,0/)
        data_ped(659,1:3)=(/659,0,0/)
        data_ped(660,1:3)=(/660,0,0/)
        data_ped(661,1:3)=(/661,0,0/)
        data_ped(662,1:3)=(/662,0,0/)
        data_ped(663,1:3)=(/663,0,0/)
        data_ped(664,1:3)=(/664,0,0/)
        data_ped(665,1:3)=(/665,0,0/)
        data_ped(666,1:3)=(/666,0,0/)
        data_ped(667,1:3)=(/667,0,0/)
        data_ped(668,1:3)=(/668,0,0/)
        data_ped(669,1:3)=(/669,0,0/)
        data_ped(670,1:3)=(/670,0,0/)
        data_ped(671,1:3)=(/671,0,0/)
        data_ped(672,1:3)=(/672,0,0/)
        data_ped(673,1:3)=(/673,0,0/)
        data_ped(674,1:3)=(/674,0,0/)
        data_ped(675,1:3)=(/675,0,0/)
        data_ped(676,1:3)=(/676,0,0/)
        data_ped(677,1:3)=(/677,0,0/)
        data_ped(678,1:3)=(/678,0,0/)
        data_ped(679,1:3)=(/679,0,0/)
        data_ped(680,1:3)=(/680,0,0/)
        data_ped(681,1:3)=(/681,0,0/)
        data_ped(682,1:3)=(/682,0,0/)
        data_ped(683,1:3)=(/683,0,0/)
        data_ped(684,1:3)=(/684,0,0/)
        data_ped(685,1:3)=(/685,0,0/)
        data_ped(686,1:3)=(/686,0,0/)
        data_ped(687,1:3)=(/687,0,0/)
        data_ped(688,1:3)=(/688,0,0/)
        data_ped(689,1:3)=(/689,0,0/)
        data_ped(690,1:3)=(/690,0,0/)
        data_ped(691,1:3)=(/691,0,0/)
        data_ped(692,1:3)=(/692,0,0/)
        data_ped(693,1:3)=(/693,0,0/)
        data_ped(694,1:3)=(/694,0,0/)
        data_ped(695,1:3)=(/695,0,0/)
        data_ped(696,1:3)=(/696,0,0/)
        data_ped(697,1:3)=(/697,0,0/)
        data_ped(698,1:3)=(/698,0,0/)
        data_ped(699,1:3)=(/699,0,0/)
        data_ped(700,1:3)=(/700,0,0/)
        data_ped(701,1:3)=(/701,0,0/)
        data_ped(702,1:3)=(/702,0,0/)
        data_ped(703,1:3)=(/703,0,0/)
        data_ped(704,1:3)=(/704,0,0/)
        data_ped(705,1:3)=(/705,0,0/)
        data_ped(706,1:3)=(/706,0,0/)
        data_ped(707,1:3)=(/707,0,0/)
        data_ped(708,1:3)=(/708,0,0/)
        data_ped(709,1:3)=(/709,0,0/)
        data_ped(710,1:3)=(/710,0,0/)
        data_ped(711,1:3)=(/711,0,0/)
        data_ped(712,1:3)=(/712,0,0/)
        data_ped(713,1:3)=(/713,0,0/)
        data_ped(714,1:3)=(/714,0,0/)
        data_ped(715,1:3)=(/715,0,0/)
        data_ped(716,1:3)=(/716,0,0/)
        data_ped(717,1:3)=(/717,0,0/)
        data_ped(718,1:3)=(/718,0,0/)
        data_ped(719,1:3)=(/719,0,0/)
        data_ped(720,1:3)=(/720,0,0/)
        data_ped(721,1:3)=(/721,0,0/)
        data_ped(722,1:3)=(/722,0,0/)
        data_ped(723,1:3)=(/723,0,0/)
        data_ped(724,1:3)=(/724,0,0/)
        data_ped(725,1:3)=(/725,0,0/)
        data_ped(726,1:3)=(/726,0,0/)
        data_ped(727,1:3)=(/727,0,0/)
        data_ped(728,1:3)=(/728,0,0/)
        data_ped(729,1:3)=(/729,0,0/)
        data_ped(730,1:3)=(/730,0,0/)
        data_ped(731,1:3)=(/731,0,0/)
        data_ped(732,1:3)=(/732,0,0/)
        data_ped(733,1:3)=(/733,0,0/)
        data_ped(734,1:3)=(/734,0,0/)
        data_ped(735,1:3)=(/735,0,0/)
        data_ped(736,1:3)=(/736,0,0/)
        data_ped(737,1:3)=(/737,0,0/)
        data_ped(738,1:3)=(/738,0,0/)
        data_ped(739,1:3)=(/739,0,0/)
        data_ped(740,1:3)=(/740,0,0/)
        data_ped(741,1:3)=(/741,0,0/)
        data_ped(742,1:3)=(/742,0,0/)
        data_ped(743,1:3)=(/743,0,0/)
        data_ped(744,1:3)=(/744,0,0/)
        data_ped(745,1:3)=(/745,0,0/)
        data_ped(746,1:3)=(/746,0,0/)
        data_ped(747,1:3)=(/747,0,0/)
        data_ped(748,1:3)=(/748,0,0/)
        data_ped(749,1:3)=(/749,0,0/)
        data_ped(750,1:3)=(/750,0,0/)
        data_ped(751,1:3)=(/751,0,0/)
        data_ped(752,1:3)=(/752,0,0/)
        data_ped(753,1:3)=(/753,0,0/)
        data_ped(754,1:3)=(/754,0,0/)
        data_ped(755,1:3)=(/755,0,0/)
        data_ped(756,1:3)=(/756,0,0/)
        data_ped(757,1:3)=(/757,0,0/)
        data_ped(758,1:3)=(/758,0,0/)
        data_ped(759,1:3)=(/759,0,0/)
        data_ped(760,1:3)=(/760,0,0/)
        data_ped(761,1:3)=(/761,0,0/)
        data_ped(762,1:3)=(/762,0,0/)
        data_ped(763,1:3)=(/763,0,0/)
        data_ped(764,1:3)=(/764,0,0/)
        data_ped(765,1:3)=(/765,0,0/)
        data_ped(766,1:3)=(/766,0,0/)
        data_ped(767,1:3)=(/767,0,0/)
        data_ped(768,1:3)=(/768,0,0/)
        data_ped(769,1:3)=(/769,0,0/)
        data_ped(770,1:3)=(/770,0,0/)
        data_ped(771,1:3)=(/771,0,0/)
        data_ped(772,1:3)=(/772,0,0/)
        data_ped(773,1:3)=(/773,0,0/)
        data_ped(774,1:3)=(/774,0,0/)
        data_ped(775,1:3)=(/775,0,0/)
        data_ped(776,1:3)=(/776,0,0/)
        data_ped(777,1:3)=(/777,0,0/)
        data_ped(778,1:3)=(/778,0,0/)
        data_ped(779,1:3)=(/779,0,0/)
        data_ped(780,1:3)=(/780,0,0/)
        data_ped(781,1:3)=(/781,0,0/)
        data_ped(782,1:3)=(/782,0,0/)
        data_ped(783,1:3)=(/783,0,0/)
        data_ped(784,1:3)=(/784,0,0/)
        data_ped(785,1:3)=(/785,0,0/)
        data_ped(786,1:3)=(/786,0,0/)
        data_ped(787,1:3)=(/787,0,0/)
        data_ped(788,1:3)=(/788,0,0/)
        data_ped(789,1:3)=(/789,0,0/)
        data_ped(790,1:3)=(/790,0,0/)
        data_ped(791,1:3)=(/791,0,0/)
        data_ped(792,1:3)=(/792,0,0/)
        data_ped(793,1:3)=(/793,0,0/)
        data_ped(794,1:3)=(/794,0,0/)
        data_ped(795,1:3)=(/795,0,0/)
        data_ped(796,1:3)=(/796,0,0/)
        data_ped(797,1:3)=(/797,0,0/)
        data_ped(798,1:3)=(/798,0,0/)
        data_ped(799,1:3)=(/799,0,0/)
        data_ped(800,1:3)=(/800,0,0/)
        data_ped(801,1:3)=(/801,0,0/)
        data_ped(802,1:3)=(/802,0,0/)
        data_ped(803,1:3)=(/803,0,0/)
        data_ped(804,1:3)=(/804,0,0/)
        data_ped(805,1:3)=(/805,0,0/)
        data_ped(806,1:3)=(/806,0,0/)
        data_ped(807,1:3)=(/807,0,0/)
        data_ped(808,1:3)=(/808,0,0/)
        data_ped(809,1:3)=(/809,0,0/)
        data_ped(810,1:3)=(/810,0,0/)
        data_ped(811,1:3)=(/811,0,0/)
        data_ped(812,1:3)=(/812,0,0/)
        data_ped(813,1:3)=(/813,0,0/)
        data_ped(814,1:3)=(/814,0,0/)
        data_ped(815,1:3)=(/815,0,0/)
        data_ped(816,1:3)=(/816,0,0/)
        data_ped(817,1:3)=(/817,0,0/)
        data_ped(818,1:3)=(/818,0,0/)
        data_ped(819,1:3)=(/819,0,0/)
        data_ped(820,1:3)=(/820,0,0/)
        data_ped(821,1:3)=(/821,0,0/)
        data_ped(822,1:3)=(/822,0,0/)
        data_ped(823,1:3)=(/823,0,0/)
        data_ped(824,1:3)=(/824,0,0/)
        data_ped(825,1:3)=(/825,0,0/)
        data_ped(826,1:3)=(/826,0,0/)
        data_ped(827,1:3)=(/827,0,0/)
        data_ped(828,1:3)=(/828,0,0/)
        data_ped(829,1:3)=(/829,0,0/)
        data_ped(830,1:3)=(/830,0,0/)
        data_ped(831,1:3)=(/831,0,0/)
        data_ped(832,1:3)=(/832,0,0/)
        data_ped(833,1:3)=(/833,0,0/)
        data_ped(834,1:3)=(/834,0,0/)
        data_ped(835,1:3)=(/835,0,0/)
        data_ped(836,1:3)=(/836,0,0/)
        data_ped(837,1:3)=(/837,0,0/)
        data_ped(838,1:3)=(/838,0,0/)
        data_ped(839,1:3)=(/839,0,0/)
        data_ped(840,1:3)=(/840,0,0/)
        data_ped(841,1:3)=(/841,0,0/)
        data_ped(842,1:3)=(/842,0,0/)
        data_ped(843,1:3)=(/843,0,0/)
        data_ped(844,1:3)=(/844,0,0/)
        data_ped(845,1:3)=(/845,0,0/)
        data_ped(846,1:3)=(/846,0,0/)
        data_ped(847,1:3)=(/847,0,0/)
        data_ped(848,1:3)=(/848,0,0/)
        data_ped(849,1:3)=(/849,0,0/)
        data_ped(850,1:3)=(/850,0,0/)
        data_ped(851,1:3)=(/851,0,0/)
        data_ped(852,1:3)=(/852,0,0/)
        data_ped(853,1:3)=(/853,0,0/)
        data_ped(854,1:3)=(/854,0,0/)
        data_ped(855,1:3)=(/855,0,0/)
        data_ped(856,1:3)=(/856,0,0/)
        data_ped(857,1:3)=(/857,0,0/)
        data_ped(858,1:3)=(/858,0,0/)
        data_ped(859,1:3)=(/859,0,0/)
        data_ped(860,1:3)=(/860,0,0/)
        data_ped(861,1:3)=(/861,0,0/)
        data_ped(862,1:3)=(/862,0,0/)
        data_ped(863,1:3)=(/863,0,0/)
        data_ped(864,1:3)=(/864,0,0/)
        data_ped(865,1:3)=(/865,0,0/)
        data_ped(866,1:3)=(/866,0,0/)
        data_ped(867,1:3)=(/867,0,0/)
        data_ped(868,1:3)=(/868,0,0/)
        data_ped(869,1:3)=(/869,0,0/)
        data_ped(870,1:3)=(/870,0,0/)
        data_ped(871,1:3)=(/871,0,0/)
        data_ped(872,1:3)=(/872,0,0/)
        data_ped(873,1:3)=(/873,0,0/)
        data_ped(874,1:3)=(/874,0,0/)
        data_ped(875,1:3)=(/875,0,0/)
        data_ped(876,1:3)=(/876,0,0/)
        data_ped(877,1:3)=(/877,0,0/)
        data_ped(878,1:3)=(/878,0,0/)
        data_ped(879,1:3)=(/879,0,0/)
        data_ped(880,1:3)=(/880,0,0/)
        data_ped(881,1:3)=(/881,0,0/)
        data_ped(882,1:3)=(/882,0,0/)
        data_ped(883,1:3)=(/883,0,0/)
        data_ped(884,1:3)=(/884,0,0/)
        data_ped(885,1:3)=(/885,0,0/)
        data_ped(886,1:3)=(/886,0,0/)
        data_ped(887,1:3)=(/887,0,0/)
        data_ped(888,1:3)=(/888,0,0/)
        data_ped(889,1:3)=(/889,0,0/)
        data_ped(890,1:3)=(/890,0,0/)
        data_ped(891,1:3)=(/891,0,0/)
        data_ped(892,1:3)=(/892,0,0/)
        data_ped(893,1:3)=(/893,0,0/)
        data_ped(894,1:3)=(/894,0,0/)
        data_ped(895,1:3)=(/895,0,0/)
        data_ped(896,1:3)=(/896,0,0/)
        data_ped(897,1:3)=(/897,0,0/)
        data_ped(898,1:3)=(/898,0,0/)
        data_ped(899,1:3)=(/899,0,0/)
        data_ped(900,1:3)=(/900,0,0/)
        data_ped(901,1:3)=(/901,0,0/)
        data_ped(902,1:3)=(/902,0,0/)
        data_ped(903,1:3)=(/903,0,0/)
        data_ped(904,1:3)=(/904,0,0/)
        data_ped(905,1:3)=(/905,0,0/)
        data_ped(906,1:3)=(/906,0,0/)
        data_ped(907,1:3)=(/907,0,0/)
        data_ped(908,1:3)=(/908,0,0/)
        data_ped(909,1:3)=(/909,0,0/)
        data_ped(910,1:3)=(/910,0,0/)
        data_ped(911,1:3)=(/911,0,0/)
        data_ped(912,1:3)=(/912,0,0/)
        data_ped(913,1:3)=(/913,0,0/)
        data_ped(914,1:3)=(/914,0,0/)
        data_ped(915,1:3)=(/915,0,0/)
        data_ped(916,1:3)=(/916,0,0/)
        data_ped(917,1:3)=(/917,0,0/)
        data_ped(918,1:3)=(/918,0,0/)
        data_ped(919,1:3)=(/919,0,0/)
        data_ped(920,1:3)=(/920,0,0/)
        data_ped(921,1:3)=(/921,0,0/)
        data_ped(922,1:3)=(/922,0,0/)
        data_ped(923,1:3)=(/923,0,0/)
        data_ped(924,1:3)=(/924,0,0/)
        data_ped(925,1:3)=(/925,0,0/)
        data_ped(926,1:3)=(/926,0,0/)
        data_ped(927,1:3)=(/927,0,0/)
        data_ped(928,1:3)=(/928,0,0/)
        data_ped(929,1:3)=(/929,0,0/)
        data_ped(930,1:3)=(/930,0,0/)
        data_ped(931,1:3)=(/931,0,0/)
        data_ped(932,1:3)=(/932,0,0/)
        data_ped(933,1:3)=(/933,0,0/)
        data_ped(934,1:3)=(/934,0,0/)
        data_ped(935,1:3)=(/935,0,0/)
        data_ped(936,1:3)=(/936,0,0/)
        data_ped(937,1:3)=(/937,0,0/)
        data_ped(938,1:3)=(/938,0,0/)
        data_ped(939,1:3)=(/939,0,0/)
        data_ped(940,1:3)=(/940,0,0/)
        data_ped(941,1:3)=(/941,0,0/)
        data_ped(942,1:3)=(/942,0,0/)
        data_ped(943,1:3)=(/943,0,0/)
        data_ped(944,1:3)=(/944,0,0/)
        data_ped(945,1:3)=(/945,0,0/)
        data_ped(946,1:3)=(/946,0,0/)
        data_ped(947,1:3)=(/947,0,0/)
        data_ped(948,1:3)=(/948,0,0/)
        data_ped(949,1:3)=(/949,0,0/)
        data_ped(950,1:3)=(/950,0,0/)
        data_ped(951,1:3)=(/951,0,0/)
        data_ped(952,1:3)=(/952,0,0/)
        data_ped(953,1:3)=(/953,0,0/)
        data_ped(954,1:3)=(/954,0,0/)
        data_ped(955,1:3)=(/955,0,0/)
        data_ped(956,1:3)=(/956,0,0/)
        data_ped(957,1:3)=(/957,0,0/)
        data_ped(958,1:3)=(/958,0,0/)
        data_ped(959,1:3)=(/959,0,0/)
        data_ped(960,1:3)=(/960,0,0/)
        data_ped(961,1:3)=(/961,0,0/)
        data_ped(962,1:3)=(/962,0,0/)
        data_ped(963,1:3)=(/963,0,0/)
        data_ped(964,1:3)=(/964,0,0/)
        data_ped(965,1:3)=(/965,0,0/)
        data_ped(966,1:3)=(/966,0,0/)
        data_ped(967,1:3)=(/967,0,0/)
        data_ped(968,1:3)=(/968,0,0/)
        data_ped(969,1:3)=(/969,0,0/)
        data_ped(970,1:3)=(/970,0,0/)
        data_ped(971,1:3)=(/971,0,0/)
        data_ped(972,1:3)=(/972,0,0/)
        data_ped(973,1:3)=(/973,0,0/)
        data_ped(974,1:3)=(/974,0,0/)
        data_ped(975,1:3)=(/975,0,0/)
        data_ped(976,1:3)=(/976,0,0/)
        data_ped(977,1:3)=(/977,0,0/)
        data_ped(978,1:3)=(/978,0,0/)
        data_ped(979,1:3)=(/979,0,0/)
        data_ped(980,1:3)=(/980,0,0/)
        data_ped(981,1:3)=(/981,0,0/)
        data_ped(982,1:3)=(/982,0,0/)
        data_ped(983,1:3)=(/983,0,0/)
        data_ped(984,1:3)=(/984,0,0/)
        data_ped(985,1:3)=(/985,0,0/)
        data_ped(986,1:3)=(/986,0,0/)
        data_ped(987,1:3)=(/987,0,0/)
        data_ped(988,1:3)=(/988,0,0/)
        data_ped(989,1:3)=(/989,0,0/)
        data_ped(990,1:3)=(/990,0,0/)
        data_ped(991,1:3)=(/991,0,0/)
        data_ped(992,1:3)=(/992,0,0/)
        data_ped(993,1:3)=(/993,0,0/)
        data_ped(994,1:3)=(/994,0,0/)
        data_ped(995,1:3)=(/995,0,0/)
        data_ped(996,1:3)=(/996,0,0/)
        data_ped(997,1:3)=(/997,0,0/)
        data_ped(998,1:3)=(/998,0,0/)
        data_ped(999,1:3)=(/999,0,0/)
        data_ped(1000,1:3)=(/1000,0,0/)
        data_ped(1001,1:3)=(/1001,0,0/)
        data_ped(1002,1:3)=(/1002,0,0/)
        data_ped(1003,1:3)=(/1003,0,0/)
        data_ped(1004,1:3)=(/1004,0,0/)
        data_ped(1005,1:3)=(/1005,0,0/)
        data_ped(1006,1:3)=(/1006,0,0/)
        data_ped(1007,1:3)=(/1007,0,0/)
        data_ped(1008,1:3)=(/1008,0,0/)
        data_ped(1009,1:3)=(/1009,0,0/)
        data_ped(1010,1:3)=(/1010,0,0/)
        data_ped(1011,1:3)=(/1011,0,0/)
        data_ped(1012,1:3)=(/1012,0,0/)
        data_ped(1013,1:3)=(/1013,0,0/)
        data_ped(1014,1:3)=(/1014,0,0/)
        data_ped(1015,1:3)=(/1015,0,0/)
        data_ped(1016,1:3)=(/1016,0,0/)
        data_ped(1017,1:3)=(/1017,0,0/)
        data_ped(1018,1:3)=(/1018,0,0/)
        data_ped(1019,1:3)=(/1019,0,0/)
        data_ped(1020,1:3)=(/1020,0,0/)
        data_ped(1021,1:3)=(/1021,0,0/)
        data_ped(1022,1:3)=(/1022,0,0/)
        data_ped(1023,1:3)=(/1023,0,0/)
        data_ped(1024,1:3)=(/1024,0,0/)
        data_ped(1025,1:3)=(/1025,0,0/)
        data_ped(1026,1:3)=(/1026,0,0/)
        data_ped(1027,1:3)=(/1027,0,0/)
        data_ped(1028,1:3)=(/1028,0,0/)
        data_ped(1029,1:3)=(/1029,0,0/)
        data_ped(1030,1:3)=(/1030,0,0/)
        data_ped(1031,1:3)=(/1031,0,0/)
        data_ped(1032,1:3)=(/1032,0,0/)
        data_ped(1033,1:3)=(/1033,0,0/)
        data_ped(1034,1:3)=(/1034,0,0/)
        data_ped(1035,1:3)=(/1035,0,0/)
        data_ped(1036,1:3)=(/1036,0,0/)
        data_ped(1037,1:3)=(/1037,0,0/)
        data_ped(1038,1:3)=(/1038,0,0/)
        data_ped(1039,1:3)=(/1039,0,0/)
        data_ped(1040,1:3)=(/1040,0,0/)
        data_ped(1041,1:3)=(/1041,0,0/)
        data_ped(1042,1:3)=(/1042,0,0/)
        data_ped(1043,1:3)=(/1043,0,0/)
        data_ped(1044,1:3)=(/1044,0,0/)
        data_ped(1045,1:3)=(/1045,0,0/)
        data_ped(1046,1:3)=(/1046,0,0/)
        data_ped(1047,1:3)=(/1047,0,0/)
        data_ped(1048,1:3)=(/1048,0,0/)
        data_ped(1049,1:3)=(/1049,0,0/)
        data_ped(1050,1:3)=(/1050,0,0/)
        data_ped(1051,1:3)=(/1051,0,0/)
        data_ped(1052,1:3)=(/1052,0,0/)
        data_ped(1053,1:3)=(/1053,0,0/)
        data_ped(1054,1:3)=(/1054,0,0/)
        data_ped(1055,1:3)=(/1055,0,0/)
        data_ped(1056,1:3)=(/1056,0,0/)
        data_ped(1057,1:3)=(/1057,0,0/)
        data_ped(1058,1:3)=(/1058,0,0/)
        data_ped(1059,1:3)=(/1059,0,0/)
        data_ped(1060,1:3)=(/1060,0,0/)
        data_ped(1061,1:3)=(/1061,0,0/)
        data_ped(1062,1:3)=(/1062,0,0/)
        data_ped(1063,1:3)=(/1063,0,0/)
        data_ped(1064,1:3)=(/1064,0,0/)
        data_ped(1065,1:3)=(/1065,0,0/)
        data_ped(1066,1:3)=(/1066,0,0/)
        data_ped(1067,1:3)=(/1067,0,0/)
        data_ped(1068,1:3)=(/1068,0,0/)
        data_ped(1069,1:3)=(/1069,0,0/)
        data_ped(1070,1:3)=(/1070,0,0/)
        data_ped(1071,1:3)=(/1071,0,0/)
        data_ped(1072,1:3)=(/1072,0,0/)
        data_ped(1073,1:3)=(/1073,0,0/)
        data_ped(1074,1:3)=(/1074,0,0/)
        data_ped(1075,1:3)=(/1075,0,0/)
        data_ped(1076,1:3)=(/1076,0,0/)
        data_ped(1077,1:3)=(/1077,0,0/)
        data_ped(1078,1:3)=(/1078,0,0/)
        data_ped(1079,1:3)=(/1079,0,0/)
        data_ped(1080,1:3)=(/1080,0,0/)
        data_ped(1081,1:3)=(/1081,0,0/)
        data_ped(1082,1:3)=(/1082,0,0/)
        data_ped(1083,1:3)=(/1083,0,0/)
        data_ped(1084,1:3)=(/1084,0,0/)
        data_ped(1085,1:3)=(/1085,0,0/)
        data_ped(1086,1:3)=(/1086,0,0/)
        data_ped(1087,1:3)=(/1087,0,0/)
        data_ped(1088,1:3)=(/1088,0,0/)
        data_ped(1089,1:3)=(/1089,0,0/)
        data_ped(1090,1:3)=(/1090,0,0/)
        data_ped(1091,1:3)=(/1091,0,0/)
        data_ped(1092,1:3)=(/1092,0,0/)
        data_ped(1093,1:3)=(/1093,0,0/)
        data_ped(1094,1:3)=(/1094,0,0/)
        data_ped(1095,1:3)=(/1095,0,0/)
        data_ped(1096,1:3)=(/1096,0,0/)
        data_ped(1097,1:3)=(/1097,0,0/)
        data_ped(1098,1:3)=(/1098,0,0/)
        data_ped(1099,1:3)=(/1099,0,0/)
        data_ped(1100,1:3)=(/1100,0,0/)
        data_ped(1101,1:3)=(/1101,0,0/)
        data_ped(1102,1:3)=(/1102,0,0/)
        data_ped(1103,1:3)=(/1103,0,0/)
        data_ped(1104,1:3)=(/1104,0,0/)
        data_ped(1105,1:3)=(/1105,0,0/)
        data_ped(1106,1:3)=(/1106,0,0/)
        data_ped(1107,1:3)=(/1107,0,0/)
        data_ped(1108,1:3)=(/1108,0,0/)
        data_ped(1109,1:3)=(/1109,0,0/)
        data_ped(1110,1:3)=(/1110,0,0/)
        data_ped(1111,1:3)=(/1111,0,0/)
        data_ped(1112,1:3)=(/1112,0,0/)
        data_ped(1113,1:3)=(/1113,0,0/)
        data_ped(1114,1:3)=(/1114,0,0/)
        data_ped(1115,1:3)=(/1115,0,0/)
        data_ped(1116,1:3)=(/1116,0,0/)
        data_ped(1117,1:3)=(/1117,0,0/)
        data_ped(1118,1:3)=(/1118,0,0/)
        data_ped(1119,1:3)=(/1119,0,0/)
        data_ped(1120,1:3)=(/1120,0,0/)
        data_ped(1121,1:3)=(/1121,0,0/)
        data_ped(1122,1:3)=(/1122,0,0/)
        data_ped(1123,1:3)=(/1123,0,0/)
        data_ped(1124,1:3)=(/1124,0,0/)
        data_ped(1125,1:3)=(/1125,0,0/)
        data_ped(1126,1:3)=(/1126,0,0/)
        data_ped(1127,1:3)=(/1127,0,0/)
        data_ped(1128,1:3)=(/1128,0,0/)
        data_ped(1129,1:3)=(/1129,0,0/)
        data_ped(1130,1:3)=(/1130,0,0/)
        data_ped(1131,1:3)=(/1131,0,0/)
        data_ped(1132,1:3)=(/1132,0,0/)
        data_ped(1133,1:3)=(/1133,0,0/)
        data_ped(1134,1:3)=(/1134,0,0/)
        data_ped(1135,1:3)=(/1135,0,0/)
        data_ped(1136,1:3)=(/1136,0,0/)
        data_ped(1137,1:3)=(/1137,0,0/)
        data_ped(1138,1:3)=(/1138,0,0/)
        data_ped(1139,1:3)=(/1139,0,0/)
        data_ped(1140,1:3)=(/1140,0,0/)
        data_ped(1141,1:3)=(/1141,0,0/)
        data_ped(1142,1:3)=(/1142,0,0/)
        data_ped(1143,1:3)=(/1143,0,0/)
        data_ped(1144,1:3)=(/1144,0,0/)
        data_ped(1145,1:3)=(/1145,0,0/)
        data_ped(1146,1:3)=(/1146,0,0/)
        data_ped(1147,1:3)=(/1147,0,0/)
        data_ped(1148,1:3)=(/1148,0,0/)
        data_ped(1149,1:3)=(/1149,0,0/)
        data_ped(1150,1:3)=(/1150,0,0/)
        data_ped(1151,1:3)=(/1151,0,0/)
        data_ped(1152,1:3)=(/1152,0,0/)
        data_ped(1153,1:3)=(/1153,0,0/)
        data_ped(1154,1:3)=(/1154,0,0/)
        data_ped(1155,1:3)=(/1155,0,0/)
        data_ped(1156,1:3)=(/1156,0,0/)
        data_ped(1157,1:3)=(/1157,0,0/)
        data_ped(1158,1:3)=(/1158,0,0/)
        data_ped(1159,1:3)=(/1159,0,0/)
        data_ped(1160,1:3)=(/1160,0,0/)
        data_ped(1161,1:3)=(/1161,0,0/)
        data_ped(1162,1:3)=(/1162,0,0/)
        data_ped(1163,1:3)=(/1163,0,0/)
        data_ped(1164,1:3)=(/1164,0,0/)
        data_ped(1165,1:3)=(/1165,0,0/)
        data_ped(1166,1:3)=(/1166,0,0/)
        data_ped(1167,1:3)=(/1167,0,0/)
        data_ped(1168,1:3)=(/1168,0,0/)
        data_ped(1169,1:3)=(/1169,0,0/)
        data_ped(1170,1:3)=(/1170,0,0/)
        data_ped(1171,1:3)=(/1171,0,0/)
        data_ped(1172,1:3)=(/1172,0,0/)
        data_ped(1173,1:3)=(/1173,0,0/)
        data_ped(1174,1:3)=(/1174,0,0/)
        data_ped(1175,1:3)=(/1175,0,0/)
        data_ped(1176,1:3)=(/1176,0,0/)
        data_ped(1177,1:3)=(/1177,0,0/)
        data_ped(1178,1:3)=(/1178,0,0/)
        data_ped(1179,1:3)=(/1179,0,0/)
        data_ped(1180,1:3)=(/1180,0,0/)
        data_ped(1181,1:3)=(/1181,0,0/)
        data_ped(1182,1:3)=(/1182,0,0/)
        data_ped(1183,1:3)=(/1183,0,0/)
        data_ped(1184,1:3)=(/1184,0,0/)
        data_ped(1185,1:3)=(/1185,0,0/)
        data_ped(1186,1:3)=(/1186,0,0/)
        data_ped(1187,1:3)=(/1187,0,0/)
        data_ped(1188,1:3)=(/1188,0,0/)
        data_ped(1189,1:3)=(/1189,0,0/)
        data_ped(1190,1:3)=(/1190,0,0/)
        data_ped(1191,1:3)=(/1191,0,0/)
        data_ped(1192,1:3)=(/1192,0,0/)
        data_ped(1193,1:3)=(/1193,0,0/)
        data_ped(1194,1:3)=(/1194,0,0/)
        data_ped(1195,1:3)=(/1195,0,0/)
        data_ped(1196,1:3)=(/1196,0,0/)
        data_ped(1197,1:3)=(/1197,0,0/)
        data_ped(1198,1:3)=(/1198,0,0/)
        data_ped(1199,1:3)=(/1199,0,0/)
        data_ped(1200,1:3)=(/1200,0,0/)
        data_ped(1201,1:3)=(/1201,0,0/)
        data_ped(1202,1:3)=(/1202,0,0/)
        data_ped(1203,1:3)=(/1203,0,0/)
        data_ped(1204,1:3)=(/1204,0,0/)
        data_ped(1205,1:3)=(/1205,0,0/)
        data_ped(1206,1:3)=(/1206,0,0/)
        data_ped(1207,1:3)=(/1207,0,0/)
        data_ped(1208,1:3)=(/1208,0,0/)
        data_ped(1209,1:3)=(/1209,0,0/)
        data_ped(1210,1:3)=(/1210,0,0/)
        data_ped(1211,1:3)=(/1211,0,0/)
        data_ped(1212,1:3)=(/1212,0,0/)
        data_ped(1213,1:3)=(/1213,0,0/)
        data_ped(1214,1:3)=(/1214,0,0/)
        data_ped(1215,1:3)=(/1215,0,0/)
        data_ped(1216,1:3)=(/1216,0,0/)
        data_ped(1217,1:3)=(/1217,0,0/)
        data_ped(1218,1:3)=(/1218,0,0/)
        data_ped(1219,1:3)=(/1219,0,0/)
        data_ped(1220,1:3)=(/1220,0,0/)
        data_ped(1221,1:3)=(/1221,0,0/)
        data_ped(1222,1:3)=(/1222,0,0/)
        data_ped(1223,1:3)=(/1223,0,0/)
        data_ped(1224,1:3)=(/1224,0,0/)
        data_ped(1225,1:3)=(/1225,0,0/)
        data_ped(1226,1:3)=(/1226,0,0/)
        data_ped(1227,1:3)=(/1227,0,0/)
        data_ped(1228,1:3)=(/1228,0,0/)
        data_ped(1229,1:3)=(/1229,0,0/)
        data_ped(1230,1:3)=(/1230,0,0/)
        data_ped(1231,1:3)=(/1231,0,0/)
        data_ped(1232,1:3)=(/1232,0,0/)
        data_ped(1233,1:3)=(/1233,0,0/)
        data_ped(1234,1:3)=(/1234,0,0/)
        data_ped(1235,1:3)=(/1235,0,0/)
        data_ped(1236,1:3)=(/1236,0,0/)
        data_ped(1237,1:3)=(/1237,0,0/)
        data_ped(1238,1:3)=(/1238,0,0/)
        data_ped(1239,1:3)=(/1239,0,0/)
        data_ped(1240,1:3)=(/1240,0,0/)
        data_ped(1241,1:3)=(/1241,0,0/)
        data_ped(1242,1:3)=(/1242,0,0/)
        data_ped(1243,1:3)=(/1243,0,0/)
        data_ped(1244,1:3)=(/1244,0,0/)
        data_ped(1245,1:3)=(/1245,0,0/)
        data_ped(1246,1:3)=(/1246,0,0/)
        data_ped(1247,1:3)=(/1247,0,0/)
        data_ped(1248,1:3)=(/1248,0,0/)
        data_ped(1249,1:3)=(/1249,0,0/)
        data_ped(1250,1:3)=(/1250,0,0/)
        data_ped(1251,1:3)=(/1251,0,0/)
        data_ped(1252,1:3)=(/1252,0,0/)
        data_ped(1253,1:3)=(/1253,0,0/)
        data_ped(1254,1:3)=(/1254,0,0/)
        data_ped(1255,1:3)=(/1255,0,0/)
        data_ped(1256,1:3)=(/1256,0,0/)
        data_ped(1257,1:3)=(/1257,0,0/)
        data_ped(1258,1:3)=(/1258,0,0/)
        data_ped(1259,1:3)=(/1259,0,0/)
        data_ped(1260,1:3)=(/1260,0,0/)
        data_ped(1261,1:3)=(/1261,0,0/)
        data_ped(1262,1:3)=(/1262,0,0/)
        data_ped(1263,1:3)=(/1263,0,0/)
        data_ped(1264,1:3)=(/1264,0,0/)
        data_ped(1265,1:3)=(/1265,0,0/)
        data_ped(1266,1:3)=(/1266,0,0/)
        data_ped(1267,1:3)=(/1267,0,0/)
        data_ped(1268,1:3)=(/1268,0,0/)
        data_ped(1269,1:3)=(/1269,0,0/)
        data_ped(1270,1:3)=(/1270,0,0/)
        data_ped(1271,1:3)=(/1271,0,0/)
        data_ped(1272,1:3)=(/1272,0,0/)
        data_ped(1273,1:3)=(/1273,0,0/)
        data_ped(1274,1:3)=(/1274,0,0/)
        data_ped(1275,1:3)=(/1275,0,0/)
        data_ped(1276,1:3)=(/1276,0,0/)
        data_ped(1277,1:3)=(/1277,0,0/)
        data_ped(1278,1:3)=(/1278,0,0/)
        data_ped(1279,1:3)=(/1279,0,0/)
        data_ped(1280,1:3)=(/1280,0,0/)
        data_ped(1281,1:3)=(/1281,0,0/)
        data_ped(1282,1:3)=(/1282,0,0/)
        data_ped(1283,1:3)=(/1283,0,0/)
        data_ped(1284,1:3)=(/1284,0,0/)
        data_ped(1285,1:3)=(/1285,0,0/)
        data_ped(1286,1:3)=(/1286,0,0/)
        data_ped(1287,1:3)=(/1287,0,0/)
        data_ped(1288,1:3)=(/1288,0,0/)
        data_ped(1289,1:3)=(/1289,0,0/)
        data_ped(1290,1:3)=(/1290,0,0/)
        data_ped(1291,1:3)=(/1291,0,0/)
        data_ped(1292,1:3)=(/1292,0,0/)
        data_ped(1293,1:3)=(/1293,0,0/)
        data_ped(1294,1:3)=(/1294,0,0/)
        data_ped(1295,1:3)=(/1295,0,0/)
        data_ped(1296,1:3)=(/1296,0,0/)
        data_ped(1297,1:3)=(/1297,0,0/)
        data_ped(1298,1:3)=(/1298,0,0/)
        data_ped(1299,1:3)=(/1299,0,0/)
        data_ped(1300,1:3)=(/1300,0,0/)
        data_ped(1301,1:3)=(/1301,0,0/)
        data_ped(1302,1:3)=(/1302,0,0/)
        data_ped(1303,1:3)=(/1303,0,0/)
        data_ped(1304,1:3)=(/1304,0,0/)
        data_ped(1305,1:3)=(/1305,0,0/)
        data_ped(1306,1:3)=(/1306,0,0/)
        data_ped(1307,1:3)=(/1307,0,0/)
        data_ped(1308,1:3)=(/1308,0,0/)
        data_ped(1309,1:3)=(/1309,0,0/)
        data_ped(1310,1:3)=(/1310,0,0/)
        data_ped(1311,1:3)=(/1311,0,0/)
        data_ped(1312,1:3)=(/1312,0,0/)
        data_ped(1313,1:3)=(/1313,0,0/)
        data_ped(1314,1:3)=(/1314,0,0/)
        data_ped(1315,1:3)=(/1315,0,0/)
        data_ped(1316,1:3)=(/1316,0,0/)
        data_ped(1317,1:3)=(/1317,0,0/)
        data_ped(1318,1:3)=(/1318,0,0/)
        data_ped(1319,1:3)=(/1319,0,0/)
        data_ped(1320,1:3)=(/1320,0,0/)
        data_ped(1321,1:3)=(/1321,0,0/)
        data_ped(1322,1:3)=(/1322,0,0/)
        data_ped(1323,1:3)=(/1323,0,0/)
        data_ped(1324,1:3)=(/1324,0,0/)
        data_ped(1325,1:3)=(/1325,0,0/)
        data_ped(1326,1:3)=(/1326,0,0/)
        data_ped(1327,1:3)=(/1327,0,0/)
        data_ped(1328,1:3)=(/1328,0,0/)
        data_ped(1329,1:3)=(/1329,0,0/)
        data_ped(1330,1:3)=(/1330,0,0/)
        data_ped(1331,1:3)=(/1331,0,0/)
        data_ped(1332,1:3)=(/1332,0,0/)
        data_ped(1333,1:3)=(/1333,0,0/)
        data_ped(1334,1:3)=(/1334,0,0/)
        data_ped(1335,1:3)=(/1335,0,0/)
        data_ped(1336,1:3)=(/1336,0,0/)
        data_ped(1337,1:3)=(/1337,0,0/)
        data_ped(1338,1:3)=(/1338,0,0/)
        data_ped(1339,1:3)=(/1339,0,0/)
        data_ped(1340,1:3)=(/1340,0,0/)
        data_ped(1341,1:3)=(/1341,0,0/)
        data_ped(1342,1:3)=(/1342,0,0/)
        data_ped(1343,1:3)=(/1343,0,0/)
        data_ped(1344,1:3)=(/1344,0,0/)
        data_ped(1345,1:3)=(/1345,0,0/)
        data_ped(1346,1:3)=(/1346,0,0/)
        data_ped(1347,1:3)=(/1347,0,0/)
        data_ped(1348,1:3)=(/1348,0,0/)
        data_ped(1349,1:3)=(/1349,0,0/)
        data_ped(1350,1:3)=(/1350,0,0/)
        data_ped(1351,1:3)=(/1351,0,0/)
        data_ped(1352,1:3)=(/1352,0,0/)
        data_ped(1353,1:3)=(/1353,0,0/)
        data_ped(1354,1:3)=(/1354,0,0/)
        data_ped(1355,1:3)=(/1355,0,0/)
        data_ped(1356,1:3)=(/1356,0,0/)
        data_ped(1357,1:3)=(/1357,0,0/)
        data_ped(1358,1:3)=(/1358,0,0/)
        data_ped(1359,1:3)=(/1359,0,0/)
        data_ped(1360,1:3)=(/1360,0,0/)
        data_ped(1361,1:3)=(/1361,0,0/)
        data_ped(1362,1:3)=(/1362,0,0/)
        data_ped(1363,1:3)=(/1363,0,0/)
        data_ped(1364,1:3)=(/1364,0,0/)
        data_ped(1365,1:3)=(/1365,0,0/)
        data_ped(1366,1:3)=(/1366,0,0/)
        data_ped(1367,1:3)=(/1367,0,0/)
        data_ped(1368,1:3)=(/1368,0,0/)
        data_ped(1369,1:3)=(/1369,0,0/)
        data_ped(1370,1:3)=(/1370,0,0/)
        data_ped(1371,1:3)=(/1371,0,0/)
        data_ped(1372,1:3)=(/1372,0,0/)
        data_ped(1373,1:3)=(/1373,0,0/)
        data_ped(1374,1:3)=(/1374,0,0/)
        data_ped(1375,1:3)=(/1375,0,0/)
        data_ped(1376,1:3)=(/1376,0,0/)
        data_ped(1377,1:3)=(/1377,0,0/)
        data_ped(1378,1:3)=(/1378,0,0/)
        data_ped(1379,1:3)=(/1379,0,0/)
        data_ped(1380,1:3)=(/1380,0,0/)
        data_ped(1381,1:3)=(/1381,0,0/)
        data_ped(1382,1:3)=(/1382,0,0/)
        data_ped(1383,1:3)=(/1383,0,0/)
        data_ped(1384,1:3)=(/1384,0,0/)
        data_ped(1385,1:3)=(/1385,0,0/)
        data_ped(1386,1:3)=(/1386,0,0/)
        data_ped(1387,1:3)=(/1387,0,0/)
        data_ped(1388,1:3)=(/1388,0,0/)
        data_ped(1389,1:3)=(/1389,0,0/)
        data_ped(1390,1:3)=(/1390,0,0/)
        data_ped(1391,1:3)=(/1391,0,0/)
        data_ped(1392,1:3)=(/1392,0,0/)
        data_ped(1393,1:3)=(/1393,0,0/)
        data_ped(1394,1:3)=(/1394,0,0/)
        data_ped(1395,1:3)=(/1395,0,0/)
        data_ped(1396,1:3)=(/1396,0,0/)
        data_ped(1397,1:3)=(/1397,0,0/)
        data_ped(1398,1:3)=(/1398,0,0/)
        data_ped(1399,1:3)=(/1399,0,0/)
        data_ped(1400,1:3)=(/1400,0,0/)
        data_ped(1401,1:3)=(/1401,0,0/)
        data_ped(1402,1:3)=(/1402,0,0/)
        data_ped(1403,1:3)=(/1403,0,0/)
        data_ped(1404,1:3)=(/1404,0,0/)
        data_ped(1405,1:3)=(/1405,0,0/)
        data_ped(1406,1:3)=(/1406,0,0/)
        data_ped(1407,1:3)=(/1407,0,0/)
        data_ped(1408,1:3)=(/1408,0,0/)
        data_ped(1409,1:3)=(/1409,0,0/)
        data_ped(1410,1:3)=(/1410,0,0/)
        data_ped(1411,1:3)=(/1411,0,0/)
        data_ped(1412,1:3)=(/1412,0,0/)
        data_ped(1413,1:3)=(/1413,0,0/)
        data_ped(1414,1:3)=(/1414,0,1254/)
        data_ped(1415,1:3)=(/1415,0,206/)
        data_ped(1416,1:3)=(/1416,1170,1085/)
        data_ped(1417,1:3)=(/1417,609,5/)
        data_ped(1418,1:3)=(/1418,0,8/)
        data_ped(1419,1:3)=(/1419,308,83/)
        data_ped(1420,1:3)=(/1420,641,65/)
        data_ped(1421,1:3)=(/1421,641,0/)
        data_ped(1422,1:3)=(/1422,0,15/)
        data_ped(1423,1:3)=(/1423,522,0/)
        data_ped(1424,1:3)=(/1424,0,18/)
        data_ped(1425,1:3)=(/1425,0,27/)
        data_ped(1426,1:3)=(/1426,748,0/)
        data_ped(1427,1:3)=(/1427,0,33/)
        data_ped(1428,1:3)=(/1428,0,35/)
        data_ped(1429,1:3)=(/1429,0,38/)
        data_ped(1430,1:3)=(/1430,0,41/)
        data_ped(1431,1:3)=(/1431,0,66/)
        data_ped(1432,1:3)=(/1432,0,644/)
        data_ped(1433,1:3)=(/1433,285,80/)
        data_ped(1434,1:3)=(/1434,302,91/)
        data_ped(1435,1:3)=(/1435,298,0/)
        data_ped(1436,1:3)=(/1436,299,0/)
        data_ped(1437,1:3)=(/1437,304,0/)
        data_ped(1438,1:3)=(/1438,0,85/)
        data_ped(1439,1:3)=(/1439,291,0/)
        data_ped(1440,1:3)=(/1440,294,0/)
        data_ped(1441,1:3)=(/1441,557,88/)
        data_ped(1442,1:3)=(/1442,300,0/)
        data_ped(1443,1:3)=(/1443,301,90/)
        data_ped(1444,1:3)=(/1444,289,0/)
        data_ped(1445,1:3)=(/1445,642,0/)
        data_ped(1446,1:3)=(/1446,0,100/)
        data_ped(1447,1:3)=(/1447,323,0/)
        data_ped(1448,1:3)=(/1448,324,0/)
        data_ped(1449,1:3)=(/1449,328,0/)
        data_ped(1450,1:3)=(/1450,1168,0/)
        data_ped(1451,1:3)=(/1451,642,0/)
        data_ped(1452,1:3)=(/1452,277,113/)
        data_ped(1453,1:3)=(/1453,410,138/)
        data_ped(1454,1:3)=(/1454,321,0/)
        data_ped(1455,1:3)=(/1455,325,0/)
        data_ped(1456,1:3)=(/1456,327,0/)
        data_ped(1457,1:3)=(/1457,330,0/)
        data_ped(1458,1:3)=(/1458,332,0/)
        data_ped(1459,1:3)=(/1459,333,0/)
        data_ped(1460,1:3)=(/1460,338,0/)
        data_ped(1461,1:3)=(/1461,341,119/)
        data_ped(1462,1:3)=(/1462,335,0/)
        data_ped(1463,1:3)=(/1463,336,0/)
        data_ped(1464,1:3)=(/1464,332,1162/)
        data_ped(1465,1:3)=(/1465,1141,0/)
        data_ped(1466,1:3)=(/1466,369,121/)
        data_ped(1467,1:3)=(/1467,349,122/)
        data_ped(1468,1:3)=(/1468,372,124/)
        data_ped(1469,1:3)=(/1469,350,126/)
        data_ped(1470,1:3)=(/1470,380,123/)
        data_ped(1471,1:3)=(/1471,386,125/)
        data_ped(1472,1:3)=(/1472,353,1471/)
        data_ped(1473,1:3)=(/1473,376,126/)
        data_ped(1474,1:3)=(/1474,362,128/)
        data_ped(1475,1:3)=(/1475,349,129/)
        data_ped(1476,1:3)=(/1476,356,130/)
        data_ped(1477,1:3)=(/1477,359,132/)
        data_ped(1478,1:3)=(/1478,361,133/)
        data_ped(1479,1:3)=(/1479,358,120/)
        data_ped(1480,1:3)=(/1480,348,0/)
        data_ped(1481,1:3)=(/1481,392,0/)
        data_ped(1482,1:3)=(/1482,394,0/)
        data_ped(1483,1:3)=(/1483,1168,0/)
        data_ped(1484,1:3)=(/1484,344,0/)
        data_ped(1485,1:3)=(/1485,345,0/)
        data_ped(1486,1:3)=(/1486,347,0/)
        data_ped(1487,1:3)=(/1487,348,0/)
        data_ped(1488,1:3)=(/1488,343,0/)
        data_ped(1489,1:3)=(/1489,344,0/)
        data_ped(1490,1:3)=(/1490,398,0/)
        data_ped(1491,1:3)=(/1491,342,0/)
        data_ped(1492,1:3)=(/1492,312,0/)
        data_ped(1493,1:3)=(/1493,396,0/)
        data_ped(1494,1:3)=(/1494,403,0/)
        data_ped(1495,1:3)=(/1495,406,0/)
        data_ped(1496,1:3)=(/1496,414,0/)
        data_ped(1497,1:3)=(/1497,418,0/)
        data_ped(1498,1:3)=(/1498,406,0/)
        data_ped(1499,1:3)=(/1499,417,0/)
        data_ped(1500,1:3)=(/1500,938,0/)
        data_ped(1501,1:3)=(/1501,431,0/)
        data_ped(1502,1:3)=(/1502,432,0/)
        data_ped(1503,1:3)=(/1503,417,0/)
        data_ped(1504,1:3)=(/1504,399,0/)
        data_ped(1505,1:3)=(/1505,435,0/)
        data_ped(1506,1:3)=(/1506,436,0/)
        data_ped(1507,1:3)=(/1507,625,61/)
        data_ped(1508,1:3)=(/1508,469,0/)
        data_ped(1509,1:3)=(/1509,799,0/)
        data_ped(1510,1:3)=(/1510,282,0/)
        data_ped(1511,1:3)=(/1511,1056,0/)
        data_ped(1512,1:3)=(/1512,498,0/)
        data_ped(1513,1:3)=(/1513,445,0/)
        data_ped(1514,1:3)=(/1514,905,27/)
        data_ped(1515,1:3)=(/1515,546,0/)
        data_ped(1516,1:3)=(/1516,306,0/)
        data_ped(1517,1:3)=(/1517,0,7/)
        data_ped(1518,1:3)=(/1518,0,10/)
        data_ped(1519,1:3)=(/1519,0,17/)
        data_ped(1520,1:3)=(/1520,0,264/)
        data_ped(1521,1:3)=(/1521,276,112/)
        data_ped(1522,1:3)=(/1522,278,109/)
        data_ped(1523,1:3)=(/1523,408,0/)
        data_ped(1524,1:3)=(/1524,315,115/)
        data_ped(1525,1:3)=(/1525,298,0/)
        data_ped(1526,1:3)=(/1526,289,74/)
        data_ped(1527,1:3)=(/1527,303,0/)
        data_ped(1528,1:3)=(/1528,299,0/)
        data_ped(1529,1:3)=(/1529,295,1440/)
        data_ped(1530,1:3)=(/1530,298,84/)
        data_ped(1531,1:3)=(/1531,293,86/)
        data_ped(1532,1:3)=(/1532,302,91/)
        data_ped(1533,1:3)=(/1533,306,0/)
        data_ped(1534,1:3)=(/1534,307,82/)
        data_ped(1535,1:3)=(/1535,0,103/)
        data_ped(1536,1:3)=(/1536,326,1449/)
        data_ped(1537,1:3)=(/1537,402,0/)
        data_ped(1538,1:3)=(/1538,409,0/)
        data_ped(1539,1:3)=(/1539,415,0/)
        data_ped(1540,1:3)=(/1540,434,0/)
        data_ped(1541,1:3)=(/1541,323,1459/)
        data_ped(1542,1:3)=(/1542,1141,1493/)
        data_ped(1543,1:3)=(/1543,339,118/)
        data_ped(1544,1:3)=(/1544,396,0/)
        data_ped(1545,1:3)=(/1545,343,1484/)
        data_ped(1546,1:3)=(/1546,344,0/)
        data_ped(1547,1:3)=(/1547,1546,0/)
        data_ped(1548,1:3)=(/1548,311,0/)
        data_ped(1549,1:3)=(/1549,346,1486/)
        data_ped(1550,1:3)=(/1550,354,1476/)
        data_ped(1551,1:3)=(/1551,380,127/)
        data_ped(1552,1:3)=(/1552,400,0/)
        data_ped(1553,1:3)=(/1553,1544,0/)
        data_ped(1554,1:3)=(/1554,1552,1488/)
        data_ped(1555,1:3)=(/1555,1537,0/)
        data_ped(1556,1:3)=(/1556,397,0/)
        data_ped(1557,1:3)=(/1557,1549,1498/)
        data_ped(1558,1:3)=(/1558,906,0/)
        data_ped(1559,1:3)=(/1559,1104,0/)
        data_ped(1560,1:3)=(/1560,417,0/)
        data_ped(1561,1:3)=(/1561,425,0/)
        data_ped(1562,1:3)=(/1562,431,0/)
        data_ped(1563,1:3)=(/1563,1557,0/)
        data_ped(1564,1:3)=(/1564,1144,0/)
        data_ped(1565,1:3)=(/1565,427,0/)
        data_ped(1566,1:3)=(/1566,429,0/)
        data_ped(1567,1:3)=(/1567,0,72/)
        data_ped(1568,1:3)=(/1568,0,63/)
        data_ped(1569,1:3)=(/1569,0,64/)
        data_ped(1570,1:3)=(/1570,0,69/)
        data_ped(1571,1:3)=(/1571,0,70/)
        data_ped(1572,1:3)=(/1572,0,448/)
        data_ped(1573,1:3)=(/1573,395,0/)
        data_ped(1574,1:3)=(/1574,288,441/)
        data_ped(1575,1:3)=(/1575,290,437/)
        data_ped(1576,1:3)=(/1576,0,77/)
        data_ped(1577,1:3)=(/1577,0,20/)
        data_ped(1578,1:3)=(/1578,291,0/)
        data_ped(1579,1:3)=(/1579,405,0/)
        data_ped(1580,1:3)=(/1580,0,54/)
        data_ped(1581,1:3)=(/1581,0,76/)
        data_ped(1582,1:3)=(/1582,0,448/)
        data_ped(1583,1:3)=(/1583,292,0/)
        data_ped(1584,1:3)=(/1584,450,465/)
        data_ped(1585,1:3)=(/1585,439,0/)
        data_ped(1586,1:3)=(/1586,1570,467/)
        data_ped(1587,1:3)=(/1587,483,0/)
        data_ped(1588,1:3)=(/1588,0,461/)
        data_ped(1589,1:3)=(/1589,305,1444/)
        data_ped(1590,1:3)=(/1590,0,464/)
        data_ped(1591,1:3)=(/1591,483,0/)
        data_ped(1592,1:3)=(/1592,1571,479/)
        data_ped(1593,1:3)=(/1593,0,442/)
        data_ped(1594,1:3)=(/1594,286,1593/)
        data_ped(1595,1:3)=(/1595,1567,73/)
        data_ped(1596,1:3)=(/1596,439,438/)
        data_ped(1597,1:3)=(/1597,284,476/)
        data_ped(1598,1:3)=(/1598,439,78/)
        data_ped(1599,1:3)=(/1599,0,456/)
        data_ped(1600,1:3)=(/1600,450,71/)
        data_ped(1601,1:3)=(/1601,443,1568/)
        data_ped(1602,1:3)=(/1602,453,0/)
        data_ped(1603,1:3)=(/1603,0,460/)
        data_ped(1604,1:3)=(/1604,286,459/)
        data_ped(1605,1:3)=(/1605,1572,478/)
        data_ped(1606,1:3)=(/1606,286,0/)
        data_ped(1607,1:3)=(/1607,482,1568/)
        data_ped(1608,1:3)=(/1608,1570,474/)
        data_ped(1609,1:3)=(/1609,1571,1576/)
        data_ped(1610,1:3)=(/1610,1575,0/)
        data_ped(1611,1:3)=(/1611,0,473/)
        data_ped(1612,1:3)=(/1612,1571,475/)
        data_ped(1613,1:3)=(/1613,286,0/)
        data_ped(1614,1:3)=(/1614,1569,471/)
        data_ped(1615,1:3)=(/1615,1582,0/)
        data_ped(1616,1:3)=(/1616,1571,0/)
        data_ped(1617,1:3)=(/1617,518,0/)
        data_ped(1618,1:3)=(/1618,518,480/)
        data_ped(1619,1:3)=(/1619,0,514/)
        data_ped(1620,1:3)=(/1620,443,472/)
        data_ped(1621,1:3)=(/1621,443,0/)
        data_ped(1622,1:3)=(/1622,1575,526/)
        data_ped(1623,1:3)=(/1623,286,0/)
        data_ped(1624,1:3)=(/1624,0,486/)
        data_ped(1625,1:3)=(/1625,0,455/)
        data_ped(1626,1:3)=(/1626,294,1439/)
        data_ped(1627,1:3)=(/1627,518,477/)
        data_ped(1628,1:3)=(/1628,457,516/)
        data_ped(1629,1:3)=(/1629,489,504/)
        data_ped(1630,1:3)=(/1630,521,506/)
        data_ped(1631,1:3)=(/1631,1582,1597/)
        data_ped(1632,1:3)=(/1632,0,487/)
        data_ped(1633,1:3)=(/1633,0,505/)
        data_ped(1634,1:3)=(/1634,292,532/)
        data_ped(1635,1:3)=(/1635,1569,543/)
        data_ped(1636,1:3)=(/1636,518,0/)
        data_ped(1637,1:3)=(/1637,1581,75/)
        data_ped(1638,1:3)=(/1638,1530,0/)
        data_ped(1639,1:3)=(/1639,0,462/)
        data_ped(1640,1:3)=(/1640,1578,0/)
        data_ped(1641,1:3)=(/1641,523,0/)
        data_ped(1642,1:3)=(/1642,483,449/)
        data_ped(1643,1:3)=(/1643,0,1583/)
        data_ped(1644,1:3)=(/1644,1526,497/)
        data_ped(1645,1:3)=(/1645,0,534/)
        data_ped(1646,1:3)=(/1646,445,0/)
        data_ped(1647,1:3)=(/1647,510,0/)
        data_ped(1648,1:3)=(/1648,286,1593/)
        data_ped(1649,1:3)=(/1649,1586,531/)
        data_ped(1650,1:3)=(/1650,0,485/)
        data_ped(1651,1:3)=(/1651,488,463/)
        data_ped(1652,1:3)=(/1652,0,533/)
        data_ped(1653,1:3)=(/1653,482,528/)
        data_ped(1654,1:3)=(/1654,1591,1596/)
        data_ped(1655,1:3)=(/1655,0,550/)
        data_ped(1656,1:3)=(/1656,1609,503/)
        data_ped(1657,1:3)=(/1657,0,1587/)
        data_ped(1658,1:3)=(/1658,0,566/)
        data_ped(1659,1:3)=(/1659,483,0/)
        data_ped(1660,1:3)=(/1660,0,1617/)
        data_ped(1661,1:3)=(/1661,558,0/)
        data_ped(1662,1:3)=(/1662,0,570/)
        data_ped(1663,1:3)=(/1663,1575,507/)
        data_ped(1664,1:3)=(/1664,1578,1574/)
        data_ped(1665,1:3)=(/1665,509,495/)
        data_ped(1666,1:3)=(/1666,1598,1589/)
        data_ped(1667,1:3)=(/1667,1601,1604/)
        data_ped(1668,1:3)=(/1668,1578,1603/)
        data_ped(1669,1:3)=(/1669,1599,512/)
        data_ped(1670,1:3)=(/1670,1598,79/)
        data_ped(1671,1:3)=(/1671,518,531/)
        data_ped(1672,1:3)=(/1672,521,0/)
        data_ped(1673,1:3)=(/1673,521,506/)
        data_ped(1674,1:3)=(/1674,0,524/)
        data_ped(1675,1:3)=(/1675,558,0/)
        data_ped(1676,1:3)=(/1676,444,0/)
        data_ped(1677,1:3)=(/1677,1591,501/)
        data_ped(1678,1:3)=(/1678,520,466/)
        data_ped(1679,1:3)=(/1679,1590,0/)
        data_ped(1680,1:3)=(/1680,482,442/)
        data_ped(1681,1:3)=(/1681,520,0/)
        data_ped(1682,1:3)=(/1682,0,1611/)
        data_ped(1683,1:3)=(/1683,1590,1592/)
        data_ped(1684,1:3)=(/1684,538,0/)
        data_ped(1685,1:3)=(/1685,1570,0/)
        data_ped(1686,1:3)=(/1686,1618,1621/)
        data_ped(1687,1:3)=(/1687,1612,0/)
        data_ped(1688,1:3)=(/1688,0,1623/)
        data_ped(1689,1:3)=(/1689,520,1610/)
        data_ped(1690,1:3)=(/1690,1619,0/)
        data_ped(1691,1:3)=(/1691,521,506/)
        data_ped(1692,1:3)=(/1692,1581,0/)
        data_ped(1693,1:3)=(/1693,1579,0/)
        data_ped(1694,1:3)=(/1694,1575,0/)
        data_ped(1695,1:3)=(/1695,0,544/)
        data_ped(1696,1:3)=(/1696,1620,0/)
        data_ped(1697,1:3)=(/1697,0,559/)
        data_ped(1698,1:3)=(/1698,1579,1573/)
        data_ped(1699,1:3)=(/1699,0,58/)
        data_ped(1700,1:3)=(/1700,0,1593/)
        data_ped(1701,1:3)=(/1701,0,564/)
        data_ped(1702,1:3)=(/1702,0,567/)
        data_ped(1703,1:3)=(/1703,1612,527/)
        data_ped(1704,1:3)=(/1704,1570,1605/)
        data_ped(1705,1:3)=(/1705,518,508/)
        data_ped(1706,1:3)=(/1706,297,0/)
        data_ped(1707,1:3)=(/1707,454,0/)
        data_ped(1708,1:3)=(/1708,458,0/)
        data_ped(1709,1:3)=(/1709,1590,563/)
        data_ped(1710,1:3)=(/1710,1609,0/)
        data_ped(1711,1:3)=(/1711,1601,0/)
        data_ped(1712,1:3)=(/1712,537,0/)
        data_ped(1713,1:3)=(/1713,1609,0/)
        data_ped(1714,1:3)=(/1714,520,0/)
        data_ped(1715,1:3)=(/1715,521,0/)
        data_ped(1716,1:3)=(/1716,440,0/)
        data_ped(1717,1:3)=(/1717,1600,1615/)
        data_ped(1718,1:3)=(/1718,1600,1595/)
        data_ped(1719,1:3)=(/1719,520,0/)
        data_ped(1720,1:3)=(/1720,1526,1629/)
        data_ped(1721,1:3)=(/1721,1581,1625/)
        data_ped(1722,1:3)=(/1722,1601,1594/)
        data_ped(1723,1:3)=(/1723,0,560/)
        data_ped(1724,1:3)=(/1724,1600,0/)
        data_ped(1725,1:3)=(/1725,294,89/)
        data_ped(1726,1:3)=(/1726,1598,530/)
        data_ped(1727,1:3)=(/1727,556,565/)
        data_ped(1728,1:3)=(/1728,1599,541/)
        data_ped(1729,1:3)=(/1729,0,1630/)
        data_ped(1730,1:3)=(/1730,447,0/)
        data_ped(1731,1:3)=(/1731,446,0/)
        data_ped(1732,1:3)=(/1732,0,562/)
        data_ped(1733,1:3)=(/1733,1619,1593/)
        data_ped(1734,1:3)=(/1734,1575,551/)
        data_ped(1735,1:3)=(/1735,1607,0/)
        data_ped(1736,1:3)=(/1736,591,1633/)
        data_ped(1737,1:3)=(/1737,538,0/)
        data_ped(1738,1:3)=(/1738,440,1602/)
        data_ped(1739,1:3)=(/1739,610,0/)
        data_ped(1740,1:3)=(/1740,511,0/)
        data_ped(1741,1:3)=(/1741,490,0/)
        data_ped(1742,1:3)=(/1742,558,0/)
        data_ped(1743,1:3)=(/1743,0,569/)
        data_ped(1744,1:3)=(/1744,1600,529/)
        data_ped(1745,1:3)=(/1745,520,0/)
        data_ped(1746,1:3)=(/1746,1607,0/)
        data_ped(1747,1:3)=(/1747,1608,0/)
        data_ped(1748,1:3)=(/1748,611,0/)
        data_ped(1749,1:3)=(/1749,538,0/)
        data_ped(1750,1:3)=(/1750,411,0/)
        data_ped(1751,1:3)=(/1751,440,0/)
        data_ped(1752,1:3)=(/1752,1579,0/)
        data_ped(1753,1:3)=(/1753,1607,1632/)
        data_ped(1754,1:3)=(/1754,520,0/)
        data_ped(1755,1:3)=(/1755,1612,0/)
        data_ped(1756,1:3)=(/1756,0,1659/)
        data_ped(1757,1:3)=(/1757,1599,481/)
        data_ped(1758,1:3)=(/1758,548,0/)
        data_ped(1759,1:3)=(/1759,494,468/)
        data_ped(1760,1:3)=(/1760,1600,1658/)
        data_ped(1761,1:3)=(/1761,1575,0/)
        data_ped(1762,1:3)=(/1762,1637,0/)
        data_ped(1763,1:3)=(/1763,581,0/)
        data_ped(1764,1:3)=(/1764,1575,560/)
        data_ped(1765,1:3)=(/1765,520,0/)
        data_ped(1766,1:3)=(/1766,491,0/)
        data_ped(1767,1:3)=(/1767,0,1597/)
        data_ped(1768,1:3)=(/1768,0,51/)
        data_ped(1769,1:3)=(/1769,1601,1648/)
        data_ped(1770,1:3)=(/1770,520,1611/)
        data_ped(1771,1:3)=(/1771,536,1660/)
        data_ped(1772,1:3)=(/1772,1599,0/)
        data_ped(1773,1:3)=(/1773,1601,1604/)
        data_ped(1774,1:3)=(/1774,1575,542/)
        data_ped(1775,1:3)=(/1775,669,0/)
        data_ped(1776,1:3)=(/1776,0,1573/)
        data_ped(1777,1:3)=(/1777,1590,1636/)
        data_ped(1778,1:3)=(/1778,0,1616/)
        data_ped(1779,1:3)=(/1779,1619,0/)
        data_ped(1780,1:3)=(/1780,0,1690/)
        data_ped(1781,1:3)=(/1781,1598,575/)
        data_ped(1782,1:3)=(/1782,609,0/)
        data_ped(1783,1:3)=(/1783,521,0/)
        data_ped(1784,1:3)=(/1784,0,596/)
        data_ped(1785,1:3)=(/1785,558,0/)
        data_ped(1786,1:3)=(/1786,1609,1670/)
        data_ped(1787,1:3)=(/1787,0,619/)
        data_ped(1788,1:3)=(/1788,0,55/)
        data_ped(1789,1:3)=(/1789,0,1422/)
        data_ped(1790,1:3)=(/1790,0,586/)
        data_ped(1791,1:3)=(/1791,0,1593/)
        data_ped(1792,1:3)=(/1792,1613,635/)
        data_ped(1793,1:3)=(/1793,1579,1693/)
        data_ped(1794,1:3)=(/1794,0,603/)
        data_ped(1795,1:3)=(/1795,629,0/)
        data_ped(1796,1:3)=(/1796,581,0/)
        data_ped(1797,1:3)=(/1797,492,0/)
        data_ped(1798,1:3)=(/1798,558,0/)
        data_ped(1799,1:3)=(/1799,0,578/)
        data_ped(1800,1:3)=(/1800,0,21/)
        data_ped(1801,1:3)=(/1801,1640,0/)
        data_ped(1802,1:3)=(/1802,1575,0/)
        data_ped(1803,1:3)=(/1803,1637,0/)
        data_ped(1804,1:3)=(/1804,1653,1657/)
        data_ped(1805,1:3)=(/1805,1619,0/)
        data_ped(1806,1:3)=(/1806,0,13/)
        data_ped(1807,1:3)=(/1807,0,554/)
        data_ped(1808,1:3)=(/1808,1698,585/)
        data_ped(1809,1:3)=(/1809,1598,1589/)
        data_ped(1810,1:3)=(/1810,1652,1588/)
        data_ped(1811,1:3)=(/1811,1653,1643/)
        data_ped(1812,1:3)=(/1812,1653,540/)
        data_ped(1813,1:3)=(/1813,1586,617/)
        data_ped(1814,1:3)=(/1814,0,1594/)
        data_ped(1815,1:3)=(/1815,482,0/)
        data_ped(1816,1:3)=(/1816,458,0/)
        data_ped(1817,1:3)=(/1817,0,571/)
        data_ped(1818,1:3)=(/1818,1612,0/)
        data_ped(1819,1:3)=(/1819,1637,1688/)
        data_ped(1820,1:3)=(/1820,0,561/)
        data_ped(1821,1:3)=(/1821,662,0/)
        data_ped(1822,1:3)=(/1822,1575,1644/)
        data_ped(1823,1:3)=(/1823,1706,612/)
        data_ped(1824,1:3)=(/1824,0,1679/)
        data_ped(1825,1:3)=(/1825,1608,0/)
        data_ped(1826,1:3)=(/1826,1653,1685/)
        data_ped(1827,1:3)=(/1827,0,21/)
        data_ped(1828,1:3)=(/1828,1612,1671/)
        data_ped(1829,1:3)=(/1829,573,502/)
        data_ped(1830,1:3)=(/1830,0,632/)
        data_ped(1831,1:3)=(/1831,1637,1630/)
        data_ped(1832,1:3)=(/1832,1667,1687/)
        data_ped(1833,1:3)=(/1833,1651,1656/)
        data_ped(1834,1:3)=(/1834,1702,614/)
        data_ped(1835,1:3)=(/1835,1575,615/)
        data_ped(1836,1:3)=(/1836,1683,1686/)
        data_ped(1837,1:3)=(/1837,611,0/)
        data_ped(1838,1:3)=(/1838,1698,0/)
        data_ped(1839,1:3)=(/1839,547,0/)
        data_ped(1840,1:3)=(/1840,1698,0/)
        data_ped(1841,1:3)=(/1841,666,0/)
        data_ped(1842,1:3)=(/1842,592,0/)
        data_ped(1843,1:3)=(/1843,1682,1624/)
        data_ped(1844,1:3)=(/1844,1581,0/)
        data_ped(1845,1:3)=(/1845,592,0/)
        data_ped(1846,1:3)=(/1846,499,0/)
        data_ped(1847,1:3)=(/1847,1723,0/)
        data_ped(1848,1:3)=(/1848,470,0/)
        data_ped(1849,1:3)=(/1849,517,0/)
        data_ped(1850,1:3)=(/1850,693,0/)
        data_ped(1851,1:3)=(/1851,0,648/)
        data_ped(1852,1:3)=(/1852,0,1692/)
        data_ped(1853,1:3)=(/1853,1600,0/)
        data_ped(1854,1:3)=(/1854,519,576/)
        data_ped(1855,1:3)=(/1855,520,0/)
        data_ped(1856,1:3)=(/1856,520,496/)
        data_ped(1857,1:3)=(/1857,1726,0/)
        data_ped(1858,1:3)=(/1858,1580,143/)
        data_ped(1859,1:3)=(/1859,1620,1677/)
        data_ped(1860,1:3)=(/1860,1704,1634/)
        data_ped(1861,1:3)=(/1861,0,555/)
        data_ped(1862,1:3)=(/1862,1580,1674/)
        data_ped(1863,1:3)=(/1863,0,580/)
        data_ped(1864,1:3)=(/1864,1669,1665/)
        data_ped(1865,1:3)=(/1865,1627,595/)
        data_ped(1866,1:3)=(/1866,1667,0/)
        data_ped(1867,1:3)=(/1867,413,0/)
        data_ped(1868,1:3)=(/1868,0,1606/)
        data_ped(1869,1:3)=(/1869,573,0/)
        data_ped(1870,1:3)=(/1870,696,0/)
        data_ped(1871,1:3)=(/1871,602,0/)
        data_ped(1872,1:3)=(/1872,1729,539/)
        data_ped(1873,1:3)=(/1873,0,616/)
        data_ped(1874,1:3)=(/1874,1701,1714/)
        data_ped(1875,1:3)=(/1875,1667,594/)
        data_ped(1876,1:3)=(/1876,1653,1754/)
        data_ped(1877,1:3)=(/1877,1601,553/)
        data_ped(1878,1:3)=(/1878,0,513/)
        data_ped(1879,1:3)=(/1879,1722,1710/)
        data_ped(1880,1:3)=(/1880,1721,0/)
        data_ped(1881,1:3)=(/1881,0,1655/)
        data_ped(1882,1:3)=(/1882,523,0/)
        data_ped(1883,1:3)=(/1883,0,632/)
        data_ped(1884,1:3)=(/1884,1721,0/)
        data_ped(1885,1:3)=(/1885,694,0/)
        data_ped(1886,1:3)=(/1886,1667,672/)
        data_ped(1887,1:3)=(/1887,0,705/)
        data_ped(1888,1:3)=(/1888,1703,634/)
        data_ped(1889,1:3)=(/1889,1726,1696/)
        data_ped(1890,1:3)=(/1890,0,624/)
        data_ped(1891,1:3)=(/1891,0,741/)
        data_ped(1892,1:3)=(/1892,1612,1648/)
        data_ped(1893,1:3)=(/1893,0,727/)
        data_ped(1894,1:3)=(/1894,1708,0/)
        data_ped(1895,1:3)=(/1895,592,0/)
        data_ped(1896,1:3)=(/1896,0,733/)
        data_ped(1897,1:3)=(/1897,0,192/)
        data_ped(1898,1:3)=(/1898,1751,0/)
        data_ped(1899,1:3)=(/1899,0,691/)
        data_ped(1900,1:3)=(/1900,1726,0/)
        data_ped(1901,1:3)=(/1901,780,0/)
        data_ped(1902,1:3)=(/1902,1639,0/)
        data_ped(1903,1:3)=(/1903,558,0/)
        data_ped(1904,1:3)=(/1904,661,0/)
        data_ped(1905,1:3)=(/1905,935,0/)
        data_ped(1906,1:3)=(/1906,0,721/)
        data_ped(1907,1:3)=(/1907,1726,1663/)
        data_ped(1908,1:3)=(/1908,0,577/)
        data_ped(1909,1:3)=(/1909,1668,1614/)
        data_ped(1910,1:3)=(/1910,1586,1777/)
        data_ped(1911,1:3)=(/1911,1599,1755/)
        data_ped(1912,1:3)=(/1912,1774,672/)
        data_ped(1913,1:3)=(/1913,665,647/)
        data_ped(1914,1:3)=(/1914,776,0/)
        data_ped(1915,1:3)=(/1915,282,0/)
        data_ped(1916,1:3)=(/1916,1653,0/)
        data_ped(1917,1:3)=(/1917,1580,0/)
        data_ped(1918,1:3)=(/1918,1729,0/)
        data_ped(1919,1:3)=(/1919,0,621/)
        data_ped(1920,1:3)=(/1920,1726,575/)
        data_ped(1921,1:3)=(/1921,694,649/)
        data_ped(1922,1:3)=(/1922,1608,1642/)
        data_ped(1923,1:3)=(/1923,1667,613/)
        data_ped(1924,1:3)=(/1924,0,590/)
        data_ped(1925,1:3)=(/1925,0,656/)
        data_ped(1926,1:3)=(/1926,1654,247/)
        data_ped(1927,1:3)=(/1927,0,737/)
        data_ped(1928,1:3)=(/1928,0,689/)
        data_ped(1929,1:3)=(/1929,0,262/)
        data_ped(1930,1:3)=(/1930,1653,1744/)
        data_ped(1931,1:3)=(/1931,1639,1589/)
        data_ped(1932,1:3)=(/1932,1726,1719/)
        data_ped(1933,1:3)=(/1933,0,658/)
        data_ped(1934,1:3)=(/1934,1653,273/)
        data_ped(1935,1:3)=(/1935,0,274/)
        data_ped(1936,1:3)=(/1936,744,23/)
        data_ped(1937,1:3)=(/1937,1653,0/)
        data_ped(1938,1:3)=(/1938,1608,1711/)
        data_ped(1939,1:3)=(/1939,1725,1443/)
        data_ped(1940,1:3)=(/1940,1727,1736/)
        data_ped(1941,1:3)=(/1941,1720,0/)
        data_ped(1942,1:3)=(/1942,592,0/)
        data_ped(1943,1:3)=(/1943,493,95/)
        data_ped(1944,1:3)=(/1944,1751,0/)
        data_ped(1945,1:3)=(/1945,0,1759/)
        data_ped(1946,1:3)=(/1946,0,618/)
        data_ped(1947,1:3)=(/1947,0,605/)
        data_ped(1948,1:3)=(/1948,0,220/)
        data_ped(1949,1:3)=(/1949,549,568/)
        data_ped(1950,1:3)=(/1950,1653,674/)
        data_ped(1951,1:3)=(/1951,775,1423/)
        data_ped(1952,1:3)=(/1952,0,821/)
        data_ped(1953,1:3)=(/1953,1751,1602/)
        data_ped(1954,1:3)=(/1954,1726,1719/)
        data_ped(1955,1:3)=(/1955,0,1735/)
        data_ped(1956,1:3)=(/1956,1726,1696/)
        data_ped(1957,1:3)=(/1957,1637,0/)
        data_ped(1958,1:3)=(/1958,1639,0/)
        data_ped(1959,1:3)=(/1959,0,689/)
        data_ped(1960,1:3)=(/1960,0,676/)
        data_ped(1961,1:3)=(/1961,610,704/)
        data_ped(1962,1:3)=(/1962,1811,1671/)
        data_ped(1963,1:3)=(/1963,0,585/)
        data_ped(1964,1:3)=(/1964,493,0/)
        data_ped(1965,1:3)=(/1965,1699,1662/)
        data_ped(1966,1:3)=(/1966,749,1513/)
        data_ped(1967,1:3)=(/1967,0,653/)
        data_ped(1968,1:3)=(/1968,0,622/)
        data_ped(1969,1:3)=(/1969,0,580/)
        data_ped(1970,1:3)=(/1970,0,624/)
        data_ped(1971,1:3)=(/1971,0,623/)
        data_ped(1972,1:3)=(/1972,1773,1649/)
        data_ped(1973,1:3)=(/1973,484,34/)
        data_ped(1974,1:3)=(/1974,1612,1757/)
        data_ped(1975,1:3)=(/1975,1791,1778/)
        data_ped(1976,1:3)=(/1976,1530,1441/)
        data_ped(1977,1:3)=(/1977,802,0/)
        data_ped(1978,1:3)=(/1978,1653,1679/)
        data_ped(1979,1:3)=(/1979,749,569/)
        data_ped(1980,1:3)=(/1980,1653,678/)
        data_ped(1981,1:3)=(/1981,1653,0/)
        data_ped(1982,1:3)=(/1982,745,6/)
        data_ped(1983,1:3)=(/1983,1810,1691/)
        data_ped(1984,1:3)=(/1984,725,0/)
        data_ped(1985,1:3)=(/1985,1768,0/)
        data_ped(1986,1:3)=(/1986,744,1752/)
        data_ped(1987,1:3)=(/1987,747,0/)
        data_ped(1988,1:3)=(/1988,0,178/)
        data_ped(1989,1:3)=(/1989,0,786/)
        data_ped(1990,1:3)=(/1990,0,633/)
        data_ped(1991,1:3)=(/1991,1772,1712/)
        data_ped(1992,1:3)=(/1992,1768,0/)
        data_ped(1993,1:3)=(/1993,1726,1673/)
        data_ped(1994,1:3)=(/1994,1726,0/)
        data_ped(1995,1:3)=(/1995,1590,758/)
        data_ped(1996,1:3)=(/1996,0,598/)
        data_ped(1997,1:3)=(/1997,1637,1622/)
        data_ped(1998,1:3)=(/1998,0,718/)
        data_ped(1999,1:3)=(/1999,1668,1635/)
        data_ped(2000,1:3)=(/2000,0,606/)
        data_ped(2001,1:3)=(/2001,0,771/)
        data_ped(2002,1:3)=(/2002,582,0/)
        data_ped(2003,1:3)=(/2003,1811,0/)
        data_ped(2004,1:3)=(/2004,1726,0/)
        data_ped(2005,1:3)=(/2005,0,757/)
        data_ped(2006,1:3)=(/2006,1771,1753/)
        data_ped(2007,1:3)=(/2007,1680,1649/)
        data_ped(2008,1:3)=(/2008,0,1761/)
        data_ped(2009,1:3)=(/2009,0,683/)
        data_ped(2010,1:3)=(/2010,1599,1784/)
        data_ped(2011,1:3)=(/2011,0,249/)
        data_ped(2012,1:3)=(/2012,0,687/)
        data_ped(2013,1:3)=(/2013,0,608/)
        data_ped(2014,1:3)=(/2014,0,601/)
        data_ped(2015,1:3)=(/2015,1811,1828/)
        data_ped(2016,1:3)=(/2016,0,49/)
        data_ped(2017,1:3)=(/2017,1726,1805/)
        data_ped(2018,1:3)=(/2018,0,40/)
        data_ped(2019,1:3)=(/2019,1702,1672/)
        data_ped(2020,1:3)=(/2020,626,1756/)
        data_ped(2021,1:3)=(/2021,1726,604/)
        data_ped(2022,1:3)=(/2022,1684,0/)
        data_ped(2023,1:3)=(/2023,407,0/)
        data_ped(2024,1:3)=(/2024,701,0/)
        data_ped(2025,1:3)=(/2025,1703,0/)
        data_ped(2026,1:3)=(/2026,755,0/)
        data_ped(2027,1:3)=(/2027,0,731/)
        data_ped(2028,1:3)=(/2028,749,1646/)
        data_ped(2029,1:3)=(/2029,0,597/)
        data_ped(2030,1:3)=(/2030,749,1513/)
        data_ped(2031,1:3)=(/2031,0,572/)
        data_ped(2032,1:3)=(/2032,0,715/)
        data_ped(2033,1:3)=(/2033,0,644/)
        data_ped(2034,1:3)=(/2034,751,525/)
        data_ped(2035,1:3)=(/2035,777,0/)
        data_ped(2036,1:3)=(/2036,592,0/)
        data_ped(2037,1:3)=(/2037,641,0/)
        data_ped(2038,1:3)=(/2038,1698,0/)
        data_ped(2039,1:3)=(/2039,404,0/)
        data_ped(2040,1:3)=(/2040,1581,770/)
        data_ped(2041,1:3)=(/2041,1811,0/)
        data_ped(2042,1:3)=(/2042,1723,1689/)
        data_ped(2043,1:3)=(/2043,1724,659/)
        data_ped(2044,1:3)=(/2044,1637,1688/)
        data_ped(2045,1:3)=(/2045,1723,0/)
        data_ped(2046,1:3)=(/2046,0,692/)
        data_ped(2047,1:3)=(/2047,1813,1671/)
        data_ped(2048,1:3)=(/2048,774,1674/)
        data_ped(2049,1:3)=(/2049,0,786/)
        data_ped(2050,1:3)=(/2050,1706,0/)
        data_ped(2051,1:3)=(/2051,1653,0/)
        data_ped(2052,1:3)=(/2052,0,718/)
        data_ped(2053,1:3)=(/2053,1684,1763/)
        data_ped(2054,1:3)=(/2054,744,4/)
        data_ped(2055,1:3)=(/2055,0,251/)
        data_ped(2056,1:3)=(/2056,0,252/)
        data_ped(2057,1:3)=(/2057,0,822/)
        data_ped(2058,1:3)=(/2058,0,790/)
        data_ped(2059,1:3)=(/2059,0,688/)
        data_ped(2060,1:3)=(/2060,1637,1717/)
        data_ped(2061,1:3)=(/2061,1626,1516/)
        data_ped(2062,1:3)=(/2062,1698,645/)
        data_ped(2063,1:3)=(/2063,1667,1713/)
        data_ped(2064,1:3)=(/2064,1773,1733/)
        data_ped(2065,1:3)=(/2065,1811,1760/)
        data_ped(2066,1:3)=(/2066,754,0/)
        data_ped(2067,1:3)=(/2067,310,0/)
        data_ped(2068,1:3)=(/2068,630,0/)
        data_ped(2069,1:3)=(/2069,668,0/)
        data_ped(2070,1:3)=(/2070,751,0/)
        data_ped(2071,1:3)=(/2071,0,735/)
        data_ped(2072,1:3)=(/2072,1577,841/)
        data_ped(2073,1:3)=(/2073,0,535/)
        data_ped(2074,1:3)=(/2074,0,1800/)
        data_ped(2075,1:3)=(/2075,0,786/)
        data_ped(2076,1:3)=(/2076,0,712/)
        data_ped(2077,1:3)=(/2077,0,765/)
        data_ped(2078,1:3)=(/2078,1703,1824/)
        data_ped(2079,1:3)=(/2079,641,671/)
        data_ped(2080,1:3)=(/2080,1700,1631/)
        data_ped(2081,1:3)=(/2081,1790,1780/)
        data_ped(2082,1:3)=(/2082,1628,1782/)
        data_ped(2083,1:3)=(/2083,0,29/)
        data_ped(2084,1:3)=(/2084,695,0/)
        data_ped(2085,1:3)=(/2085,1840,0/)
        data_ped(2086,1:3)=(/2086,1832,758/)
        data_ped(2087,1:3)=(/2087,1669,0/)
        data_ped(2088,1:3)=(/2088,1816,1758/)
        data_ped(2089,1:3)=(/2089,829,0/)
        data_ped(2090,1:3)=(/2090,0,46/)
        data_ped(2091,1:3)=(/2091,879,0/)
        data_ped(2092,1:3)=(/2092,0,1844/)
        data_ped(2093,1:3)=(/2093,0,733/)
        data_ped(2094,1:3)=(/2094,750,0/)
        data_ped(2095,1:3)=(/2095,860,0/)
        data_ped(2096,1:3)=(/2096,1681,552/)
        data_ped(2097,1:3)=(/2097,803,848/)
        data_ped(2098,1:3)=(/2098,1832,0/)
        data_ped(2099,1:3)=(/2099,0,785/)
        data_ped(2100,1:3)=(/2100,0,622/)
        data_ped(2101,1:3)=(/2101,1653,1819/)
        data_ped(2102,1:3)=(/2102,1865,1843/)
        data_ped(2103,1:3)=(/2103,0,1655/)
        data_ped(2104,1:3)=(/2104,666,1641/)
        data_ped(2105,1:3)=(/2105,1640,0/)
        data_ped(2106,1:3)=(/2106,0,1647/)
        data_ped(2107,1:3)=(/2107,1702,1762/)
        data_ped(2108,1:3)=(/2108,500,781/)
        data_ped(2109,1:3)=(/2109,0,1868/)
        data_ped(2110,1:3)=(/2110,0,655/)
        data_ped(2111,1:3)=(/2111,0,739/)
        data_ped(2112,1:3)=(/2112,0,1785/)
        data_ped(2113,1:3)=(/2113,801,1883/)
        data_ped(2114,1:3)=(/2114,0,640/)
        data_ped(2115,1:3)=(/2115,0,743/)
        data_ped(2116,1:3)=(/2116,1667,1829/)
        data_ped(2117,1:3)=(/2117,1792,1589/)
        data_ped(2118,1:3)=(/2118,0,659/)
        data_ped(2119,1:3)=(/2119,0,761/)
        data_ped(2120,1:3)=(/2120,0,703/)
        data_ped(2121,1:3)=(/2121,0,631/)
        data_ped(2122,1:3)=(/2122,1769,1433/)
        data_ped(2123,1:3)=(/2123,1608,0/)
        data_ped(2124,1:3)=(/2124,695,0/)
        data_ped(2125,1:3)=(/2125,1577,0/)
        data_ped(2126,1:3)=(/2126,745,0/)
        data_ped(2127,1:3)=(/2127,668,0/)
        data_ped(2128,1:3)=(/2128,1895,0/)
        data_ped(2129,1:3)=(/2129,864,0/)
        data_ped(2130,1:3)=(/2130,0,1903/)
        data_ped(2131,1:3)=(/2131,0,1661/)
        data_ped(2132,1:3)=(/2132,0,570/)
        data_ped(2133,1:3)=(/2133,778,0/)
        data_ped(2134,1:3)=(/2134,0,545/)
        data_ped(2135,1:3)=(/2135,1720,1822/)
        data_ped(2136,1:3)=(/2136,779,892/)
        data_ped(2137,1:3)=(/2137,1721,1874/)
        data_ped(2138,1:3)=(/2138,1640,850/)
        data_ped(2139,1:3)=(/2139,1627,1877/)
        data_ped(2140,1:3)=(/2140,0,221/)
        data_ped(2141,1:3)=(/2141,1706,224/)
        data_ped(2142,1:3)=(/2142,0,1855/)
        data_ped(2143,1:3)=(/2143,779,0/)
        data_ped(2144,1:3)=(/2144,643,0/)
        data_ped(2145,1:3)=(/2145,0,836/)
        data_ped(2146,1:3)=(/2146,751,0/)
        data_ped(2147,1:3)=(/2147,753,0/)
        data_ped(2148,1:3)=(/2148,0,574/)
        data_ped(2149,1:3)=(/2149,749,0/)
        data_ped(2150,1:3)=(/2150,0,712/)
        data_ped(2151,1:3)=(/2151,1832,1878/)
        data_ped(2152,1:3)=(/2152,1653,0/)
        data_ped(2153,1:3)=(/2153,1640,0/)
        data_ped(2154,1:3)=(/2154,1892,0/)
        data_ped(2155,1:3)=(/2155,1705,1645/)
        data_ped(2156,1:3)=(/2156,1723,0/)
        data_ped(2157,1:3)=(/2157,1811,0/)
        data_ped(2158,1:3)=(/2158,1699,1662/)
        data_ped(2159,1:3)=(/2159,1830,1798/)
        data_ped(2160,1:3)=(/2160,1838,1/)
        data_ped(2161,1:3)=(/2161,0,68/)
        data_ped(2162,1:3)=(/2162,0,639/)
        data_ped(2163,1:3)=(/2163,1831,1678/)
        data_ped(2164,1:3)=(/2164,0,1647/)
        data_ped(2165,1:3)=(/2165,0,901/)
        data_ped(2166,1:3)=(/2166,0,579/)
        data_ped(2167,1:3)=(/2167,0,796/)
        data_ped(2168,1:3)=(/2168,1773,722/)
        data_ped(2169,1:3)=(/2169,1667,1718/)
        data_ped(2170,1:3)=(/2170,1807,723/)
        data_ped(2171,1:3)=(/2171,1726,1818/)
        data_ped(2172,1:3)=(/2172,626,1825/)
        data_ped(2173,1:3)=(/2173,0,515/)
        data_ped(2174,1:3)=(/2174,1894,0/)
        data_ped(2175,1:3)=(/2175,592,0/)
        data_ped(2176,1:3)=(/2176,584,0/)
        data_ped(2177,1:3)=(/2177,699,0/)
        data_ped(2178,1:3)=(/2178,413,0/)
        data_ped(2179,1:3)=(/2179,831,0/)
        data_ped(2180,1:3)=(/2180,1862,1675/)
        data_ped(2181,1:3)=(/2181,1893,1988/)
        data_ped(2182,1:3)=(/2182,0,675/)
        data_ped(2183,1:3)=(/2183,1830,1801/)
        data_ped(2184,1:3)=(/2184,1930,1873/)
        data_ped(2185,1:3)=(/2185,1830,1798/)
        data_ped(2186,1:3)=(/2186,0,1694/)
        data_ped(2187,1:3)=(/2187,1708,840/)
        data_ped(2188,1:3)=(/2188,0,835/)
        data_ped(2189,1:3)=(/2189,0,717/)
        data_ped(2190,1:3)=(/2190,670,0/)
        data_ped(2191,1:3)=(/2191,695,0/)
        data_ped(2192,1:3)=(/2192,695,0/)
        data_ped(2193,1:3)=(/2193,695,0/)
        data_ped(2194,1:3)=(/2194,865,0/)
        data_ped(2195,1:3)=(/2195,907,0/)
        data_ped(2196,1:3)=(/2196,0,1906/)
        data_ped(2197,1:3)=(/2197,0,1661/)
        data_ped(2198,1:3)=(/2198,0,1858/)
        data_ped(2199,1:3)=(/2199,931,1850/)
        data_ped(2200,1:3)=(/2200,0,784/)
        data_ped(2201,1:3)=(/2201,775,0/)
        data_ped(2202,1:3)=(/2202,932,0/)
        data_ped(2203,1:3)=(/2203,858,0/)
        data_ped(2204,1:3)=(/2204,825,0/)
        data_ped(2205,1:3)=(/2205,824,0/)
        data_ped(2206,1:3)=(/2206,0,1855/)
        data_ped(2207,1:3)=(/2207,1930,0/)
        data_ped(2208,1:3)=(/2208,0,1783/)
        data_ped(2209,1:3)=(/2209,1938,0/)
        data_ped(2210,1:3)=(/2210,0,1779/)
        data_ped(2211,1:3)=(/2211,1703,1962/)
        data_ped(2212,1:3)=(/2212,1812,0/)
        data_ped(2213,1:3)=(/2213,1668,728/)
        data_ped(2214,1:3)=(/2214,1731,1602/)
        data_ped(2215,1:3)=(/2215,0,1907/)
        data_ped(2216,1:3)=(/2216,0,98/)
        data_ped(2217,1:3)=(/2217,798,1848/)
        data_ped(2218,1:3)=(/2218,0,849/)
        data_ped(2219,1:3)=(/2219,1812,1734/)
        data_ped(2220,1:3)=(/2220,1832,1878/)
        data_ped(2221,1:3)=(/2221,1654,229/)
        data_ped(2222,1:3)=(/2222,1834,1997/)
        data_ped(2223,1:3)=(/2223,0,817/)
        data_ped(2224,1:3)=(/2224,1967,684/)
        data_ped(2225,1:3)=(/2225,1892,1709/)
        data_ped(2226,1:3)=(/2226,1862,1743/)
        data_ped(2227,1:3)=(/2227,1720,1764/)
        data_ped(2228,1:3)=(/2228,0,855/)
        data_ped(2229,1:3)=(/2229,0,821/)
        data_ped(2230,1:3)=(/2230,0,607/)
        data_ped(2231,1:3)=(/2231,1702,1765/)
        data_ped(2232,1:3)=(/2232,0,688/)
        data_ped(2233,1:3)=(/2233,0,793/)
        data_ped(2234,1:3)=(/2234,0,742/)
        data_ped(2235,1:3)=(/2235,0,263/)
        data_ped(2236,1:3)=(/2236,1731,857/)
        data_ped(2237,1:3)=(/2237,1831,1786/)
        data_ped(2238,1:3)=(/2238,1937,930/)
        data_ped(2239,1:3)=(/2239,0,837/)
        data_ped(2240,1:3)=(/2240,830,941/)
        data_ped(2241,1:3)=(/2241,0,1752/)
        data_ped(2242,1:3)=(/2242,0,673/)
        data_ped(2243,1:3)=(/2243,1530,1781/)
        data_ped(2244,1:3)=(/2244,0,104/)
        data_ped(2245,1:3)=(/2245,593,0/)
        data_ped(2246,1:3)=(/2246,826,0/)
        data_ped(2247,1:3)=(/2247,697,0/)
        data_ped(2248,1:3)=(/2248,862,0/)
        data_ped(2249,1:3)=(/2249,0,149/)
        data_ped(2250,1:3)=(/2250,1969,729/)
        data_ped(2251,1:3)=(/2251,0,587/)
        data_ped(2252,1:3)=(/2252,1924,707/)
        data_ped(2253,1:3)=(/2253,1789,1827/)
        data_ped(2254,1:3)=(/2254,0,589/)
        data_ped(2255,1:3)=(/2255,747,1510/)
        data_ped(2256,1:3)=(/2256,0,176/)
        data_ped(2257,1:3)=(/2257,1924,808/)
        data_ped(2258,1:3)=(/2258,0,764/)
        data_ped(2259,1:3)=(/2259,0,719/)
        data_ped(2260,1:3)=(/2260,0,713/)
        data_ped(2261,1:3)=(/2261,0,194/)
        data_ped(2262,1:3)=(/2262,0,846/)
        data_ped(2263,1:3)=(/2263,0,918/)
        data_ped(2264,1:3)=(/2264,1789,767/)
        data_ped(2265,1:3)=(/2265,1830,1715/)
        data_ped(2266,1:3)=(/2266,0,680/)
        data_ped(2267,1:3)=(/2267,0,1802/)
        data_ped(2268,1:3)=(/2268,0,1996/)
        data_ped(2269,1:3)=(/2269,0,1948/)
        data_ped(2270,1:3)=(/2270,0,219/)
        data_ped(2271,1:3)=(/2271,0,1803/)
        data_ped(2272,1:3)=(/2272,0,756/)
        data_ped(2273,1:3)=(/2273,1973,0/)
        data_ped(2274,1:3)=(/2274,1808,1641/)
        data_ped(2275,1:3)=(/2275,1942,0/)
        data_ped(2276,1:3)=(/2276,700,0/)
        data_ped(2277,1:3)=(/2277,642,0/)
        data_ped(2278,1:3)=(/2278,866,0/)
        data_ped(2279,1:3)=(/2279,420,0/)
        data_ped(2280,1:3)=(/2280,1966,0/)
        data_ped(2281,1:3)=(/2281,773,0/)
        data_ped(2282,1:3)=(/2282,861,0/)
        data_ped(2283,1:3)=(/2283,1830,1987/)
        data_ped(2284,1:3)=(/2284,858,0/)
        data_ped(2285,1:3)=(/2285,903,0/)
        data_ped(2286,1:3)=(/2286,773,0/)
        data_ped(2287,1:3)=(/2287,932,0/)
        data_ped(2288,1:3)=(/2288,0,782/)
        data_ped(2289,1:3)=(/2289,1891,1732/)
        data_ped(2290,1:3)=(/2290,1938,815/)
        data_ped(2291,1:3)=(/2291,1937,1886/)
        data_ped(2292,1:3)=(/2292,1698,0/)
        data_ped(2293,1:3)=(/2293,1866,0/)
        data_ped(2294,1:3)=(/2294,1773,0/)
        data_ped(2295,1:3)=(/2295,592,0/)
        data_ped(2296,1:3)=(/2296,1861,816/)
        data_ped(2297,1:3)=(/2297,1928,681/)
        data_ped(2298,1:3)=(/2298,0,796/)
        data_ped(2299,1:3)=(/2299,694,682/)
        data_ped(2300,1:3)=(/2300,0,620/)
        data_ped(2301,1:3)=(/2301,1728,1916/)
        data_ped(2302,1:3)=(/2302,641,2033/)
        data_ped(2303,1:3)=(/2303,797,12/)
        data_ped(2304,1:3)=(/2304,0,36/)
        data_ped(2305,1:3)=(/2305,0,241/)
        data_ped(2306,1:3)=(/2306,0,908/)
        data_ped(2307,1:3)=(/2307,0,246/)
        data_ped(2308,1:3)=(/2308,0,686/)
        data_ped(2309,1:3)=(/2309,0,250/)
        data_ped(2310,1:3)=(/2310,0,856/)
        data_ped(2311,1:3)=(/2311,0,606/)
        data_ped(2312,1:3)=(/2312,0,740/)
        data_ped(2313,1:3)=(/2313,1862,1766/)
        data_ped(2314,1:3)=(/2314,0,847/)
        data_ped(2315,1:3)=(/2315,2023,1716/)
        data_ped(2316,1:3)=(/2316,1892,1923/)
        data_ped(2317,1:3)=(/2317,1892,1664/)
        data_ped(2318,1:3)=(/2318,1626,1515/)
        data_ped(2319,1:3)=(/2319,1830,1961/)
        data_ped(2320,1:3)=(/2320,2023,1898/)
        data_ped(2321,1:3)=(/2321,2069,216/)
        data_ped(2322,1:3)=(/2322,609,0/)
        data_ped(2323,1:3)=(/2323,752,515/)
        data_ped(2324,1:3)=(/2324,2023,0/)
        data_ped(2325,1:3)=(/2325,1894,0/)
        data_ped(2326,1:3)=(/2326,832,0/)
        data_ped(2327,1:3)=(/2327,984,2091/)
        data_ped(2328,1:3)=(/2328,0,888/)
        data_ped(2329,1:3)=(/2329,0,805/)
        data_ped(2330,1:3)=(/2330,1789,606/)
        data_ped(2331,1:3)=(/2331,0,783/)
        data_ped(2332,1:3)=(/2332,592,0/)
        data_ped(2333,1:3)=(/2333,641,1794/)
        data_ped(2334,1:3)=(/2334,0,709/)
        data_ped(2335,1:3)=(/2335,0,637/)
        data_ped(2336,1:3)=(/2336,0,891/)
        data_ped(2337,1:3)=(/2337,0,915/)
        data_ped(2338,1:3)=(/2338,0,652/)
        data_ped(2339,1:3)=(/2339,0,710/)
        data_ped(2340,1:3)=(/2340,1699,1851/)
        data_ped(2341,1:3)=(/2341,0,185/)
        data_ped(2342,1:3)=(/2342,0,193/)
        data_ped(2343,1:3)=(/2343,2025,925/)
        data_ped(2344,1:3)=(/2344,952,208/)
        data_ped(2345,1:3)=(/2345,0,814/)
        data_ped(2346,1:3)=(/2346,0,813/)
        data_ped(2347,1:3)=(/2347,1728,1876/)
        data_ped(2348,1:3)=(/2348,0,714/)
        data_ped(2349,1:3)=(/2349,1776,0/)
        data_ped(2350,1:3)=(/2350,322,0/)
        data_ped(2351,1:3)=(/2351,0,24/)
        data_ped(2352,1:3)=(/2352,583,0/)
        data_ped(2353,1:3)=(/2353,609,911/)
        data_ped(2354,1:3)=(/2354,2023,52/)
        data_ped(2355,1:3)=(/2355,0,56/)
        data_ped(2356,1:3)=(/2356,862,892/)
        data_ped(2357,1:3)=(/2357,0,706/)
        data_ped(2358,1:3)=(/2358,2023,1738/)
        data_ped(2359,1:3)=(/2359,0,1785/)
        data_ped(2360,1:3)=(/2360,879,0/)
        data_ped(2361,1:3)=(/2361,958,0/)
        data_ped(2362,1:3)=(/2362,0,720/)
        data_ped(2363,1:3)=(/2363,1866,1899/)
        data_ped(2364,1:3)=(/2364,2015,0/)
        data_ped(2365,1:3)=(/2365,1703,0/)
        data_ped(2366,1:3)=(/2366,0,599/)
        data_ped(2367,1:3)=(/2367,0,787/)
        data_ped(2368,1:3)=(/2368,1812,920/)
        data_ped(2369,1:3)=(/2369,0,875/)
        data_ped(2370,1:3)=(/2370,1943,0/)
        data_ped(2371,1:3)=(/2371,1640,1852/)
        data_ped(2372,1:3)=(/2372,1938,0/)
        data_ped(2373,1:3)=(/2373,2065,0/)
        data_ped(2374,1:3)=(/2374,1930,922/)
        data_ped(2375,1:3)=(/2375,2010,1910/)
        data_ped(2376,1:3)=(/2376,0,256/)
        data_ped(2377,1:3)=(/2377,309,1913/)
        data_ped(2378,1:3)=(/2378,2021,948/)
        data_ped(2379,1:3)=(/2379,1831,685/)
        data_ped(2380,1:3)=(/2380,1972,14/)
        data_ped(2381,1:3)=(/2381,0,1002/)
        data_ped(2382,1:3)=(/2382,0,654/)
        data_ped(2383,1:3)=(/2383,0,1918/)
        data_ped(2384,1:3)=(/2384,0,899/)
        data_ped(2385,1:3)=(/2385,0,255/)
        data_ped(2386,1:3)=(/2386,0,600/)
        data_ped(2387,1:3)=(/2387,2021,1663/)
        data_ped(2388,1:3)=(/2388,1866,1888/)
        data_ped(2389,1:3)=(/2389,0,1884/)
        data_ped(2390,1:3)=(/2390,1940,1860/)
        data_ped(2391,1:3)=(/2391,0,2051/)
        data_ped(2392,1:3)=(/2392,667,0/)
        data_ped(2393,1:3)=(/2393,1936,0/)
        data_ped(2394,1:3)=(/2394,0,766/)
        data_ped(2395,1:3)=(/2395,0,912/)
        data_ped(2396,1:3)=(/2396,1891,730/)
        data_ped(2397,1:3)=(/2397,0,162/)
        data_ped(2398,1:3)=(/2398,957,1508/)
        data_ped(2399,1:3)=(/2399,0,834/)
        data_ped(2400,1:3)=(/2400,0,588/)
        data_ped(2401,1:3)=(/2401,1807,870/)
        data_ped(2402,1:3)=(/2402,0,890/)
        data_ped(2403,1:3)=(/2403,1830,871/)
        data_ped(2404,1:3)=(/2404,1789,1946/)
        data_ped(2405,1:3)=(/2405,0,708/)
        data_ped(2406,1:3)=(/2406,0,676/)
        data_ped(2407,1:3)=(/2407,0,171/)
        data_ped(2408,1:3)=(/2408,1830,1961/)
        data_ped(2409,1:3)=(/2409,0,872/)
        data_ped(2410,1:3)=(/2410,0,638/)
        data_ped(2411,1:3)=(/2411,859,183/)
        data_ped(2412,1:3)=(/2412,0,679/)
        data_ped(2413,1:3)=(/2413,983,196/)
        data_ped(2414,1:3)=(/2414,0,810/)
        data_ped(2415,1:3)=(/2415,2015,1824/)
        data_ped(2416,1:3)=(/2416,0,1948/)
        data_ped(2417,1:3)=(/2417,1703,1823/)
        data_ped(2418,1:3)=(/2418,0,2080/)
        data_ped(2419,1:3)=(/2419,0,874/)
        data_ped(2420,1:3)=(/2420,1931,1886/)
        data_ped(2421,1:3)=(/2421,2127,867/)
        data_ped(2422,1:3)=(/2422,0,941/)
        data_ped(2423,1:3)=(/2423,939,646/)
        data_ped(2424,1:3)=(/2424,936,0/)
        data_ped(2425,1:3)=(/2425,956,0/)
        data_ped(2426,1:3)=(/2426,0,967/)
        data_ped(2427,1:3)=(/2427,663,0/)
        data_ped(2428,1:3)=(/2428,664,0/)
        data_ped(2429,1:3)=(/2429,1707,0/)
        data_ped(2430,1:3)=(/2430,751,0/)
        data_ped(2431,1:3)=(/2431,936,0/)
        data_ped(2432,1:3)=(/2432,2023,0/)
        data_ped(2433,1:3)=(/2433,433,0/)
        data_ped(2434,1:3)=(/2434,2119,0/)
        data_ped(2435,1:3)=(/2435,1944,1846/)
        data_ped(2436,1:3)=(/2436,1832,0/)
        data_ped(2437,1:3)=(/2437,0,1018/)
        data_ped(2438,1:3)=(/2438,2176,1/)
        data_ped(2439,1:3)=(/2439,1973,0/)
        data_ped(2440,1:3)=(/2440,0,942/)
        data_ped(2441,1:3)=(/2441,0,2086/)
        data_ped(2442,1:3)=(/2442,0,1919/)
        data_ped(2443,1:3)=(/2443,1720,2101/)
        data_ped(2444,1:3)=(/2444,0,1857/)
        data_ped(2445,1:3)=(/2445,1833,1999/)
        data_ped(2446,1:3)=(/2446,1891,1915/)
        data_ped(2447,1:3)=(/2447,1978,1784/)
        data_ped(2448,1:3)=(/2448,1934,925/)
        data_ped(2449,1:3)=(/2449,2161,0/)
        data_ped(2450,1:3)=(/2450,1770,876/)
        data_ped(2451,1:3)=(/2451,749,819/)
        data_ped(2452,1:3)=(/2452,0,772/)
        data_ped(2453,1:3)=(/2453,1789,877/)
        data_ped(2454,1:3)=(/2454,0,900/)
        data_ped(2455,1:3)=(/2455,0,657/)
        data_ped(2456,1:3)=(/2456,2116,1962/)
        data_ped(2457,1:3)=(/2457,0,1745/)
        data_ped(2458,1:3)=(/2458,0,1747/)
        data_ped(2459,1:3)=(/2459,2021,1805/)
        data_ped(2460,1:3)=(/2460,1776,990/)
        data_ped(2461,1:3)=(/2461,1933,1900/)
        data_ped(2462,1:3)=(/2462,2176,1841/)
        data_ped(2463,1:3)=(/2463,0,1869/)
        data_ped(2464,1:3)=(/2464,1808,2085/)
        data_ped(2465,1:3)=(/2465,2128,0/)
        data_ped(2466,1:3)=(/2466,2176,909/)
        data_ped(2467,1:3)=(/2467,884,0/)
        data_ped(2468,1:3)=(/2468,987,0/)
        data_ped(2469,1:3)=(/2469,2108,0/)
        data_ped(2470,1:3)=(/2470,0,1739/)
        data_ped(2471,1:3)=(/2471,2013,977/)
        data_ped(2472,1:3)=(/2472,1924,636/)
        data_ped(2473,1:3)=(/2473,0,761/)
        data_ped(2474,1:3)=(/2474,0,651/)
        data_ped(2475,1:3)=(/2475,1970,873/)
        data_ped(2476,1:3)=(/2476,0,1989/)
        data_ped(2477,1:3)=(/2477,0,186/)
        data_ped(2478,1:3)=(/2478,660,1797/)
        data_ped(2479,1:3)=(/2479,1038,0/)
        data_ped(2480,1:3)=(/2480,0,762/)
        data_ped(2481,1:3)=(/2481,1830,1821/)
        data_ped(2482,1:3)=(/2482,0,677/)
        data_ped(2483,1:3)=(/2483,0,184/)
        data_ped(2484,1:3)=(/2484,0,916/)
        data_ped(2485,1:3)=(/2485,0,2073/)
        data_ped(2486,1:3)=(/2486,0,190/)
        data_ped(2487,1:3)=(/2487,931,1850/)
        data_ped(2488,1:3)=(/2488,0,768/)
        data_ped(2489,1:3)=(/2489,2017,811/)
        data_ped(2490,1:3)=(/2490,1703,1954/)
        data_ped(2491,1:3)=(/2491,1941,2208/)
        data_ped(2492,1:3)=(/2492,0,215/)
        data_ped(2493,1:3)=(/2493,1932,926/)
        data_ped(2494,1:3)=(/2494,641,67/)
        data_ped(2495,1:3)=(/2495,2054,25/)
        data_ped(2496,1:3)=(/2496,2116,999/)
        data_ped(2497,1:3)=(/2497,751,868/)
        data_ped(2498,1:3)=(/2498,1973,2147/)
        data_ped(2499,1:3)=(/2499,2119,0/)
        data_ped(2500,1:3)=(/2500,2022,0/)
        data_ped(2501,1:3)=(/2501,2023,0/)
        data_ped(2502,1:3)=(/2502,1748,1737/)
        data_ped(2503,1:3)=(/2503,627,0/)
        data_ped(2504,1:3)=(/2504,2022,0/)
        data_ped(2505,1:3)=(/2505,747,1742/)
        data_ped(2506,1:3)=(/2506,0,2049/)
        data_ped(2507,1:3)=(/2507,2015,1029/)
        data_ped(2508,1:3)=(/2508,1004,2204/)
        data_ped(2509,1:3)=(/2509,1830,0/)
        data_ped(2510,1:3)=(/2510,2119,794/)
        data_ped(2511,1:3)=(/2511,0,1995/)
        data_ped(2512,1:3)=(/2512,1730,0/)
        data_ped(2513,1:3)=(/2513,0,1829/)
        data_ped(2514,1:3)=(/2514,2060,997/)
        data_ped(2515,1:3)=(/2515,0,1881/)
        data_ped(2516,1:3)=(/2516,0,48/)
        data_ped(2517,1:3)=(/2517,1974,0/)
        data_ped(2518,1:3)=(/2518,661,0/)
        data_ped(2519,1:3)=(/2519,1861,1919/)
        data_ped(2520,1:3)=(/2520,1773,1909/)
        data_ped(2521,1:3)=(/2521,1934,974/)
        data_ped(2522,1:3)=(/2522,0,2155/)
        data_ped(2523,1:3)=(/2523,1833,2102/)
        data_ped(2524,1:3)=(/2524,0,1000/)
        data_ped(2525,1:3)=(/2525,2106,1979/)
        data_ped(2526,1:3)=(/2526,0,852/)
        data_ped(2527,1:3)=(/2527,0,719/)
        data_ped(2528,1:3)=(/2528,2117,234/)
        data_ped(2529,1:3)=(/2529,2016,2216/)
        data_ped(2530,1:3)=(/2530,1933,2156/)
        data_ped(2531,1:3)=(/2531,2161,2037/)
        data_ped(2532,1:3)=(/2532,0,39/)
        data_ped(2533,1:3)=(/2533,1654,244/)
        data_ped(2534,1:3)=(/2534,0,253/)
        data_ped(2535,1:3)=(/2535,0,738/)
        data_ped(2536,1:3)=(/2536,0,2004/)
        data_ped(2537,1:3)=(/2537,1654,257/)
        data_ped(2538,1:3)=(/2538,1730,1964/)
        data_ped(2539,1:3)=(/2539,0,1746/)
        data_ped(2540,1:3)=(/2540,1892,795/)
        data_ped(2541,1:3)=(/2541,0,2050/)
        data_ped(2542,1:3)=(/2542,0,269/)
        data_ped(2543,1:3)=(/2543,1748,2036/)
        data_ped(2544,1:3)=(/2544,1038,867/)
        data_ped(2545,1:3)=(/2545,1650,2035/)
        data_ped(2546,1:3)=(/2546,627,1655/)
        data_ped(2547,1:3)=(/2547,1831,1872/)
        data_ped(2548,1:3)=(/2548,1809,199/)
        data_ped(2549,1:3)=(/2549,1834,1991/)
        data_ped(2550,1:3)=(/2550,2169,1920/)
        data_ped(2551,1:3)=(/2551,1942,650/)
        data_ped(2552,1:3)=(/2552,2062,0/)
        data_ped(2553,1:3)=(/2553,310,0/)
        data_ped(2554,1:3)=(/2554,412,0/)
        data_ped(2555,1:3)=(/2555,2116,0/)
        data_ped(2556,1:3)=(/2556,882,0/)
        data_ped(2557,1:3)=(/2557,1892,0/)
        data_ped(2558,1:3)=(/2558,0,1046/)
        data_ped(2559,1:3)=(/2559,1970,1695/)
        data_ped(2560,1:3)=(/2560,0,759/)
        data_ped(2561,1:3)=(/2561,0,151/)
        data_ped(2562,1:3)=(/2562,2019,760/)
        data_ped(2563,1:3)=(/2563,0,156/)
        data_ped(2564,1:3)=(/2564,2019,792/)
        data_ped(2565,1:3)=(/2565,0,1421/)
        data_ped(2566,1:3)=(/2566,954,1509/)
        data_ped(2567,1:3)=(/2567,1008,1741/)
        data_ped(2568,1:3)=(/2568,1970,656/)
        data_ped(2569,1:3)=(/2569,0,2029/)
        data_ped(2570,1:3)=(/2570,2059,820/)
        data_ped(2571,1:3)=(/2571,2009,809/)
        data_ped(2572,1:3)=(/2572,1970,845/)
        data_ped(2573,1:3)=(/2573,1057,2285/)
        data_ped(2574,1:3)=(/2574,0,189/)
        data_ped(2575,1:3)=(/2575,1008,1676/)
        data_ped(2576,1:3)=(/2576,2113,2251/)
        data_ped(2577,1:3)=(/2577,746,1426/)
        data_ped(2578,1:3)=(/2578,0,812/)
        data_ped(2579,1:3)=(/2579,1721,1853/)
        data_ped(2580,1:3)=(/2580,1723,1876/)
        data_ped(2581,1:3)=(/2581,1932,769/)
        data_ped(2582,1:3)=(/2582,2109,1073/)
        data_ped(2583,1:3)=(/2583,2056,225/)
        data_ped(2584,1:3)=(/2584,2161,702/)
        data_ped(2585,1:3)=(/2585,2054,1982/)
        data_ped(2586,1:3)=(/2586,641,62/)
        data_ped(2587,1:3)=(/2587,1838,988/)
        data_ped(2588,1:3)=(/2588,1059,0/)
        data_ped(2589,1:3)=(/2589,827,0/)
        data_ped(2590,1:3)=(/2590,881,0/)
        data_ped(2591,1:3)=(/2591,985,0/)
        data_ped(2592,1:3)=(/2592,1894,2195/)
        data_ped(2593,1:3)=(/2593,1040,0/)
        data_ped(2594,1:3)=(/2594,1892,0/)
        data_ped(2595,1:3)=(/2595,724,0/)
        data_ped(2596,1:3)=(/2596,0,47/)
        data_ped(2597,1:3)=(/2597,862,1984/)
        data_ped(2598,1:3)=(/2598,0,1062/)
        data_ped(2599,1:3)=(/2599,980,0/)
        data_ped(2600,1:3)=(/2600,979,0/)
        data_ped(2601,1:3)=(/2601,862,0/)
        data_ped(2602,1:3)=(/2602,2064,1990/)
        data_ped(2603,1:3)=(/2603,1750,0/)
        data_ped(2604,1:3)=(/2604,0,966/)
        data_ped(2605,1:3)=(/2605,2017,1019/)
        data_ped(2606,1:3)=(/2606,1943,993/)
        data_ped(2607,1:3)=(/2607,2236,0/)
        data_ped(2608,1:3)=(/2608,2022,885/)
        data_ped(2609,1:3)=(/2609,1835,1804/)
        data_ped(2610,1:3)=(/2610,2236,2143/)
        data_ped(2611,1:3)=(/2611,2315,0/)
        data_ped(2612,1:3)=(/2612,1942,726/)
        data_ped(2613,1:3)=(/2613,1006,226/)
        data_ped(2614,1:3)=(/2614,2116,2211/)
        data_ped(2615,1:3)=(/2615,1972,1911/)
        data_ped(2616,1:3)=(/2616,1967,1958/)
        data_ped(2617,1:3)=(/2617,0,924/)
        data_ped(2618,1:3)=(/2618,0,1917/)
        data_ped(2619,1:3)=(/2619,2008,2003/)
        data_ped(2620,1:3)=(/2620,0,2219/)
        data_ped(2621,1:3)=(/2621,0,1885/)
        data_ped(2622,1:3)=(/2622,2010,2220/)
        data_ped(2623,1:3)=(/2623,1767,1075/)
        data_ped(2624,1:3)=(/2624,2116,1922/)
        data_ped(2625,1:3)=(/2625,1942,1796/)
        data_ped(2626,1:3)=(/2626,1943,1795/)
        data_ped(2627,1:3)=(/2627,1650,2083/)
        data_ped(2628,1:3)=(/2628,1891,0/)
        data_ped(2629,1:3)=(/2629,1895,0/)
        data_ped(2630,1:3)=(/2630,828,1462/)
        data_ped(2631,1:3)=(/2631,2054,0/)
        data_ped(2632,1:3)=(/2632,1011,0/)
        data_ped(2633,1:3)=(/2633,1789,2076/)
        data_ped(2634,1:3)=(/2634,0,1847/)
        data_ped(2635,1:3)=(/2635,1924,2099/)
        data_ped(2636,1:3)=(/2636,2233,2250/)
        data_ped(2637,1:3)=(/2637,0,163/)
        data_ped(2638,1:3)=(/2638,0,1068/)
        data_ped(2639,1:3)=(/2639,1928,913/)
        data_ped(2640,1:3)=(/2640,2057,927/)
        data_ped(2641,1:3)=(/2641,0,167/)
        data_ped(2642,1:3)=(/2642,0,172/)
        data_ped(2643,1:3)=(/2643,980,1849/)
        data_ped(2644,1:3)=(/2644,0,175/)
        data_ped(2645,1:3)=(/2645,0,180/)
        data_ped(2646,1:3)=(/2646,2059,1091/)
        data_ped(2647,1:3)=(/2647,0,1919/)
        data_ped(2648,1:3)=(/2648,0,2027/)
        data_ped(2649,1:3)=(/2649,0,732/)
        data_ped(2650,1:3)=(/2650,1863,711/)
        data_ped(2651,1:3)=(/2651,0,188/)
        data_ped(2652,1:3)=(/2652,800,2094/)
        data_ped(2653,1:3)=(/2653,0,2286/)
        data_ped(2654,1:3)=(/2654,982,2287/)
        data_ped(2655,1:3)=(/2655,0,1512/)
        data_ped(2656,1:3)=(/2656,0,996/)
        data_ped(2657,1:3)=(/2657,981,2095/)
        data_ped(2658,1:3)=(/2658,0,947/)
        data_ped(2659,1:3)=(/2659,1863,202/)
        data_ped(2660,1:3)=(/2660,2013,2041/)
        data_ped(2661,1:3)=(/2661,2064,2219/)
        data_ped(2662,1:3)=(/2662,2019,2135/)
        data_ped(2663,1:3)=(/2663,2017,851/)
        data_ped(2664,1:3)=(/2664,1931,2045/)
        data_ped(2665,1:3)=(/2665,2063,2210/)
        data_ped(2666,1:3)=(/2666,2124,989/)
        data_ped(2667,1:3)=(/2667,2022,1842/)
        data_ped(2668,1:3)=(/2668,2066,0/)
        data_ped(2669,1:3)=(/2669,0,42/)
        data_ped(2670,1:3)=(/2670,1698,2354/)
        data_ped(2671,1:3)=(/2671,883,0/)
        data_ped(2672,1:3)=(/2672,2066,0/)
        data_ped(2673,1:3)=(/2673,1059,0/)
        data_ped(2674,1:3)=(/2674,804,886/)
        data_ped(2675,1:3)=(/2675,310,0/)
        data_ped(2676,1:3)=(/2676,826,0/)
        data_ped(2677,1:3)=(/2677,960,0/)
        data_ped(2678,1:3)=(/2678,2175,0/)
        data_ped(2679,1:3)=(/2679,628,0/)
        data_ped(2680,1:3)=(/2680,1815,0/)
        data_ped(2681,1:3)=(/2681,0,887/)
        data_ped(2682,1:3)=(/2682,880,0/)
        data_ped(2683,1:3)=(/2683,774,0/)
        data_ped(2684,1:3)=(/2684,1789,844/)
        data_ped(2685,1:3)=(/2685,2124,1901/)
        data_ped(2686,1:3)=(/2686,318,0/)
        data_ped(2687,1:3)=(/2687,1078,0/)
        data_ped(2688,1:3)=(/2688,1939,991/)
        data_ped(2689,1:3)=(/2689,1972,0/)
        data_ped(2690,1:3)=(/2690,1789,0/)
        data_ped(2691,1:3)=(/2691,2013,1955/)
        data_ped(2692,1:3)=(/2692,1943,970/)
        data_ped(2693,1:3)=(/2693,2123,0/)
        data_ped(2694,1:3)=(/2694,1834,2087/)
        data_ped(2695,1:3)=(/2695,2168,0/)
        data_ped(2696,1:3)=(/2696,2062,1087/)
        data_ped(2697,1:3)=(/2697,0,1064/)
        data_ped(2698,1:3)=(/2698,2108,0/)
        data_ped(2699,1:3)=(/2699,0,716/)
        data_ped(2700,1:3)=(/2700,1927,2001/)
        data_ped(2701,1:3)=(/2701,1975,1856/)
        data_ped(2702,1:3)=(/2702,1924,1695/)
        data_ped(2703,1:3)=(/2703,1892,230/)
        data_ped(2704,1:3)=(/2704,1970,897/)
        data_ped(2705,1:3)=(/2705,1862,1914/)
        data_ped(2706,1:3)=(/2706,0,1880/)
        data_ped(2707,1:3)=(/2707,2017,1945/)
        data_ped(2708,1:3)=(/2708,0,1858/)
        data_ped(2709,1:3)=(/2709,2064,2291/)
        data_ped(2710,1:3)=(/2710,1807,789/)
        data_ped(2711,1:3)=(/2711,0,2216/)
        data_ped(2712,1:3)=(/2712,1929,1055/)
        data_ped(2713,1:3)=(/2713,0,794/)
        data_ped(2714,1:3)=(/2714,1833,2005/)
        data_ped(2715,1:3)=(/2715,0,2294/)
        data_ped(2716,1:3)=(/2716,0,2157/)
        data_ped(2717,1:3)=(/2717,0,2214/)
        data_ped(2718,1:3)=(/2718,1840,0/)
        data_ped(2719,1:3)=(/2719,2062,0/)
        data_ped(2720,1:3)=(/2720,2245,0/)
        data_ped(2721,1:3)=(/2721,2324,0/)
        data_ped(2722,1:3)=(/2722,1730,1870/)
        data_ped(2723,1:3)=(/2723,1010,0/)
        data_ped(2724,1:3)=(/2724,2108,0/)
        data_ped(2725,1:3)=(/2725,1107,0/)
        data_ped(2726,1:3)=(/2726,2009,2027/)
        data_ped(2727,1:3)=(/2727,1965,2259/)
        data_ped(2728,1:3)=(/2728,2112,2280/)
        data_ped(2729,1:3)=(/2729,2169,1950/)
        data_ped(2730,1:3)=(/2730,1861,1919/)
        data_ped(2731,1:3)=(/2731,0,944/)
        data_ped(2732,1:3)=(/2732,1968,842/)
        data_ped(2733,1:3)=(/2733,2058,2150/)
        data_ped(2734,1:3)=(/2734,0,157/)
        data_ped(2735,1:3)=(/2735,2165,1799/)
        data_ped(2736,1:3)=(/2736,0,1998/)
        data_ped(2737,1:3)=(/2737,0,173/)
        data_ped(2738,1:3)=(/2738,1862,2283/)
        data_ped(2739,1:3)=(/2739,0,2202/)
        data_ped(2740,1:3)=(/2740,0,763/)
        data_ped(2741,1:3)=(/2741,1926,182/)
        data_ped(2742,1:3)=(/2742,1862,2289/)
        data_ped(2743,1:3)=(/2743,2233,734/)
        data_ped(2744,1:3)=(/2744,2163,902/)
        data_ped(2745,1:3)=(/2745,2011,204/)
        data_ped(2746,1:3)=(/2746,2235,1028/)
        data_ped(2747,1:3)=(/2747,0,1030/)
        data_ped(2748,1:3)=(/2748,0,1093/)
        data_ped(2749,1:3)=(/2749,2237,1956/)
        data_ped(2750,1:3)=(/2750,0,2209/)
        data_ped(2751,1:3)=(/2751,2161,11/)
        data_ped(2752,1:3)=(/2752,1007,0/)
        data_ped(2753,1:3)=(/2753,2054,1982/)
        data_ped(2754,1:3)=(/2754,2066,1/)
        data_ped(2755,1:3)=(/2755,1125,57/)
        data_ped(2756,1:3)=(/2756,0,1062/)
        data_ped(2757,1:3)=(/2757,2067,2349/)
        data_ped(2758,1:3)=(/2758,2022,2295/)
        data_ped(2759,1:3)=(/2759,2066,0/)
        data_ped(2760,1:3)=(/2760,2460,988/)
        data_ped(2761,1:3)=(/2761,1839,0/)
        data_ped(2762,1:3)=(/2762,986,0/)
        data_ped(2763,1:3)=(/2763,1124,0/)
        data_ped(2764,1:3)=(/2764,0,1017/)
        data_ped(2765,1:3)=(/2765,0,1981/)
        data_ped(2766,1:3)=(/2766,1894,0/)
        data_ped(2767,1:3)=(/2767,1698,0/)
        data_ped(2768,1:3)=(/2768,2320,2188/)
        data_ped(2769,1:3)=(/2769,1892,1887/)
        data_ped(2770,1:3)=(/2770,2303,2333/)
        data_ped(2771,1:3)=(/2771,1971,894/)
        data_ped(2772,1:3)=(/2772,1861,1908/)
        data_ped(2773,1:3)=(/2773,2237,1879/)
        data_ped(2774,1:3)=(/2774,2171,2078/)
        data_ped(2775,1:3)=(/2775,2062,2104/)
        data_ped(2776,1:3)=(/2776,1972,237/)
        data_ped(2777,1:3)=(/2777,2019,975/)
        data_ped(2778,1:3)=(/2778,2469,2133/)
        data_ped(2779,1:3)=(/2779,0,242/)
        data_ped(2780,1:3)=(/2780,2221,243/)
        data_ped(2781,1:3)=(/2781,2309,245/)
        data_ped(2782,1:3)=(/2782,1862,2183/)
        data_ped(2783,1:3)=(/2783,978,2048/)
        data_ped(2784,1:3)=(/2784,2052,1035/)
        data_ped(2785,1:3)=(/2785,2124,950/)
        data_ped(2786,1:3)=(/2786,0,2254/)
        data_ped(2787,1:3)=(/2787,1891,929/)
        data_ped(2788,1:3)=(/2788,1830,878/)
        data_ped(2789,1:3)=(/2789,0,2268/)
        data_ped(2790,1:3)=(/2790,0,1963/)
        data_ped(2791,1:3)=(/2791,2116,1922/)
        data_ped(2792,1:3)=(/2792,2064,1101/)
        data_ped(2793,1:3)=(/2793,2316,2007/)
        data_ped(2794,1:3)=(/2794,2242,1921/)
        data_ped(2795,1:3)=(/2795,1943,951/)
        data_ped(2796,1:3)=(/2796,1076,836/)
        data_ped(2797,1:3)=(/2797,663,1445/)
        data_ped(2798,1:3)=(/2798,1977,0/)
        data_ped(2799,1:3)=(/2799,833,0/)
        data_ped(2800,1:3)=(/2800,2247,0/)
        data_ped(2801,1:3)=(/2801,2167,1817/)
        data_ped(2802,1:3)=(/2802,2298,2405/)
        data_ped(2803,1:3)=(/2803,1938,2343/)
        data_ped(2804,1:3)=(/2804,1830,2505/)
        data_ped(2805,1:3)=(/2805,2107,2251/)
        data_ped(2806,1:3)=(/2806,0,889/)
        data_ped(2807,1:3)=(/2807,2014,2200/)
        data_ped(2808,1:3)=(/2808,2385,2261/)
        data_ped(2809,1:3)=(/2809,0,1047/)
        data_ped(2810,1:3)=(/2810,1928,807/)
        data_ped(2811,1:3)=(/2811,1965,2000/)
        data_ped(2812,1:3)=(/2812,2112,2030/)
        data_ped(2813,1:3)=(/2813,1830,2149/)
        data_ped(2814,1:3)=(/2814,2112,2509/)
        data_ped(2815,1:3)=(/2815,0,1069/)
        data_ped(2816,1:3)=(/2816,2300,976/)
        data_ped(2817,1:3)=(/2817,0,876/)
        data_ped(2818,1:3)=(/2818,1830,2203/)
        data_ped(2819,1:3)=(/2819,746,2205/)
        data_ped(2820,1:3)=(/2820,2223,2074/)
        data_ped(2821,1:3)=(/2821,0,2346/)
        data_ped(2822,1:3)=(/2822,2019,1092/)
        data_ped(2823,1:3)=(/2823,2124,2435/)
        data_ped(2824,1:3)=(/2824,0,2032/)
        data_ped(2825,1:3)=(/2825,0,207/)
        data_ped(2826,1:3)=(/2826,2325,0/)
        data_ped(2827,1:3)=(/2827,1837,1109/)
        data_ped(2828,1:3)=(/2828,1943,222/)
        data_ped(2829,1:3)=(/2829,0,2427/)
        data_ped(2830,1:3)=(/2830,1942,2504/)
        data_ped(2831,1:3)=(/2831,2068,2136/)
        data_ped(2832,1:3)=(/2832,2118,0/)
        data_ped(2833,1:3)=(/2833,2237,2040/)
        data_ped(2834,1:3)=(/2834,0,796/)
        data_ped(2835,1:3)=(/2835,0,53/)
        data_ped(2836,1:3)=(/2836,1943,839/)
        data_ped(2837,1:3)=(/2837,1723,0/)
        data_ped(2838,1:3)=(/2838,0,2347/)
        data_ped(2839,1:3)=(/2839,2308,2516/)
        data_ped(2840,1:3)=(/2840,0,2212/)
        data_ped(2841,1:3)=(/2841,2065,2154/)
        data_ped(2842,1:3)=(/2842,2062,2438/)
        data_ped(2843,1:3)=(/2843,0,818/)
        data_ped(2844,1:3)=(/2844,0,2031/)
        data_ped(2845,1:3)=(/2845,1698,2214/)
        data_ped(2846,1:3)=(/2846,2456,236/)
        data_ped(2847,1:3)=(/2847,0,1043/)
        data_ped(2848,1:3)=(/2848,1966,1961/)
        data_ped(2849,1:3)=(/2849,1965,2253/)
        data_ped(2850,1:3)=(/2850,0,174/)
        data_ped(2851,1:3)=(/2851,2237,2138/)
        data_ped(2852,1:3)=(/2852,933,2217/)
        data_ped(2853,1:3)=(/2853,2237,2218/)
        data_ped(2854,1:3)=(/2854,2237,2291/)
        data_ped(2855,1:3)=(/2855,2013,1037/)
        data_ped(2856,1:3)=(/2856,1106,45/)
        data_ped(2857,1:3)=(/2857,2237,1993/)
        data_ped(2858,1:3)=(/2858,2161,1806/)
        data_ped(2859,1:3)=(/2859,2245,1044/)
        data_ped(2860,1:3)=(/2860,1837,2190/)
        data_ped(2861,1:3)=(/2861,2067,2349/)
        data_ped(2862,1:3)=(/2862,1840,2423/)
        data_ped(2863,1:3)=(/2863,2070,2039/)
        data_ped(2864,1:3)=(/2864,1078,0/)
        data_ped(2865,1:3)=(/2865,1932,1066/)
        data_ped(2866,1:3)=(/2866,0,2473/)
        data_ped(2867,1:3)=(/2867,2057,2258/)
        data_ped(2868,1:3)=(/2868,2456,943/)
        data_ped(2869,1:3)=(/2869,2221,2596/)
        data_ped(2870,1:3)=(/2870,0,1115/)
        data_ped(2871,1:3)=(/2871,0,155/)
        data_ped(2872,1:3)=(/2872,955,2281/)
        data_ped(2873,1:3)=(/2873,1158,2358/)
        data_ped(2874,1:3)=(/2874,1140,2508/)
        data_ped(2875,1:3)=(/2875,982,1740/)
        data_ped(2876,1:3)=(/2876,1924,1024/)
        data_ped(2877,1:3)=(/2877,1889,2185/)
        data_ped(2878,1:3)=(/2878,2234,2264/)
        data_ped(2879,1:3)=(/2879,823,1149/)
        data_ped(2880,1:3)=(/2880,1970,2337/)
        data_ped(2881,1:3)=(/2881,0,768/)
        data_ped(2882,1:3)=(/2882,0,1096/)
        data_ped(2883,1:3)=(/2883,0,174/)
        data_ped(2884,1:3)=(/2884,0,1026/)
        data_ped(2885,1:3)=(/2885,1928,843/)
        data_ped(2886,1:3)=(/2886,0,690/)
        data_ped(2887,1:3)=(/2887,2008,2328/)
        data_ped(2888,1:3)=(/2888,2170,2394/)
        data_ped(2889,1:3)=(/2889,0,919/)
        data_ped(2890,1:3)=(/2890,0,1116/)
        data_ped(2891,1:3)=(/2891,0,209/)
        data_ped(2892,1:3)=(/2892,0,1071/)
        data_ped(2893,1:3)=(/2893,0,2098/)
        data_ped(2894,1:3)=(/2894,2238,1072/)
        data_ped(2895,1:3)=(/2895,0,1095/)
        data_ped(2896,1:3)=(/2896,0,222/)
        data_ped(2897,1:3)=(/2897,904,0/)
        data_ped(2898,1:3)=(/2898,2054,96/)
        data_ped(2899,1:3)=(/2899,2120,941/)
        data_ped(2900,1:3)=(/2900,0,1145/)
        data_ped(2901,1:3)=(/2901,0,2582/)
        data_ped(2902,1:3)=(/2902,1840,0/)
        data_ped(2903,1:3)=(/2903,985,911/)
        data_ped(2904,1:3)=(/2904,2015,0/)
        data_ped(2905,1:3)=(/2905,1750,0/)
        data_ped(2906,1:3)=(/2906,2322,0/)
        data_ped(2907,1:3)=(/2907,1698,910/)
        data_ped(2908,1:3)=(/2908,2054,205/)
        data_ped(2909,1:3)=(/2909,2320,0/)
        data_ped(2910,1:3)=(/2910,1837,2350/)
        data_ped(2911,1:3)=(/2911,1749,2332/)
        data_ped(2912,1:3)=(/2912,2054,1752/)
        data_ped(2913,1:3)=(/2913,0,1044/)
        data_ped(2914,1:3)=(/2914,961,0/)
        data_ped(2915,1:3)=(/2915,962,0/)
        data_ped(2916,1:3)=(/2916,863,0/)
        data_ped(2917,1:3)=(/2917,2462,0/)
        data_ped(2918,1:3)=(/2918,1698,838/)
        data_ped(2919,1:3)=(/2919,1102,0/)
        data_ped(2920,1:3)=(/2920,2064,0/)
        data_ped(2921,1:3)=(/2921,2168,0/)
        data_ped(2922,1:3)=(/2922,2379,2206/)
        data_ped(2923,1:3)=(/2923,1836,1983/)
        data_ped(2924,1:3)=(/2924,1942,2500/)
        data_ped(2925,1:3)=(/2925,2239,2499/)
        data_ped(2926,1:3)=(/2926,2174,2501/)
        data_ped(2927,1:3)=(/2927,0,28/)
        data_ped(2928,1:3)=(/2928,2315,2610/)
        data_ped(2929,1:3)=(/2929,2322,0/)
        data_ped(2930,1:3)=(/2930,1041,0/)
        data_ped(2931,1:3)=(/2931,1775,0/)
        data_ped(2932,1:3)=(/2932,2052,2252/)
        data_ped(2933,1:3)=(/2933,2115,2437/)
        data_ped(2934,1:3)=(/2934,2158,1120/)
        data_ped(2935,1:3)=(/2935,2223,2367/)
        data_ped(2936,1:3)=(/2936,2316,2047/)
        data_ped(2937,1:3)=(/2937,0,788/)
        data_ped(2938,1:3)=(/2938,0,1053/)
        data_ped(2939,1:3)=(/2939,2321,1020/)
        data_ped(2940,1:3)=(/2940,0,2142/)
        data_ped(2941,1:3)=(/2941,2321,1013/)
        data_ped(2942,1:3)=(/2942,2303,2333/)
        data_ped(2943,1:3)=(/2943,2534,2256/)
        data_ped(2944,1:3)=(/2944,2012,1959/)
        data_ped(2945,1:3)=(/2945,2300,1960/)
        data_ped(2946,1:3)=(/2946,2300,2073/)
        data_ped(2947,1:3)=(/2947,774,1904/)
        data_ped(2948,1:3)=(/2948,0,1097/)
        data_ped(2949,1:3)=(/2949,933,2440/)
        data_ped(2950,1:3)=(/2950,2163,2371/)
        data_ped(2951,1:3)=(/2951,2379,2345/)
        data_ped(2952,1:3)=(/2952,2116,2006/)
        data_ped(2953,1:3)=(/2953,2464,2276/)
        data_ped(2954,1:3)=(/2954,2322,2495/)
        data_ped(2955,1:3)=(/2955,2175,2424/)
        data_ped(2956,1:3)=(/2956,700,992/)
        data_ped(2957,1:3)=(/2957,2544,1088/)
        data_ped(2958,1:3)=(/2958,1009,0/)
        data_ped(2959,1:3)=(/2959,2066,0/)
        data_ped(2960,1:3)=(/2960,1080,0/)
        data_ped(2961,1:3)=(/2961,698,0/)
        data_ped(2962,1:3)=(/2962,2326,0/)
        data_ped(2963,1:3)=(/2963,0,1045/)
        data_ped(2964,1:3)=(/2964,1861,1902/)
        data_ped(2965,1:3)=(/2965,2054,1986/)
        data_ped(2966,1:3)=(/2966,1965,2253/)
        data_ped(2967,1:3)=(/2967,1767,153/)
        data_ped(2968,1:3)=(/2968,1768,1820/)
        data_ped(2969,1:3)=(/2969,2112,2028/)
        data_ped(2970,1:3)=(/2970,1787,1998/)
        data_ped(2971,1:3)=(/2971,0,1877/)
        data_ped(2972,1:3)=(/2972,979,2478/)
        data_ped(2973,1:3)=(/2973,982,2199/)
        data_ped(2974,1:3)=(/2974,2232,806/)
        data_ped(2975,1:3)=(/2975,2009,168/)
        data_ped(2976,1:3)=(/2976,0,998/)
        data_ped(2977,1:3)=(/2977,1039,2434/)
        data_ped(2978,1:3)=(/2978,1830,1951/)
        data_ped(2979,1:3)=(/2979,1058,1905/)
        data_ped(2980,1:3)=(/2980,2309,177/)
        data_ped(2981,1:3)=(/2981,2547,2213/)
        data_ped(2982,1:3)=(/2982,2169,2003/)
        data_ped(2983,1:3)=(/2983,0,2338/)
        data_ped(2984,1:3)=(/2984,1938,200/)
        data_ped(2985,1:3)=(/2985,1157,203/)
        data_ped(2986,1:3)=(/2986,749,2183/)
        data_ped(2987,1:3)=(/2987,2115,1152/)
        data_ped(2988,1:3)=(/2988,2238,2207/)
        data_ped(2989,1:3)=(/2989,1926,218/)
        data_ped(2990,1:3)=(/2990,1159,2604/)
        data_ped(2991,1:3)=(/2991,0,1015/)
        data_ped(2992,1:3)=(/2992,2552,0/)
        data_ped(2993,1:3)=(/2993,643,0/)
        data_ped(2994,1:3)=(/2994,2120,2422/)
        data_ped(2995,1:3)=(/2995,319,1146/)
        data_ped(2996,1:3)=(/2996,2315,965/)
        data_ped(2997,1:3)=(/2997,0,1086/)
        data_ped(2998,1:3)=(/2998,985,1982/)
        data_ped(2999,1:3)=(/2999,2320,2192/)
        data_ped(3000,1:3)=(/3000,319,1063/)
        data_ped(3001,1:3)=(/3001,0,2606/)
        data_ped(3002,1:3)=(/3002,1973,1428/)
        data_ped(3003,1:3)=(/3003,2320,2498/)
        data_ped(3004,1:3)=(/3004,1973,2358/)
        data_ped(3005,1:3)=(/3005,1698,0/)
        data_ped(3006,1:3)=(/3006,2175,0/)
        data_ped(3007,1:3)=(/3007,310,0/)
        data_ped(3008,1:3)=(/3008,2247,0/)
        data_ped(3009,1:3)=(/3009,1012,0/)
        data_ped(3010,1:3)=(/3010,2015,81/)
        data_ped(3011,1:3)=(/3011,934,0/)
        data_ped(3012,1:3)=(/3012,2462,2696/)
        data_ped(3013,1:3)=(/3013,0,2141/)
        data_ped(3014,1:3)=(/3014,2624,2365/)
        data_ped(3015,1:3)=(/3015,2462,0/)
        data_ped(3016,1:3)=(/3016,2117,973/)
        data_ped(3017,1:3)=(/3017,2296,1034/)
        data_ped(3018,1:3)=(/3018,1968,2515/)
        data_ped(3019,1:3)=(/3019,1965,2330/)
        data_ped(3020,1:3)=(/3020,2549,2609/)
        data_ped(3021,1:3)=(/3021,1972,231/)
        data_ped(3022,1:3)=(/3022,2116,2517/)
        data_ped(3023,1:3)=(/3023,0,2347/)
        data_ped(3024,1:3)=(/3024,2161,2565/)
        data_ped(3025,1:3)=(/3025,2054,1417/)
        data_ped(3026,1:3)=(/3026,2303,2586/)
        data_ped(3027,1:3)=(/3027,0,791/)
        data_ped(3028,1:3)=(/3028,2169,1920/)
        data_ped(3029,1:3)=(/3029,749,2518/)
        data_ped(3030,1:3)=(/3030,1770,2184/)
        data_ped(3031,1:3)=(/3031,1139,2049/)
        data_ped(3032,1:3)=(/3032,2467,2611/)
        data_ped(3033,1:3)=(/3033,0,265/)
        data_ped(3034,1:3)=(/3034,1811,272/)
        data_ped(3035,1:3)=(/3035,2724,2698/)
        data_ped(3036,1:3)=(/3036,2230,2480/)
        data_ped(3037,1:3)=(/3037,2067,2085/)
        data_ped(3038,1:3)=(/3038,2315,2084/)
        data_ped(3039,1:3)=(/3039,1943,0/)
        data_ped(3040,1:3)=(/3040,1895,0/)
        data_ped(3041,1:3)=(/3041,2462,0/)
        data_ped(3042,1:3)=(/3042,2467,0/)
        data_ped(3043,1:3)=(/3043,2231,1980/)
        data_ped(3044,1:3)=(/3044,1930,2137/)
        data_ped(3045,1:3)=(/3045,2298,2569/)
        data_ped(3046,1:3)=(/3046,2225,148/)
        data_ped(3047,1:3)=(/3047,1862,2183/)
        data_ped(3048,1:3)=(/3048,2055,152/)
        data_ped(3049,1:3)=(/3049,2315,2597/)
        data_ped(3050,1:3)=(/3050,2379,1994/)
        data_ped(3051,1:3)=(/3051,1974,161/)
        data_ped(3052,1:3)=(/3052,2707,2359/)
        data_ped(3053,1:3)=(/3053,1862,1904/)
        data_ped(3054,1:3)=(/3054,2321,1128/)
        data_ped(3055,1:3)=(/3055,2164,2361/)
        data_ped(3056,1:3)=(/3056,2309,170/)
        data_ped(3057,1:3)=(/3057,2381,853/)
        data_ped(3058,1:3)=(/3058,1970,1952/)
        data_ped(3059,1:3)=(/3059,2297,914/)
        data_ped(3060,1:3)=(/3060,2167,2485/)
        data_ped(3061,1:3)=(/3061,2386,2733/)
        data_ped(3062,1:3)=(/3062,2550,2594/)
        data_ped(3063,1:3)=(/3063,2172,1114/)
        data_ped(3064,1:3)=(/3064,2166,1100/)
        data_ped(3065,1:3)=(/3065,2628,2510/)
        data_ped(3066,1:3)=(/3066,0,972/)
        data_ped(3067,1:3)=(/3067,2319,2599/)
        data_ped(3068,1:3)=(/3068,2172,1119/)
        data_ped(3069,1:3)=(/3069,2459,1151/)
        data_ped(3070,1:3)=(/3070,0,1953/)
        data_ped(3071,1:3)=(/3071,0,1957/)
        data_ped(3072,1:3)=(/3072,2380,1118/)
        data_ped(3073,1:3)=(/3073,2463,2513/)
        data_ped(3074,1:3)=(/3074,2523,1153/)
        data_ped(3075,1:3)=(/3075,2237,2043/)
        data_ped(3076,1:3)=(/3076,1889,2289/)
        data_ped(3077,1:3)=(/3077,1840,0/)
        data_ped(3078,1:3)=(/3078,937,1016/)
        data_ped(3079,1:3)=(/3079,2245,0/)
        data_ped(3080,1:3)=(/3080,1698,2214/)
        data_ped(3081,1:3)=(/3081,2160,0/)
        data_ped(3082,1:3)=(/3082,2160,0/)
        data_ped(3083,1:3)=(/3083,1942,1796/)
        data_ped(3084,1:3)=(/3084,2790,0/)
        data_ped(3085,1:3)=(/3085,2174,0/)
        data_ped(3086,1:3)=(/3086,1808,1175/)
        data_ped(3087,1:3)=(/3087,280,0/)
        data_ped(3088,1:3)=(/3088,2790,0/)
        data_ped(3089,1:3)=(/3089,2315,2610/)
        data_ped(3090,1:3)=(/3090,1105,2762/)
        data_ped(3091,1:3)=(/3091,0,2561/)
        data_ped(3092,1:3)=(/3092,0,2146/)
        data_ped(3093,1:3)=(/3093,749,2284/)
        data_ped(3094,1:3)=(/3094,2320,2498/)
        data_ped(3095,1:3)=(/3095,0,2144/)
        data_ped(3096,1:3)=(/3096,2020,2490/)
        data_ped(3097,1:3)=(/3097,2707,2408/)
        data_ped(3098,1:3)=(/3098,2448,2662/)
        data_ped(3099,1:3)=(/3099,2311,1049/)
        data_ped(3100,1:3)=(/3100,2457,2482/)
        data_ped(3101,1:3)=(/3101,0,1121/)
        data_ped(3102,1:3)=(/3102,1965,2406/)
        data_ped(3103,1:3)=(/3103,0,896/)
        data_ped(3104,1:3)=(/3104,2457,2140/)
        data_ped(3105,1:3)=(/3105,2447,2689/)
        data_ped(3106,1:3)=(/3106,2456,2006/)
        data_ped(3107,1:3)=(/3107,2311,2260/)
        data_ped(3108,1:3)=(/3108,2444,898/)
        data_ped(3109,1:3)=(/3109,2303,2494/)
        data_ped(3110,1:3)=(/3110,1969,2410/)
        data_ped(3111,1:3)=(/3111,1864,1859/)
        data_ped(3112,1:3)=(/3112,1167,2481/)
        data_ped(3113,1:3)=(/3113,2309,1122/)
        data_ped(3114,1:3)=(/3114,2307,248/)
        data_ped(3115,1:3)=(/3115,2169,2215/)
        data_ped(3116,1:3)=(/3116,0,949/)
        data_ped(3117,1:3)=(/3117,2714,2496/)
        data_ped(3118,1:3)=(/3118,0,268/)
        data_ped(3119,1:3)=(/3119,0,2479/)
        data_ped(3120,1:3)=(/3120,0,909/)
        data_ped(3121,1:3)=(/3121,1775,2757/)
        data_ped(3122,1:3)=(/3122,2465,2676/)
        data_ped(3123,1:3)=(/3123,2462,0/)
        data_ped(3124,1:3)=(/3124,2062,2759/)
        data_ped(3125,1:3)=(/3125,2244,0/)
        data_ped(3126,1:3)=(/3126,1169,0/)
        data_ped(3127,1:3)=(/3127,2326,2279/)
        data_ped(3128,1:3)=(/3128,2523,2135/)
        data_ped(3129,1:3)=(/3129,0,2580/)
        data_ped(3130,1:3)=(/3130,1965,854/)
        data_ped(3131,1:3)=(/3131,0,147/)
        data_ped(3132,1:3)=(/3132,0,2640/)
        data_ped(3133,1:3)=(/3133,2232,1111/)
        data_ped(3134,1:3)=(/3134,2619,2744/)
        data_ped(3135,1:3)=(/3135,1890,2650/)
        data_ped(3136,1:3)=(/3136,2701,1117/)
        data_ped(3137,1:3)=(/3137,1830,2282/)
        data_ped(3138,1:3)=(/3138,1861,2335/)
        data_ped(3139,1:3)=(/3139,1861,2093/)
        data_ped(3140,1:3)=(/3140,0,2342/)
        data_ped(3141,1:3)=(/3141,2711,2808/)
        data_ped(3142,1:3)=(/3142,1924,2570/)
        data_ped(3143,1:3)=(/3143,2229,994/)
        data_ped(3144,1:3)=(/3144,2106,1113/)
        data_ped(3145,1:3)=(/3145,2778,0/)
        data_ped(3146,1:3)=(/3146,2170,971/)
        data_ped(3147,1:3)=(/3147,2009,2074/)
        data_ped(3148,1:3)=(/3148,2384,1137/)
        data_ped(3149,1:3)=(/3149,0,191/)
        data_ped(3150,1:3)=(/3150,2383,2514/)
        data_ped(3151,1:3)=(/3151,2116,921/)
        data_ped(3152,1:3)=(/3152,0,2489/)
        data_ped(3153,1:3)=(/3153,2315,2601/)
        data_ped(3154,1:3)=(/3154,1078,210/)
        data_ped(3155,1:3)=(/3155,2556,2671/)
        data_ped(3156,1:3)=(/3156,2466,2598/)
        data_ped(3157,1:3)=(/3157,2320,1022/)
        data_ped(3158,1:3)=(/3158,317,1063/)
        data_ped(3159,1:3)=(/3159,0,2034/)
        data_ped(3160,1:3)=(/3160,2320,2673/)
        data_ped(3161,1:3)=(/3161,1169,2424/)
        data_ped(3162,1:3)=(/3162,0,885/)
        data_ped(3163,1:3)=(/3163,985,0/)
        data_ped(3164,1:3)=(/3164,2244,1985/)
        data_ped(3165,1:3)=(/3165,1082,1820/)
        data_ped(3166,1:3)=(/3166,1707,2428/)
        data_ped(3167,1:3)=(/3167,2067,2349/)
        data_ped(3168,1:3)=(/3168,2315,0/)
        data_ped(3169,1:3)=(/3169,2127,0/)
        data_ped(3170,1:3)=(/3170,2175,0/)
        data_ped(3171,1:3)=(/3171,2315,0/)
        data_ped(3172,1:3)=(/3172,0,1112/)
        data_ped(3173,1:3)=(/3173,2320,2588/)
        data_ped(3174,1:3)=(/3174,1189,1160/)
        data_ped(3175,1:3)=(/3175,2112,2028/)
        data_ped(3176,1:3)=(/3176,2444,2031/)
        data_ped(3177,1:3)=(/3177,2443,2343/)
        data_ped(3178,1:3)=(/3178,2158,2366/)
        data_ped(3179,1:3)=(/3179,2528,2765/)
        data_ped(3180,1:3)=(/3180,1965,1050/)
        data_ped(3181,1:3)=(/3181,2389,2660/)
        data_ped(3182,1:3)=(/3182,2067,1052/)
        data_ped(3183,1:3)=(/3183,2172,2419/)
        data_ped(3184,1:3)=(/3184,2309,238/)
        data_ped(3185,1:3)=(/3185,2312,2558/)
        data_ped(3186,1:3)=(/3186,1862,2197/)
        data_ped(3187,1:3)=(/3187,2303,1420/)
        data_ped(3188,1:3)=(/3188,2531,60/)
        data_ped(3189,1:3)=(/3189,2320,2610/)
        data_ped(3190,1:3)=(/3190,2121,2002/)
        data_ped(3191,1:3)=(/3191,937,1074/)
        data_ped(3192,1:3)=(/3192,2557,2372/)
        data_ped(3193,1:3)=(/3193,2550,2293/)
        data_ped(3194,1:3)=(/3194,0,2576/)
        data_ped(3195,1:3)=(/3195,2624,258/)
        data_ped(3196,1:3)=(/3196,2544,2752/)
        data_ped(3197,1:3)=(/3197,1943,2585/)
        data_ped(3198,1:3)=(/3198,2315,2685/)
        data_ped(3199,1:3)=(/3199,0,50/)
        data_ped(3200,1:3)=(/3200,1808,1184/)
        data_ped(3201,1:3)=(/3201,1943,2495/)
        data_ped(3202,1:3)=(/3202,2315,0/)
        data_ped(3203,1:3)=(/3203,419,0/)
        data_ped(3204,1:3)=(/3204,2062,0/)
        data_ped(3205,1:3)=(/3205,1793,2088/)
        data_ped(3206,1:3)=(/3206,1103,0/)
        data_ped(3207,1:3)=(/3207,1123,0/)
        data_ped(3208,1:3)=(/3208,1171,0/)
        data_ped(3209,1:3)=(/3209,2245,2089/)
        data_ped(3210,1:3)=(/3210,2775,0/)
        data_ped(3211,1:3)=(/3211,2008,2328/)
        data_ped(3212,1:3)=(/3212,2297,2394/)
        data_ped(3213,1:3)=(/3213,2300,2073/)
        data_ped(3214,1:3)=(/3214,2009,2472/)
        data_ped(3215,1:3)=(/3215,2309,2641/)
        data_ped(3216,1:3)=(/3216,2114,1150/)
        data_ped(3217,1:3)=(/3217,2709,2328/)
        data_ped(3218,1:3)=(/3218,2378,2562/)
        data_ped(3219,1:3)=(/3219,2009,2027/)
        data_ped(3220,1:3)=(/3220,2444,2148/)
        data_ped(3221,1:3)=(/3221,2052,2357/)
        data_ped(3222,1:3)=(/3222,2314,2198/)
        data_ped(3223,1:3)=(/3223,1861,1896/)
        data_ped(3224,1:3)=(/3224,2623,2397/)
        data_ped(3225,1:3)=(/3225,2116,2364/)
        data_ped(3226,1:3)=(/3226,1931,1003/)
        data_ped(3227,1:3)=(/3227,2779,164/)
        data_ped(3228,1:3)=(/3228,2315,2685/)
        data_ped(3229,1:3)=(/3229,1891,2201/)
        data_ped(3230,1:3)=(/3230,2786,1202/)
        data_ped(3231,1:3)=(/3231,1862,2728/)
        data_ped(3232,1:3)=(/3232,2772,1135/)
        data_ped(3233,1:3)=(/3233,2116,181/)
        data_ped(3234,1:3)=(/3234,2158,2362/)
        data_ped(3235,1:3)=(/3235,1862,1961/)
        data_ped(3236,1:3)=(/3236,2388,1875/)
        data_ped(3237,1:3)=(/3237,2301,2184/)
        data_ped(3238,1:3)=(/3238,2230,141/)
        data_ped(3239,1:3)=(/3239,0,917/)
        data_ped(3240,1:3)=(/3240,1892,2080/)
        data_ped(3241,1:3)=(/3241,0,2097/)
        data_ped(3242,1:3)=(/3242,2448,2747/)
        data_ped(3243,1:3)=(/3243,2540,1098/)
        data_ped(3244,1:3)=(/3244,2709,1031/)
        data_ped(3245,1:3)=(/3245,0,1177/)
        data_ped(3246,1:3)=(/3246,0,211/)
        data_ped(3247,1:3)=(/3247,2387,2833/)
        data_ped(3248,1:3)=(/3248,0,2151/)
        data_ped(3249,1:3)=(/3249,1943,217/)
        data_ped(3250,1:3)=(/3250,2457,2803/)
        data_ped(3251,1:3)=(/3251,2625,2911/)
        data_ped(3252,1:3)=(/3252,963,0/)
        data_ped(3253,1:3)=(/3253,0,964/)
        data_ped(3254,1:3)=(/3254,2320,2588/)
        data_ped(3255,1:3)=(/3255,2790,2686/)
        data_ped(3256,1:3)=(/3256,2315,2136/)
        data_ped(3257,1:3)=(/3257,2464,1161/)
        data_ped(3258,1:3)=(/3258,2775,0/)
        data_ped(3259,1:3)=(/3259,0,909/)
        data_ped(3260,1:3)=(/3260,1077,2430/)
        data_ped(3261,1:3)=(/3261,1169,0/)
        data_ped(3262,1:3)=(/3262,2798,0/)
        data_ped(3263,1:3)=(/3263,1141,0/)
        data_ped(3264,1:3)=(/3264,2551,0/)
        data_ped(3265,1:3)=(/3265,1968,0/)
        data_ped(3266,1:3)=(/3266,0,1132/)
        data_ped(3267,1:3)=(/3267,1138,1156/)
        data_ped(3268,1:3)=(/3268,2060,2042/)
        data_ped(3269,1:3)=(/3269,2315,2425/)
        data_ped(3270,1:3)=(/3270,2067,1084/)
        data_ped(3271,1:3)=(/3271,0,2497/)
        data_ped(3272,1:3)=(/3272,2238,2368/)
        data_ped(3273,1:3)=(/3273,1973,965/)
        data_ped(3274,1:3)=(/3274,2062,2759/)
        data_ped(3275,1:3)=(/3275,2774,2581/)
        data_ped(3276,1:3)=(/3276,2228,2646/)
        data_ped(3277,1:3)=(/3277,2521,1897/)
        data_ped(3278,1:3)=(/3278,2453,2648/)
        data_ped(3279,1:3)=(/3279,2521,2835/)
        data_ped(3280,1:3)=(/3280,2523,2213/)
        data_ped(3281,1:3)=(/3281,2052,895/)
        data_ped(3282,1:3)=(/3282,2296,2475/)
        data_ped(3283,1:3)=(/3283,2613,2688/)
        data_ped(3284,1:3)=(/3284,2311,2046/)
        data_ped(3285,1:3)=(/3285,1862,2265/)
        data_ped(3286,1:3)=(/3286,2524,2663/)
        data_ped(3287,1:3)=(/3287,1944,1054/)
        data_ped(3288,1:3)=(/3288,2448,2837/)
        data_ped(3289,1:3)=(/3289,1965,2571/)
        data_ped(3290,1:3)=(/3290,2783,2813/)
        data_ped(3291,1:3)=(/3291,2382,2369/)
        data_ped(3292,1:3)=(/3292,2298,1155/)
        data_ped(3293,1:3)=(/3293,2379,2564/)
        data_ped(3294,1:3)=(/3294,2063,1099/)
        data_ped(3295,1:3)=(/3295,2620,2373/)
        data_ped(3296,1:3)=(/3296,0,2347/)
        data_ped(3297,1:3)=(/3297,0,267/)
        data_ped(3298,1:3)=(/3298,2456,2484/)
        data_ped(3299,1:3)=(/3299,2127,1014/)
        data_ped(3300,1:3)=(/3300,2462,2668/)
        data_ped(3301,1:3)=(/3301,2845,2352/)
        data_ped(3302,1:3)=(/3302,0,1044/)
        data_ped(3303,1:3)=(/3303,1895,1871/)
        data_ped(3304,1:3)=(/3304,2315,0/)
        data_ped(3305,1:3)=(/3305,2388,928/)
        data_ped(3306,1:3)=(/3306,1862,2131/)
        data_ped(3307,1:3)=(/3307,2301,2922/)
        data_ped(3308,1:3)=(/3308,2388,2363/)
        data_ped(3309,1:3)=(/3309,0,1090/)
        data_ped(3310,1:3)=(/3310,0,2196/)
        data_ped(3311,1:3)=(/3311,2538,993/)
        data_ped(3312,1:3)=(/3312,1138,1156/)
        data_ped(3313,1:3)=(/3313,1808,1200/)
        data_ped(3314,1:3)=(/3314,2444,2682/)
        data_ped(3315,1:3)=(/3315,2773,2290/)
        data_ped(3316,1:3)=(/3316,2312,2558/)
        data_ped(3317,1:3)=(/3317,0,951/)
        data_ped(3318,1:3)=(/3318,1182,159/)
        data_ped(3319,1:3)=(/3319,2162,2579/)
        data_ped(3320,1:3)=(/3320,2519,2834/)
        data_ped(3321,1:3)=(/3321,1005,2683/)
        data_ped(3322,1:3)=(/3322,2112,3011/)
        data_ped(3323,1:3)=(/3323,1927,995/)
        data_ped(3324,1:3)=(/3324,2783,2255/)
        data_ped(3325,1:3)=(/3325,0,1183/)
        data_ped(3326,1:3)=(/3326,2772,2571/)
        data_ped(3327,1:3)=(/3327,2229,946/)
        data_ped(3328,1:3)=(/3328,0,2882/)
        data_ped(3329,1:3)=(/3329,0,1214/)
        data_ped(3330,1:3)=(/3330,1965,2764/)
        data_ped(3331,1:3)=(/3331,1965,2733/)
        data_ped(3332,1:3)=(/3332,2852,2600/)
        data_ped(3333,1:3)=(/3333,2703,2436/)
        data_ped(3334,1:3)=(/3334,1972,2749/)
        data_ped(3335,1:3)=(/3335,2162,2267/)
        data_ped(3336,1:3)=(/3336,2536,1165/)
        data_ped(3337,1:3)=(/3337,2115,2493/)
        data_ped(3338,1:3)=(/3338,0,2100/)
        data_ped(3339,1:3)=(/3339,2378,2748/)
        data_ped(3340,1:3)=(/3340,2379,2044/)
        data_ped(3341,1:3)=(/3341,2792,1032/)
        data_ped(3342,1:3)=(/3342,1009,16/)
        data_ped(3343,1:3)=(/3343,2626,2351/)
        data_ped(3344,1:3)=(/3344,2315,1984/)
        data_ped(3345,1:3)=(/3345,2320,2187/)
        data_ped(3346,1:3)=(/3346,2320,2588/)
        data_ped(3347,1:3)=(/3347,2775,0/)
        data_ped(3348,1:3)=(/3348,2175,1193/)
        data_ped(3349,1:3)=(/3349,2775,0/)
        data_ped(3350,1:3)=(/3350,2322,2828/)
        data_ped(3351,1:3)=(/3351,1808,0/)
        data_ped(3352,1:3)=(/3352,2720,2994/)
        data_ped(3353,1:3)=(/3353,2790,0/)
        data_ped(3354,1:3)=(/3354,2175,0/)
        data_ped(3355,1:3)=(/3355,2241,1021/)
        data_ped(3356,1:3)=(/3356,2790,2266/)
        data_ped(3357,1:3)=(/3357,319,2145/)
        data_ped(3358,1:3)=(/3358,1078,0/)
        data_ped(3359,1:3)=(/3359,1973,0/)
        data_ped(3360,1:3)=(/3360,2320,0/)
        data_ped(3361,1:3)=(/3361,1867,0/)
        data_ped(3362,1:3)=(/3362,2179,2278/)
        data_ped(3363,1:3)=(/3363,414,0/)
        data_ped(3364,1:3)=(/3364,0,949/)
        data_ped(3365,1:3)=(/3365,2615,2081/)
        data_ped(3366,1:3)=(/3366,2120,2997/)
        data_ped(3367,1:3)=(/3367,1079,0/)
        data_ped(3368,1:3)=(/3368,1060,0/)
        data_ped(3369,1:3)=(/3369,2126,2591/)
        data_ped(3370,1:3)=(/3370,1944,2273/)
        data_ped(3371,1:3)=(/3371,2320,0/)
        data_ped(3372,1:3)=(/3372,0,1147/)
        data_ped(3373,1:3)=(/3373,2850,2975/)
        data_ped(3374,1:3)=(/3374,2932,2259/)
        data_ped(3375,1:3)=(/3375,2299,2651/)
        data_ped(3376,1:3)=(/3376,1965,2876/)
        data_ped(3377,1:3)=(/3377,2782,2573/)
        data_ped(3378,1:3)=(/3378,2052,2336/)
        data_ped(3379,1:3)=(/3379,2844,2605/)
        data_ped(3380,1:3)=(/3380,2932,2649/)
        data_ped(3381,1:3)=(/3381,1965,2412/)
        data_ped(3382,1:3)=(/3382,2774,2135/)
        data_ped(3383,1:3)=(/3383,2614,2420/)
        data_ped(3384,1:3)=(/3384,2703,2368/)
        data_ped(3385,1:3)=(/3385,0,2103/)
        data_ped(3386,1:3)=(/3386,1944,2439/)
        data_ped(3387,1:3)=(/3387,2107,2251/)
        data_ped(3388,1:3)=(/3388,2719,1882/)
        data_ped(3389,1:3)=(/3389,2774,2744/)
        data_ped(3390,1:3)=(/3390,2456,260/)
        data_ped(3391,1:3)=(/3391,2614,261/)
        data_ped(3392,1:3)=(/3392,2322,2692/)
        data_ped(3393,1:3)=(/3393,0,266/)
        data_ped(3394,1:3)=(/3394,2456,2695/)
        data_ped(3395,1:3)=(/3395,1808,2900/)
        data_ped(3396,1:3)=(/3396,0,2754/)
        data_ped(3397,1:3)=(/3397,2317,2904/)
        data_ped(3398,1:3)=(/3398,1808,2423/)
        data_ped(3399,1:3)=(/3399,2303,37/)
        data_ped(3400,1:3)=(/3400,0,1065/)
        data_ped(3401,1:3)=(/3401,1936,0/)
        data_ped(3402,1:3)=(/3402,2062,0/)
        data_ped(3403,1:3)=(/3403,2959,2917/)
        data_ped(3404,1:3)=(/3404,2320,0/)
        data_ped(3405,1:3)=(/3405,2320,2498/)
        data_ped(3406,1:3)=(/3406,1973,0/)
        data_ped(3407,1:3)=(/3407,1867,0/)
        data_ped(3408,1:3)=(/3408,2840,1868/)
        data_ped(3409,1:3)=(/3409,2524,1945/)
        data_ped(3410,1:3)=(/3410,1925,2132/)
        data_ped(3411,1:3)=(/3411,1965,2400/)
        data_ped(3412,1:3)=(/3412,2945,2976/)
        data_ped(3413,1:3)=(/3413,0,2805/)
        data_ped(3414,1:3)=(/3414,2445,1067/)
        data_ped(3415,1:3)=(/3415,3033,195/)
        data_ped(3416,1:3)=(/3416,2699,1136/)
        data_ped(3417,1:3)=(/3417,2772,1051/)
        data_ped(3418,1:3)=(/3418,2848,2814/)
        data_ped(3419,1:3)=(/3419,0,1025/)
        data_ped(3420,1:3)=(/3420,2159,2360/)
        data_ped(3421,1:3)=(/3421,2009,2442/)
        data_ped(3422,1:3)=(/3422,0,187/)
        data_ped(3423,1:3)=(/3423,2782,1511/)
        data_ped(3424,1:3)=(/3424,0,205/)
        data_ped(3425,1:3)=(/3425,0,1176/)
        data_ped(3426,1:3)=(/3426,2842,1187/)
        data_ped(3427,1:3)=(/3427,0,214/)
        data_ped(3428,1:3)=(/3428,0,1094/)
        data_ped(3429,1:3)=(/3429,2716,1166/)
        data_ped(3430,1:3)=(/3430,3021,1033/)
        data_ped(3431,1:3)=(/3431,2066,1084/)
        data_ped(3432,1:3)=(/3432,2546,3000/)
        data_ped(3433,1:3)=(/3433,2538,3001/)
        data_ped(3434,1:3)=(/3434,2627,1209/)
        data_ped(3435,1:3)=(/3435,3037,0/)
        data_ped(3436,1:3)=(/3436,1808,0/)
        data_ped(3437,1:3)=(/3437,2244,0/)
        data_ped(3438,1:3)=(/3438,1190,0/)
        data_ped(3439,1:3)=(/3439,2173,0/)
        data_ped(3440,1:3)=(/3440,2053,0/)
        data_ped(3441,1:3)=(/3441,2177,893/)
        data_ped(3442,1:3)=(/3442,1235,0/)
        data_ped(3443,1:3)=(/3443,3125,0/)
        data_ped(3444,1:3)=(/3444,2790,1131/)
        data_ped(3445,1:3)=(/3445,2066,0/)
        data_ped(3446,1:3)=(/3446,2721,969/)
        data_ped(3447,1:3)=(/3447,1973,2193/)
        data_ped(3448,1:3)=(/3448,2554,2677/)
        data_ped(3449,1:3)=(/3449,2467,0/)
        data_ped(3450,1:3)=(/3450,1973,0/)
        data_ped(3451,1:3)=(/3451,2627,2034/)
        data_ped(3452,1:3)=(/3452,2315,2356/)
        data_ped(3453,1:3)=(/3453,2795,0/)
        data_ped(3454,1:3)=(/3454,0,2927/)
        data_ped(3455,1:3)=(/3455,2723,2897/)
        data_ped(3456,1:3)=(/3456,2859,2902/)
        data_ped(3457,1:3)=(/3457,2772,2572/)
        data_ped(3458,1:3)=(/3458,1891,2182/)
        data_ped(3459,1:3)=(/3459,2230,2339/)
        data_ped(3460,1:3)=(/3460,2524,923/)
        data_ped(3461,1:3)=(/3461,2849,1048/)
        data_ped(3462,1:3)=(/3462,2537,2638/)
        data_ped(3463,1:3)=(/3463,2789,2824/)
        data_ped(3464,1:3)=(/3464,2773,1912/)
        data_ped(3465,1:3)=(/3465,2794,2665/)
        data_ped(3466,1:3)=(/3466,2783,2183/)
        data_ped(3467,1:3)=(/3467,2538,2606/)
        data_ped(3468,1:3)=(/3468,2053,2608/)
        data_ped(3469,1:3)=(/3469,2859,3015/)
        data_ped(3470,1:3)=(/3470,0,235/)
        data_ped(3471,1:3)=(/3471,2707,2746/)
        data_ped(3472,1:3)=(/3472,0,239/)
        data_ped(3473,1:3)=(/3473,0,2401/)
        data_ped(3474,1:3)=(/3474,2298,1155/)
        data_ped(3475,1:3)=(/3475,2320,2610/)
        data_ped(3476,1:3)=(/3476,2456,259/)
        data_ped(3477,1:3)=(/3477,1931,2691/)
        data_ped(3478,1:3)=(/3478,2065,2154/)
        data_ped(3479,1:3)=(/3479,3035,1224/)
        data_ped(3480,1:3)=(/3480,2053,1109/)
        data_ped(3481,1:3)=(/3481,2244,2968/)
        data_ped(3482,1:3)=(/3482,1973,2996/)
        data_ped(3483,1:3)=(/3483,2790,2354/)
        data_ped(3484,1:3)=(/3484,1936,31/)
        data_ped(3485,1:3)=(/3485,0,1211/)
        data_ped(3486,1:3)=(/3486,2629,2593/)
        data_ped(3487,1:3)=(/3487,3037,3086/)
        data_ped(3488,1:3)=(/3488,829,0/)
        data_ped(3489,1:3)=(/3489,3200,2913/)
        data_ped(3490,1:3)=(/3490,2800,3008/)
        data_ped(3491,1:3)=(/3491,1081,0/)
        data_ped(3492,1:3)=(/3492,940,0/)
        data_ped(3493,1:3)=(/3493,2299,2249/)
        data_ped(3494,1:3)=(/3494,1966,2131/)
        data_ped(3495,1:3)=(/3495,2699,2409/)
        data_ped(3496,1:3)=(/3496,2780,2090/)
        data_ped(3497,1:3)=(/3497,2617,2813/)
        data_ped(3498,1:3)=(/3498,0,2134/)
        data_ped(3499,1:3)=(/3499,2026,160/)
        data_ped(3500,1:3)=(/3500,2121,945/)
        data_ped(3501,1:3)=(/3501,2241,1514/)
        data_ped(3502,1:3)=(/3502,3022,1148/)
        data_ped(3503,1:3)=(/3503,2781,165/)
        data_ped(3504,1:3)=(/3504,1973,2685/)
        data_ped(3505,1:3)=(/3505,2452,2690/)
        data_ped(3506,1:3)=(/3506,2379,1203/)
        data_ped(3507,1:3)=(/3507,3128,1027/)
        data_ped(3508,1:3)=(/3508,2223,1164/)
        data_ped(3509,1:3)=(/3509,2631,1188/)
        data_ped(3510,1:3)=(/3510,0,1221/)
        data_ped(3511,1:3)=(/3511,2701,2602/)
        data_ped(3512,1:3)=(/3512,3128,2920/)
        data_ped(3513,1:3)=(/3513,2458,1154/)
        data_ped(3514,1:3)=(/3514,0,893/)
        data_ped(3515,1:3)=(/3515,2520,223/)
        data_ped(3516,1:3)=(/3516,2523,2271/)
        data_ped(3517,1:3)=(/3517,2859,3077/)
        data_ped(3518,1:3)=(/3518,2626,2903/)
        data_ped(3519,1:3)=(/3519,2546,2497/)
        data_ped(3520,1:3)=(/3520,2958,0/)
        data_ped(3521,1:3)=(/3521,2053,2190/)
        data_ped(3522,1:3)=(/3522,2230,2560/)
        data_ped(3523,1:3)=(/3523,1973,0/)
        data_ped(3524,1:3)=(/3524,2790,2686/)
        data_ped(3525,1:3)=(/3525,2053,2502/)
        data_ped(3526,1:3)=(/3526,3124,1212/)
        data_ped(3527,1:3)=(/3527,1973,2143/)
        data_ped(3528,1:3)=(/3528,1181,0/)
        data_ped(3529,1:3)=(/3529,2845,0/)
        data_ped(3530,1:3)=(/3530,2053,1109/)
        data_ped(3531,1:3)=(/3531,3037,0/)
        data_ped(3532,1:3)=(/3532,2244,0/)
        data_ped(3533,1:3)=(/3533,3032,2435/)
        data_ped(3534,1:3)=(/3534,2053,0/)
        data_ped(3535,1:3)=(/3535,2320,0/)
        data_ped(3536,1:3)=(/3536,2554,0/)
        data_ped(3537,1:3)=(/3537,2790,0/)
        data_ped(3538,1:3)=(/3538,2241,1210/)
        data_ped(3539,1:3)=(/3539,2855,2345/)
        data_ped(3540,1:3)=(/3540,2538,2755/)
        data_ped(3541,1:3)=(/3541,2793,1134/)
        data_ped(3542,1:3)=(/3542,1935,2896/)
        data_ped(3543,1:3)=(/3543,1697,2077/)
        data_ped(3544,1:3)=(/3544,2699,2568/)
        data_ped(3545,1:3)=(/3545,2715,2186/)
        data_ped(3546,1:3)=(/3546,2379,2491/)
        data_ped(3547,1:3)=(/3547,2375,2341/)
        data_ped(3548,1:3)=(/3548,1965,2412/)
        data_ped(3549,1:3)=(/3549,2839,2256/)
        data_ped(3550,1:3)=(/3550,2844,2331/)
        data_ped(3551,1:3)=(/3551,2790,2266/)
        data_ped(3552,1:3)=(/3552,3018,2414/)
        data_ped(3553,1:3)=(/3553,3032,2823/)
        data_ped(3554,1:3)=(/3554,3020,2661/)
        data_ped(3555,1:3)=(/3555,2699,2886/)
        data_ped(3556,1:3)=(/3556,2783,2183/)
        data_ped(3557,1:3)=(/3557,3032,2607/)
        data_ped(3558,1:3)=(/3558,2524,1223/)
        data_ped(3559,1:3)=(/3559,0,2105/)
        data_ped(3560,1:3)=(/3560,3018,2895/)
        data_ped(3561,1:3)=(/3561,2313,2262/)
        data_ped(3562,1:3)=(/3562,2627,3000/)
        data_ped(3563,1:3)=(/3563,2532,2770/)
        data_ped(3564,1:3)=(/3564,3037,2696/)
        data_ped(3565,1:3)=(/3565,2053,2612/)
        data_ped(3566,1:3)=(/3566,2236,2666/)
        data_ped(3567,1:3)=(/3567,2462,2274/)
        data_ped(3568,1:3)=(/3568,2315,2999/)
        data_ped(3569,1:3)=(/3569,1973,2928/)
        data_ped(3570,1:3)=(/3570,2173,2426/)
        data_ped(3571,1:3)=(/3571,2718,2038/)
        data_ped(3572,1:3)=(/3572,3037,2349/)
        data_ped(3573,1:3)=(/3573,2320,0/)
        data_ped(3574,1:3)=(/3574,1143,0/)
        data_ped(3575,1:3)=(/3575,2778,2603/)
        data_ped(3576,1:3)=(/3576,2555,0/)
        data_ped(3577,1:3)=(/3577,1182,0/)
        data_ped(3578,1:3)=(/3578,2053,0/)
        data_ped(3579,1:3)=(/3579,829,0/)
        data_ped(3580,1:3)=(/3580,1142,0/)
        data_ped(3581,1:3)=(/3581,3021,2729/)
        data_ped(3582,1:3)=(/3582,2110,1947/)
        data_ped(3583,1:3)=(/3583,2801,2964/)
        data_ped(3584,1:3)=(/3584,2521,2980/)
        data_ped(3585,1:3)=(/3585,2702,1198/)
        data_ped(3586,1:3)=(/3586,3114,169/)
        data_ped(3587,1:3)=(/3587,2707,2877/)
        data_ped(3588,1:3)=(/3588,2789,2507/)
        data_ped(3589,1:3)=(/3589,2158,1998/)
        data_ped(3590,1:3)=(/3590,2305,2477/)
        data_ped(3591,1:3)=(/3591,2859,1163/)
        data_ped(3592,1:3)=(/3592,3097,2879/)
        data_ped(3593,1:3)=(/3593,2957,1013/)
        data_ped(3594,1:3)=(/3594,3194,3067/)
        data_ped(3595,1:3)=(/3595,0,2878/)
        data_ped(3596,1:3)=(/3596,2538,1178/)
        data_ped(3597,1:3)=(/3597,3101,2473/)
        data_ped(3598,1:3)=(/3598,2784,1129/)
        data_ped(3599,1:3)=(/3599,2713,2919/)
        data_ped(3600,1:3)=(/3600,2458,1222/)
        data_ped(3601,1:3)=(/3601,2709,3050/)
        data_ped(3602,1:3)=(/3602,2631,2687/)
        data_ped(3603,1:3)=(/3603,0,213/)
        data_ped(3604,1:3)=(/3604,2847,2272/)
        data_ped(3605,1:3)=(/3605,3188,2079/)
        data_ped(3606,1:3)=(/3606,3197,0/)
        data_ped(3607,1:3)=(/3607,2795,2912/)
        data_ped(3608,1:3)=(/3608,2462,0/)
        data_ped(3609,1:3)=(/3609,2775,2756/)
        data_ped(3610,1:3)=(/3610,3037,1226/)
        data_ped(3611,1:3)=(/3611,2315,0/)
        data_ped(3612,1:3)=(/3612,1944,2273/)
        data_ped(3613,1:3)=(/3613,3039,1110/)
        data_ped(3614,1:3)=(/3614,2024,0/)
        data_ped(3615,1:3)=(/3615,2462,3270/)
        data_ped(3616,1:3)=(/3616,2053,2911/)
        data_ped(3617,1:3)=(/3617,1103,0/)
        data_ped(3618,1:3)=(/3618,3126,2431/)
        data_ped(3619,1:3)=(/3619,3032,2435/)
        data_ped(3620,1:3)=(/3620,2790,0/)
        data_ped(3621,1:3)=(/3621,2790,2432/)
        data_ped(3622,1:3)=(/3622,3037,0/)
        data_ped(3623,1:3)=(/3623,3036,3091/)
        data_ped(3624,1:3)=(/3624,0,1174/)
        data_ped(3625,1:3)=(/3625,1232,2685/)
        data_ped(3626,1:3)=(/3626,3197,2353/)
        data_ped(3627,1:3)=(/3627,2553,3259/)
        data_ped(3628,1:3)=(/3628,2946,2334/)
        data_ped(3629,1:3)=(/3629,2300,2983/)
        data_ped(3630,1:3)=(/3630,2774,2139/)
        data_ped(3631,1:3)=(/3631,2298,3060/)
        data_ped(3632,1:3)=(/3632,2710,2740/)
        data_ped(3633,1:3)=(/3633,2933,2644/)
        data_ped(3634,1:3)=(/3634,2943,2574/)
        data_ped(3635,1:3)=(/3635,2521,3149/)
        data_ped(3636,1:3)=(/3636,2521,3227/)
        data_ped(3637,1:3)=(/3637,2446,2818/)
        data_ped(3638,1:3)=(/3638,2454,2269/)
        data_ped(3639,1:3)=(/3639,3097,2411/)
        data_ped(3640,1:3)=(/3640,3104,2567/)
        data_ped(3641,1:3)=(/3641,3101,2182/)
        data_ped(3642,1:3)=(/3642,1966,3053/)
        data_ped(3643,1:3)=(/3643,2451,2643/)
        data_ped(3644,1:3)=(/3644,3018,2346/)
        data_ped(3645,1:3)=(/3645,2948,2583/)
        data_ped(3646,1:3)=(/3646,2065,3096/)
        data_ped(3647,1:3)=(/3647,2117,228/)
        data_ped(3648,1:3)=(/3648,2227,3141/)
        data_ped(3649,1:3)=(/3649,2390,2492/)
        data_ped(3650,1:3)=(/3650,1788,2751/)
        data_ped(3651,1:3)=(/3651,0,1127/)
        data_ped(3652,1:3)=(/3652,2445,2511/)
        data_ped(3653,1:3)=(/3653,2796,1001/)
        data_ped(3654,1:3)=(/3654,2716,2822/)
        data_ped(3655,1:3)=(/3655,2113,2901/)
        data_ped(3656,1:3)=(/3656,2631,1188/)
        data_ped(3657,1:3)=(/3657,2715,3016/)
        data_ped(3658,1:3)=(/3658,2854,2694/)
        data_ped(3659,1:3)=(/3659,3197,2929/)
        data_ped(3660,1:3)=(/3660,2116,2769/)
        data_ped(3661,1:3)=(/3661,2160,3325/)
        data_ped(3662,1:3)=(/3662,3024,3174/)
        data_ped(3663,1:3)=(/3663,3197,2898/)
        data_ped(3664,1:3)=(/3664,2861,2423/)
        data_ped(3665,1:3)=(/3665,2790,2686/)
        data_ped(3666,1:3)=(/3666,2324,0/)
        data_ped(3667,1:3)=(/3667,2236,0/)
        data_ped(3668,1:3)=(/3668,2861,2438/)
        data_ped(3669,1:3)=(/3669,3040,0/)
        data_ped(3670,1:3)=(/3670,1808,0/)
        data_ped(3671,1:3)=(/3671,2296,2726/)
        data_ped(3672,1:3)=(/3672,3044,3129/)
        data_ped(3673,1:3)=(/3673,2374,2865/)
        data_ped(3674,1:3)=(/3674,2522,3046/)
        data_ped(3675,1:3)=(/3675,2932,2966/)
        data_ped(3676,1:3)=(/3676,3044,2868/)
        data_ped(3677,1:3)=(/3677,2933,3048/)
        data_ped(3678,1:3)=(/3678,3098,3136/)
        data_ped(3679,1:3)=(/3679,2617,2396/)
        data_ped(3680,1:3)=(/3680,2106,3137/)
        data_ped(3681,1:3)=(/3681,0,2684/)
        data_ped(3682,1:3)=(/3682,2313,2344/)
        data_ped(3683,1:3)=(/3683,2707,3322/)
        data_ped(3684,1:3)=(/3684,2533,179/)
        data_ped(3685,1:3)=(/3685,2301,3240/)
        data_ped(3686,1:3)=(/3686,0,1218/)
        data_ped(3687,1:3)=(/3687,0,3367/)
        data_ped(3688,1:3)=(/3688,2376,212/)
        data_ped(3689,1:3)=(/3689,2939,0/)
        data_ped(3690,1:3)=(/3690,0,1230/)
        data_ped(3691,1:3)=(/3691,3117,227/)
        data_ped(3692,1:3)=(/3692,2939,1192/)
        data_ped(3693,1:3)=(/3693,2236,1219/)
        data_ped(3694,1:3)=(/3694,3036,3091/)
        data_ped(3695,1:3)=(/3695,2955,2678/)
        data_ped(3696,1:3)=(/3696,3388,2992/)
        data_ped(3697,1:3)=(/3697,2244,2968/)
        data_ped(3698,1:3)=(/3698,2859,0/)
        data_ped(3699,1:3)=(/3699,2121,3253/)
        data_ped(3700,1:3)=(/3700,2796,2049/)
        data_ped(3701,1:3)=(/3701,2322,2692/)
        data_ped(3702,1:3)=(/3702,0,44/)
        data_ped(3703,1:3)=(/3703,2790,2670/)
        data_ped(3704,1:3)=(/3704,2842,0/)
        data_ped(3705,1:3)=(/3705,2053,2827/)
        data_ped(3706,1:3)=(/3706,3287,2909/)
        data_ped(3707,1:3)=(/3707,2720,2899/)
        data_ped(3708,1:3)=(/3708,2053,2667/)
        data_ped(3709,1:3)=(/3709,2236,3049/)
        data_ped(3710,1:3)=(/3710,2053,2830/)
        data_ped(3711,1:3)=(/3711,3123,0/)
        data_ped(3712,1:3)=(/3712,3191,2503/)
        data_ped(3713,1:3)=(/3713,2721,2191/)
        data_ped(3714,1:3)=(/3714,2543,0/)
        data_ped(3715,1:3)=(/3715,2323,3313/)
        data_ped(3716,1:3)=(/3716,2466,0/)
        data_ped(3717,1:3)=(/3717,3037,2696/)
        data_ped(3718,1:3)=(/3718,3040,0/)
        data_ped(3719,1:3)=(/3719,2954,0/)
        data_ped(3720,1:3)=(/3720,959,3003/)
        data_ped(3721,1:3)=(/3721,3025,1425/)
        data_ped(3722,1:3)=(/3722,2315,1130/)
        data_ped(3723,1:3)=(/3723,2778,2905/)
        data_ped(3724,1:3)=(/3724,3020,3071/)
        data_ped(3725,1:3)=(/3725,3097,2875/)
        data_ped(3726,1:3)=(/3726,2787,3238/)
        data_ped(3727,1:3)=(/3727,2622,2417/)
        data_ped(3728,1:3)=(/3728,2296,2811/)
        data_ped(3729,1:3)=(/3729,2773,3226/)
        data_ped(3730,1:3)=(/3730,3102,3147/)
        data_ped(3731,1:3)=(/3731,2932,3138/)
        data_ped(3732,1:3)=(/3732,1926,3277/)
        data_ped(3733,1:3)=(/3733,1138,2476/)
        data_ped(3734,1:3)=(/3734,1966,2403/)
        data_ped(3735,1:3)=(/3735,3018,2805/)
        data_ped(3736,1:3)=(/3736,2455,2262/)
        data_ped(3737,1:3)=(/3737,2536,2263/)
        data_ped(3738,1:3)=(/3738,2621,2987/)
        data_ped(3739,1:3)=(/3739,2236,2072/)
        data_ped(3740,1:3)=(/3740,1861,2412/)
        data_ped(3741,1:3)=(/3741,2158,2736/)
        data_ped(3742,1:3)=(/3742,2523,1204/)
        data_ped(3743,1:3)=(/3743,2796,3364/)
        data_ped(3744,1:3)=(/3744,2615,233/)
        data_ped(3745,1:3)=(/3745,3196,1173/)
        data_ped(3746,1:3)=(/3746,2111,1179/)
        data_ped(3747,1:3)=(/3747,2841,2441/)
        data_ped(3748,1:3)=(/3748,3128,3272/)
        data_ped(3749,1:3)=(/3749,3287,2928/)
        data_ped(3750,1:3)=(/3750,1231,271/)
        data_ped(3751,1:3)=(/3751,3116,2480/)
        data_ped(3752,1:3)=(/3752,2464,1949/)
        data_ped(3753,1:3)=(/3753,3121,0/)
        data_ped(3754,1:3)=(/3754,1180,1451/)
        data_ped(3755,1:3)=(/3755,2392,1133/)
        data_ped(3756,1:3)=(/3756,2053,2504/)
        data_ped(3757,1:3)=(/3757,1042,0/)
        data_ped(3758,1:3)=(/3758,2443,2395/)
        data_ped(3759,1:3)=(/3759,2222,2471/)
        data_ped(3760,1:3)=(/3760,2106,2130/)
        data_ped(3761,1:3)=(/3761,2525,3220/)
        data_ped(3762,1:3)=(/3762,2158,2633/)
        data_ped(3763,1:3)=(/3763,2527,2636/)
        data_ped(3764,1:3)=(/3764,2932,2820/)
        data_ped(3765,1:3)=(/3765,2106,3052/)
        data_ped(3766,1:3)=(/3766,3184,2397/)
        data_ped(3767,1:3)=(/3767,3191,3265/)
        data_ped(3768,1:3)=(/3768,3185,3061/)
        data_ped(3769,1:3)=(/3769,2795,2604/)
        data_ped(3770,1:3)=(/3770,2305,1126/)
        data_ped(3771,1:3)=(/3771,2306,2421/)
        data_ped(3772,1:3)=(/3772,2461,2566/)
        data_ped(3773,1:3)=(/3773,2947,2654/)
        data_ped(3774,1:3)=(/3774,0,3276/)
        data_ped(3775,1:3)=(/3775,2240,1229/)
        data_ped(3776,1:3)=(/3776,3030,2347/)
        data_ped(3777,1:3)=(/3777,1139,2476/)
        data_ped(3778,1:3)=(/3778,2240,2994/)
        data_ped(3779,1:3)=(/3779,0,1239/)
        data_ped(3780,1:3)=(/3780,2625,2608/)
        data_ped(3781,1:3)=(/3781,2389,0/)
        data_ped(3782,1:3)=(/3782,0,1227/)
        data_ped(3783,1:3)=(/3783,3191,3158/)
        data_ped(3784,1:3)=(/3784,2960,1244/)
        data_ped(3785,1:3)=(/3785,2790,2767/)
        data_ped(3786,1:3)=(/3786,2625,2667/)
        data_ped(3787,1:3)=(/3787,0,2993/)
        data_ped(3788,1:3)=(/3788,3197,205/)
        data_ped(3789,1:3)=(/3789,2958,3006/)
        data_ped(3790,1:3)=(/3790,3039,0/)
        data_ped(3791,1:3)=(/3791,2240,3352/)
        data_ped(3792,1:3)=(/3792,3191,3271/)
        data_ped(3793,1:3)=(/3793,2625,0/)
        data_ped(3794,1:3)=(/3794,3197,2495/)
        data_ped(3795,1:3)=(/3795,2315,0/)
        data_ped(3796,1:3)=(/3796,2173,0/)
        data_ped(3797,1:3)=(/3797,3300,2675/)
        data_ped(3798,1:3)=(/3798,2625,0/)
        data_ped(3799,1:3)=(/3799,2066,2917/)
        data_ped(3800,1:3)=(/3800,1108,0/)
        data_ped(3801,1:3)=(/3801,3479,0/)
        data_ped(3802,1:3)=(/3802,2322,2828/)
        data_ped(3803,1:3)=(/3803,2790,2354/)
        data_ped(3804,1:3)=(/3804,0,3157/)
        data_ped(3805,1:3)=(/3805,959,3003/)
        data_ped(3806,1:3)=(/3806,3197,2353/)
        data_ped(3807,1:3)=(/3807,2468,2831/)
        data_ped(3808,1:3)=(/3808,2960,0/)
        data_ped(3809,1:3)=(/3809,2315,2768/)
        data_ped(3810,1:3)=(/3810,2849,2810/)
        data_ped(3811,1:3)=(/3811,3390,2893/)
        data_ped(3812,1:3)=(/3812,1520,1415/)
        data_ped(3813,1:3)=(/3813,2310,2811/)
        data_ped(3814,1:3)=(/3814,3023,3074/)
        data_ped(3815,1:3)=(/3815,2164,2972/)
        data_ped(3816,1:3)=(/3816,3177,3244/)
        data_ped(3817,1:3)=(/3817,2938,2735/)
        data_ped(3818,1:3)=(/3818,2777,2825/)
        data_ped(3819,1:3)=(/3819,2946,2257/)
        data_ped(3820,1:3)=(/3820,2375,3373/)
        data_ped(3821,1:3)=(/3821,3097,2739/)
        data_ped(3822,1:3)=(/3822,1139,2075/)
        data_ped(3823,1:3)=(/3823,2106,3231/)
        data_ped(3824,1:3)=(/3824,2848,2978/)
        data_ped(3825,1:3)=(/3825,3111,3057/)
        data_ped(3826,1:3)=(/3826,3186,3055/)
        data_ped(3827,1:3)=(/3827,3112,2977/)
        data_ped(3828,1:3)=(/3828,3101,3458/)
        data_ped(3829,1:3)=(/3829,2315,2136/)
        data_ped(3830,1:3)=(/3830,2236,3070/)
        data_ped(3831,1:3)=(/3831,3285,2183/)
        data_ped(3832,1:3)=(/3832,2793,3014/)
        data_ped(3833,1:3)=(/3833,2462,3270/)
        data_ped(3834,1:3)=(/3834,1965,3045/)
        data_ped(3835,1:3)=(/3835,3177,2416/)
        data_ped(3836,1:3)=(/3836,3197,951/)
        data_ped(3837,1:3)=(/3837,3018,2805/)
        data_ped(3838,1:3)=(/3838,2631,1178/)
        data_ped(3839,1:3)=(/3839,2795,2990/)
        data_ped(3840,1:3)=(/3840,3479,3145/)
        data_ped(3841,1:3)=(/3841,2772,2404/)
        data_ped(3842,1:3)=(/3842,2539,254/)
        data_ped(3843,1:3)=(/3843,3201,2370/)
        data_ped(3844,1:3)=(/3844,0,3173/)
        data_ped(3845,1:3)=(/3845,2960,1036/)
        data_ped(3846,1:3)=(/3846,3106,3051/)
        data_ped(3847,1:3)=(/3847,1943,1514/)
        data_ped(3848,1:3)=(/3848,2053,2275/)
        data_ped(3849,1:3)=(/3849,2625,2350/)
        data_ped(3850,1:3)=(/3850,2953,3796/)
        data_ped(3851,1:3)=(/3851,2320,2498/)
        data_ped(3852,1:3)=(/3852,1837,0/)
        data_ped(3853,1:3)=(/3853,1180,2277/)
        data_ped(3854,1:3)=(/3854,1103,2914/)
        data_ped(3855,1:3)=(/3855,3037,3012/)
        data_ped(3856,1:3)=(/3856,2775,0/)
        data_ped(3857,1:3)=(/3857,3304,0/)
        data_ped(3858,1:3)=(/3858,1083,0/)
        data_ped(3859,1:3)=(/3859,1837,0/)
        data_ped(3860,1:3)=(/3860,1083,0/)
        data_ped(3861,1:3)=(/3861,2771,2963/)
        data_ped(3862,1:3)=(/3862,2934,3216/)
        data_ped(3863,1:3)=(/3863,1966,2180/)
        data_ped(3864,1:3)=(/3864,2377,2965/)
        data_ped(3865,1:3)=(/3865,3279,2638/)
        data_ped(3866,1:3)=(/3866,3183,3275/)
        data_ped(3867,1:3)=(/3867,2521,2483/)
        data_ped(3868,1:3)=(/3868,3177,2887/)
        data_ped(3869,1:3)=(/3869,2932,3139/)
        data_ped(3870,1:3)=(/3870,0,146/)
        data_ped(3871,1:3)=(/3871,3036,2560/)
        data_ped(3872,1:3)=(/3872,0,154/)
        data_ped(3873,1:3)=(/3873,0,158/)
        data_ped(3874,1:3)=(/3874,2701,3272/)
        data_ped(3875,1:3)=(/3875,3099,3374/)
        data_ped(3876,1:3)=(/3876,2125,3370/)
        data_ped(3877,1:3)=(/3877,3283,3010/)
        data_ped(3878,1:3)=(/3878,0,166/)
        data_ped(3879,1:3)=(/3879,3103,2559/)
        data_ped(3880,1:3)=(/3880,2841,1186/)
        data_ped(3881,1:3)=(/3881,3468,1992/)
        data_ped(3882,1:3)=(/3882,3466,2875/)
        data_ped(3883,1:3)=(/3883,2790,198/)
        data_ped(3884,1:3)=(/3884,2444,2832/)
        data_ped(3885,1:3)=(/3885,2712,2664/)
        data_ped(3886,1:3)=(/3886,2541,3454/)
        data_ped(3887,1:3)=(/3887,3477,3075/)
        data_ped(3888,1:3)=(/3888,3301,2187/)
        data_ped(3889,1:3)=(/3889,3396,2756/)
        data_ped(3890,1:3)=(/3890,2790,2354/)
        data_ped(3891,1:3)=(/3891,0,3349/)
        data_ped(3892,1:3)=(/3892,3123,2668/)
        data_ped(3893,1:3)=(/3893,2053,2612/)
        data_ped(3894,1:3)=(/3894,3287,2826/)
        data_ped(3895,1:3)=(/3895,2542,0/)
        data_ped(3896,1:3)=(/3896,3201,3311/)
        data_ped(3897,1:3)=(/3897,2625,3521/)
        data_ped(3898,1:3)=(/3898,3041,3351/)
        data_ped(3899,1:3)=(/3899,2248,1428/)
        data_ped(3900,1:3)=(/3900,2861,3313/)
        data_ped(3901,1:3)=(/3901,3125,0/)
        data_ped(3902,1:3)=(/3902,2053,2924/)
        data_ped(3903,1:3)=(/3903,2053,0/)
        data_ped(3904,1:3)=(/3904,2790,2214/)
        data_ped(3905,1:3)=(/3905,2247,0/)
        data_ped(3906,1:3)=(/3906,2053,0/)
        data_ped(3907,1:3)=(/3907,2393,0/)
        data_ped(3908,1:3)=(/3908,423,0/)
        data_ped(3909,1:3)=(/3909,1253,0/)
        data_ped(3910,1:3)=(/3910,2236,1953/)
        data_ped(3911,1:3)=(/3911,3381,2402/)
        data_ped(3912,1:3)=(/3912,3019,3232/)
        data_ped(3913,1:3)=(/3913,2226,2637/)
        data_ped(3914,1:3)=(/3914,2951,2894/)
        data_ped(3915,1:3)=(/3915,3181,2988/)
        data_ped(3916,1:3)=(/3916,2168,3237/)
        data_ped(3917,1:3)=(/3917,2943,3415/)
        data_ped(3918,1:3)=(/3918,3027,2974/)
        data_ped(3919,1:3)=(/3919,2944,2639/)
        data_ped(3920,1:3)=(/3920,2940,2985/)
        data_ped(3921,1:3)=(/3921,3284,3330/)
        data_ped(3922,1:3)=(/3922,2521,2483/)
        data_ped(3923,1:3)=(/3923,2299,2642/)
        data_ped(3924,1:3)=(/3924,2935,3316/)
        data_ped(3925,1:3)=(/3925,3280,3413/)
        data_ped(3926,1:3)=(/3926,3280,2567/)
        data_ped(3927,1:3)=(/3927,2707,3322/)
        data_ped(3928,1:3)=(/3928,3097,2487/)
        data_ped(3929,1:3)=(/3929,2788,2030/)
        data_ped(3930,1:3)=(/3930,2932,2727/)
        data_ped(3931,1:3)=(/3931,3408,2348/)
        data_ped(3932,1:3)=(/3932,2714,3233/)
        data_ped(3933,1:3)=(/3933,3199,232/)
        data_ped(3934,1:3)=(/3934,2053,2911/)
        data_ped(3935,1:3)=(/3935,3101,3458/)
        data_ped(3936,1:3)=(/3936,3214,2890/)
        data_ped(3937,1:3)=(/3937,2158,2816/)
        data_ped(3938,1:3)=(/3938,3285,3093/)
        data_ped(3939,1:3)=(/3939,0,2838/)
        data_ped(3940,1:3)=(/3940,2793,2923/)
        data_ped(3941,1:3)=(/3941,3394,2930/)
        data_ped(3942,1:3)=(/3942,3191,2697/)
        data_ped(3943,1:3)=(/3943,3301,3003/)
        data_ped(3944,1:3)=(/3944,3300,3082/)
        data_ped(3945,1:3)=(/3945,2175,2590/)
        data_ped(3946,1:3)=(/3946,2538,968/)
        data_ped(3947,1:3)=(/3947,2845,3084/)
        data_ped(3948,1:3)=(/3948,0,1245/)
        data_ped(3949,1:3)=(/3949,2053,2589/)
        data_ped(3950,1:3)=(/3950,2053,3083/)
        data_ped(3951,1:3)=(/3951,3303,0/)
        data_ped(3952,1:3)=(/3952,3398,3167/)
        data_ped(3953,1:3)=(/3953,2320,0/)
        data_ped(3954,1:3)=(/3954,2790,3085/)
        data_ped(3955,1:3)=(/3955,2845,2926/)
        data_ped(3956,1:3)=(/3956,2955,0/)
        data_ped(3957,1:3)=(/3957,2306,3169/)
        data_ped(3958,1:3)=(/3958,2236,3171/)
        data_ped(3959,1:3)=(/3959,1234,0/)
        data_ped(3960,1:3)=(/3960,1257,0/)
        data_ped(3961,1:3)=(/3961,2236,0/)
        data_ped(3962,1:3)=(/3962,2795,0/)
        data_ped(3963,1:3)=(/3963,3040,0/)
        data_ped(3964,1:3)=(/3964,3179,3217/)
        data_ped(3965,1:3)=(/3965,2106,2470/)
        data_ped(3966,1:3)=(/3966,2701,3305/)
        data_ped(3967,1:3)=(/3967,2106,2327/)
        data_ped(3968,1:3)=(/3968,3018,2805/)
        data_ped(3969,1:3)=(/3969,3191,3428/)
        data_ped(3970,1:3)=(/3970,3099,2867/)
        data_ped(3971,1:3)=(/3971,2932,2071/)
        data_ped(3972,1:3)=(/3972,2704,2329/)
        data_ped(3973,1:3)=(/3973,3468,2968/)
        data_ped(3974,1:3)=(/3974,3214,2733/)
        data_ped(3975,1:3)=(/3975,2801,3223/)
        data_ped(3976,1:3)=(/3976,3382,3218/)
        data_ped(3977,1:3)=(/3977,2010,3225/)
        data_ped(3978,1:3)=(/3978,2790,2918/)
        data_ped(3979,1:3)=(/3979,2315,2681/)
        data_ped(3980,1:3)=(/3980,2236,3452/)
        data_ped(3981,1:3)=(/3981,3296,1070/)
        data_ped(3982,1:3)=(/3982,2379,3539/)
        data_ped(3983,1:3)=(/3983,3282,3064/)
        data_ped(3984,1:3)=(/3984,0,3267/)
        data_ped(3985,1:3)=(/3985,3379,3076/)
        data_ped(3986,1:3)=(/3986,3201,2512/)
        data_ped(3987,1:3)=(/3987,2936,3151/)
        data_ped(3988,1:3)=(/3988,2941,0/)
        data_ped(3989,1:3)=(/3989,0,3501/)
        data_ped(3990,1:3)=(/3990,2538,2692/)
        data_ped(3991,1:3)=(/3991,3481,3446/)
        data_ped(3992,1:3)=(/3992,3209,2757/)
        data_ped(3993,1:3)=(/3993,2545,3092/)
        data_ped(3994,1:3)=(/3994,3300,3081/)
        data_ped(3995,1:3)=(/3995,2460,3007/)
        data_ped(3996,1:3)=(/3996,3201,3540/)
        data_ped(3997,1:3)=(/3997,2053,2830/)
        data_ped(3998,1:3)=(/3998,3482,3160/)
        data_ped(3999,1:3)=(/3999,3467,3358/)
        data_ped(4000,1:3)=(/4000,2863,0/)
        data_ped(4001,1:3)=(/4001,2538,2591/)
        data_ped(4002,1:3)=(/4002,2053,3083/)
        data_ped(4003,1:3)=(/4003,2790,2354/)
        data_ped(4004,1:3)=(/4004,3482,1244/)
        data_ped(4005,1:3)=(/4005,3469,3258/)
        data_ped(4006,1:3)=(/4006,1225,3354/)
        data_ped(4007,1:3)=(/4007,2538,3369/)
        data_ped(4008,1:3)=(/4008,3289,3331/)
        data_ped(4009,1:3)=(/4009,3487,3526/)
        data_ped(4010,1:3)=(/4010,1205,3263/)
        data_ped(4011,1:3)=(/4011,1541,1460/)
        data_ped(4012,1:3)=(/4012,3042,0/)
        data_ped(4013,1:3)=(/4013,2797,0/)
        data_ped(4014,1:3)=(/4014,3301,0/)
        data_ped(4015,1:3)=(/4015,2053,0/)
        data_ped(4016,1:3)=(/4016,3301,3438/)
        data_ped(4017,1:3)=(/4017,2938,2735/)
        data_ped(4018,1:3)=(/4018,3098,3247/)
        data_ped(4019,1:3)=(/4019,3190,3338/)
        data_ped(4020,1:3)=(/4020,2795,3249/)
        data_ped(4021,1:3)=(/4021,3385,3514/)
        data_ped(4022,1:3)=(/4022,3394,2270/)
        data_ped(4023,1:3)=(/4023,3293,2578/)
        data_ped(4024,1:3)=(/4024,3099,2892/)
        data_ped(4025,1:3)=(/4025,2937,3146/)
        data_ped(4026,1:3)=(/4026,3214,2635/)
        data_ped(4027,1:3)=(/4027,1965,2970/)
        data_ped(4028,1:3)=(/4028,2944,3143/)
        data_ped(4029,1:3)=(/4029,3214,3544/)
        data_ped(4030,1:3)=(/4030,2779,2407/)
        data_ped(4031,1:3)=(/4031,2521,2645/)
        data_ped(4032,1:3)=(/4032,3279,2638/)
        data_ped(4033,1:3)=(/4033,3493,1172/)
        data_ped(4034,1:3)=(/4034,3387,3418/)
        data_ped(4035,1:3)=(/4035,3112,2655/)
        data_ped(4036,1:3)=(/4036,2848,2413/)
        data_ped(4037,1:3)=(/4037,3118,1061/)
        data_ped(4038,1:3)=(/4038,2947,3332/)
        data_ped(4039,1:3)=(/4039,3281,2815/)
        data_ped(4040,1:3)=(/4040,3029,2819/)
        data_ped(4041,1:3)=(/4041,3292,3148/)
        data_ped(4042,1:3)=(/4042,3184,3375/)
        data_ped(4043,1:3)=(/4043,3116,2561/)
        data_ped(4044,1:3)=(/4044,3128,2868/)
        data_ped(4045,1:3)=(/4045,3197,2836/)
        data_ped(4046,1:3)=(/4046,3028,3013/)
        data_ped(4047,1:3)=(/4047,3477,3241/)
        data_ped(4048,1:3)=(/4048,3042,3318/)
        data_ped(4049,1:3)=(/4049,2316,3340/)
        data_ped(4050,1:3)=(/4050,0,240/)
        data_ped(4051,1:3)=(/4051,3561,3069/)
        data_ped(4052,1:3)=(/4052,3467,3154/)
        data_ped(4053,1:3)=(/4053,3483,3080/)
        data_ped(4054,1:3)=(/4054,0,2910/)
        data_ped(4055,1:3)=(/4055,2954,1514/)
        data_ped(4056,1:3)=(/4056,2053,2667/)
        data_ped(4057,1:3)=(/4057,0,2766/)
        data_ped(4058,1:3)=(/4058,3469,3274/)
        data_ped(4059,1:3)=(/4059,2722,30/)
        data_ped(4060,1:3)=(/4060,2538,2082/)
        data_ped(4061,1:3)=(/4061,3182,2913/)
        data_ped(4062,1:3)=(/4062,3468,0/)
        data_ped(4063,1:3)=(/4063,2053,2758/)
        data_ped(4064,1:3)=(/4064,2790,2686/)
        data_ped(4065,1:3)=(/4065,2053,2502/)
        data_ped(4066,1:3)=(/4066,1251,0/)
        data_ped(4067,1:3)=(/4067,1256,3531/)
        data_ped(4068,1:3)=(/4068,2861,0/)
        data_ped(4069,1:3)=(/4069,3406,3360/)
        data_ped(4070,1:3)=(/4070,2632,2829/)
        data_ped(4071,1:3)=(/4071,3181,3243/)
        data_ped(4072,1:3)=(/4072,3214,2731/)
        data_ped(4073,1:3)=(/4073,3471,3329/)
        data_ped(4074,1:3)=(/4074,2298,2730/)
        data_ped(4075,1:3)=(/4075,2933,2563/)
        data_ped(4076,1:3)=(/4076,3286,2473/)
        data_ped(4077,1:3)=(/4077,2443,2474/)
        data_ped(4078,1:3)=(/4078,2526,3498/)
        data_ped(4079,1:3)=(/4079,3020,3506/)
        data_ped(4080,1:3)=(/4080,2950,3546/)
        data_ped(4081,1:3)=(/4081,3381,3219/)
        data_ped(4082,1:3)=(/4082,3653,3364/)
        data_ped(4083,1:3)=(/4083,2954,3538/)
        data_ped(4084,1:3)=(/4084,3116,2480/)
        data_ped(4085,1:3)=(/4085,3116,3623/)
        data_ped(4086,1:3)=(/4086,2545,3092/)
        data_ped(4087,1:3)=(/4087,3466,3066/)
        data_ped(4088,1:3)=(/4088,3290,3643/)
        data_ped(4089,1:3)=(/4089,3397,201/)
        data_ped(4090,1:3)=(/4090,3482,0/)
        data_ped(4091,1:3)=(/4091,2864,3453/)
        data_ped(4092,1:3)=(/4092,2776,2921/)
        data_ped(4093,1:3)=(/4093,3120,1208/)
        data_ped(4094,1:3)=(/4094,2775,3627/)
        data_ped(4095,1:3)=(/4095,3201,3433/)
        data_ped(4096,1:3)=(/4096,3190,3432/)
        data_ped(4097,1:3)=(/4097,2546,3434/)
        data_ped(4098,1:3)=(/4098,3562,3078/)
        data_ped(4099,1:3)=(/4099,2538,0/)
        data_ped(4100,1:3)=(/4100,3557,2873/)
        data_ped(4101,1:3)=(/4101,0,2672/)
        data_ped(4102,1:3)=(/4102,3190,3271/)
        data_ped(4103,1:3)=(/4103,2240,2994/)
        data_ped(4104,1:3)=(/4104,0,3443/)
        data_ped(4105,1:3)=(/4105,2543,3348/)
        data_ped(4106,1:3)=(/4106,2625,1195/)
        data_ped(4107,1:3)=(/4107,3197,2908/)
        data_ped(4108,1:3)=(/4108,3557,3523/)
        data_ped(4109,1:3)=(/4109,3120,3610/)
        data_ped(4110,1:3)=(/4110,2538,2692/)
        data_ped(4111,1:3)=(/4111,3025,3613/)
        data_ped(4112,1:3)=(/4112,2464,2674/)
        data_ped(4113,1:3)=(/4113,3201,2512/)
        data_ped(4114,1:3)=(/4114,3485,3355/)
        data_ped(4115,1:3)=(/4115,3197,3606/)
        data_ped(4116,1:3)=(/4116,3482,3085/)
        data_ped(4117,1:3)=(/4117,3562,0/)
        data_ped(4118,1:3)=(/4118,3038,2610/)
        data_ped(4119,1:3)=(/4119,2551,3525/)
        data_ped(4120,1:3)=(/4120,3488,3622/)
        data_ped(4121,1:3)=(/4121,2775,3608/)
        data_ped(4122,1:3)=(/4122,2955,3534/)
        data_ped(4123,1:3)=(/4123,2551,0/)
        data_ped(4124,1:3)=(/4124,2246,0/)
        data_ped(4125,1:3)=(/4125,2775,0/)
        data_ped(4126,1:3)=(/4126,401,137/)
        data_ped(4127,1:3)=(/4127,1215,0/)
        data_ped(4128,1:3)=(/4128,3566,3551/)
        data_ped(4129,1:3)=(/4129,2053,2827/)
        data_ped(4130,1:3)=(/4130,3405,3533/)
        data_ped(4131,1:3)=(/4131,3566,3346/)
        data_ped(4132,1:3)=(/4132,3478,3365/)
        data_ped(4133,1:3)=(/4133,3474,3328/)
        data_ped(4134,1:3)=(/4134,2707,3229/)
        data_ped(4135,1:3)=(/4135,3107,3323/)
        data_ped(4136,1:3)=(/4136,3296,3333/)
        data_ped(4137,1:3)=(/4137,3190,2884/)
        data_ped(4138,1:3)=(/4138,2841,2418/)
        data_ped(4139,1:3)=(/4139,2843,3309/)
        data_ped(4140,1:3)=(/4140,3112,2566/)
        data_ped(4141,1:3)=(/4141,2616,3503/)
        data_ped(4142,1:3)=(/4142,3466,2973/)
        data_ped(4143,1:3)=(/4143,3112,3065/)
        data_ped(4144,1:3)=(/4144,3565,3687/)
        data_ped(4145,1:3)=(/4145,3286,3738/)
        data_ped(4146,1:3)=(/4146,2949,2979/)
        data_ped(4147,1:3)=(/4147,2538,3509/)
        data_ped(4148,1:3)=(/4148,2706,3513/)
        data_ped(4149,1:3)=(/4149,3098,3136/)
        data_ped(4150,1:3)=(/4150,3119,2991/)
        data_ped(4151,1:3)=(/4151,3190,2697/)
        data_ped(4152,1:3)=(/4152,3383,1218/)
        data_ped(4153,1:3)=(/4153,2722,2929/)
        data_ped(4154,1:3)=(/4154,2799,1265/)
        data_ped(4155,1:3)=(/4155,3285,2872/)
        data_ped(4156,1:3)=(/4156,3465,3541/)
        data_ped(4157,1:3)=(/4157,2789,3237/)
        data_ped(4158,1:3)=(/4158,3301,3273/)
        data_ped(4159,1:3)=(/4159,2845,2926/)
        data_ped(4160,1:3)=(/4160,2854,2984/)
        data_ped(4161,1:3)=(/4161,3557,3371/)
        data_ped(4162,1:3)=(/4162,3575,3723/)
        data_ped(4163,1:3)=(/4163,0,3372/)
        data_ped(4164,1:3)=(/4164,0,3520/)
        data_ped(4165,1:3)=(/4165,3566,3003/)
        data_ped(4166,1:3)=(/4166,3201,3540/)
        data_ped(4167,1:3)=(/4167,2863,1228/)
        data_ped(4168,1:3)=(/4168,3557,2823/)
        data_ped(4169,1:3)=(/4169,2785,3089/)
        data_ped(4170,1:3)=(/4170,3574,2916/)
        data_ped(4171,1:3)=(/4171,2053,0/)
        data_ped(4172,1:3)=(/4172,3491,0/)
        data_ped(4173,1:3)=(/4173,2725,2194/)
        data_ped(4174,1:3)=(/4174,1238,0/)
        data_ped(4175,1:3)=(/4175,3044,3630/)
        data_ped(4176,1:3)=(/4176,3100,3307/)
        data_ped(4177,1:3)=(/4177,3379,2805/)
        data_ped(4178,1:3)=(/4178,3383,3414/)
        data_ped(4179,1:3)=(/4179,3284,3132/)
        data_ped(4180,1:3)=(/4180,3385,2732/)
        data_ped(4181,1:3)=(/4181,3214,2340/)
        data_ped(4182,1:3)=(/4182,2375,2871/)
        data_ped(4183,1:3)=(/4183,3379,3640/)
        data_ped(4184,1:3)=(/4184,3653,3500/)
        data_ped(4185,1:3)=(/4185,3672,3502/)
        data_ped(4186,1:3)=(/4186,2221,2967/)
        data_ped(4187,1:3)=(/4187,2932,2253/)
        data_ped(4188,1:3)=(/4188,0,3501/)
        data_ped(4189,1:3)=(/4189,3646,1199/)
        data_ped(4190,1:3)=(/4190,3105,3172/)
        data_ped(4191,1:3)=(/4191,0,1258/)
        data_ped(4192,1:3)=(/4192,3097,3772/)
        data_ped(4193,1:3)=(/4193,2615,3268/)
        data_ped(4194,1:3)=(/4194,3023,3335/)
        data_ped(4195,1:3)=(/4195,3750,3542/)
        data_ped(4196,1:3)=(/4196,3298,3248/)
        data_ped(4197,1:3)=(/4197,3197,2753/)
        data_ped(4198,1:3)=(/4198,3031,2995/)
        data_ped(4199,1:3)=(/4199,3579,3609/)
        data_ped(4200,1:3)=(/4200,3041,3517/)
        data_ped(4201,1:3)=(/4201,2722,2998/)
        data_ped(4202,1:3)=(/4202,2546,3239/)
        data_ped(4203,1:3)=(/4203,2538,2692/)
        data_ped(4204,1:3)=(/4204,3557,3611/)
        data_ped(4205,1:3)=(/4205,3301,3706/)
        data_ped(4206,1:3)=(/4206,3031,3519/)
        data_ped(4207,1:3)=(/4207,3565,0/)
        data_ped(4208,1:3)=(/4208,3482,2354/)
        data_ped(4209,1:3)=(/4209,3197,1194/)
        data_ped(4210,1:3)=(/4210,3566,0/)
        data_ped(4211,1:3)=(/4211,3300,0/)
        data_ped(4212,1:3)=(/4212,3401,94/)
        data_ped(4213,1:3)=(/4213,3120,1208/)
        data_ped(4214,1:3)=(/4214,3856,0/)
        data_ped(4215,1:3)=(/4215,1271,0/)
        data_ped(4216,1:3)=(/4216,2630,1465/)
        data_ped(4217,1:3)=(/4217,3482,3005/)
        data_ped(4218,1:3)=(/4218,2785,3529/)
        data_ped(4219,1:3)=(/4219,3468,3437/)
        data_ped(4220,1:3)=(/4220,2545,3357/)
        data_ped(4221,1:3)=(/4221,3182,2992/)
        data_ped(4222,1:3)=(/4222,2955,0/)
        data_ped(4223,1:3)=(/4223,3664,0/)
        data_ped(4224,1:3)=(/4224,3198,0/)
        data_ped(4225,1:3)=(/4225,3301,3359/)
        data_ped(4226,1:3)=(/4226,1262,0/)
        data_ped(4227,1:3)=(/4227,1216,0/)
        data_ped(4228,1:3)=(/4228,3295,0/)
        data_ped(4229,1:3)=(/4229,3745,3988/)
        data_ped(4230,1:3)=(/4230,3280,2658/)
        data_ped(4231,1:3)=(/4231,3750,2990/)
        data_ped(4232,1:3)=(/4232,3043,2659/)
        data_ped(4233,1:3)=(/4233,3765,2812/)
        data_ped(4234,1:3)=(/4234,3381,2472/)
        data_ped(4235,1:3)=(/4235,2772,2880/)
        data_ped(4236,1:3)=(/4236,2705,2470/)
        data_ped(4237,1:3)=(/4237,2785,3504/)
        data_ped(4238,1:3)=(/4238,3018,2821/)
        data_ped(4239,1:3)=(/4239,3765,2738/)
        data_ped(4240,1:3)=(/4240,3840,3455/)
        data_ped(4241,1:3)=(/4241,3566,2768/)
        data_ped(4242,1:3)=(/4242,2627,3432/)
        data_ped(4243,1:3)=(/4243,3197,3369/)
        data_ped(4244,1:3)=(/4244,2546,3271/)
        data_ped(4245,1:3)=(/4245,3557,2873/)
        data_ped(4246,1:3)=(/4246,2302,2584/)
        data_ped(4247,1:3)=(/4247,1518,1418/)
        data_ped(4248,1:3)=(/4248,3380,3234/)
        data_ped(4249,1:3)=(/4249,3301,3370/)
        data_ped(4250,1:3)=(/4250,3566,3254/)
        data_ped(4251,1:3)=(/4251,3299,2399/)
        data_ped(4252,1:3)=(/4252,2856,3345/)
        data_ped(4253,1:3)=(/4253,3200,3431/)
        data_ped(4254,1:3)=(/4254,2546,3451/)
        data_ped(4255,1:3)=(/4255,3669,0/)
        data_ped(4256,1:3)=(/4256,3403,0/)
        data_ped(4257,1:3)=(/4257,2721,0/)
        data_ped(4258,1:3)=(/4258,2700,3130/)
        data_ped(4259,1:3)=(/4259,3214,2964/)
        data_ped(4260,1:3)=(/4260,3742,3584/)
        data_ped(4261,1:3)=(/4261,2526,2810/)
        data_ped(4262,1:3)=(/4262,3760,3235/)
        data_ped(4263,1:3)=(/4263,3211,3308/)
        data_ped(4264,1:3)=(/4264,3557,3049/)
        data_ped(4265,1:3)=(/4265,3744,3638/)
        data_ped(4266,1:3)=(/4266,2705,2327/)
        data_ped(4267,1:3)=(/4267,3463,3236/)
        data_ped(4268,1:3)=(/4268,3676,2982/)
        data_ped(4269,1:3)=(/4269,2844,3737/)
        data_ped(4270,1:3)=(/4270,3181,3588/)
        data_ped(4271,1:3)=(/4271,2702,2807/)
        data_ped(4272,1:3)=(/4272,3676,3315/)
        data_ped(4273,1:3)=(/4273,2932,2726/)
        data_ped(4274,1:3)=(/4274,2221,2734/)
        data_ped(4275,1:3)=(/4275,3462,3732/)
        data_ped(4276,1:3)=(/4276,3201,3802/)
        data_ped(4277,1:3)=(/4277,3655,2577/)
        data_ped(4278,1:3)=(/4278,3038,3625/)
        data_ped(4279,1:3)=(/4279,3654,2737/)
        data_ped(4280,1:3)=(/4280,3191,3451/)
        data_ped(4281,1:3)=(/4281,0,2993/)
        data_ped(4282,1:3)=(/4282,0,1196/)
        data_ped(4283,1:3)=(/4283,3665,2999/)
        data_ped(4284,1:3)=(/4284,2546,3364/)
        data_ped(4285,1:3)=(/4285,2863,3893/)
        data_ped(4286,1:3)=(/4286,3200,0/)
        data_ped(4287,1:3)=(/4287,3670,3436/)
        data_ped(4288,1:3)=(/4288,3031,3159/)
        data_ped(4289,1:3)=(/4289,3489,3435/)
        data_ped(4290,1:3)=(/4290,3562,2884/)
        data_ped(4291,1:3)=(/4291,2546,3434/)
        data_ped(4292,1:3)=(/4292,3031,3699/)
        data_ped(4293,1:3)=(/4293,3303,97/)
        data_ped(4294,1:3)=(/4294,3749,1175/)
        data_ped(4295,1:3)=(/4295,3201,2906/)
        data_ped(4296,1:3)=(/4296,3839,3542/)
        data_ped(4297,1:3)=(/4297,2955,3440/)
        data_ped(4298,1:3)=(/4298,2795,1259/)
        data_ped(4299,1:3)=(/4299,2546,2995/)
        data_ped(4300,1:3)=(/4300,3665,3002/)
        data_ped(4301,1:3)=(/4301,3201,3350/)
        data_ped(4302,1:3)=(/4302,2956,1161/)
        data_ped(4303,1:3)=(/4303,3483,2686/)
        data_ped(4304,1:3)=(/4304,3201,3806/)
        data_ped(4305,1:3)=(/4305,3182,3715/)
        data_ped(4306,1:3)=(/4306,2175,0/)
        data_ped(4307,1:3)=(/4307,0,1243/)
        data_ped(4308,1:3)=(/4308,2955,0/)
        data_ped(4309,1:3)=(/4309,1269,0/)
        data_ped(4310,1:3)=(/4310,3665,0/)
        data_ped(4311,1:3)=(/4311,320,0/)
        data_ped(4312,1:3)=(/4312,2955,3903/)
        data_ped(4313,1:3)=(/4313,3038,2685/)
        data_ped(4314,1:3)=(/4314,3301,0/)
        data_ped(4315,1:3)=(/4315,3649,0/)
        data_ped(4316,1:3)=(/4316,3403,0/)
        data_ped(4317,1:3)=(/4317,3833,0/)
        data_ped(4318,1:3)=(/4318,1559,0/)
        data_ped(4319,1:3)=(/4319,3040,0/)
        data_ped(4320,1:3)=(/4320,1236,0/)
        data_ped(4321,1:3)=(/4321,2178,0/)
        data_ped(4322,1:3)=(/4322,1558,0/)
        data_ped(4323,1:3)=(/4323,2240,0/)
        data_ped(4324,1:3)=(/4324,428,0/)
        data_ped(4325,1:3)=(/4325,2324,0/)
        data_ped(4326,1:3)=(/4326,2859,3610/)
        data_ped(4327,1:3)=(/4327,2772,3548/)
        data_ped(4328,1:3)=(/4328,3023,3429/)
        data_ped(4329,1:3)=(/4329,2851,3150/)
        data_ped(4330,1:3)=(/4330,3116,2480/)
        data_ped(4331,1:3)=(/4331,3097,3827/)
        data_ped(4332,1:3)=(/4332,3464,3724/)
        data_ped(4333,1:3)=(/4333,3197,2753/)
        data_ped(4334,1:3)=(/4334,3282,3425/)
        data_ped(4335,1:3)=(/4335,3394,2693/)
        data_ped(4336,1:3)=(/4336,3101,3310/)
        data_ped(4337,1:3)=(/4337,2443,3250/)
        data_ped(4338,1:3)=(/4338,3665,3808/)
        data_ped(4339,1:3)=(/4339,3837,3306/)
        data_ped(4340,1:3)=(/4340,2775,3095/)
        data_ped(4341,1:3)=(/4341,3660,3877/)
        data_ped(4342,1:3)=(/4342,3567,3456/)
        data_ped(4343,1:3)=(/4343,3557,3344/)
        data_ped(4344,1:3)=(/4344,3952,1248/)
        data_ped(4345,1:3)=(/4345,2956,3257/)
        data_ped(4346,1:3)=(/4346,2462,3313/)
        data_ped(4347,1:3)=(/4347,3121,3627/)
        data_ped(4348,1:3)=(/4348,2625,3261/)
        data_ped(4349,1:3)=(/4349,3667,3003/)
        data_ped(4350,1:3)=(/4350,1537,136/)
        data_ped(4351,1:3)=(/4351,2785,0/)
        data_ped(4352,1:3)=(/4352,3665,3535/)
        data_ped(4353,1:3)=(/4353,3182,2763/)
        data_ped(4354,1:3)=(/4354,2962,2433/)
        data_ped(4355,1:3)=(/4355,3762,3768/)
        data_ped(4356,1:3)=(/4356,2772,2802/)
        data_ped(4357,1:3)=(/4357,3675,3681/)
        data_ped(4358,1:3)=(/4358,3097,2742/)
        data_ped(4359,1:3)=(/4359,3128,3134/)
        data_ped(4360,1:3)=(/4360,3470,2971/)
        data_ped(4361,1:3)=(/4361,2846,3499/)
        data_ped(4362,1:3)=(/4362,3191,3451/)
        data_ped(4363,1:3)=(/4363,3482,3803/)
        data_ped(4364,1:3)=(/4364,3557,3805/)
        data_ped(4365,1:3)=(/4365,3113,3422/)
        data_ped(4366,1:3)=(/4366,0,2878/)
        data_ped(4367,1:3)=(/4367,3482,3883/)
        data_ped(4368,1:3)=(/4368,3201,3806/)
        data_ped(4369,1:3)=(/4369,3946,3453/)
        data_ped(4370,1:3)=(/4370,3197,3626/)
        data_ped(4371,1:3)=(/4371,3182,3012/)
        data_ped(4372,1:3)=(/4372,3557,3807/)
        data_ped(4373,1:3)=(/4373,0,2928/)
        data_ped(4374,1:3)=(/4374,3025,2585/)
        data_ped(4375,1:3)=(/4375,3121,3892/)
        data_ped(4376,1:3)=(/4376,2775,3079/)
        data_ped(4377,1:3)=(/4377,3562,3702/)
        data_ped(4378,1:3)=(/4378,3482,2907/)
        data_ped(4379,1:3)=(/4379,2538,3607/)
        data_ped(4380,1:3)=(/4380,2625,0/)
        data_ped(4381,1:3)=(/4381,3961,0/)
        data_ped(4382,1:3)=(/4382,2954,3163/)
        data_ped(4383,1:3)=(/4383,3483,3080/)
        data_ped(4384,1:3)=(/4384,3557,3049/)
        data_ped(4385,1:3)=(/4385,1233,1255/)
        data_ped(4386,1:3)=(/4386,3857,0/)
        data_ped(4387,1:3)=(/4387,2538,2828/)
        data_ped(4388,1:3)=(/4388,3482,3703/)
        data_ped(4389,1:3)=(/4389,3962,2990/)
        data_ped(4390,1:3)=(/4390,3665,0/)
        data_ped(4391,1:3)=(/4391,3482,3890/)
        data_ped(4392,1:3)=(/4392,3756,0/)
        data_ped(4393,1:3)=(/4393,3665,2768/)
        data_ped(4394,1:3)=(/4394,3665,0/)
        data_ped(4395,1:3)=(/4395,3301,3370/)
        data_ped(4396,1:3)=(/4396,3031,0/)
        data_ped(4397,1:3)=(/4397,3402,0/)
        data_ped(4398,1:3)=(/4398,1237,0/)
        data_ped(4399,1:3)=(/4399,0,3801/)
        data_ped(4400,1:3)=(/4400,3464,3916/)
        data_ped(4401,1:3)=(/4401,3017,3598/)
        data_ped(4402,1:3)=(/4402,2226,3770/)
        data_ped(4403,1:3)=(/4403,3291,3419/)
        data_ped(4404,1:3)=(/4404,3742,3242/)
        data_ped(4405,1:3)=(/4405,3495,3912/)
        data_ped(4406,1:3)=(/4406,2450,4033/)
        data_ped(4407,1:3)=(/4407,3213,3278/)
        data_ped(4408,1:3)=(/4408,3938,2653/)
        data_ped(4409,1:3)=(/4409,3760,2728/)
        data_ped(4410,1:3)=(/4410,3556,3321/)
        data_ped(4411,1:3)=(/4411,2952,3729/)
        data_ped(4412,1:3)=(/4412,3018,3312/)
        data_ped(4413,1:3)=(/4413,3483,2767/)
        data_ped(4414,1:3)=(/4414,2625,2924/)
        data_ped(4415,1:3)=(/4415,2775,3591/)
        data_ped(4416,1:3)=(/4416,3196,0/)
        data_ped(4417,1:3)=(/4417,3557,3318/)
        data_ped(4418,1:3)=(/4418,2245,3274/)
        data_ped(4419,1:3)=(/4419,3386,3809/)
        data_ped(4420,1:3)=(/4420,3300,3517/)
        data_ped(4421,1:3)=(/4421,3562,3158/)
        data_ped(4422,1:3)=(/4422,2538,2692/)
        data_ped(4423,1:3)=(/4423,2859,3898/)
        data_ped(4424,1:3)=(/4424,2775,3615/)
        data_ped(4425,1:3)=(/4425,3301,2439/)
        data_ped(4426,1:3)=(/4426,3833,3346/)
        data_ped(4427,1:3)=(/4427,3204,0/)
        data_ped(4428,1:3)=(/4428,2375,3131/)
        data_ped(4429,1:3)=(/4429,3671,3817/)
        data_ped(4430,1:3)=(/4430,2526,2635/)
        data_ped(4431,1:3)=(/4431,3837,3823/)
        data_ped(4432,1:3)=(/4432,2772,3376/)
        data_ped(4433,1:3)=(/4433,2932,3133/)
        data_ped(4434,1:3)=(/4434,3839,3317/)
        data_ped(4435,1:3)=(/4435,3837,3587/)
        data_ped(4436,1:3)=(/4436,3556,3928/)
        data_ped(4437,1:3)=(/4437,3835,3915/)
        data_ped(4438,1:3)=(/4438,3283,3427/)
        data_ped(4439,1:3)=(/4439,3940,150/)
        data_ped(4440,1:3)=(/4440,2787,3091/)
        data_ped(4441,1:3)=(/4441,3749,3804/)
        data_ped(4442,1:3)=(/4442,2462,3910/)
        data_ped(4443,1:3)=(/4443,3957,4150/)
        data_ped(4444,1:3)=(/4444,3386,3693/)
        data_ped(4445,1:3)=(/4445,4067,3156/)
        data_ped(4446,1:3)=(/4446,2538,3607/)
        data_ped(4447,1:3)=(/4447,3392,1427/)
        data_ped(4448,1:3)=(/4448,2859,2668/)
        data_ped(4449,1:3)=(/4449,3665,3808/)
        data_ped(4450,1:3)=(/4450,3190,3700/)
        data_ped(4451,1:3)=(/4451,3190,3159/)
        data_ped(4452,1:3)=(/4452,3482,3444/)
        data_ped(4453,1:3)=(/4453,3301,3876/)
        data_ped(4454,1:3)=(/4454,3848,4000/)
        data_ped(4455,1:3)=(/4455,2546,3364/)
        data_ped(4456,1:3)=(/4456,2625,0/)
        data_ped(4457,1:3)=(/4457,2955,2911/)
        data_ped(4458,1:3)=(/4458,3037,3155/)
        data_ped(4459,1:3)=(/4459,2955,3708/)
        data_ped(4460,1:3)=(/4460,2955,3525/)
        data_ped(4461,1:3)=(/4461,3960,3707/)
        data_ped(4462,1:3)=(/4462,3743,3441/)
        data_ped(4463,1:3)=(/4463,2543,3618/)
        data_ped(4464,1:3)=(/4464,3749,3612/)
        data_ped(4465,1:3)=(/4465,2863,0/)
        data_ped(4466,1:3)=(/4466,1252,0/)
        data_ped(4467,1:3)=(/4467,3557,3450/)
        data_ped(4468,1:3)=(/4468,3557,4014/)
        data_ped(4469,1:3)=(/4469,3845,3005/)
        data_ped(4470,1:3)=(/4470,1261,0/)
        data_ped(4471,1:3)=(/4471,3557,0/)
        data_ped(4472,1:3)=(/4472,3958,3909/)
        data_ped(4473,1:3)=(/4473,393,0/)
        data_ped(4474,1:3)=(/4474,3564,3347/)
        data_ped(4475,1:3)=(/4475,2935,3327/)
        data_ped(4476,1:3)=(/4476,3288,3337/)
        data_ped(4477,1:3)=(/4477,2375,2989/)
        data_ped(4478,1:3)=(/4478,2789,3334/)
        data_ped(4479,1:3)=(/4479,3192,3339/)
        data_ped(4480,1:3)=(/4480,3657,3072/)
        data_ped(4481,1:3)=(/4481,3581,3068/)
        data_ped(4482,1:3)=(/4482,3555,3326/)
        data_ped(4483,1:3)=(/4483,3381,2647/)
        data_ped(4484,1:3)=(/4484,3583,3731/)
        data_ped(4485,1:3)=(/4485,3867,3056/)
        data_ped(4486,1:3)=(/4486,3766,4030/)
        data_ped(4487,1:3)=(/4487,3460,2486/)
        data_ped(4488,1:3)=(/4488,3674,3684/)
        data_ped(4489,1:3)=(/4489,3473,3505/)
        data_ped(4490,1:3)=(/4490,3967,2969/)
        data_ped(4491,1:3)=(/4491,2707,3047/)
        data_ped(4492,1:3)=(/4492,3968,2874/)
        data_ped(4493,1:3)=(/4493,2618,2656/)
        data_ped(4494,1:3)=(/4494,3392,3424/)
        data_ped(4495,1:3)=(/4495,3195,3691/)
        data_ped(4496,1:3)=(/4496,3932,3430/)
        data_ped(4497,1:3)=(/4497,3299,1242/)
        data_ped(4498,1:3)=(/4498,3562,3519/)
        data_ped(4499,1:3)=(/4499,2839,4042/)
        data_ped(4500,1:3)=(/4500,3386,3809/)
        data_ped(4501,1:3)=(/4501,3198,3888/)
        data_ped(4502,1:3)=(/4502,3386,2768/)
        data_ped(4503,1:3)=(/4503,3177,3887/)
        data_ped(4504,1:3)=(/4504,3020,3341/)
        data_ped(4505,1:3)=(/4505,3557,2928/)
        data_ped(4506,1:3)=(/4506,3106,1250/)
        data_ped(4507,1:3)=(/4507,4049,3886/)
        data_ped(4508,1:3)=(/4508,2955,3521/)
        data_ped(4509,1:3)=(/4509,3557,3004/)
        data_ped(4510,1:3)=(/4510,3482,3703/)
        data_ped(4511,1:3)=(/4511,3469,3086/)
        data_ped(4512,1:3)=(/4512,3665,2358/)
        data_ped(4513,1:3)=(/4513,3557,3273/)
        data_ped(4514,1:3)=(/4514,3557,0/)
        data_ped(4515,1:3)=(/4515,1270,0/)
        data_ped(4516,1:3)=(/4516,3557,3370/)
        data_ped(4517,1:3)=(/4517,421,0/)
        data_ped(4518,1:3)=(/4518,2842,0/)
        data_ped(4519,1:3)=(/4519,3977,4032/)
        data_ped(4520,1:3)=(/4520,3761,3736/)
        data_ped(4521,1:3)=(/4521,3214,3919/)
        data_ped(4522,1:3)=(/4522,3378,3585/)
        data_ped(4523,1:3)=(/4523,3837,3642/)
        data_ped(4524,1:3)=(/4524,3119,3689/)
        data_ped(4525,1:3)=(/4525,3378,3728/)
        data_ped(4526,1:3)=(/4526,3214,3918/)
        data_ped(4527,1:3)=(/4527,3108,3735/)
        data_ped(4528,1:3)=(/4528,3678,3319/)
        data_ped(4529,1:3)=(/4529,3128,3874/)
        data_ped(4530,1:3)=(/4530,3044,2889/)
        data_ped(4531,1:3)=(/4531,3462,3634/)
        data_ped(4532,1:3)=(/4532,3746,4041/)
        data_ped(4533,1:3)=(/4533,3557,3894/)
        data_ped(4534,1:3)=(/4534,2845,3269/)
        data_ped(4535,1:3)=(/4535,4154,2603/)
        data_ped(4536,1:3)=(/4536,3400,4282/)
        data_ped(4537,1:3)=(/4537,3483,3080/)
        data_ped(4538,1:3)=(/4538,3031,4102/)
        data_ped(4539,1:3)=(/4539,3562,3092/)
        data_ped(4540,1:3)=(/4540,3385,3792/)
        data_ped(4541,1:3)=(/4541,3482,3255/)
        data_ped(4542,1:3)=(/4542,4048,2685/)
        data_ped(4543,1:3)=(/4543,3121,4005/)
        data_ped(4544,1:3)=(/4544,4061,4109/)
        data_ped(4545,1:3)=(/4545,0,3049/)
        data_ped(4546,1:3)=(/4546,3198,4283/)
        data_ped(4547,1:3)=(/4547,3197,3369/)
        data_ped(4548,1:3)=(/4548,1280,4214/)
        data_ped(4549,1:3)=(/4549,3198,4128/)
        data_ped(4550,1:3)=(/4550,3845,3356/)
        data_ped(4551,1:3)=(/4551,2845,3551/)
        data_ped(4552,1:3)=(/4552,2845,2831/)
        data_ped(4553,1:3)=(/4553,3743,3159/)
        data_ped(4554,1:3)=(/4554,0,2754/)
        data_ped(4555,1:3)=(/4555,3483,3087/)
        data_ped(4556,1:3)=(/4556,3385,3260/)
        data_ped(4557,1:3)=(/4557,3848,3695/)
        data_ped(4558,1:3)=(/4558,2775,3892/)
        data_ped(4559,1:3)=(/4559,3848,3714/)
        data_ped(4560,1:3)=(/4560,3203,3448/)
        data_ped(4561,1:3)=(/4561,1274,0/)
        data_ped(4562,1:3)=(/4562,2129,3905/)
        data_ped(4563,1:3)=(/4563,1246,0/)
        data_ped(4564,1:3)=(/4564,3557,4016/)
        data_ped(4565,1:3)=(/4565,418,0/)
        data_ped(4566,1:3)=(/4566,3557,0/)
        data_ped(4567,1:3)=(/4567,3389,0/)
        data_ped(4568,1:3)=(/4568,1566,0/)
        data_ped(4569,1:3)=(/4569,2226,3590/)
        data_ped(4570,1:3)=(/4570,4049,3776/)
        data_ped(4571,1:3)=(/4571,2932,2888/)
        data_ped(4572,1:3)=(/4572,4157,3816/)
        data_ped(4573,1:3)=(/4573,3671,3135/)
        data_ped(4574,1:3)=(/4574,3764,3234/)
        data_ped(4575,1:3)=(/4575,3193,3931/)
        data_ped(4576,1:3)=(/4576,3843,3602/)
        data_ped(4577,1:3)=(/4577,4052,3986/)
        data_ped(4578,1:3)=(/4578,3097,3882/)
        data_ped(4579,1:3)=(/4579,3559,3600/)
        data_ped(4580,1:3)=(/4580,3494,3927/)
        data_ped(4581,1:3)=(/4581,3837,3175/)
        data_ped(4582,1:3)=(/4582,4076,3734/)
        data_ped(4583,1:3)=(/4583,3760,3497/)
        data_ped(4584,1:3)=(/4584,3843,3596/)
        data_ped(4585,1:3)=(/4585,3494,3683/)
        data_ped(4586,1:3)=(/4586,4046,3688/)
        data_ped(4587,1:3)=(/4587,3761,3594/)
        data_ped(4588,1:3)=(/4588,4051,4088/)
        data_ped(4589,1:3)=(/4589,2618,4040/)
        data_ped(4590,1:3)=(/4590,2853,3152/)
        data_ped(4591,1:3)=(/4591,4073,3420/)
        data_ped(4592,1:3)=(/4592,3935,4084/)
        data_ped(4593,1:3)=(/4593,2375,3140/)
        data_ped(4594,1:3)=(/4594,2845,3094/)
        data_ped(4595,1:3)=(/4595,3973,4144/)
        data_ped(4596,1:3)=(/4596,3119,1217/)
        data_ped(4597,1:3)=(/4597,3944,2587/)
        data_ped(4598,1:3)=(/4598,2462,3910/)
        data_ped(4599,1:3)=(/4599,3571,4207/)
        data_ped(4600,1:3)=(/4600,3943,3439/)
        data_ped(4601,1:3)=(/4601,2460,3693/)
        data_ped(4602,1:3)=(/4602,3037,3608/)
        data_ped(4603,1:3)=(/4603,3392,4091/)
        data_ped(4604,1:3)=(/4604,3743,2995/)
        data_ped(4605,1:3)=(/4605,3743,3078/)
        data_ped(4606,1:3)=(/4606,0,1272/)
        data_ped(4607,1:3)=(/4607,3197,4001/)
        data_ped(4608,1:3)=(/4608,2775,3615/)
        data_ped(4609,1:3)=(/4609,2845,4100/)
        data_ped(4610,1:3)=(/4610,3848,3710/)
        data_ped(4611,1:3)=(/4611,2955,3521/)
        data_ped(4612,1:3)=(/4612,2955,3708/)
        data_ped(4613,1:3)=(/4613,3564,4121/)
        data_ped(4614,1:3)=(/4614,2863,3893/)
        data_ped(4615,1:3)=(/4615,3198,0/)
        data_ped(4616,1:3)=(/4616,3018,3767/)
        data_ped(4617,1:3)=(/4617,3966,3215/)
        data_ped(4618,1:3)=(/4618,3765,3920/)
        data_ped(4619,1:3)=(/4619,4155,3929/)
        data_ped(4620,1:3)=(/4620,3128,3685/)
        data_ped(4621,1:3)=(/4621,3748,4149/)
        data_ped(4622,1:3)=(/4622,3863,2872/)
        data_ped(4623,1:3)=(/4623,4048,2873/)
        data_ped(4624,1:3)=(/4624,3410,3320/)
        data_ped(4625,1:3)=(/4625,3869,3589/)
        data_ped(4626,1:3)=(/4626,3557,3876/)
        data_ped(4627,1:3)=(/4627,4162,3690/)
        data_ped(4628,1:3)=(/4628,3022,4092/)
        data_ped(4629,1:3)=(/4629,2845,3739/)
        data_ped(4630,1:3)=(/4630,3843,3540/)
        data_ped(4631,1:3)=(/4631,3864,4091/)
        data_ped(4632,1:3)=(/4632,0,3155/)
        data_ped(4633,1:3)=(/4633,3482,3890/)
        data_ped(4634,1:3)=(/4634,4057,3696/)
        data_ped(4635,1:3)=(/4635,0,2899/)
        data_ped(4636,1:3)=(/4636,4058,4287/)
        data_ped(4637,1:3)=(/4637,3392,4107/)
        data_ped(4638,1:3)=(/4638,3467,3802/)
        data_ped(4639,1:3)=(/4639,3475,3437/)
        data_ped(4640,1:3)=(/4640,2323,3995/)
        data_ped(4641,1:3)=(/4641,3197,3369/)
        data_ped(4642,1:3)=(/4642,3843,3602/)
        data_ped(4643,1:3)=(/4643,4060,4007/)
        data_ped(4644,1:3)=(/4644,3392,3790/)
        data_ped(4645,1:3)=(/4645,4347,3352/)
        data_ped(4646,1:3)=(/4646,319,4102/)
        data_ped(4647,1:3)=(/4647,3838,3001/)
        data_ped(4648,1:3)=(/4648,3302,4120/)
        data_ped(4649,1:3)=(/4649,3743,4117/)
        data_ped(4650,1:3)=(/4650,1273,1249/)
        data_ped(4651,1:3)=(/4651,3475,3165/)
        data_ped(4652,1:3)=(/4652,3203,2429/)
        data_ped(4653,1:3)=(/4653,3557,3228/)
        data_ped(4654,1:3)=(/4654,3121,2668/)
        data_ped(4655,1:3)=(/4655,3845,2670/)
        data_ped(4656,1:3)=(/4656,3202,0/)
        data_ped(4657,1:3)=(/4657,3950,3798/)
        data_ped(4658,1:3)=(/4658,2845,3168/)
        data_ped(4659,1:3)=(/4659,3953,0/)
        data_ped(4660,1:3)=(/4660,1252,3170/)
        data_ped(4661,1:3)=(/4661,4159,3524/)
        data_ped(4662,1:3)=(/4662,4048,2928/)
        data_ped(4663,1:3)=(/4663,3859,0/)
        data_ped(4664,1:3)=(/4664,3475,0/)
        data_ped(4665,1:3)=(/4665,1275,0/)
        data_ped(4666,1:3)=(/4666,4172,4227/)
        data_ped(4667,1:3)=(/4667,3764,3983/)
        data_ped(4668,1:3)=(/4668,2772,4029/)
        data_ped(4669,1:3)=(/4669,3381,3058/)
        data_ped(4670,1:3)=(/4670,3462,3922/)
        data_ped(4671,1:3)=(/4671,3556,3773/)
        data_ped(4672,1:3)=(/4672,3838,3154/)
        data_ped(4673,1:3)=(/4673,3097,4035/)
        data_ped(4674,1:3)=(/4674,3297,3603/)
        data_ped(4675,1:3)=(/4675,3018,3550/)
        data_ped(4676,1:3)=(/4676,3299,4191/)
        data_ped(4677,1:3)=(/4677,3467,4276/)
        data_ped(4678,1:3)=(/4678,3658,4189/)
        data_ped(4679,1:3)=(/4679,3743,3699/)
        data_ped(4680,1:3)=(/4680,3935,3694/)
        data_ped(4681,1:3)=(/4681,4412,2881/)
        data_ped(4682,1:3)=(/4682,3198,4131/)
        data_ped(4683,1:3)=(/4683,3386,3980/)
        data_ped(4684,1:3)=(/4684,3651,3054/)
        data_ped(4685,1:3)=(/4685,3743,4086/)
        data_ped(4686,1:3)=(/4686,3557,4016/)
        data_ped(4687,1:3)=(/4687,3037,3627/)
        data_ped(4688,1:3)=(/4688,2776,3880/)
        data_ped(4689,1:3)=(/4689,4251,3593/)
        data_ped(4690,1:3)=(/4690,4154,4399/)
        data_ped(4691,1:3)=(/4691,3392,3701/)
        data_ped(4692,1:3)=(/4692,3564,4286/)
        data_ped(4693,1:3)=(/4693,0,3775/)
        data_ped(4694,1:3)=(/4694,3569,4372/)
        data_ped(4695,1:3)=(/4695,3851,3698/)
        data_ped(4696,1:3)=(/4696,4052,3246/)
        data_ped(4697,1:3)=(/4697,4060,3990/)
        data_ped(4698,1:3)=(/4698,3392,4197/)
        data_ped(4699,1:3)=(/4699,319,3451/)
        data_ped(4700,1:3)=(/4700,4415,4305/)
        data_ped(4701,1:3)=(/4701,3197,4001/)
        data_ped(4702,1:3)=(/4702,4060,2692/)
        data_ped(4703,1:3)=(/4703,3121,0/)
        data_ped(4704,1:3)=(/4704,3198,3619/)
        data_ped(4705,1:3)=(/4705,3756,3902/)
        data_ped(4706,1:3)=(/4706,4342,0/)
        data_ped(4707,1:3)=(/4707,3214,3412/)
        data_ped(4708,1:3)=(/4708,3097,3377/)
        data_ped(4709,1:3)=(/4709,3104,3413/)
        data_ped(4710,1:3)=(/4710,3834,4017/)
        data_ped(4711,1:3)=(/4711,3214,3631/)
        data_ped(4712,1:3)=(/4712,3106,3545/)
        data_ped(4713,1:3)=(/4713,2845,3318/)
        data_ped(4714,1:3)=(/4714,2708,3639/)
        data_ped(4715,1:3)=(/4715,3475,3881/)
        data_ped(4716,1:3)=(/4716,4050,3501/)
        data_ped(4717,1:3)=(/4717,3577,3720/)
        data_ped(4718,1:3)=(/4718,3177,1279/)
        data_ped(4719,1:3)=(/4719,3950,3342/)
        data_ped(4720,1:3)=(/4720,0,4195/)
        data_ped(4721,1:3)=(/4721,3294,2987/)
        data_ped(4722,1:3)=(/4722,4252,4199/)
        data_ped(4723,1:3)=(/4723,3567,3456/)
        data_ped(4724,1:3)=(/4724,4426,3991/)
        data_ped(4725,1:3)=(/4725,3842,3895/)
        data_ped(4726,1:3)=(/4726,4059,4201/)
        data_ped(4727,1:3)=(/4727,3564,3787/)
        data_ped(4728,1:3)=(/4728,3037,4094/)
        data_ped(4729,1:3)=(/4729,4052,3896/)
        data_ped(4730,1:3)=(/4730,4415,2760/)
        data_ped(4731,1:3)=(/4731,3833,3704/)
        data_ped(4732,1:3)=(/4732,4056,3255/)
        data_ped(4733,1:3)=(/4733,4159,3444/)
        data_ped(4734,1:3)=(/4734,4245,3627/)
        data_ped(4735,1:3)=(/4735,4340,4093/)
        data_ped(4736,1:3)=(/4736,3197,3369/)
        data_ped(4737,1:3)=(/4737,0,102/)
        data_ped(4738,1:3)=(/4738,3037,3445/)
        data_ped(4739,1:3)=(/4739,3571,3164/)
        data_ped(4740,1:3)=(/4740,3848,2924/)
        data_ped(4741,1:3)=(/4741,1281,1197/)
        data_ped(4742,1:3)=(/4742,2721,3901/)
        data_ped(4743,1:3)=(/4743,3753,4214/)
        data_ped(4744,1:3)=(/4744,3121,3797/)
        data_ped(4745,1:3)=(/4745,3287,3157/)
        data_ped(4746,1:3)=(/4746,3287,3899/)
        data_ped(4747,1:3)=(/4747,3666,2915/)
        data_ped(4748,1:3)=(/4748,4056,0/)
        data_ped(4749,1:3)=(/4749,3287,4108/)
        data_ped(4750,1:3)=(/4750,4256,3262/)
        data_ped(4751,1:3)=(/4751,3564,0/)
        data_ped(4752,1:3)=(/4752,2955,3521/)
        data_ped(4753,1:3)=(/4753,3577,4241/)
        data_ped(4754,1:3)=(/4754,3951,3009/)
        data_ped(4755,1:3)=(/4755,4515,4226/)
        data_ped(4756,1:3)=(/4756,2797,2429/)
        data_ped(4757,1:3)=(/4757,2845,3739/)
        data_ped(4758,1:3)=(/4758,4060,3794/)
        data_ped(4759,1:3)=(/4759,3962,0/)
        data_ped(4760,1:3)=(/4760,3210,4215/)
        data_ped(4761,1:3)=(/4761,4427,0/)
        data_ped(4762,1:3)=(/4762,3127,0/)
        data_ped(4763,1:3)=(/4763,3861,3230/)
        data_ped(4764,1:3)=(/4764,3974,2885/)
        data_ped(4765,1:3)=(/4765,3740,3142/)
        data_ped(4766,1:3)=(/4766,3967,3053/)
        data_ped(4767,1:3)=(/4767,3180,3416/)
        data_ped(4768,1:3)=(/4768,3763,1284/)
        data_ped(4769,1:3)=(/4769,4044,3601/)
        data_ped(4770,1:3)=(/4770,2375,3131/)
        data_ped(4771,1:3)=(/4771,4187,3921/)
        data_ped(4772,1:3)=(/4772,3869,3508/)
        data_ped(4773,1:3)=(/4773,3975,3879/)
        data_ped(4774,1:3)=(/4774,3462,3635/)
        data_ped(4775,1:3)=(/4775,4263,3885/)
        data_ped(4776,1:3)=(/4776,4052,3246/)
        data_ped(4777,1:3)=(/4777,3556,3599/)
        data_ped(4778,1:3)=(/4778,4184,3822/)
        data_ped(4779,1:3)=(/4779,3760,2728/)
        data_ped(4780,1:3)=(/4780,3097,3423/)
        data_ped(4781,1:3)=(/4781,3868,3336/)
        data_ped(4782,1:3)=(/4782,3749,3153/)
        data_ped(4783,1:3)=(/4783,3747,4196/)
        data_ped(4784,1:3)=(/4784,4072,3930/)
        data_ped(4785,1:3)=(/4785,3935,4084/)
        data_ped(4786,1:3)=(/4786,3583,3421/)
        data_ped(4787,1:3)=(/4787,0,1268/)
        data_ped(4788,1:3)=(/4788,3467,4374/)
        data_ped(4789,1:3)=(/4789,3022,3686/)
        data_ped(4790,1:3)=(/4790,3832,4132/)
        data_ped(4791,1:3)=(/4791,2546,3783/)
        data_ped(4792,1:3)=(/4792,4245,3891/)
        data_ped(4793,1:3)=(/4793,3941,3781/)
        data_ped(4794,1:3)=(/4794,1522,110/)
        data_ped(4795,1:3)=(/4795,279,108/)
        data_ped(4796,1:3)=(/4796,3482,2670/)
        data_ped(4797,1:3)=(/4797,2543,3348/)
        data_ped(4798,1:3)=(/4798,3565,3437/)
        data_ped(4799,1:3)=(/4799,3564,4213/)
        data_ped(4800,1:3)=(/4800,4056,3353/)
        data_ped(4801,1:3)=(/4801,4060,3788/)
        data_ped(4802,1:3)=(/4802,2546,3699/)
        data_ped(4803,1:3)=(/4803,313,106/)
        data_ped(4804,1:3)=(/4804,2862,2761/)
        data_ped(4805,1:3)=(/4805,4063,4129/)
        data_ped(4806,1:3)=(/4806,4336,3735/)
        data_ped(4807,1:3)=(/4807,3097,4408/)
        data_ped(4808,1:3)=(/4808,3462,3277/)
        data_ped(4809,1:3)=(/4809,4243,3769/)
        data_ped(4810,1:3)=(/4810,4151,3552/)
        data_ped(4811,1:3)=(/4811,3379,3310/)
        data_ped(4812,1:3)=(/4812,3284,3924/)
        data_ped(4813,1:3)=(/4813,3967,2969/)
        data_ped(4814,1:3)=(/4814,4071,3507/)
        data_ped(4815,1:3)=(/4815,4186,3547/)
        data_ped(4816,1:3)=(/4816,3119,1416/)
        data_ped(4817,1:3)=(/4817,4347,4445/)
        data_ped(4818,1:3)=(/4818,3191,2145/)
        data_ped(4819,1:3)=(/4819,4055,3542/)
        data_ped(4820,1:3)=(/4820,3743,4291/)
        data_ped(4821,1:3)=(/4821,4597,3693/)
        data_ped(4822,1:3)=(/4822,3482,3703/)
        data_ped(4823,1:3)=(/4823,3743,2697/)
        data_ped(4824,1:3)=(/4824,2546,3782/)
        data_ped(4825,1:3)=(/4825,4416,0/)
        data_ped(4826,1:3)=(/4826,3659,3996/)
        data_ped(4827,1:3)=(/4827,0,3442/)
        data_ped(4828,1:3)=(/4828,4250,3256/)
        data_ped(4829,1:3)=(/4829,4060,3626/)
        data_ped(4830,1:3)=(/4830,4062,3713/)
        data_ped(4831,1:3)=(/4831,4159,3524/)
        data_ped(4832,1:3)=(/4832,3392,3806/)
        data_ped(4833,1:3)=(/4833,4153,4115/)
        data_ped(4834,1:3)=(/4834,2546,3500/)
        data_ped(4835,1:3)=(/4835,3848,3789/)
        data_ped(4836,1:3)=(/4836,2845,4012/)
        data_ped(4837,1:3)=(/4837,4415,4305/)
        data_ped(4838,1:3)=(/4838,3482,0/)
        data_ped(4839,1:3)=(/4839,4055,4020/)
        data_ped(4840,1:3)=(/4840,2863,3264/)
        data_ped(4841,1:3)=(/4841,4351,3537/)
        data_ped(4842,1:3)=(/4842,4060,0/)
        data_ped(4843,1:3)=(/4843,4244,4362/)
        data_ped(4844,1:3)=(/4844,4249,2873/)
        data_ped(4845,1:3)=(/4845,3564,4326/)
        data_ped(4846,1:3)=(/4846,4351,0/)
        data_ped(4847,1:3)=(/4847,3564,4121/)
        data_ped(4848,1:3)=(/4848,4257,0/)
        data_ped(4849,1:3)=(/4849,4055,0/)
        data_ped(4850,1:3)=(/4850,3862,4135/)
        data_ped(4851,1:3)=(/4851,3178,3730/)
        data_ped(4852,1:3)=(/4852,3462,3633/)
        data_ped(4853,1:3)=(/4853,3972,3059/)
        data_ped(4854,1:3)=(/4854,3097,3377/)
        data_ped(4855,1:3)=(/4855,4274,1276/)
        data_ped(4856,1:3)=(/4856,3285,3682/)
        data_ped(4857,1:3)=(/4857,4358,3637/)
        data_ped(4858,1:3)=(/4858,4249,4237/)
        data_ped(4859,1:3)=(/4859,4161,3829/)
        data_ped(4860,1:3)=(/4860,4250,3553/)
        data_ped(4861,1:3)=(/4861,3934,3871/)
        data_ped(4862,1:3)=(/4862,4412,3/)
        data_ped(4863,1:3)=(/4863,4597,4444/)
        data_ped(4864,1:3)=(/4864,3743,4097/)
        data_ped(4865,1:3)=(/4865,3565,3881/)
        data_ped(4866,1:3)=(/4866,4244,3993/)
        data_ped(4867,1:3)=(/4867,2543,3348/)
        data_ped(4868,1:3)=(/4868,4250,3094/)
        data_ped(4869,1:3)=(/4869,3743,4086/)
        data_ped(4870,1:3)=(/4870,3198,3254/)
        data_ped(4871,1:3)=(/4871,3943,4094/)
        data_ped(4872,1:3)=(/4872,4502,1287/)
        data_ped(4873,1:3)=(/4873,3848,4106/)
        data_ped(4874,1:3)=(/4874,2845,4300/)
        data_ped(4875,1:3)=(/4875,3482,4303/)
        data_ped(4876,1:3)=(/4876,3851,4210/)
        data_ped(4877,1:3)=(/4877,0,101/)
        data_ped(4878,1:3)=(/4878,4601,3722/)
        data_ped(4879,1:3)=(/4879,2797,3617/)
        data_ped(4880,1:3)=(/4880,3208,4316/)
        data_ped(4881,1:3)=(/4881,3943,0/)
        data_ped(4882,1:3)=(/4882,3953,0/)
        data_ped(4883,1:3)=(/4883,3468,3532/)
        data_ped(4884,1:3)=(/4884,3484,4212/)
        data_ped(4885,1:3)=(/4885,3393,4228/)
        data_ped(4886,1:3)=(/4886,2933,4141/)
        data_ped(4887,1:3)=(/4887,3864,4369/)
        data_ped(4888,1:3)=(/4888,3865,3913/)
        data_ped(4889,1:3)=(/4889,3183,3981/)
        data_ped(4890,1:3)=(/4890,3287,4364/)
        data_ped(4891,1:3)=(/4891,3472,3873/)
        data_ped(4892,1:3)=(/4892,3214,3457/)
        data_ped(4893,1:3)=(/4893,4412,2881/)
        data_ped(4894,1:3)=(/4894,2775,4375/)
        data_ped(4895,1:3)=(/4895,4414,3368/)
        data_ped(4896,1:3)=(/4896,4502,4278/)
        data_ped(4897,1:3)=(/4897,4255,1292/)
        data_ped(4898,1:3)=(/4898,3838,3986/)
        data_ped(4899,1:3)=(/4899,3672,4496/)
        data_ped(4900,1:3)=(/4900,3198,4283/)
        data_ped(4901,1:3)=(/4901,2546,4098/)
        data_ped(4902,1:3)=(/4902,4502,3784/)
        data_ped(4903,1:3)=(/4903,3743,4096/)
        data_ped(4904,1:3)=(/4904,3838,4298/)
        data_ped(4905,1:3)=(/4905,4245,3012/)
        data_ped(4906,1:3)=(/4906,3953,1267/)
        data_ped(4907,1:3)=(/4907,3467,4209/)
        data_ped(4908,1:3)=(/4908,4056,4452/)
        data_ped(4909,1:3)=(/4909,4056,3703/)
        data_ped(4910,1:3)=(/4910,4166,3999/)
        data_ped(4911,1:3)=(/4911,3749,3612/)
        data_ped(4912,1:3)=(/4912,3848,3780/)
        data_ped(4913,1:3)=(/4913,3467,4197/)
        data_ped(4914,1:3)=(/4914,4414,0/)
        data_ped(4915,1:3)=(/4915,4056,4550/)
        data_ped(4916,1:3)=(/4916,3743,4202/)
        data_ped(4917,1:3)=(/4917,4053,0/)
        data_ped(4918,1:3)=(/4918,4056,3803/)
        data_ped(4919,1:3)=(/4919,1282,4311/)
        data_ped(4920,1:3)=(/4920,4502,4310/)
        data_ped(4921,1:3)=(/4921,4502,0/)
        data_ped(4922,1:3)=(/4922,4056,0/)
        data_ped(4923,1:3)=(/4923,4056,4378/)
        data_ped(4924,1:3)=(/4924,4250,3795/)
        data_ped(4925,1:3)=(/4925,3398,2754/)
        data_ped(4926,1:3)=(/4926,4058,0/)
        data_ped(4927,1:3)=(/4927,4608,3717/)
        data_ped(4928,1:3)=(/4928,4056,3080/)
        data_ped(4929,1:3)=(/4929,2860,4105/)
        data_ped(4930,1:3)=(/4930,3483,4116/)
        data_ped(4931,1:3)=(/4931,1191,0/)
        data_ped(4932,1:3)=(/4932,3386,0/)
        data_ped(4933,1:3)=(/4933,430,0/)
        data_ped(4934,1:3)=(/4934,4427,4325/)
        data_ped(4935,1:3)=(/4935,4359,4480/)
        data_ped(4936,1:3)=(/4936,3758,4230/)
        data_ped(4937,1:3)=(/4937,3841,3063/)
        data_ped(4938,1:3)=(/4938,4337,4194/)
        data_ped(4939,1:3)=(/4939,3964,1264/)
        data_ped(4940,1:3)=(/4940,4243,4370/)
        data_ped(4941,1:3)=(/4941,3189,3228/)
        data_ped(4942,1:3)=(/4942,4052,3878/)
        data_ped(4943,1:3)=(/4943,4184,3822/)
        data_ped(4944,1:3)=(/4944,4412,4019/)
        data_ped(4945,1:3)=(/4945,3936,3632/)
        data_ped(4946,1:3)=(/4946,3392,3510/)
        data_ped(4947,1:3)=(/4947,3656,4368/)
        data_ped(4948,1:3)=(/4948,4502,4629/)
        data_ped(4949,1:3)=(/4949,3461,2743/)
        data_ped(4950,1:3)=(/4950,4081,3810/)
        data_ped(4951,1:3)=(/4951,3661,3604/)
        data_ped(4952,1:3)=(/4952,4074,4026/)
        data_ped(4953,1:3)=(/4953,3935,4085/)
        data_ped(4954,1:3)=(/4954,4437,4479/)
        data_ped(4955,1:3)=(/4955,3189,3894/)
        data_ped(4956,1:3)=(/4956,4420,4474/)
        data_ped(4957,1:3)=(/4957,2941,3771/)
        data_ped(4958,1:3)=(/4958,4353,3992/)
        data_ped(4959,1:3)=(/4959,3743,4086/)
        data_ped(4960,1:3)=(/4960,4344,3994/)
        data_ped(4961,1:3)=(/4961,3467,3701/)
        data_ped(4962,1:3)=(/4962,4060,4110/)
        data_ped(4963,1:3)=(/4963,4159,4208/)
        data_ped(4964,1:3)=(/4964,3569,2873/)
        data_ped(4965,1:3)=(/4965,3848,4457/)
        data_ped(4966,1:3)=(/4966,4502,4552/)
        data_ped(4967,1:3)=(/4967,4060,3990/)
        data_ped(4968,1:3)=(/4968,3851,4468/)
        data_ped(4969,1:3)=(/4969,4063,4561/)
        data_ped(4970,1:3)=(/4970,3480,0/)
        data_ped(4971,1:3)=(/4971,1240,0/)
        data_ped(4972,1:3)=(/4972,3963,4320/)
        data_ped(4973,1:3)=(/4973,3204,0/)
        data_ped(4974,1:3)=(/4974,4077,3458/)
        data_ped(4975,1:3)=(/4975,4270,4439/)
        data_ped(4976,1:3)=(/4976,3384,4232/)
        data_ped(4977,1:3)=(/4977,4412,2881/)
        data_ped(4978,1:3)=(/4978,4175,4478/)
        data_ped(4979,1:3)=(/4979,3018,3312/)
        data_ped(4980,1:3)=(/4980,4245,3313/)
        data_ped(4981,1:3)=(/4981,4181,3819/)
        data_ped(4982,1:3)=(/4982,4245,4371/)
        data_ped(4983,1:3)=(/4983,4245,4441/)
        data_ped(4984,1:3)=(/4984,3097,4408/)
        data_ped(4985,1:3)=(/4985,3671,4133/)
        data_ped(4986,1:3)=(/4986,3741,3875/)
        data_ped(4987,1:3)=(/4987,3097,4143/)
        data_ped(4988,1:3)=(/4988,3476,4089/)
        data_ped(4989,1:3)=(/4989,4502,4534/)
        data_ped(4990,1:3)=(/4990,4690,4535/)
        data_ped(4991,1:3)=(/4991,4412,4082/)
        data_ped(4992,1:3)=(/4992,4502,3979/)
        data_ped(4993,1:3)=(/4993,4606,3095/)
        data_ped(4994,1:3)=(/4994,3301,3252/)
        data_ped(4995,1:3)=(/4995,4244,4362/)
        data_ped(4996,1:3)=(/4996,4601,3722/)
        data_ped(4997,1:3)=(/4997,4421,4202/)
        data_ped(4998,1:3)=(/4998,4502,3739/)
        data_ped(4999,1:3)=(/4999,3836,4295/)
        data_ped(5000,1:3)=(/5000,3749,3345/)
        data_ped(5001,1:3)=(/5001,4252,4103/)
        data_ped(5002,1:3)=(/5002,4421,4540/)
        data_ped(5003,1:3)=(/5003,3943,4640/)
        data_ped(5004,1:3)=(/5004,4167,4380/)
        data_ped(5005,1:3)=(/5005,3756,3902/)
        data_ped(5006,1:3)=(/5006,4056,3524/)
        data_ped(5007,1:3)=(/5007,3659,4584/)
        data_ped(5008,1:3)=(/5008,3467,4547/)
        data_ped(5009,1:3)=(/5009,3565,3709/)
        data_ped(5010,1:3)=(/5010,4420,2754/)
        data_ped(5011,1:3)=(/5011,3564,4121/)
        data_ped(5012,1:3)=(/5012,4418,0/)
        data_ped(5013,1:3)=(/5013,4705,4392/)
        data_ped(5014,1:3)=(/5014,3848,0/)
        data_ped(5015,1:3)=(/5015,4056,3703/)
        data_ped(5016,1:3)=(/5016,1290,0/)
        data_ped(5017,1:3)=(/5017,3663,4333/)
        data_ped(5018,1:3)=(/5018,4060,4379/)
        data_ped(5019,1:3)=(/5019,3486,0/)
        data_ped(5020,1:3)=(/5020,2961,4317/)
        data_ped(5021,1:3)=(/5021,4165,0/)
        data_ped(5022,1:3)=(/5022,4159,0/)
        data_ped(5023,1:3)=(/5023,3755,3536/)
        data_ped(5024,1:3)=(/5024,2797,0/)
        data_ped(5025,1:3)=(/5025,3467,4209/)
        data_ped(5026,1:3)=(/5026,3492,142/)
        data_ped(5027,1:3)=(/5027,3492,0/)
        data_ped(5028,1:3)=(/5028,4258,4235/)
        data_ped(5029,1:3)=(/5029,3110,3982/)
        data_ped(5030,1:3)=(/5030,3212,3813/)
        data_ped(5031,1:3)=(/5031,4076,1283/)
        data_ped(5032,1:3)=(/5032,4047,4411/)
        data_ped(5033,1:3)=(/5033,3462,3636/)
        data_ped(5034,1:3)=(/5034,4075,4279/)
        data_ped(5035,1:3)=(/5035,3844,4533/)
        data_ped(5036,1:3)=(/5036,4681,2881/)
        data_ped(5037,1:3)=(/5037,3494,3324/)
        data_ped(5038,1:3)=(/5038,3661,1414/)
        data_ped(5039,1:3)=(/5039,4261,4667/)
        data_ped(5040,1:3)=(/5040,3677,3645/)
        data_ped(5041,1:3)=(/5041,4158,3979/)
        data_ped(5042,1:3)=(/5042,3751,3694/)
        data_ped(5043,1:3)=(/5043,3743,4290/)
        data_ped(5044,1:3)=(/5044,4335,4190/)
        data_ped(5045,1:3)=(/5045,3467,3701/)
        data_ped(5046,1:3)=(/5046,3206,1286/)
        data_ped(5047,1:3)=(/5047,3948,4008/)
        data_ped(5048,1:3)=(/5048,4602,4375/)
        data_ped(5049,1:3)=(/5049,4060,4446/)
        data_ped(5050,1:3)=(/5050,3567,4543/)
        data_ped(5051,1:3)=(/5051,3659,3996/)
        data_ped(5052,1:3)=(/5052,4243,4113/)
        data_ped(5053,1:3)=(/5053,4159,4003/)
        data_ped(5054,1:3)=(/5054,2862,4009/)
        data_ped(5055,1:3)=(/5055,3122,4124/)
        data_ped(5056,1:3)=(/5056,4068,4376/)
        data_ped(5057,1:3)=(/5057,4517,4321/)
        data_ped(5058,1:3)=(/5058,3490,4564/)
        data_ped(5059,1:3)=(/5059,3933,0/)
        data_ped(5060,1:3)=(/5060,4502,0/)
        data_ped(5061,1:3)=(/5061,3976,3870/)
        data_ped(5062,1:3)=(/5062,4622,4673/)
        data_ped(5063,1:3)=(/5063,4681,4082/)
        data_ped(5064,1:3)=(/5064,3189,3228/)
        data_ped(5065,1:3)=(/5065,3648,4485/)
        data_ped(5066,1:3)=(/5066,4341,4193/)
        data_ped(5067,1:3)=(/5067,4595,3873/)
        data_ped(5068,1:3)=(/5068,3097,4408/)
        data_ped(5069,1:3)=(/5069,4151,3597/)
        data_ped(5070,1:3)=(/5070,4260,4486/)
        data_ped(5071,1:3)=(/5071,4505,3767/)
        data_ped(5072,1:3)=(/5072,4153,4494/)
        data_ped(5073,1:3)=(/5073,4245,3889/)
        data_ped(5074,1:3)=(/5074,3743,4198/)
        data_ped(5075,1:3)=(/5075,4252,4635/)
        data_ped(5076,1:3)=(/5076,4693,3779/)
        data_ped(5077,1:3)=(/5077,4415,4305/)
        data_ped(5078,1:3)=(/5078,4159,4541/)
        data_ped(5079,1:3)=(/5079,4413,4731/)
        data_ped(5080,1:3)=(/5080,4056,4383/)
        data_ped(5081,1:3)=(/5081,3833,3614/)
        data_ped(5082,1:3)=(/5082,3182,0/)
        data_ped(5083,1:3)=(/5083,4346,4445/)
        data_ped(5084,1:3)=(/5084,3388,4735/)
        data_ped(5085,1:3)=(/5085,2955,4456/)
        data_ped(5086,1:3)=(/5086,4060,4446/)
        data_ped(5087,1:3)=(/5087,3562,3500/)
        data_ped(5088,1:3)=(/5088,4687,4386/)
        data_ped(5089,1:3)=(/5089,4877,4737/)
        data_ped(5090,1:3)=(/5090,3659,3996/)
        data_ped(5091,1:3)=(/5091,3189,4131/)
        data_ped(5092,1:3)=(/5092,4069,3527/)
        data_ped(5093,1:3)=(/5093,4594,3551/)
        data_ped(5094,1:3)=(/5094,3182,4558/)
        data_ped(5095,1:3)=(/5095,4167,0/)
        data_ped(5096,1:3)=(/5096,2955,3530/)
        data_ped(5097,1:3)=(/5097,3189,3254/)
        data_ped(5098,1:3)=(/5098,4056,0/)
        data_ped(5099,1:3)=(/5099,4352,4224/)
        data_ped(5100,1:3)=(/5100,4414,0/)
        data_ped(5101,1:3)=(/5101,4805,0/)
        data_ped(5102,1:3)=(/5102,1289,0/)
        data_ped(5103,1:3)=(/5103,4060,3369/)
        data_ped(5104,1:3)=(/5104,4060,4001/)
        data_ped(5105,1:3)=(/5105,4414,0/)
        data_ped(5106,1:3)=(/5106,3851,0/)
        data_ped(5107,1:3)=(/5107,2845,4566/)
        data_ped(5108,1:3)=(/5108,3953,0/)
        data_ped(5109,1:3)=(/5109,3560,1285/)
        data_ped(5110,1:3)=(/5110,4437,4328/)
        data_ped(5111,1:3)=(/5111,4158,4278/)
        data_ped(5112,1:3)=(/5112,4269,4578/)
        data_ped(5113,1:3)=(/5113,3659,4630/)
        data_ped(5114,1:3)=(/5114,3108,4585/)
        data_ped(5115,1:3)=(/5115,4079,4145/)
        data_ped(5116,1:3)=(/5116,4505,3829/)
        data_ped(5117,1:3)=(/5117,4417,3228/)
        data_ped(5118,1:3)=(/5118,4505,3553/)
        data_ped(5119,1:3)=(/5119,4505,4237/)
        data_ped(5120,1:3)=(/5120,3937,3595/)
        data_ped(5121,1:3)=(/5121,3214,3221/)
        data_ped(5122,1:3)=(/5122,3108,3825/)
        data_ped(5123,1:3)=(/5123,4360,3923/)
        data_ped(5124,1:3)=(/5124,4627,4240/)
        data_ped(5125,1:3)=(/5125,4689,4229/)
        data_ped(5126,1:3)=(/5126,4152,3872/)
        data_ped(5127,1:3)=(/5127,3663,4333/)
        data_ped(5128,1:3)=(/5128,4708,4582/)
        data_ped(5129,1:3)=(/5129,4245,4632/)
        data_ped(5130,1:3)=(/5130,3965,3725/)
        data_ped(5131,1:3)=(/5131,4245,4442/)
        data_ped(5132,1:3)=(/5132,4698,4729/)
        data_ped(5133,1:3)=(/5133,4419,4204/)
        data_ped(5134,1:3)=(/5134,3480,4306/)
        data_ped(5135,1:3)=(/5135,3847,4296/)
        data_ped(5136,1:3)=(/5136,3851,4004/)
        data_ped(5137,1:3)=(/5137,3847,4389/)
        data_ped(5138,1:3)=(/5138,4167,3893/)
        data_ped(5139,1:3)=(/5139,4423,3439/)
        data_ped(5140,1:3)=(/5140,3949,4307/)
        data_ped(5141,1:3)=(/5141,3301,4309/)
        data_ped(5142,1:3)=(/5142,3565,3713/)
        data_ped(5143,1:3)=(/5143,4158,4381/)
        data_ped(5144,1:3)=(/5144,3483,3080/)
        data_ped(5145,1:3)=(/5145,4500,4314/)
        data_ped(5146,1:3)=(/5146,4500,4118/)
        data_ped(5147,1:3)=(/5147,4419,0/)
        data_ped(5148,1:3)=(/5148,4601,3002/)
        data_ped(5149,1:3)=(/5149,4425,4472/)
        data_ped(5150,1:3)=(/5150,4412,3726/)
        data_ped(5151,1:3)=(/5151,3973,4144/)
        data_ped(5152,1:3)=(/5152,4619,4493/)
        data_ped(5153,1:3)=(/5153,4682,3830/)
        data_ped(5154,1:3)=(/5154,3838,3311/)
        data_ped(5155,1:3)=(/5155,3018,4277/)
        data_ped(5156,1:3)=(/5156,4175,4572/)
        data_ped(5157,1:3)=(/5157,4688,4332/)
        data_ped(5158,1:3)=(/5158,4412,3459/)
        data_ped(5159,1:3)=(/5159,4622,3821/)
        data_ped(5160,1:3)=(/5160,2845,4373/)
        data_ped(5161,1:3)=(/5161,3462,4031/)
        data_ped(5162,1:3)=(/5162,2933,4593/)
        data_ped(5163,1:3)=(/5163,4184,3543/)
        data_ped(5164,1:3)=(/5164,3847,4231/)
        data_ped(5165,1:3)=(/5165,1519,1424/)
        data_ped(5166,1:3)=(/5166,3743,4242/)
        data_ped(5167,1:3)=(/5167,4694,4278/)
        data_ped(5168,1:3)=(/5168,4060,3806/)
        data_ped(5169,1:3)=(/5169,2953,4640/)
        data_ped(5170,1:3)=(/5170,4414,4454/)
        data_ped(5171,1:3)=(/5171,4158,0/)
        data_ped(5172,1:3)=(/5172,3855,4821/)
        data_ped(5173,1:3)=(/5173,4687,3445/)
        data_ped(5174,1:3)=(/5174,4878,4385/)
        data_ped(5175,1:3)=(/5175,4153,4643/)
        data_ped(5176,1:3)=(/5176,4510,4635/)
        data_ped(5177,1:3)=(/5177,4804,4009/)
        data_ped(5178,1:3)=(/5178,4510,4743/)
        data_ped(5179,1:3)=(/5179,4421,4206/)
        data_ped(5180,1:3)=(/5180,4244,4553/)
        data_ped(5181,1:3)=(/5181,3483,4116/)
        data_ped(5182,1:3)=(/5182,4254,4462/)
        data_ped(5183,1:3)=(/5183,3565,4219/)
        data_ped(5184,1:3)=(/5184,3572,4744/)
        data_ped(5185,1:3)=(/5185,3189,4283/)
        data_ped(5186,1:3)=(/5186,2775,4448/)
        data_ped(5187,1:3)=(/5187,3647,4315/)
        data_ped(5188,1:3)=(/5188,4056,4391/)
        data_ped(5189,1:3)=(/5189,4257,4659/)
        data_ped(5190,1:3)=(/5190,4615,4218/)
        data_ped(5191,1:3)=(/5191,4413,4838/)
        data_ped(5192,1:3)=(/5192,4056,3803/)
        data_ped(5193,1:3)=(/5193,4056,4661/)
        data_ped(5194,1:3)=(/5194,4500,4395/)
        data_ped(5195,1:3)=(/5195,4500,4314/)
        data_ped(5196,1:3)=(/5196,4500,4662/)
        data_ped(5197,1:3)=(/5197,4421,3159/)
        data_ped(5198,1:3)=(/5198,3570,4302/)
        data_ped(5199,1:3)=(/5199,3743,4539/)
        data_ped(5200,1:3)=(/5200,4419,4384/)
        data_ped(5201,1:3)=(/5201,3569,4471/)
        data_ped(5202,1:3)=(/5202,4516,4397/)
        data_ped(5203,1:3)=(/5203,1207,4223/)
        data_ped(5204,1:3)=(/5204,4698,4642/)
        data_ped(5205,1:3)=(/5205,3407,3361/)
        data_ped(5206,1:3)=(/5206,4173,3362/)
        data_ped(5207,1:3)=(/5207,3933,3800/)
        data_ped(5208,1:3)=(/5208,3845,4090/)
        data_ped(5209,1:3)=(/5209,3395,0/)
        data_ped(5210,1:3)=(/5210,3580,4567/)
        data_ped(5211,1:3)=(/5211,4063,0/)
        data_ped(5212,1:3)=(/5212,4244,4451/)
        data_ped(5213,1:3)=(/5213,2962,144/)
        data_ped(5214,1:3)=(/5214,3673,4476/)
        data_ped(5215,1:3)=(/5215,4434,4020/)
        data_ped(5216,1:3)=(/5216,4262,3884/)
        data_ped(5217,1:3)=(/5217,3659,4672/)
        data_ped(5218,1:3)=(/5218,3864,3828/)
        data_ped(5219,1:3)=(/5219,4434,4231/)
        data_ped(5220,1:3)=(/5220,4682,4533/)
        data_ped(5221,1:3)=(/5221,4076,4581/)
        data_ped(5222,1:3)=(/5222,3973,4144/)
        data_ped(5223,1:3)=(/5223,4682,3830/)
        data_ped(5224,1:3)=(/5224,4523,4087/)
        data_ped(5225,1:3)=(/5225,4159,4367/)
        data_ped(5226,1:3)=(/5226,3845,3883/)
        data_ped(5227,1:3)=(/5227,4180,4037/)
        data_ped(5228,1:3)=(/5228,4180,4238/)
        data_ped(5229,1:3)=(/5229,3864,4494/)
        data_ped(5230,1:3)=(/5230,4709,4410/)
        data_ped(5231,1:3)=(/5231,4045,3510/)
        data_ped(5232,1:3)=(/5232,3411,4484/)
        data_ped(5233,1:3)=(/5233,4616,3812/)
        data_ped(5234,1:3)=(/5234,3554,4570/)
        data_ped(5235,1:3)=(/5235,4433,3417/)
        data_ped(5236,1:3)=(/5236,4338,4717/)
        data_ped(5237,1:3)=(/5237,4686,3979/)
        data_ped(5238,1:3)=(/5238,4078,4039/)
        data_ped(5239,1:3)=(/5239,3197,4720/)
        data_ped(5240,1:3)=(/5240,3567,3778/)
        data_ped(5241,1:3)=(/5241,2546,4823/)
        data_ped(5242,1:3)=(/5242,3569,4453/)
        data_ped(5243,1:3)=(/5243,4686,4363/)
        data_ped(5244,1:3)=(/5244,4816,4951/)
        data_ped(5245,1:3)=(/5245,4792,4093/)
        data_ped(5246,1:3)=(/5246,3189,4131/)
        data_ped(5247,1:3)=(/5247,4692,4326/)
        data_ped(5248,1:3)=(/5248,4244,3993/)
        data_ped(5249,1:3)=(/5249,4245,4289/)
        data_ped(5250,1:3)=(/5250,4153,4203/)
        data_ped(5251,1:3)=(/5251,4414,4459/)
        data_ped(5252,1:3)=(/5252,2845,4542/)
        data_ped(5253,1:3)=(/5253,4159,4003/)
        data_ped(5254,1:3)=(/5254,4423,3711/)
        data_ped(5255,1:3)=(/5255,3189,4753/)
        data_ped(5256,1:3)=(/5256,3851,4836/)
        data_ped(5257,1:3)=(/5257,2797,4560/)
        data_ped(5258,1:3)=(/5258,4704,3980/)
        data_ped(5259,1:3)=(/5259,4056,4208/)
        data_ped(5260,1:3)=(/5260,3121,3517/)
        data_ped(5261,1:3)=(/5261,416,0/)
        data_ped(5262,1:3)=(/5262,3933,3907/)
        data_ped(5263,1:3)=(/5263,4170,4126/)
        data_ped(5264,1:3)=(/5264,416,0/)
        data_ped(5265,1:3)=(/5265,4973,4761/)
        data_ped(5266,1:3)=(/5266,3409,3644/)
        data_ped(5267,1:3)=(/5267,4275,3820/)
        data_ped(5268,1:3)=(/5268,4418,3426/)
        data_ped(5269,1:3)=(/5269,4622,4192/)
        data_ped(5270,1:3)=(/5270,4158,4372/)
        data_ped(5271,1:3)=(/5271,4273,4024/)
        data_ped(5272,1:3)=(/5272,4184,4019/)
        data_ped(5273,1:3)=(/5273,4259,3774/)
        data_ped(5274,1:3)=(/5274,275,2/)
        data_ped(5275,1:3)=(/5275,4684,4676/)
        data_ped(5276,1:3)=(/5276,3939,4404/)
        data_ped(5277,1:3)=(/5277,4355,4366/)
        data_ped(5278,1:3)=(/5278,4080,4495/)
        data_ped(5279,1:3)=(/5279,4788,4374/)
        data_ped(5280,1:3)=(/5280,3663,3518/)
        data_ped(5281,1:3)=(/5281,2845,3998/)
        data_ped(5282,1:3)=(/5282,3467,3788/)
        data_ped(5283,1:3)=(/5283,3398,4821/)
        data_ped(5284,1:3)=(/5284,1293,93/)
        data_ped(5285,1:3)=(/5285,3659,3999/)
        data_ped(5286,1:3)=(/5286,4423,4548/)
        data_ped(5287,1:3)=(/5287,4594,3551/)
        data_ped(5288,1:3)=(/5288,4161,4390/)
        data_ped(5289,1:3)=(/5289,2845,4118/)
        data_ped(5290,1:3)=(/5290,4510,4103/)
        data_ped(5291,1:3)=(/5291,4414,4122/)
        data_ped(5292,1:3)=(/5292,3756,3902/)
        data_ped(5293,1:3)=(/5293,4056,3904/)
        data_ped(5294,1:3)=(/5294,4159,4732/)
        data_ped(5295,1:3)=(/5295,1300,5016/)
        data_ped(5296,1:3)=(/5296,3855,3799/)
        data_ped(5297,1:3)=(/5297,3949,4463/)
        data_ped(5298,1:3)=(/5298,2845,0/)
        data_ped(5299,1:3)=(/5299,4882,4848/)
        data_ped(5300,1:3)=(/5300,4414,0/)
        data_ped(5301,1:3)=(/5301,4594,0/)
        data_ped(5302,1:3)=(/5302,426,0/)
        data_ped(5303,1:3)=(/5303,4813,1301/)
        data_ped(5304,1:3)=(/5304,3933,4083/)
        data_ped(5305,1:3)=(/5305,4716,4577/)
        data_ped(5306,1:3)=(/5306,3680,4781/)
        data_ped(5307,1:3)=(/5307,4716,4940/)
        data_ped(5308,1:3)=(/5308,4716,4576/)
        data_ped(5309,1:3)=(/5309,4356,4571/)
        data_ped(5310,1:3)=(/5310,4618,3818/)
        data_ped(5311,1:3)=(/5311,3864,4946/)
        data_ped(5312,1:3)=(/5312,4714,4038/)
        data_ped(5313,1:3)=(/5313,4076,4581/)
        data_ped(5314,1:3)=(/5314,4713,3641/)
        data_ped(5315,1:3)=(/5315,3378,4028/)
        data_ped(5316,1:3)=(/5316,4521,4767/)
        data_ped(5317,1:3)=(/5317,4267,4403/)
        data_ped(5318,1:3)=(/5318,4528,4721/)
        data_ped(5319,1:3)=(/5319,4894,4722/)
        data_ped(5320,1:3)=(/5320,4254,4288/)
        data_ped(5321,1:3)=(/5321,4158,4896/)
        data_ped(5322,1:3)=(/5322,4060,4301/)
        data_ped(5323,1:3)=(/5323,4060,3369/)
        data_ped(5324,1:3)=(/5324,4698,4379/)
        data_ped(5325,1:3)=(/5325,4056,3890/)
        data_ped(5326,1:3)=(/5326,4056,4822/)
        data_ped(5327,1:3)=(/5327,3743,4450/)
        data_ped(5328,1:3)=(/5328,4594,4108/)
        data_ped(5329,1:3)=(/5329,3567,4326/)
        data_ped(5330,1:3)=(/5330,3659,4095/)
        data_ped(5331,1:3)=(/5331,3467,4832/)
        data_ped(5332,1:3)=(/5332,4700,4648/)
        data_ped(5333,1:3)=(/5333,4421,4649/)
        data_ped(5334,1:3)=(/5334,3743,4242/)
        data_ped(5335,1:3)=(/5335,4257,4659/)
        data_ped(5336,1:3)=(/5336,2797,4560/)
        data_ped(5337,1:3)=(/5337,3864,4147/)
        data_ped(5338,1:3)=(/5338,3864,4494/)
        data_ped(5339,1:3)=(/5339,4177,4589/)
        data_ped(5340,1:3)=(/5340,4175,4628/)
        data_ped(5341,1:3)=(/5341,3973,4144/)
        data_ped(5342,1:3)=(/5342,4434,3542/)
        data_ped(5343,1:3)=(/5343,3565,4739/)
        data_ped(5344,1:3)=(/5344,3934,4953/)
        data_ped(5345,1:3)=(/5345,4509,4294/)
        data_ped(5346,1:3)=(/5346,0,4825/)
        data_ped(5347,1:3)=(/5347,5046,4906/)
        data_ped(5348,1:3)=(/5348,4699,4455/)
        data_ped(5349,1:3)=(/5349,4872,4545/)
        data_ped(5350,1:3)=(/5350,4697,4641/)
        data_ped(5351,1:3)=(/5351,4697,3788/)
        data_ped(5352,1:3)=(/5352,4798,4830/)
        data_ped(5353,1:3)=(/5353,3743,4498/)
        data_ped(5354,1:3)=(/5354,3468,3705/)
        data_ped(5355,1:3)=(/5355,5054,4650/)
        data_ped(5356,1:3)=(/5356,283,111/)
        data_ped(5357,1:3)=(/5357,1521,1452/)
        data_ped(5358,1:3)=(/5358,4612,3530/)
        data_ped(5359,1:3)=(/5359,4424,4654/)
        data_ped(5360,1:3)=(/5360,2845,4656/)
        data_ped(5361,1:3)=(/5361,1260,0/)
        data_ped(5362,1:3)=(/5362,4703,5012/)
        data_ped(5363,1:3)=(/5363,4056,4208/)
        data_ped(5364,1:3)=(/5364,5052,5018/)
        data_ped(5365,1:3)=(/5365,4969,3090/)
        data_ped(5366,1:3)=(/5366,1277,3716/)
        data_ped(5367,1:3)=(/5367,4881,4926/)
        data_ped(5368,1:3)=(/5368,4064,4633/)
        data_ped(5369,1:3)=(/5369,4992,4278/)
        data_ped(5370,1:3)=(/5370,3569,4841/)
        data_ped(5371,1:3)=(/5371,3303,0/)
        data_ped(5372,1:3)=(/5372,281,105/)
        data_ped(5373,1:3)=(/5373,4070,4013/)
        data_ped(5374,1:3)=(/5374,3564,4847/)
        data_ped(5375,1:3)=(/5375,4697,0/)
        data_ped(5376,1:3)=(/5376,4884,0/)
        data_ped(5377,1:3)=(/5377,4064,0/)
        data_ped(5378,1:3)=(/5378,4158,4948/)
        data_ped(5379,1:3)=(/5379,4713,5035/)
        data_ped(5380,1:3)=(/5380,4163,4776/)
        data_ped(5381,1:3)=(/5381,4713,4782/)
        data_ped(5382,1:3)=(/5382,4886,4569/)
        data_ped(5383,1:3)=(/5383,4696,3540/)
        data_ped(5384,1:3)=(/5384,4793,4725/)
        data_ped(5385,1:3)=(/5385,4624,4771/)
        data_ped(5386,1:3)=(/5386,4871,4200/)
        data_ped(5387,1:3)=(/5387,4163,4577/)
        data_ped(5388,1:3)=(/5388,5064,3553/)
        data_ped(5389,1:3)=(/5389,3864,4147/)
        data_ped(5390,1:3)=(/5390,4254,4292/)
        data_ped(5391,1:3)=(/5391,4713,4719/)
        data_ped(5392,1:3)=(/5392,4429,4475/)
        data_ped(5393,1:3)=(/5393,4686,4998/)
        data_ped(5394,1:3)=(/5394,4792,3366/)
        data_ped(5395,1:3)=(/5395,4522,4937/)
        data_ped(5396,1:3)=(/5396,3189,4283/)
        data_ped(5397,1:3)=(/5397,4960,4821/)
        data_ped(5398,1:3)=(/5398,4594,4449/)
        data_ped(5399,1:3)=(/5399,4695,4994/)
        data_ped(5400,1:3)=(/5400,4796,4909/)
        data_ped(5401,1:3)=(/5401,3945,5004/)
        data_ped(5402,1:3)=(/5402,3189,4546/)
        data_ped(5403,1:3)=(/5403,3836,4729/)
        data_ped(5404,1:3)=(/5404,4796,5015/)
        data_ped(5405,1:3)=(/5405,2842,4738/)
        data_ped(5406,1:3)=(/5406,3743,4646/)
        data_ped(5407,1:3)=(/5407,314,107/)
        data_ped(5408,1:3)=(/5408,4613,4221/)
        data_ped(5409,1:3)=(/5409,4423,4125/)
        data_ped(5410,1:3)=(/5410,5050,4751/)
        data_ped(5411,1:3)=(/5411,3564,4211/)
        data_ped(5412,1:3)=(/5412,4870,0/)
        data_ped(5413,1:3)=(/5413,4179,4401/)
        data_ped(5414,1:3)=(/5414,4524,5038/)
        data_ped(5415,1:3)=(/5415,4436,3824/)
        data_ped(5416,1:3)=(/5416,4181,4772/)
        data_ped(5417,1:3)=(/5417,4181,3911/)
        data_ped(5418,1:3)=(/5418,4525,4851/)
        data_ped(5419,1:3)=(/5419,4525,4234/)
        data_ped(5420,1:3)=(/5420,3671,3629/)
        data_ped(5421,1:3)=(/5421,4248,4405/)
        data_ped(5422,1:3)=(/5422,3743,4290/)
        data_ped(5423,1:3)=(/5423,3749,4205/)
        data_ped(5424,1:3)=(/5424,4605,4377/)
        data_ped(5425,1:3)=(/5425,4697,4637/)
        data_ped(5426,1:3)=(/5426,3668,5010/)
        data_ped(5427,1:3)=(/5427,4797,4827/)
        data_ped(5428,1:3)=(/5428,4970,5172/)
        data_ped(5429,1:3)=(/5429,4415,5083/)
        data_ped(5430,1:3)=(/5430,4697,4644/)
        data_ped(5431,1:3)=(/5431,4510,3791/)
        data_ped(5432,1:3)=(/5432,4594,4241/)
        data_ped(5433,1:3)=(/5433,5052,4387/)
        data_ped(5434,1:3)=(/5434,4594,4717/)
        data_ped(5435,1:3)=(/5435,4056,0/)
        data_ped(5436,1:3)=(/5436,4594,4749/)
        data_ped(5437,1:3)=(/5437,4880,4750/)
        data_ped(5438,1:3)=(/5438,4874,4924/)
        data_ped(5439,1:3)=(/5439,4698,4842/)
        data_ped(5440,1:3)=(/5440,4796,4928/)
        data_ped(5441,1:3)=(/5441,1304,0/)
        data_ped(5442,1:3)=(/5442,3189,4393/)
        data_ped(5443,1:3)=(/5443,3743,4396/)
        data_ped(5444,1:3)=(/5444,4518,4665/)
        data_ped(5445,1:3)=(/5445,4870,0/)
        data_ped(5446,1:3)=(/5446,4964,3370/)
        data_ped(5447,1:3)=(/5447,4609,4461/)
        data_ped(5448,1:3)=(/5448,3946,4829/)
        data_ped(5449,1:3)=(/5449,3757,4473/)
        data_ped(5450,1:3)=(/5450,4343,4932/)
        data_ped(5451,1:3)=(/5451,4524,4497/)
        data_ped(5452,1:3)=(/5452,3849,4898/)
        data_ped(5453,1:3)=(/5453,5131,4445/)
        data_ped(5454,1:3)=(/5454,4415,5129/)
        data_ped(5455,1:3)=(/5455,5049,3343/)
        data_ped(5456,1:3)=(/5456,5049,4447/)
        data_ped(5457,1:3)=(/5457,4253,4281/)
        data_ped(5458,1:3)=(/5458,3189,4900/)
        data_ped(5459,1:3)=(/5459,3743,4901/)
        data_ped(5460,1:3)=(/5460,5165,22/)
        data_ped(5461,1:3)=(/5461,5073,4736/)
        data_ped(5462,1:3)=(/5462,4065,4740/)
        data_ped(5463,1:3)=(/5463,4865,3697/)
        data_ped(5464,1:3)=(/5464,4697,4631/)
        data_ped(5465,1:3)=(/5465,5254,5077/)
        data_ped(5466,1:3)=(/5466,5050,4727/)
        data_ped(5467,1:3)=(/5467,4056,4822/)
        data_ped(5468,1:3)=(/5468,4706,5082/)
        data_ped(5469,1:3)=(/5469,4697,4197/)
        data_ped(5470,1:3)=(/5470,4609,4915/)
        data_ped(5471,1:3)=(/5471,4795,1453/)
        data_ped(5472,1:3)=(/5472,5140,4009/)
        data_ped(5473,1:3)=(/5473,4805,0/)
        data_ped(5474,1:3)=(/5474,4594,4108/)
        data_ped(5475,1:3)=(/5475,4698,4304/)
        data_ped(5476,1:3)=(/5476,4063,4015/)
        data_ped(5477,1:3)=(/5477,4158,4757/)
        data_ped(5478,1:3)=(/5478,4686,4896/)
        data_ped(5479,1:3)=(/5479,4151,4947/)
        data_ped(5480,1:3)=(/5480,3970,4483/)
        data_ped(5481,1:3)=(/5481,4713,3733/)
        data_ped(5482,1:3)=(/5482,5066,4400/)
        data_ped(5483,1:3)=(/5483,4435,3914/)
        data_ped(5484,1:3)=(/5484,4984,4589/)
        data_ped(5485,1:3)=(/5485,5070,5109/)
        data_ped(5486,1:3)=(/5486,4163,4940/)
        data_ped(5487,1:3)=(/5487,4434,3459/)
        data_ped(5488,1:3)=(/5488,5064,4859/)
        data_ped(5489,1:3)=(/5489,4713,4941/)
        data_ped(5490,1:3)=(/5490,4339,4146/)
        data_ped(5491,1:3)=(/5491,3582,4489/)
        data_ped(5492,1:3)=(/5492,4708,4580/)
        data_ped(5493,1:3)=(/5493,4891,4144/)
        data_ped(5494,1:3)=(/5494,4527,4587/)
        data_ped(5495,1:3)=(/5495,5069,3828/)
        data_ped(5496,1:3)=(/5496,4626,4895/)
        data_ped(5497,1:3)=(/5497,4713,3458/)
        data_ped(5498,1:3)=(/5498,4713,4778/)
        data_ped(5499,1:3)=(/5499,4180,3822/)
        data_ped(5500,1:3)=(/5500,3864,4147/)
        data_ped(5501,1:3)=(/5501,4789,5110/)
        data_ped(5502,1:3)=(/5502,4431,4140/)
        data_ped(5503,1:3)=(/5503,5139,4897/)
        data_ped(5504,1:3)=(/5504,3934,3522/)
        data_ped(5505,1:3)=(/5505,4992,5171/)
        data_ped(5506,1:3)=(/5506,4958,3900/)
        data_ped(5507,1:3)=(/5507,4963,5081/)
        data_ped(5508,1:3)=(/5508,0,1302/)
        data_ped(5509,1:3)=(/5509,4056,3703/)
        data_ped(5510,1:3)=(/5510,4801,5175/)
        data_ped(5511,1:3)=(/5511,4056,4217/)
        data_ped(5512,1:3)=(/5512,4420,5186/)
        data_ped(5513,1:3)=(/5513,3743,3712/)
        data_ped(5514,1:3)=(/5514,4594,4108/)
        data_ped(5515,1:3)=(/5515,4594,4364/)
        data_ped(5516,1:3)=(/5516,4971,4563/)
        data_ped(5517,1:3)=(/5517,3955,4922/)
        data_ped(5518,1:3)=(/5518,5261,4756/)
        data_ped(5519,1:3)=(/5519,4501,0/)
        data_ped(5520,1:3)=(/5520,4964,0/)
        data_ped(5521,1:3)=(/5521,4415,4536/)
        data_ped(5522,1:3)=(/5522,4244,4818/)
        data_ped(5523,1:3)=(/5523,4699,4820/)
        data_ped(5524,1:3)=(/5524,5247,5076/)
        data_ped(5525,1:3)=(/5525,4604,4997/)
        data_ped(5526,1:3)=(/5526,4874,4828/)
        data_ped(5527,1:3)=(/5527,5073,4641/)
        data_ped(5528,1:3)=(/5528,5073,5008/)
        data_ped(5529,1:3)=(/5529,5247,5084/)
        data_ped(5530,1:3)=(/5530,4600,4101/)
        data_ped(5531,1:3)=(/5531,4686,4902/)
        data_ped(5532,1:3)=(/5532,4796,5006/)
        data_ped(5533,1:3)=(/5533,4956,4730/)
        data_ped(5534,1:3)=(/5534,3567,5088/)
        data_ped(5535,1:3)=(/5535,3189,4549/)
        data_ped(5536,1:3)=(/5536,3121,5094/)
        data_ped(5537,1:3)=(/5537,4611,4559/)
        data_ped(5538,1:3)=(/5538,3573,4464/)
        data_ped(5539,1:3)=(/5539,4698,5086/)
        data_ped(5540,1:3)=(/5540,3860,3908/)
        data_ped(5541,1:3)=(/5541,3568,5041/)
        data_ped(5542,1:3)=(/5542,5044,4954/)
        data_ped(5543,1:3)=(/5543,3743,4995/)
        data_ped(5544,1:3)=(/5544,4713,4943/)
        data_ped(5545,1:3)=(/5545,4808,4477/)
        data_ped(5546,1:3)=(/5546,4713,5220/)
        data_ped(5547,1:3)=(/5547,4713,5223/)
        data_ped(5548,1:3)=(/5548,5064,5117/)
        data_ped(5549,1:3)=(/5549,5064,3830/)
        data_ped(5550,1:3)=(/5550,4889,4774/)
        data_ped(5551,1:3)=(/5551,3760,4855/)
        data_ped(5552,1:3)=(/5552,5154,4576/)
        data_ped(5553,1:3)=(/5553,4811,4329/)
        data_ped(5554,1:3)=(/5554,4810,4021/)
        data_ped(5555,1:3)=(/5555,4713,4778/)
        data_ped(5556,1:3)=(/5556,4713,3641/)
        data_ped(5557,1:3)=(/5557,5067,5222/)
        data_ped(5558,1:3)=(/5558,4626,5226/)
        data_ped(5559,1:3)=(/5559,4626,4948/)
        data_ped(5560,1:3)=(/5560,4626,5225/)
        data_ped(5561,1:3)=(/5561,3864,4147/)
        data_ped(5562,1:3)=(/5562,3864,5229/)
        data_ped(5563,1:3)=(/5563,3864,4946/)
        data_ped(5564,1:3)=(/5564,5069,3828/)
        data_ped(5565,1:3)=(/5565,4621,4850/)
        data_ped(5566,1:3)=(/5566,4523,5112/)
        data_ped(5567,1:3)=(/5567,3864,4947/)
        data_ped(5568,1:3)=(/5568,4975,4938/)
        data_ped(5569,1:3)=(/5569,4710,4763/)
        data_ped(5570,1:3)=(/5570,4259,3628/)
        data_ped(5571,1:3)=(/5571,4532,4945/)
        data_ped(5572,1:3)=(/5572,4271,4668/)
        data_ped(5573,1:3)=(/5573,4868,5075/)
        data_ped(5574,1:3)=(/5574,4702,5007/)
        data_ped(5575,1:3)=(/5575,5280,4644/)
        data_ped(5576,1:3)=(/5576,4594,5091/)
        data_ped(5577,1:3)=(/5577,4683,4746/)
        data_ped(5578,1:3)=(/5578,4866,4455/)
        data_ped(5579,1:3)=(/5579,1308,5095/)
        data_ped(5580,1:3)=(/5580,5134,4837/)
        data_ped(5581,1:3)=(/5581,3569,4900/)
        data_ped(5582,1:3)=(/5582,4064,5080/)
        data_ped(5583,1:3)=(/5583,5256,0/)
        data_ped(5584,1:3)=(/5584,4796,5022/)
        data_ped(5585,1:3)=(/5585,1291,0/)
        data_ped(5586,1:3)=(/5586,3578,0/)
        data_ped(5587,1:3)=(/5587,4970,5283/)
        data_ped(5588,1:3)=(/5588,5262,4398/)
        data_ped(5589,1:3)=(/5589,5245,4993/)
        data_ped(5590,1:3)=(/5590,4249,5167/)
        data_ped(5591,1:3)=(/5591,5131,4863/)
        data_ped(5592,1:3)=(/5592,5280,5168/)
        data_ped(5593,1:3)=(/5593,4869,4284/)
        data_ped(5594,1:3)=(/5594,4340,4634/)
        data_ped(5595,1:3)=(/5595,4244,4451/)
        data_ped(5596,1:3)=(/5596,4699,4538/)
        data_ped(5597,1:3)=(/5597,4685,4280/)
        data_ped(5598,1:3)=(/5598,4605,4834/)
        data_ped(5599,1:3)=(/5599,4056,4537/)
        data_ped(5600,1:3)=(/5600,4168,4917/)
        data_ped(5601,1:3)=(/5601,4698,3369/)
        data_ped(5602,1:3)=(/5602,4698,4304/)
        data_ped(5603,1:3)=(/5603,3569,4364/)
        data_ped(5604,1:3)=(/5604,4968,4747/)
        data_ped(5605,1:3)=(/5605,3486,5019/)
        data_ped(5606,1:3)=(/5606,3851,0/)
        data_ped(5607,1:3)=(/5607,4961,5104/)
        data_ped(5608,1:3)=(/5608,3487,4760/)
        data_ped(5609,1:3)=(/5609,4796,5192/)
        data_ped(5610,1:3)=(/5610,4706,5209/)
        data_ped(5611,1:3)=(/5611,1241,5027/)
        data_ped(5612,1:3)=(/5612,5247,5076/)
        data_ped(5613,1:3)=(/5613,4594,4131/)
        data_ped(5614,1:3)=(/5614,5045,4826/)
        data_ped(5615,1:3)=(/5615,5248,4288/)
        data_ped(5616,1:3)=(/5616,4713,4778/)
        data_ped(5617,1:3)=(/5617,4713,5219/)
        data_ped(5618,1:3)=(/5618,4806,4777/)
        data_ped(5619,1:3)=(/5619,4180,4943/)
        data_ped(5620,1:3)=(/5620,3761,4591/)
        data_ped(5621,1:3)=(/5621,4183,4575/)
        data_ped(5622,1:3)=(/5622,4259,5028/)
        data_ped(5623,1:3)=(/5623,4526,4025/)
        data_ped(5624,1:3)=(/5624,4985,4764/)
        data_ped(5625,1:3)=(/5625,3864,4147/)
        data_ped(5626,1:3)=(/5626,3864,4494/)
        data_ped(5627,1:3)=(/5627,5152,4588/)
        data_ped(5628,1:3)=(/5628,4524,5125/)
        data_ped(5629,1:3)=(/5629,5064,4860/)
        data_ped(5630,1:3)=(/5630,5270,5225/)
        data_ped(5631,1:3)=(/5631,4890,3551/)
        data_ped(5632,1:3)=(/5632,3973,4592/)
        data_ped(5633,1:3)=(/5633,3973,4785/)
        data_ped(5634,1:3)=(/5634,5067,5222/)
        data_ped(5635,1:3)=(/5635,4982,5308/)
        data_ped(5636,1:3)=(/5636,4977,5218/)
        data_ped(5637,1:3)=(/5637,5469,5168/)
        data_ped(5638,1:3)=(/5638,3570,4302/)
        data_ped(5639,1:3)=(/5639,4876,5284/)
        data_ped(5640,1:3)=(/5640,5143,4921/)
        data_ped(5641,1:3)=(/5641,4056,4661/)
        data_ped(5642,1:3)=(/5642,4796,5098/)
        data_ped(5643,1:3)=(/5643,1309,0/)
        data_ped(5644,1:3)=(/5644,4796,4928/)
        data_ped(5645,1:3)=(/5645,4342,5374/)
        data_ped(5646,1:3)=(/5646,5148,5203/)
        data_ped(5647,1:3)=(/5647,5253,4922/)
        data_ped(5648,1:3)=(/5648,4883,5208/)
        data_ped(5649,1:3)=(/5649,5055,0/)
        data_ped(5650,1:3)=(/5650,1565,4933/)
        data_ped(5651,1:3)=(/5651,5265,4934/)
        data_ped(5652,1:3)=(/5652,5249,4636/)
        data_ped(5653,1:3)=(/5653,4598,4996/)
        data_ped(5654,1:3)=(/5654,5466,5345/)
        data_ped(5655,1:3)=(/5655,4604,5002/)
        data_ped(5656,1:3)=(/5656,4249,4278/)
        data_ped(5657,1:3)=(/5657,4691,4904/)
        data_ped(5658,1:3)=(/5658,5051,4642/)
        data_ped(5659,1:3)=(/5659,4603,4631/)
        data_ped(5660,1:3)=(/5660,4802,4556/)
        data_ped(5661,1:3)=(/5661,4345,4112/)
        data_ped(5662,1:3)=(/5662,5397,4640/)
        data_ped(5663,1:3)=(/5663,4064,4555/)
        data_ped(5664,1:3)=(/5664,4796,5080/)
        data_ped(5665,1:3)=(/5665,4796,3890/)
        data_ped(5666,1:3)=(/5666,3568,4363/)
        data_ped(5667,1:3)=(/5667,4064,4655/)
        data_ped(5668,1:3)=(/5668,4800,5362/)
        data_ped(5669,1:3)=(/5669,4594,4549/)
        data_ped(5670,1:3)=(/5670,4698,3369/)
        data_ped(5671,1:3)=(/5671,5405,5173/)
        data_ped(5672,1:3)=(/5672,5143,5107/)
        data_ped(5673,1:3)=(/5673,5045,3719/)
        data_ped(5674,1:3)=(/5674,5244,4443/)
        data_ped(5675,1:3)=(/5675,5280,4447/)
        data_ped(5676,1:3)=(/5676,4249,4902/)
        data_ped(5677,1:3)=(/5677,5132,4907/)
        data_ped(5678,1:3)=(/5678,4056,4208/)
        data_ped(5679,1:3)=(/5679,5408,4458/)
        data_ped(5680,1:3)=(/5680,4686,5369/)
        data_ped(5681,1:3)=(/5681,5246,4130/)
        data_ped(5682,1:3)=(/5682,3680,5230/)
        data_ped(5683,1:3)=(/5683,3866,5031/)
        data_ped(5684,1:3)=(/5684,4806,4671/)
        data_ped(5685,1:3)=(/5685,4180,4943/)
        data_ped(5686,1:3)=(/5686,5266,5221/)
        data_ped(5687,1:3)=(/5687,5159,4939/)
        data_ped(5688,1:3)=(/5688,4520,4780/)
        data_ped(5689,1:3)=(/5689,4809,5215/)
        data_ped(5690,1:3)=(/5690,4890,5226/)
        data_ped(5691,1:3)=(/5691,5064,3829/)
        data_ped(5692,1:3)=(/5692,5064,4859/)
        data_ped(5693,1:3)=(/5693,5160,5308/)
        data_ped(5694,1:3)=(/5694,4979,4148/)
        data_ped(5695,1:3)=(/5695,4271,5121/)
        data_ped(5696,1:3)=(/5696,5156,4023/)
        data_ped(5697,1:3)=(/5697,5154,4940/)
        data_ped(5698,1:3)=(/5698,3973,4584/)
        data_ped(5699,1:3)=(/5699,3973,5218/)
        data_ped(5700,1:3)=(/5700,3864,4946/)
        data_ped(5701,1:3)=(/5701,3864,4494/)
        data_ped(5702,1:3)=(/5702,4887,5500/)
        data_ped(5703,1:3)=(/5703,4892,4574/)
        data_ped(5704,1:3)=(/5704,5270,5391/)
        data_ped(5705,1:3)=(/5705,4623,3543/)
        data_ped(5706,1:3)=(/5706,5150,4137/)
        data_ped(5707,1:3)=(/5707,5273,5039/)
        data_ped(5708,1:3)=(/5708,5468,5429/)
        data_ped(5709,1:3)=(/5709,5330,5285/)
        data_ped(5710,1:3)=(/5710,5246,5288/)
        data_ped(5711,1:3)=(/5711,3850,5169/)
        data_ped(5712,1:3)=(/5712,4420,5359/)
        data_ped(5713,1:3)=(/5713,4612,4559/)
        data_ped(5714,1:3)=(/5714,3851,5361/)
        data_ped(5715,1:3)=(/5715,5409,5295/)
        data_ped(5716,1:3)=(/5716,4514,4127/)
        data_ped(5717,1:3)=(/5717,5333,5182/)
        data_ped(5718,1:3)=(/5718,4799,4734/)
        data_ped(5719,1:3)=(/5719,4796,5006/)
        data_ped(5720,1:3)=(/5720,5408,4845/)
        data_ped(5721,1:3)=(/5721,5405,4927/)
        data_ped(5722,1:3)=(/5722,5060,4931/)
        data_ped(5723,1:3)=(/5723,5146,0/)
        data_ped(5724,1:3)=(/5724,5059,5376/)
        data_ped(5725,1:3)=(/5725,3851,0/)
        data_ped(5726,1:3)=(/5726,4699,4903/)
        data_ped(5727,1:3)=(/5727,4699,4299/)
        data_ped(5728,1:3)=(/5728,4415,5394/)
        data_ped(5729,1:3)=(/5729,5135,3769/)
        data_ped(5730,1:3)=(/5730,4699,4679/)
        data_ped(5731,1:3)=(/5731,4866,5043/)
        data_ped(5732,1:3)=(/5732,4699,4997/)
        data_ped(5733,1:3)=(/5733,5399,4905/)
        data_ped(5734,1:3)=(/5734,5137,4114/)
        data_ped(5735,1:3)=(/5735,4698,5025/)
        data_ped(5736,1:3)=(/5736,4698,5351/)
        data_ped(5737,1:3)=(/5737,5330,3986/)
        data_ped(5738,1:3)=(/5738,4796,4831/)
        data_ped(5739,1:3)=(/5739,5472,5355/)
        data_ped(5740,1:3)=(/5740,4868,5184/)
        data_ped(5741,1:3)=(/5741,4796,5293/)
        data_ped(5742,1:3)=(/5742,5332,5191/)
        data_ped(5743,1:3)=(/5743,4791,4916/)
        data_ped(5744,1:3)=(/5744,5147,5102/)
        data_ped(5745,1:3)=(/5745,5057,4560/)
        data_ped(5746,1:3)=(/5746,4961,5204/)
        data_ped(5747,1:3)=(/5747,4244,4451/)
        data_ped(5748,1:3)=(/5748,4056,4378/)
        data_ped(5749,1:3)=(/5749,4962,4913/)
        data_ped(5750,1:3)=(/5750,4265,4675/)
        data_ped(5751,1:3)=(/5751,4617,1296/)
        data_ped(5752,1:3)=(/5752,4986,4853/)
        data_ped(5753,1:3)=(/5753,5274,1298/)
        data_ped(5754,1:3)=(/5754,5062,4239/)
        data_ped(5755,1:3)=(/5755,4981,4784/)
        data_ped(5756,1:3)=(/5756,5266,5122/)
        data_ped(5757,1:3)=(/5757,4626,5041/)
        data_ped(5758,1:3)=(/5758,4626,5227/)
        data_ped(5759,1:3)=(/5759,4985,5030/)
        data_ped(5760,1:3)=(/5760,5072,5479/)
        data_ped(5761,1:3)=(/5761,3969,5217/)
        data_ped(5762,1:3)=(/5762,5154,5113/)
        data_ped(5763,1:3)=(/5763,3973,5218/)
        data_ped(5764,1:3)=(/5764,3973,5231/)
        data_ped(5765,1:3)=(/5765,5063,5036/)
        data_ped(5766,1:3)=(/5766,4430,5040/)
        data_ped(5767,1:3)=(/5767,4990,5124/)
        data_ped(5768,1:3)=(/5768,4357,4407/)
        data_ped(5769,1:3)=(/5769,3864,4584/)
        data_ped(5770,1:3)=(/5770,5337,4863/)
        data_ped(5771,1:3)=(/5771,3971,4334/)
        data_ped(5772,1:3)=(/5772,4432,5316/)
        data_ped(5773,1:3)=(/5773,5267,5123/)
        data_ped(5774,1:3)=(/5774,4983,4672/)
        data_ped(5775,1:3)=(/5775,5275,5451/)
        data_ped(5776,1:3)=(/5776,4181,4027/)
        data_ped(5777,1:3)=(/5777,4598,4955/)
        data_ped(5778,1:3)=(/5778,3569,4384/)
        data_ped(5779,1:3)=(/5779,5329,5506/)
        data_ped(5780,1:3)=(/5780,4702,5007/)
        data_ped(5781,1:3)=(/5781,5328,5347/)
        data_ped(5782,1:3)=(/5782,4961,5090/)
        data_ped(5783,1:3)=(/5783,5464,5562/)
        data_ped(5784,1:3)=(/5784,5053,4909/)
        data_ped(5785,1:3)=(/5785,5465,5010/)
        data_ped(5786,1:3)=(/5786,4894,5290/)
        data_ped(5787,1:3)=(/5787,5141,4920/)
        data_ped(5788,1:3)=(/5788,4422,5439/)
        data_ped(5789,1:3)=(/5789,5145,5298/)
        data_ped(5790,1:3)=(/5790,4802,5199/)
        data_ped(5791,1:3)=(/5791,5133,0/)
        data_ped(5792,1:3)=(/5792,5246,5200/)
        data_ped(5793,1:3)=(/5793,4594,5195/)
        data_ped(5794,1:3)=(/5794,5261,4565/)
        data_ped(5795,1:3)=(/5795,4603,5375/)
        data_ped(5796,1:3)=(/5796,5048,5524/)
        data_ped(5797,1:3)=(/5797,5133,5321/)
        data_ped(5798,1:3)=(/5798,4873,5427/)
        data_ped(5799,1:3)=(/5799,5399,5003/)
        data_ped(5800,1:3)=(/5800,5399,5001/)
        data_ped(5801,1:3)=(/5801,5464,5461/)
        data_ped(5802,1:3)=(/5802,4685,5180/)
        data_ped(5803,1:3)=(/5803,5533,4544/)
        data_ped(5804,1:3)=(/5804,5470,5079/)
        data_ped(5805,1:3)=(/5805,4796,4733/)
        data_ped(5806,1:3)=(/5806,4611,4912/)
        data_ped(5807,1:3)=(/5807,5055,4919/)
        data_ped(5808,1:3)=(/5808,4878,4645/)
        data_ped(5809,1:3)=(/5809,5246,4449/)
        data_ped(5810,1:3)=(/5810,5411,5296/)
        data_ped(5811,1:3)=(/5811,5475,4301/)
        data_ped(5812,1:3)=(/5812,4604,5197/)
        data_ped(5813,1:3)=(/5813,5336,5024/)
        data_ped(5814,1:3)=(/5814,5053,5440/)
        data_ped(5815,1:3)=(/5815,4698,4999/)
        data_ped(5816,1:3)=(/5816,4868,4364/)
        data_ped(5817,1:3)=(/5817,4698,5430/)
        data_ped(5818,1:3)=(/5818,4264,4238/)
        data_ped(5819,1:3)=(/5819,4264,5236/)
        data_ped(5820,1:3)=(/5820,4974,1313/)
        data_ped(5821,1:3)=(/5821,5270,1318/)
        data_ped(5822,1:3)=(/5822,5415,4331/)
        data_ped(5823,1:3)=(/5823,4180,5042/)
        data_ped(5824,1:3)=(/5824,4711,4773/)
        data_ped(5825,1:3)=(/5825,4815,5034/)
        data_ped(5826,1:3)=(/5826,5277,4949/)
        data_ped(5827,1:3)=(/5827,4992,5378/)
        data_ped(5828,1:3)=(/5828,5272,5311/)
        data_ped(5829,1:3)=(/5829,5067,3828/)
        data_ped(5830,1:3)=(/5830,5067,5495/)
        data_ped(5831,1:3)=(/5831,3973,5231/)
        data_ped(5832,1:3)=(/5832,4887,5546/)
        data_ped(5833,1:3)=(/5833,5276,4935/)
        data_ped(5834,1:3)=(/5834,4992,5391/)
        data_ped(5835,1:3)=(/5835,4992,5386/)
        data_ped(5836,1:3)=(/5836,5478,3739/)
        data_ped(5837,1:3)=(/5837,5478,4948/)
        data_ped(5838,1:3)=(/5838,5478,5226/)
        data_ped(5839,1:3)=(/5839,5478,5379/)
        data_ped(5840,1:3)=(/5840,5154,5305/)
        data_ped(5841,1:3)=(/5841,4271,4669/)
        data_ped(5842,1:3)=(/5842,4890,5237/)
        data_ped(5843,1:3)=(/5843,3864,4946/)
        data_ped(5844,1:3)=(/5844,3864,4494/)
        data_ped(5845,1:3)=(/5845,5151,5228/)
        data_ped(5846,1:3)=(/5846,4989,5111/)
        data_ped(5847,1:3)=(/5847,5425,5455/)
        data_ped(5848,1:3)=(/5848,5243,5505/)
        data_ped(5849,1:3)=(/5849,4348,5291/)
        data_ped(5850,1:3)=(/5850,5396,5436/)
        data_ped(5851,1:3)=(/5851,5144,4925/)
        data_ped(5852,1:3)=(/5852,5056,5367/)
        data_ped(5853,1:3)=(/5853,5474,5360/)
        data_ped(5854,1:3)=(/5854,5476,4754/)
        data_ped(5855,1:3)=(/5855,4802,5443/)
        data_ped(5856,1:3)=(/5856,5606,5585/)
        data_ped(5857,1:3)=(/5857,5532,4930/)
        data_ped(5858,1:3)=(/5858,4885,5210/)
        data_ped(5859,1:3)=(/5859,422,0/)
        data_ped(5860,1:3)=(/5860,5412,5377/)
        data_ped(5861,1:3)=(/5861,3864,5452/)
        data_ped(5862,1:3)=(/5862,4508,5503/)
        data_ped(5863,1:3)=(/5863,4957,5346/)
        data_ped(5864,1:3)=(/5864,5154,5217/)
        data_ped(5865,1:3)=(/5865,3864,4494/)
        data_ped(5866,1:3)=(/5866,3864,4946/)
        data_ped(5867,1:3)=(/5867,3864,5127/)
        data_ped(5868,1:3)=(/5868,5333,5182/)
        data_ped(5869,1:3)=(/5869,5679,5508/)
        data_ped(5870,1:3)=(/5870,5053,5509/)
        data_ped(5871,1:3)=(/5871,4610,4460/)
        data_ped(5872,1:3)=(/5872,4870,5185/)
        data_ped(5873,1:3)=(/5873,5426,5187/)
        data_ped(5874,1:3)=(/5874,5474,5194/)
        data_ped(5875,1:3)=(/5875,4961,4197/)
        data_ped(5876,1:3)=(/5876,424,1305/)
        data_ped(5877,1:3)=(/5877,5263,0/)
        data_ped(5878,1:3)=(/5878,5399,5001/)
        data_ped(5879,1:3)=(/5879,4697,4631/)
        data_ped(5880,1:3)=(/5880,3851,5438/)
        data_ped(5881,1:3)=(/5881,5053,5509/)
        data_ped(5882,1:3)=(/5882,3679,4768/)
        data_ped(5883,1:3)=(/5883,4806,4854/)
        data_ped(5884,1:3)=(/5884,4180,5544/)
        data_ped(5885,1:3)=(/5885,4180,4592/)
        data_ped(5886,1:3)=(/5886,4180,4785/)
        data_ped(5887,1:3)=(/5887,4180,4861/)
        data_ped(5888,1:3)=(/5888,4891,5222/)
        data_ped(5889,1:3)=(/5889,5341,4782/)
        data_ped(5890,1:3)=(/5890,5341,5223/)
        data_ped(5891,1:3)=(/5891,5341,5547/)
        data_ped(5892,1:3)=(/5892,5341,4955/)
        data_ped(5893,1:3)=(/5893,4887,5500/)
        data_ped(5894,1:3)=(/5894,5154,5387/)
        data_ped(5895,1:3)=(/5895,5337,5454/)
        data_ped(5896,1:3)=(/5896,5072,5489/)
        data_ped(5897,1:3)=(/5897,5072,5479/)
        data_ped(5898,1:3)=(/5898,5072,4947/)
        data_ped(5899,1:3)=(/5899,5072,5567/)
        data_ped(5900,1:3)=(/5900,4890,5226/)
        data_ped(5901,1:3)=(/5901,4890,5386/)
        data_ped(5902,1:3)=(/5902,4890,5035/)
        data_ped(5903,1:3)=(/5903,4890,3829/)
        data_ped(5904,1:3)=(/5904,5478,5119/)
        data_ped(5905,1:3)=(/5905,4985,4952/)
        data_ped(5906,1:3)=(/5906,5413,5232/)
        data_ped(5907,1:3)=(/5907,4890,5379/)
        data_ped(5908,1:3)=(/5908,4890,3739/)
        data_ped(5909,1:3)=(/5909,4992,5116/)
        data_ped(5910,1:3)=(/5910,4992,5236/)
        data_ped(5911,1:3)=(/5911,5279,5456/)
        data_ped(5912,1:3)=(/5912,5279,5455/)
        data_ped(5913,1:3)=(/5913,5398,4241/)
        data_ped(5914,1:3)=(/5914,5531,5289/)
        data_ped(5915,1:3)=(/5915,4349,5190/)
        data_ped(5916,1:3)=(/5916,4791,5596/)
        data_ped(5917,1:3)=(/5917,5326,5368/)
        data_ped(5918,1:3)=(/5918,5053,5511/)
        data_ped(5919,1:3)=(/5919,5053,5363/)
        data_ped(5920,1:3)=(/5920,5473,5643/)
        data_ped(5921,1:3)=(/5921,5605,5174/)
        data_ped(5922,1:3)=(/5922,3752,5198/)
        data_ped(5923,1:3)=(/5923,5744,4755/)
        data_ped(5924,1:3)=(/5924,5048,5011/)
        data_ped(5925,1:3)=(/5925,4511,5444/)
        data_ped(5926,1:3)=(/5926,5531,5445/)
        data_ped(5927,1:3)=(/5927,1320,0/)
        data_ped(5928,1:3)=(/5928,5396,5656/)
        data_ped(5929,1:3)=(/5929,5324,5168/)
        data_ped(5930,1:3)=(/5930,5599,5573/)
        data_ped(5931,1:3)=(/5931,5045,5282/)
        data_ped(5932,1:3)=(/5932,5425,4111/)
        data_ped(5933,1:3)=(/5933,3864,5383/)
        data_ped(5934,1:3)=(/5934,4604,4824/)
        data_ped(5935,1:3)=(/5935,4243,4833/)
        data_ped(5936,1:3)=(/5936,5701,5561/)
        data_ped(5937,1:3)=(/5937,4601,5000/)
        data_ped(5938,1:3)=(/5938,4961,5350/)
        data_ped(5939,1:3)=(/5939,5535,5576/)
        data_ped(5940,1:3)=(/5940,5465,4554/)
        data_ped(5941,1:3)=(/5941,4868,4364/)
        data_ped(5942,1:3)=(/5942,5531,5196/)
        data_ped(5943,1:3)=(/5943,4513,5583/)
        data_ped(5944,1:3)=(/5944,5531,4313/)
        data_ped(5945,1:3)=(/5945,424,1294/)
        data_ped(5946,1:3)=(/5946,5532,4303/)
        data_ped(5947,1:3)=(/5947,5531,5450/)
        data_ped(5948,1:3)=(/5948,3864,4631/)
        data_ped(5949,1:3)=(/5949,3864,4577/)
        data_ped(5950,1:3)=(/5950,5701,5527/)
        data_ped(5951,1:3)=(/5951,3864,5659/)
        data_ped(5952,1:3)=(/5952,5531,4653/)
        data_ped(5953,1:3)=(/5953,5416,5238/)
        data_ped(5954,1:3)=(/5954,4716,1332/)
        data_ped(5955,1:3)=(/5955,4625,4765/)
        data_ped(5956,1:3)=(/5956,4264,5236/)
        data_ped(5957,1:3)=(/5957,5068,5494/)
        data_ped(5958,1:3)=(/5958,5166,5240/)
        data_ped(5959,1:3)=(/5959,5166,5499/)
        data_ped(5960,1:3)=(/5960,4626,5631/)
        data_ped(5961,1:3)=(/5961,4992,5458/)
        data_ped(5962,1:3)=(/5962,3973,4782/)
        data_ped(5963,1:3)=(/5963,5271,5623/)
        data_ped(5964,1:3)=(/5964,5072,5626/)
        data_ped(5965,1:3)=(/5965,5072,5697/)
        data_ped(5966,1:3)=(/5966,4989,5456/)
        data_ped(5967,1:3)=(/5967,4989,5380/)
        data_ped(5968,1:3)=(/5968,4989,5391/)
        data_ped(5969,1:3)=(/5969,4989,5225/)
        data_ped(5970,1:3)=(/5970,5063,5632/)
        data_ped(5971,1:3)=(/5971,5069,5564/)
        data_ped(5972,1:3)=(/5972,5064,5548/)
        data_ped(5973,1:3)=(/5973,5273,5235/)
        data_ped(5974,1:3)=(/5974,4967,5637/)
        data_ped(5975,1:3)=(/5975,4803,5372/)
        data_ped(5976,1:3)=(/5976,4599,5108/)
        data_ped(5977,1:3)=(/5977,4961,5103/)
        data_ped(5978,1:3)=(/5978,5613,5446/)
        data_ped(5979,1:3)=(/5979,3864,4494/)
        data_ped(5980,1:3)=(/5980,4697,4631/)
        data_ped(5981,1:3)=(/5981,3864,5217/)
        data_ped(5982,1:3)=(/5982,3864,5528/)
        data_ped(5983,1:3)=(/5983,3864,5323/)
        data_ped(5984,1:3)=(/5984,4697,4577/)
        data_ped(5985,1:3)=(/5985,5403,5510/)
        data_ped(5986,1:3)=(/5986,5604,5714/)
        data_ped(5987,1:3)=(/5987,4159,5078/)
        data_ped(5988,1:3)=(/5988,5665,5193/)
        data_ped(5989,1:3)=(/5989,4870,5514/)
        data_ped(5990,1:3)=(/5990,5663,5642/)
        data_ped(5991,1:3)=(/5991,5807,5649/)
        data_ped(5992,1:3)=(/5992,3127,5026/)
        data_ped(5993,1:3)=(/5993,3752,5661/)
        data_ped(5994,1:3)=(/5994,4870,5442/)
        data_ped(5995,1:3)=(/5995,4358,5620/)
        data_ped(5996,1:3)=(/5996,4626,5564/)
        data_ped(5997,1:3)=(/5997,4623,5616/)
        data_ped(5998,1:3)=(/5998,5065,1303/)
        data_ped(5999,1:3)=(/5999,5556,4936/)
        data_ped(6000,1:3)=(/6000,5166,5381/)
        data_ped(6001,1:3)=(/6001,4713,5626/)
        data_ped(6002,1:3)=(/6002,5700,5453/)
        data_ped(6003,1:3)=(/6003,4715,5689/)
        data_ped(6004,1:3)=(/6004,4715,5652/)
        data_ped(6005,1:3)=(/6005,5154,5761/)
        data_ped(6006,1:3)=(/6006,4887,5546/)
        data_ped(6007,1:3)=(/6007,5072,5479/)
        data_ped(6008,1:3)=(/6008,5072,5457/)
        data_ped(6009,1:3)=(/6009,5341,5489/)
        data_ped(6010,1:3)=(/6010,4989,5391/)
        data_ped(6011,1:3)=(/6011,5493,5728/)
        data_ped(6012,1:3)=(/6012,5063,5625/)
        data_ped(6013,1:3)=(/6013,5419,5392/)
        data_ped(6014,1:3)=(/6014,5701,5305/)
        data_ped(6015,1:3)=(/6015,5543,5593/)
        data_ped(6016,1:3)=(/6016,4959,5241/)
        data_ped(6017,1:3)=(/6017,5666,5092/)
        data_ped(6018,1:3)=(/6018,5743,4843/)
        data_ped(6019,1:3)=(/6019,5531,5093/)
        data_ped(6020,1:3)=(/6020,5607,4759/)
        data_ped(6021,1:3)=(/6021,5531,5515/)
        data_ped(6022,1:3)=(/6022,4701,5282/)
        data_ped(6023,1:3)=(/6023,5406,5348/)
        data_ped(6024,1:3)=(/6024,5534,5803/)
        data_ped(6025,1:3)=(/6025,5252,5349/)
        data_ped(6026,1:3)=(/6026,5844,5452/)
        data_ped(6027,1:3)=(/6027,5601,5780/)
        data_ped(6028,1:3)=(/6028,5407,5356/)
        data_ped(6029,1:3)=(/6029,4966,5177/)
        data_ped(6030,1:3)=(/6030,4959,5578/)
        data_ped(6031,1:3)=(/6031,3864,5801/)
        data_ped(6032,1:3)=(/6032,5260,5296/)
        data_ped(6033,1:3)=(/6033,5251,4929/)
        data_ped(6034,1:3)=(/6034,5531,5185/)
        data_ped(6035,1:3)=(/6035,4354,5213/)
        data_ped(6036,1:3)=(/6036,5611,0/)
        data_ped(6037,1:3)=(/6037,5326,4908/)
        data_ped(6038,1:3)=(/6038,5402,5432/)
        data_ped(6039,1:3)=(/6039,5740,4817/)
        data_ped(6040,1:3)=(/6040,5543,5593/)
        data_ped(6041,1:3)=(/6041,5635,5307/)
        data_ped(6042,1:3)=(/6042,5635,5486/)
        data_ped(6043,1:3)=(/6043,5341,5728/)
        data_ped(6044,1:3)=(/6044,5072,5697/)
        data_ped(6045,1:3)=(/6045,4992,5393/)
        data_ped(6046,1:3)=(/6046,5675,5544/)
        data_ped(6047,1:3)=(/6047,4715,5589/)
        data_ped(6048,1:3)=(/6048,5337,5043/)
        data_ped(6049,1:3)=(/6049,5063,5594/)
        data_ped(6050,1:3)=(/6050,4616,5730/)
        data_ped(6051,1:3)=(/6051,4616,5320/)
        data_ped(6052,1:3)=(/6052,4361,5568/)
        data_ped(6053,1:3)=(/6053,5541,5656/)
        data_ped(6054,1:3)=(/6054,5478,5692/)
        data_ped(6055,1:3)=(/6055,4977,4944/)
        data_ped(6056,1:3)=(/6056,5071,5691/)
        data_ped(6057,1:3)=(/6057,4623,5616/)
        data_ped(6058,1:3)=(/6058,4626,3553/)
        data_ped(6059,1:3)=(/6059,5339,1335/)
        data_ped(6060,1:3)=(/6060,5166,5497/)
        data_ped(6061,1:3)=(/6061,5556,4864/)
        data_ped(6062,1:3)=(/6062,5531,5097/)
        data_ped(6063,1:3)=(/6063,4868,5436/)
        data_ped(6064,1:3)=(/6064,5538,5791/)
        data_ped(6065,1:3)=(/6065,5258,5516/)
        data_ped(6066,1:3)=(/6066,4794,5357/)
        data_ped(6067,1:3)=(/6067,5813,5518/)
        data_ped(6068,1:3)=(/6068,5681,4844/)
        data_ped(6069,1:3)=(/6069,5603,0/)
        data_ped(6070,1:3)=(/6070,5325,5507/)
        data_ped(6071,1:3)=(/6071,5869,5089/)
        data_ped(6072,1:3)=(/6072,5612,5779/)
        data_ped(6073,1:3)=(/6073,3864,4576/)
        data_ped(6074,1:3)=(/6074,3752,5711/)
        data_ped(6075,1:3)=(/6075,5670,5575/)
        data_ped(6076,1:3)=(/6076,5670,4907/)
        data_ped(6077,1:3)=(/6077,1543,117/)
        data_ped(6078,1:3)=(/6078,5410,5366/)
        data_ped(6079,1:3)=(/6079,5665,5181/)
        data_ped(6080,1:3)=(/6080,5745,5373/)
        data_ped(6081,1:3)=(/6081,4972,4666/)
        data_ped(6082,1:3)=(/6082,4174,5205/)
        data_ped(6083,1:3)=(/6083,5136,5287/)
        data_ped(6084,1:3)=(/6084,3864,5323/)
        data_ped(6085,1:3)=(/6085,5402,5710/)
        data_ped(6086,1:3)=(/6086,5665,5784/)
        data_ped(6087,1:3)=(/6087,5337,5319/)
        data_ped(6088,1:3)=(/6088,4992,5547/)
        data_ped(6089,1:3)=(/6089,4992,5393/)
        data_ped(6090,1:3)=(/6090,5675,5912/)
        data_ped(6091,1:3)=(/6091,4707,4950/)
        data_ped(6092,1:3)=(/6092,4887,5500/)
        data_ped(6093,1:3)=(/6093,4985,5772/)
        data_ped(6094,1:3)=(/6094,5422,5732/)
        data_ped(6095,1:3)=(/6095,5565,5483/)
        data_ped(6096,1:3)=(/6096,5342,5702/)
        data_ped(6097,1:3)=(/6097,5815,5898/)
        data_ped(6098,1:3)=(/6098,5815,5626/)
        data_ped(6099,1:3)=(/6099,4977,5766/)
        data_ped(6100,1:3)=(/6100,5747,4498/)
        data_ped(6101,1:3)=(/6101,5748,5797/)
        data_ped(6102,1:3)=(/6102,5748,5698/)
        data_ped(6103,1:3)=(/6103,4989,5558/)
        data_ped(6104,1:3)=(/6104,5496,4586/)
        data_ped(6105,1:3)=(/6105,5556,5459/)
        data_ped(6106,1:3)=(/6106,5400,5910/)
        data_ped(6107,1:3)=(/6107,5336,5794/)
        data_ped(6108,1:3)=(/6108,5943,5725/)
        data_ped(6109,1:3)=(/6109,1328,0/)
        data_ped(6110,1:3)=(/6110,5737,5782/)
        data_ped(6111,1:3)=(/6111,5425,5575/)
        data_ped(6112,1:3)=(/6112,5399,5431/)
        data_ped(6113,1:3)=(/6113,4962,5433/)
        data_ped(6114,1:3)=(/6114,5940,5785/)
        data_ped(6115,1:3)=(/6115,5746,4758/)
        data_ped(6116,1:3)=(/6116,5668,5920/)
        data_ped(6117,1:3)=(/6117,5810,5010/)
        data_ped(6118,1:3)=(/6118,1317,0/)
        data_ped(6119,1:3)=(/6119,5737,5709/)
        data_ped(6120,1:3)=(/6120,5531,5848/)
        data_ped(6121,1:3)=(/6121,5879,4577/)
        data_ped(6122,1:3)=(/6122,4056,4388/)
        data_ped(6123,1:3)=(/6123,5341,5525/)
        data_ped(6124,1:3)=(/6124,5341,5241/)
        data_ped(6125,1:3)=(/6125,5074,1333/)
        data_ped(6126,1:3)=(/6126,4891,5653/)
        data_ped(6127,1:3)=(/6127,5903,5910/)
        data_ped(6128,1:3)=(/6128,5615,5595/)
        data_ped(6129,1:3)=(/6129,5678,5962/)
        data_ped(6130,1:3)=(/6130,5678,5834/)
        data_ped(6131,1:3)=(/6131,5678,5838/)
        data_ped(6132,1:3)=(/6132,5678,5692/)
        data_ped(6133,1:3)=(/6133,5420,5120/)
        data_ped(6134,1:3)=(/6134,5747,5390/)
        data_ped(6135,1:3)=(/6135,4810,5958/)
        data_ped(6136,1:3)=(/6136,4616,5655/)
        data_ped(6137,1:3)=(/6137,5061,4862/)
        data_ped(6138,1:3)=(/6138,5274,1316/)
        data_ped(6139,1:3)=(/6139,5688,5682/)
        data_ped(6140,1:3)=(/6140,5417,5315/)
        data_ped(6141,1:3)=(/6141,5400,5630/)
        data_ped(6142,1:3)=(/6142,0,1305/)
        data_ped(6143,1:3)=(/6143,5742,5715/)
        data_ped(6144,1:3)=(/6144,5985,5735/)
        data_ped(6145,1:3)=(/6145,5331,5574/)
        data_ped(6146,1:3)=(/6146,5058,5924/)
        data_ped(6147,1:3)=(/6147,5335,5976/)
        data_ped(6148,1:3)=(/6148,5877,5859/)
        data_ped(6149,1:3)=(/6149,5602,5734/)
        data_ped(6150,1:3)=(/6150,5425,5974/)
        data_ped(6151,1:3)=(/6151,4893,5619/)
        data_ped(6152,1:3)=(/6152,5067,4942/)
        data_ped(6153,1:3)=(/6153,5633,6016/)
        data_ped(6154,1:3)=(/6154,5847,5322/)
        data_ped(6155,1:3)=(/6155,4887,5890/)
        data_ped(6156,1:3)=(/6156,5338,5760/)
        data_ped(6157,1:3)=(/6157,5903,5704/)
        data_ped(6158,1:3)=(/6158,5903,5656/)
        data_ped(6159,1:3)=(/6159,5903,5041/)
        data_ped(6160,1:3)=(/6160,5903,5797/)
        data_ped(6161,1:3)=(/6161,5903,5961/)
        data_ped(6162,1:3)=(/6162,5759,5570/)
        data_ped(6163,1:3)=(/6163,5622,5480/)
        data_ped(6164,1:3)=(/6164,4812,5752/)
        data_ped(6165,1:3)=(/6165,5478,5458/)
        data_ped(6166,1:3)=(/6166,5635,5486/)
        data_ped(6167,1:3)=(/6167,5279,6007/)
        data_ped(6168,1:3)=(/6168,5279,5567/)
        data_ped(6169,1:3)=(/6169,5423,5488/)
        data_ped(6170,1:3)=(/6170,5747,5730/)
        data_ped(6171,1:3)=(/6171,5956,5834/)
        data_ped(6172,1:3)=(/6172,4623,6000/)
        data_ped(6173,1:3)=(/6173,5418,5703/)
        data_ped(6174,1:3)=(/6174,5481,5895/)
        data_ped(6175,1:3)=(/6175,5684,5768/)
        data_ped(6176,1:3)=(/6176,5776,5771/)
        data_ped(6177,1:3)=(/6177,5938,5658/)
        data_ped(6178,1:3)=(/6178,5400,5908/)
        data_ped(6179,1:3)=(/6179,5400,5630/)
        data_ped(6180,1:3)=(/6180,5531,5242/)
        data_ped(6181,1:3)=(/6181,5951,5461/)
        data_ped(6182,1:3)=(/6182,5673,5448/)
        data_ped(6183,1:3)=(/6183,5536,5512/)
        data_ped(6184,1:3)=(/6184,5984,6031/)
        data_ped(6185,1:3)=(/6185,5136,5914/)
        data_ped(6186,1:3)=(/6186,5942,5644/)
        data_ped(6187,1:3)=(/6187,0,140/)
        data_ped(6188,1:3)=(/6188,5739,6029/)
        data_ped(6189,1:3)=(/6189,5876,6142/)
        data_ped(6190,1:3)=(/6190,5477,5092/)
        data_ped(6191,1:3)=(/6191,5160,5242/)
        data_ped(6192,1:3)=(/6192,5160,5704/)
        data_ped(6193,1:3)=(/6193,5067,5233/)
        data_ped(6194,1:3)=(/6194,5815,6044/)
        data_ped(6195,1:3)=(/6195,5614,5864/)
        data_ped(6196,1:3)=(/6196,5749,5893/)
        data_ped(6197,1:3)=(/6197,4887,5832/)
        data_ped(6198,1:3)=(/6198,5707,5824/)
        data_ped(6199,1:3)=(/6199,6010,6056/)
        data_ped(6200,1:3)=(/6200,5816,5035/)
        data_ped(6201,1:3)=(/6201,5816,5458/)
        data_ped(6202,1:3)=(/6202,5956,6053/)
        data_ped(6203,1:3)=(/6203,5956,5757/)
        data_ped(6204,1:3)=(/6204,5278,5553/)
        data_ped(6205,1:3)=(/6205,4980,5882/)
        data_ped(6206,1:3)=(/6206,5467,5930/)
        data_ped(6207,1:3)=(/6207,5984,5801/)
        data_ped(6208,1:3)=(/6208,5741,5793/)
        data_ped(6209,1:3)=(/6209,6036,4762/)
        data_ped(6210,1:3)=(/6210,5939,6017/)
        data_ped(6211,1:3)=(/6211,5937,5638/)
        data_ped(6212,1:3)=(/6212,5665,5787/)
        data_ped(6213,1:3)=(/6213,5984,5867/)
        data_ped(6214,1:3)=(/6214,5945,139/)
        data_ped(6215,1:3)=(/6215,5808,5977/)
        data_ped(6216,1:3)=(/6216,5942,5719/)
        data_ped(6217,1:3)=(/6217,5331,5780/)
        data_ped(6218,1:3)=(/6218,5988,5917/)
        data_ped(6219,1:3)=(/6219,5608,5925/)
        data_ped(6220,1:3)=(/6220,5158,4680/)
        data_ped(6221,1:3)=(/6221,5677,6027/)
        data_ped(6222,1:3)=(/6222,5949,5461/)
        data_ped(6223,1:3)=(/6223,5624,5841/)
        data_ped(6224,1:3)=(/6224,5422,5732/)
        data_ped(6225,1:3)=(/6225,5633,5896/)
        data_ped(6226,1:3)=(/6226,5903,5960/)
        data_ped(6227,1:3)=(/6227,5635,5693/)
        data_ped(6228,1:3)=(/6228,5749,5455/)
        data_ped(6229,1:3)=(/6229,5069,5479/)
        data_ped(6230,1:3)=(/6230,5613,6088/)
        data_ped(6231,1:3)=(/6231,5613,5764/)
        data_ped(6232,1:3)=(/6232,5613,5901/)
        data_ped(6233,1:3)=(/6233,5613,5458/)
        data_ped(6234,1:3)=(/6234,5613,5904/)
        data_ped(6235,1:3)=(/6235,5675,5796/)
        data_ped(6236,1:3)=(/6236,5675,5929/)
        data_ped(6237,1:3)=(/6237,4616,5971/)
        data_ped(6238,1:3)=(/6238,5269,5395/)
        data_ped(6239,1:3)=(/6239,5871,6004/)
        data_ped(6240,1:3)=(/6240,5805,5799/)
        data_ped(6241,1:3)=(/6241,5671,5720/)
        data_ped(6242,1:3)=(/6242,5669,6019/)
        data_ped(6243,1:3)=(/6243,5946,5851/)
        data_ped(6244,1:3)=(/6244,5672,5853/)
        data_ped(6245,1:3)=(/6245,6034,6069/)
        data_ped(6246,1:3)=(/6246,5874,5921/)
        data_ped(6247,1:3)=(/6247,5736,5204/)
        data_ped(6248,1:3)=(/6248,6078,5852/)
        data_ped(6249,1:3)=(/6249,6034,5792/)
        data_ped(6250,1:3)=(/6250,5334,5855/)
        data_ped(6251,1:3)=(/6251,5668,5923/)
        data_ped(6252,1:3)=(/6252,5741,5850/)
        data_ped(6253,1:3)=(/6253,5327,5717/)
        data_ped(6254,1:3)=(/6254,4979,5726/)
        data_ped(6255,1:3)=(/6255,5847,5840/)
        data_ped(6256,1:3)=(/6256,5847,5239/)
        data_ped(6257,1:3)=(/6257,5956,5909/)
        data_ped(6258,1:3)=(/6258,5956,6101/)
        data_ped(6259,1:3)=(/6259,5676,5819/)
        data_ped(6260,1:3)=(/6260,5827,5590/)
        data_ped(6261,1:3)=(/6261,5158,4680/)
        data_ped(6262,1:3)=(/6262,5424,6051/)
        data_ped(6263,1:3)=(/6263,6106,5526/)
        data_ped(6264,1:3)=(/6264,6106,6130/)
        data_ped(6265,1:3)=(/6265,6098,6113/)
        data_ped(6266,1:3)=(/6266,4989,5557/)
        data_ped(6267,1:3)=(/6267,5270,5836/)
        data_ped(6268,1:3)=(/6268,5496,5953/)
        data_ped(6269,1:3)=(/6269,5498,1361/)
        data_ped(6270,1:3)=(/6270,5550,5687/)
        data_ped(6271,1:3)=(/6271,5684,1339/)
        data_ped(6272,1:3)=(/6272,4979,5685/)
        data_ped(6273,1:3)=(/6273,5905,5569/)
        data_ped(6274,1:3)=(/6274,5047,5906/)
        data_ped(6275,1:3)=(/6275,5663,5584/)
        data_ped(6276,1:3)=(/6276,6039,5529/)
        data_ped(6277,1:3)=(/6277,5989,6062/)
        data_ped(6278,1:3)=(/6278,5406,6018/)
        data_ped(6279,1:3)=(/6279,5465,5712/)
        data_ped(6280,1:3)=(/6280,5988,5918/)
        data_ped(6281,1:3)=(/6281,5665,5641/)
        data_ped(6282,1:3)=(/6282,5272,5829/)
        data_ped(6283,1:3)=(/6283,5674,5863/)
        data_ped(6284,1:3)=(/6284,5878,5930/)
        data_ped(6285,1:3)=(/6285,6083,5819/)
        data_ped(6286,1:3)=(/6286,5843,6152/)
        data_ped(6287,1:3)=(/6287,5956,6058/)
        data_ped(6288,1:3)=(/6288,5973,5571/)
        data_ped(6289,1:3)=(/6289,6010,5962/)
        data_ped(6290,1:3)=(/6290,6047,6124/)
        data_ped(6291,1:3)=(/6291,6106,6054/)
        data_ped(6292,1:3)=(/6292,6106,5928/)
        data_ped(6293,1:3)=(/6293,6120,6112/)
        data_ped(6294,1:3)=(/6294,5847,6071/)
        data_ped(6295,1:3)=(/6295,5963,6173/)
        data_ped(6296,1:3)=(/6296,6119,6097/)
        data_ped(6297,1:3)=(/6297,5749,5911/)
        data_ped(6298,1:3)=(/6298,6086,5690/)
        data_ped(6299,1:3)=(/6299,5343,6092/)
        data_ped(6300,1:3)=(/6300,1336,1315/)
        data_ped(6301,1:3)=(/6301,4989,6043/)
        data_ped(6302,1:3)=(/6302,6038,5964/)
        data_ped(6303,1:3)=(/6303,5555,1360/)
        data_ped(6304,1:3)=(/6304,5868,5598/)
        data_ped(6305,1:3)=(/6305,6187,1322/)
        data_ped(6306,1:3)=(/6306,6147,5299/)
        data_ped(6307,1:3)=(/6307,6118,6109/)
        data_ped(6308,1:3)=(/6308,5812,5790/)
        data_ped(6309,1:3)=(/6309,5984,5527/)
        data_ped(6310,1:3)=(/6310,6117,5721/)
        data_ped(6311,1:3)=(/6311,5048,5645/)
        data_ped(6312,1:3)=(/6312,5327,5179/)
        data_ped(6313,1:3)=(/6313,5941,5926/)
        data_ped(6314,1:3)=(/6314,5665,5294/)
        data_ped(6315,1:3)=(/6315,5471,5975/)
        data_ped(6316,1:3)=(/6316,5067,6170/)
        data_ped(6317,1:3)=(/6317,5160,5902/)
        data_ped(6318,1:3)=(/6318,5879,5865/)
        data_ped(6319,1:3)=(/6319,5844,5979/)
        data_ped(6320,1:3)=(/6320,5846,5590/)
        data_ped(6321,1:3)=(/6321,5888,5597/)
        data_ped(6322,1:3)=(/6322,5675,5966/)
        data_ped(6323,1:3)=(/6323,6119,6196/)
        data_ped(6324,1:3)=(/6324,5749,5912/)
        data_ped(6325,1:3)=(/6325,5158,5501/)
        data_ped(6326,1:3)=(/6326,5965,1347/)
        data_ped(6327,1:3)=(/6327,6085,5797/)
        data_ped(6328,1:3)=(/6328,6120,5388/)
        data_ped(6329,1:3)=(/6329,5986,6064/)
        data_ped(6330,1:3)=(/6330,5327,5916/)
        data_ped(6331,1:3)=(/6331,5149,5651/)
        data_ped(6332,1:3)=(/6332,6079,5647/)
        data_ped(6333,1:3)=(/6333,5609,0/)
        data_ped(6334,1:3)=(/6334,5811,6076/)
        data_ped(6335,1:3)=(/6335,5875,6182/)
        data_ped(6336,1:3)=(/6336,5270,6203/)
        data_ped(6337,1:3)=(/6337,5618,5694/)
        data_ped(6338,1:3)=(/6338,5164,5657/)
        data_ped(6339,1:3)=(/6339,5421,6198/)
        data_ped(6340,1:3)=(/6340,6083,6089/)
        data_ped(6341,1:3)=(/6341,5770,6072/)
        data_ped(6342,1:3)=(/6342,5956,5900/)
        data_ped(6343,1:3)=(/6343,5675,6167/)
        data_ped(6344,1:3)=(/6344,6149,6022/)
        data_ped(6345,1:3)=(/6345,5765,5887/)
        data_ped(6346,1:3)=(/6346,4979,5685/)
        data_ped(6347,1:3)=(/6347,6039,5662/)
        data_ped(6348,1:3)=(/6348,5669,6063/)
        data_ped(6349,1:3)=(/6349,6081,5206/)
        data_ped(6350,1:3)=(/6350,5809,5915/)
        data_ped(6351,1:3)=(/6351,5773,6104/)
        data_ped(6352,1:3)=(/6352,4976,5317/)
        data_ped(6353,1:3)=(/6353,5993,6074/)
        data_ped(6354,1:3)=(/6354,6001,5756/)
        data_ped(6355,1:3)=(/6355,6047,6124/)
        data_ped(6356,1:3)=(/6356,6048,6023/)
        data_ped(6357,1:3)=(/6357,6049,6174/)
        data_ped(6358,1:3)=(/6358,6055,5523/)
        data_ped(6359,1:3)=(/6359,6093,6091/)
        data_ped(6360,1:3)=(/6360,6154,5322/)
        data_ped(6361,1:3)=(/6361,6221,6265/)
        data_ped(6362,1:3)=(/6362,6264,6054/)
        data_ped(6363,1:3)=(/6363,6150,6075/)
        data_ped(6364,1:3)=(/6364,5400,6025/)
        data_ped(6365,1:3)=(/6365,6215,6020/)
        data_ped(6366,1:3)=(/6366,1365,0/)
        data_ped(6367,1:3)=(/6367,6214,6305/)
        data_ped(6368,1:3)=(/6368,5944,5792/)
        data_ped(6369,1:3)=(/6369,5158,1324/)
        data_ped(6370,1:3)=(/6370,6037,5837/)
        data_ped(6371,1:3)=(/6371,5952,6089/)
        data_ped(6372,1:3)=(/6372,5956,6212/)
        data_ped(6373,1:3)=(/6373,6181,6026/)
        data_ped(6374,1:3)=(/6374,5612,6114/)
        data_ped(6375,1:3)=(/6375,6116,6178/)
        data_ped(6376,1:3)=(/6376,5992,6209/)
        data_ped(6377,1:3)=(/6377,6079,6275/)
        data_ped(6378,1:3)=(/6378,5158,1353/)
        data_ped(6379,1:3)=(/6379,1321,6300/)
        data_ped(6380,1:3)=(/6380,5612,6183/)
        data_ped(6381,1:3)=(/6381,6150,6236/)
        data_ped(6382,1:3)=(/6382,6283,1371/)
        data_ped(6383,1:3)=(/6383,6055,5886/)
        data_ped(6384,1:3)=(/6384,5881,6103/)
        data_ped(6385,1:3)=(/6385,5665,5851/)
        data_ped(6386,1:3)=(/6386,6313,6068/)
        data_ped(6387,1:3)=(/6387,5770,6072/)
        data_ped(6388,1:3)=(/6388,6001,6302/)
        data_ped(6389,1:3)=(/6389,4979,5685/)
        data_ped(6390,1:3)=(/6390,5879,5980/)
        data_ped(6391,1:3)=(/6391,6218,6025/)
        data_ped(6392,1:3)=(/6392,5993,5820/)
        data_ped(6393,1:3)=(/6393,5680,6340/)
        data_ped(6394,1:3)=(/6394,5993,1334/)
        data_ped(6395,1:3)=(/6395,5993,6325/)
        data_ped(6396,1:3)=(/6396,6252,6277/)
        data_ped(6397,1:3)=(/6397,6150,6324/)
        data_ped(6398,1:3)=(/6398,6106,5704/)
        data_ped(6399,1:3)=(/6399,6049,6032/)
        data_ped(6400,1:3)=(/6400,5846,5590/)
        data_ped(6401,1:3)=(/6401,6274,5755/)
        data_ped(6402,1:3)=(/6402,6280,5839/)
        data_ped(6403,1:3)=(/6403,6280,6132/)
        data_ped(6404,1:3)=(/6404,6264,6192/)
        data_ped(6405,1:3)=(/6405,6346,6095/)
        data_ped(6406,1:3)=(/6406,6177,6145/)
        data_ped(6407,1:3)=(/6407,6290,6301/)
        data_ped(6408,1:3)=(/6408,6047,6185/)
        data_ped(6409,1:3)=(/6409,5993,6204/)
        data_ped(6410,1:3)=(/6410,6258,5957/)
        data_ped(6411,1:3)=(/6411,6013,6295/)
        data_ped(6412,1:3)=(/6412,6049,5899/)
        data_ped(6413,1:3)=(/6413,6188,6276/)
        data_ped(6414,1:3)=(/6414,5993,6369/)
        data_ped(6415,1:3)=(/6415,6315,1369/)
        data_ped(6416,1:3)=(/6416,5994,6371/)
        data_ped(6417,1:3)=(/6417,6230,6327/)
        data_ped(6418,1:3)=(/6418,5969,6127/)
        data_ped(6419,1:3)=(/6419,6235,6140/)
        data_ped(6420,1:3)=(/6420,6346,6383/)
        data_ped(6421,1:3)=(/6421,6122,6231/)
        data_ped(6422,1:3)=(/6422,5681,6342/)
        data_ped(6423,1:3)=(/6423,6049,5899/)
        data_ped(6424,1:3)=(/6424,6281,6186/)
        data_ped(6425,1:3)=(/6425,6260,6192/)
        data_ped(6426,1:3)=(/6426,6188,6347/)
        data_ped(6427,1:3)=(/6427,6189,6367/)
        data_ped(6428,1:3)=(/6428,6338,6293/)
        data_ped(6429,1:3)=(/6429,6253,1355/)
        data_ped(6430,1:3)=(/6430,6100,6137/)
        data_ped(6431,1:3)=(/6431,5993,1373/)
        data_ped(6432,1:3)=(/6432,6258,6319/)
        data_ped(6433,1:3)=(/6433,6396,6292/)
        data_ped(6434,1:3)=(/6434,6281,6364/)
        data_ped(6435,1:3)=(/6435,6003,5999/)
        data_ped(6436,1:3)=(/6436,6217,6296/)
        data_ped(6437,1:3)=(/6437,6297,1386/)
        data_ped(6438,1:3)=(/6438,6255,1396/)
        data_ped(6439,1:3)=(/6439,6155,6299/)
        data_ped(6440,1:3)=(/6440,6281,6146/)
        data_ped(6441,1:3)=(/6441,6049,6248/)
        data_ped(6442,1:3)=(/6442,6258,1398/)
        data_ped(6443,1:3)=(/6443,6235,1393/)
        data_ped(6444,1:3)=(/6444,6049,6310/)
        data_ped(6445,1:3)=(/6445,6253,6352/)
        data_ped(6446,1:3)=(/6446,6194,5935/)
        data_ped(6447,1:3)=(/6447,5815,6139/)
        data_ped(6448,1:3)=(/6448,6333,1505/)
        data_ped(6449,1:3)=(/6449,6235,1388/)
        data_ped(6450,1:3)=(/6450,6235,6339/)
        data_ped(6451,1:3)=(/6451,5956,6205/)
        data_ped(6452,1:3)=(/6452,6035,6376/)
        data_ped(6453,1:3)=(/6453,6219,6050/)
        data_ped(6454,1:3)=(/6454,6202,1399/)
        data_ped(6455,1:3)=(/6455,6219,6288/)
        data_ped(6456,1:3)=(/6456,6335,6397/)
        data_ped(6457,1:3)=(/6457,6219,6354/)
        data_ped(6458,1:3)=(/6458,6190,6254/)
        data_ped(6459,1:3)=(/6459,6403,6328/)
        data_ped(6460,1:3)=(/6460,6374,6024/)
        data_ped(6461,1:3)=(/6461,6403,1383/)
        data_ped(6462,1:3)=(/6462,6374,1413/)
        data_ped(6463,1:3)=(/6463,6219,6387/)
        data_ped(6464,1:3)=(/6464,6219,6228/)
        data_ped(6465,1:3)=(/6465,6237,6389/)
        data_ped(6466,1:3)=(/6466,6216,1387/)
        data_ped(6467,1:3)=(/6467,2842,2931/)
        data_ped(6468,1:3)=(/6468,3026,1507/)
        data_ped(6469,1:3)=(/6469,3662,6468/)
        data_ped(6470,1:3)=(/6470,2449,0/)
        data_ped(6471,1:3)=(/6471,2018,1430/)
        data_ped(6472,1:3)=(/6472,0,1432/)
        data_ped(6473,1:3)=(/6473,3299,3692/)
        data_ped(6474,1:3)=(/6474,1517,9/)
        data_ped(6475,1:3)=(/6475,4158,4372/)
        data_ped(6476,1:3)=(/6476,3392,19/)
        data_ped(6477,1:3)=(/6477,4246,3605/)
        data_ped(6478,1:3)=(/6478,2054,0/)
        data_ped(6479,1:3)=(/6479,1935,26/)
        data_ped(6480,1:3)=(/6480,2119,0/)
        data_ped(6481,1:3)=(/6481,2010,0/)
        data_ped(6482,1:3)=(/6482,0,2669/)
        data_ped(6483,1:3)=(/6483,3116,0/)
        data_ped(6484,1:3)=(/6484,4066,43/)
        data_ped(6485,1:3)=(/6485,3836,4099/)
        data_ped(6486,1:3)=(/6486,0,1585/)
        data_ped(6487,1:3)=(/6487,1528,1442/)
        data_ped(6488,1:3)=(/6488,2128,0/)
        data_ped(6489,1:3)=(/6489,0,3162/)
        data_ped(6490,1:3)=(/6490,797,1806/)
        data_ped(6491,1:3)=(/6491,2054,0/)
        data_ped(6492,1:3)=(/6492,1809,1435/)
        data_ped(6493,1:3)=(/6493,287,1584/)
        data_ped(6494,1:3)=(/6494,1527,1437/)
        data_ped(6495,1:3)=(/6495,1666,0/)
        data_ped(6496,1:3)=(/6496,0,1438/)
        data_ped(6497,1:3)=(/6497,1618,0/)
        data_ped(6498,1:3)=(/6498,1533,92/)
        data_ped(6499,1:3)=(/6499,1561,1448/)
        data_ped(6500,1:3)=(/6500,3040,6488/)
        data_ped(6501,1:3)=(/6501,312,1450/)
        data_ped(6502,1:3)=(/6502,1523,1461/)
        data_ped(6503,1:3)=(/6503,316,1454/)
        data_ped(6504,1:3)=(/6504,1537,0/)
        data_ped(6505,1:3)=(/6505,1523,116/)
        data_ped(6506,1:3)=(/6506,4512,5716/)
        data_ped(6507,1:3)=(/6507,3303,3718/)
        data_ped(6508,1:3)=(/6508,329,1457/)
        data_ped(6509,1:3)=(/6509,1539,0/)
        data_ped(6510,1:3)=(/6510,0,1463/)
        data_ped(6511,1:3)=(/6511,331,1458/)
        data_ped(6512,1:3)=(/6512,334,3528/)
        data_ped(6513,1:3)=(/6513,1535,0/)
        data_ped(6514,1:3)=(/6514,4169,6512/)
        data_ped(6515,1:3)=(/6515,337,1482/)
        data_ped(6516,1:3)=(/6516,3854,4216/)
        data_ped(6517,1:3)=(/6517,3303,4319/)
        data_ped(6518,1:3)=(/6518,368,1466/)
        data_ped(6519,1:3)=(/6519,375,1473/)
        data_ped(6520,1:3)=(/6520,376,1469/)
        data_ped(6521,1:3)=(/6521,371,1468/)
        data_ped(6522,1:3)=(/6522,379,1470/)
        data_ped(6523,1:3)=(/6523,385,1471/)
        data_ped(6524,1:3)=(/6524,3851,4467/)
        data_ped(6525,1:3)=(/6525,1550,131/)
        data_ped(6526,1:3)=(/6526,357,1477/)
        data_ped(6527,1:3)=(/6527,360,1478/)
        data_ped(6528,1:3)=(/6528,363,1479/)
        data_ped(6529,1:3)=(/6529,1546,0/)
        data_ped(6530,1:3)=(/6530,1546,0/)
        data_ped(6531,1:3)=(/6531,1544,0/)
        data_ped(6532,1:3)=(/6532,1546,0/)
        data_ped(6533,1:3)=(/6533,1548,0/)
        data_ped(6534,1:3)=(/6534,342,1483/)
        data_ped(6535,1:3)=(/6535,1546,1485/)
        data_ped(6536,1:3)=(/6536,3404,135/)
        data_ped(6537,1:3)=(/6537,1544,1489/)
        data_ped(6538,1:3)=(/6538,1556,1490/)
        data_ped(6539,1:3)=(/6539,4350,1491/)
        data_ped(6540,1:3)=(/6540,1549,0/)
        data_ped(6541,1:3)=(/6541,1545,1495/)
        data_ped(6542,1:3)=(/6542,337,1494/)
        data_ped(6543,1:3)=(/6543,3858,4322/)
        data_ped(6544,1:3)=(/6544,1554,0/)
        data_ped(6545,1:3)=(/6545,2797,4318/)
        data_ped(6546,1:3)=(/6546,1556,0/)
        data_ped(6547,1:3)=(/6547,1554,0/)
        data_ped(6548,1:3)=(/6548,1563,6531/)
        data_ped(6549,1:3)=(/6549,4350,1501/)
        data_ped(6550,1:3)=(/6550,3203,3363/)
        data_ped(6551,1:3)=(/6551,5610,5858/)
        data_ped(6552,1:3)=(/6552,1555,0/)
        data_ped(6553,1:3)=(/6553,1815,0/)
        data_ped(6554,1:3)=(/6554,2717,2355/)
        data_ped(6555,1:3)=(/6555,2053,0/)
        data_ped(6556,1:3)=(/6556,2055,0/)
        data_ped(6557,1:3)=(/6557,2116,1887/)
        data_ped(6558,1:3)=(/6558,1830,0/)
        data_ped(6559,1:3)=(/6559,1654,0/)
        data_ped(6560,1:3)=(/6560,2528,0/)
        data_ped(6561,1:3)=(/6561,1814,0/)
        data_ped(6562,1:3)=(/6562,0,1962/)
        data_ped(6563,1:3)=(/6563,3399,1429/)
        data_ped(6564,1:3)=(/6564,1530,1438/)
        data_ped(6565,1:3)=(/6565,0,1514/)
        data_ped(6566,1:3)=(/6566,0,2152/)
        data_ped(6567,1:3)=(/6567,1811,0/)
        data_ped(6568,1:3)=(/6568,2055,0/)
        data_ped(6569,1:3)=(/6569,2055,0/)
        data_ped(6570,1:3)=(/6570,2054,0/)
        data_ped(6571,1:3)=(/6571,2061,6492/)
        data_ped(6572,1:3)=(/6572,296,6497/)
        data_ped(6573,1:3)=(/6573,2303,2033/)
        data_ped(6574,1:3)=(/6574,2858,1431/)
        data_ped(6575,1:3)=(/6575,1534,1419/)
        data_ped(6576,1:3)=(/6576,3036,2506/)
        data_ped(6577,1:3)=(/6577,3040,59/)
        data_ped(6578,1:3)=(/6578,2054,25/)
        data_ped(6579,1:3)=(/6579,6575,6482/)
        data_ped(6580,1:3)=(/6580,1524,6503/)
        data_ped(6581,1:3)=(/6581,0,1845/)
        data_ped(6582,1:3)=(/6582,1525,1436/)
        data_ped(6583,1:3)=(/6583,1654,99/)
        data_ped(6584,1:3)=(/6584,1538,1447/)
        data_ped(6585,1:3)=(/6585,1707,0/)
        data_ped(6586,1:3)=(/6586,1707,0/)
        data_ped(6587,1:3)=(/6587,3207,0/)
        data_ped(6588,1:3)=(/6588,1536,6509/)
        data_ped(6589,1:3)=(/6589,1540,1455/)
        data_ped(6590,1:3)=(/6590,1206,2679/)
        data_ped(6591,1:3)=(/6591,2630,6510/)
        data_ped(6592,1:3)=(/6592,1542,1464/)
        data_ped(6593,1:3)=(/6593,1546,1481/)
        data_ped(6594,1:3)=(/6594,1545,6532/)
        data_ped(6595,1:3)=(/6595,1547,6533/)
        data_ped(6596,1:3)=(/6596,3853,0/)
        data_ped(6597,1:3)=(/6597,1545,6535/)
        data_ped(6598,1:3)=(/6598,1537,6529/)
        data_ped(6599,1:3)=(/6599,1553,134/)
        data_ped(6600,1:3)=(/6600,1552,1480/)
        data_ped(6601,1:3)=(/6601,3959,3449/)
        data_ped(6602,1:3)=(/6602,6593,6537/)
        data_ped(6603,1:3)=(/6603,6593,1492/)
        data_ped(6604,1:3)=(/6604,3853,0/)
        data_ped(6605,1:3)=(/6605,6600,1503/)
        data_ped(6606,1:3)=(/6606,6604,6541/)
        data_ped(6607,1:3)=(/6607,1562,1502/)
        data_ped(6608,1:3)=(/6608,1564,4568/)
        data_ped(6609,1:3)=(/6609,3205,0/)
        data_ped(6610,1:3)=(/6610,6600,1499/)
        data_ped(6611,1:3)=(/6611,1566,3536/)
        data_ped(6612,1:3)=(/6612,5057,4560/)
        data_ped(6613,1:3)=(/6613,1708,0/)
        data_ped(6614,1:3)=(/6614,1654,1506/)
        data_ped(6615,1:3)=(/6615,365,6528/)
        data_ped(6616,1:3)=(/6616,367,6518/)
        data_ped(6617,1:3)=(/6617,3942,0/)
        data_ped(6618,1:3)=(/6618,378,6616/)
        data_ped(6619,1:3)=(/6619,1638,6496/)
        data_ped(6620,1:3)=(/6620,1666,6493/)
        data_ped(6621,1:3)=(/6621,1666,6486/)
        data_ped(6622,1:3)=(/6622,2864,0/)
        data_ped(6623,1:3)=(/6623,1666,6493/)
        data_ped(6624,1:3)=(/6624,1809,6494/)
        data_ped(6625,1:3)=(/6625,1532,6498/)
        data_ped(6626,1:3)=(/6626,0,6617/)
        data_ped(6627,1:3)=(/6627,1529,6495/)
        data_ped(6628,1:3)=(/6628,6622,95/)
        data_ped(6629,1:3)=(/6629,6621,1826/)
        data_ped(6630,1:3)=(/6630,953,6558/)
        data_ped(6631,1:3)=(/6631,6620,0/)
        data_ped(6632,1:3)=(/6632,6622,0/)
        data_ped(6633,1:3)=(/6633,1531,6572/)
        data_ped(6634,1:3)=(/6634,6621,1875/)
        data_ped(6635,1:3)=(/6635,6627,1854/)
        data_ped(6636,1:3)=(/6636,6621,2092/)
        data_ped(6637,1:3)=(/6637,6614,197/)
        data_ped(6638,1:3)=(/6638,0,6561/)
        data_ped(6639,1:3)=(/6639,6583,1446/)
        data_ped(6640,1:3)=(/6640,6621,0/)
        data_ped(6641,1:3)=(/6641,6621,2189/)
        data_ped(6642,1:3)=(/6642,6621,2096/)
        data_ped(6643,1:3)=(/6643,0,6631/)
        data_ped(6644,1:3)=(/6644,1939,6625/)
        data_ped(6645,1:3)=(/6645,749,6630/)
        data_ped(6646,1:3)=(/6646,1976,6487/)
        data_ped(6647,1:3)=(/6647,2011,6559/)
        data_ped(6648,1:3)=(/6648,6621,2153/)
        data_ped(6649,1:3)=(/6649,6613,869/)
        data_ped(6650,1:3)=(/6650,6621,2420/)
        data_ped(6651,1:3)=(/6651,2318,6564/)
        data_ped(6652,1:3)=(/6652,6624,0/)
        data_ped(6653,1:3)=(/6653,2309,6569/)
        data_ped(6654,1:3)=(/6654,6623,6571/)
        data_ped(6655,1:3)=(/6655,2529,6639/)
        data_ped(6656,1:3)=(/6656,0,6556/)
        data_ped(6657,1:3)=(/6657,6624,0/)
        data_ped(6658,1:3)=(/6658,6627,270/)
        data_ped(6659,1:3)=(/6659,6633,32/)
        data_ped(6660,1:3)=(/6660,2161,6472/)
        data_ped(6661,1:3)=(/6661,2387,6641/)
        data_ped(6662,1:3)=(/6662,1139,6626/)
        data_ped(6663,1:3)=(/6663,6581,0/)
        data_ped(6664,1:3)=(/6664,6624,6567/)
        data_ped(6665,1:3)=(/6665,2535,6647/)
        data_ped(6666,1:3)=(/6666,1926,6568/)
        data_ped(6667,1:3)=(/6667,2549,6648/)
        data_ped(6668,1:3)=(/6668,6581,1796/)
        data_ped(6669,1:3)=(/6669,2161,6490/)
        data_ped(6670,1:3)=(/6670,2523,6642/)
        data_ped(6671,1:3)=(/6671,6581,2924/)
        data_ped(6672,1:3)=(/6672,2222,6640/)
        data_ped(6673,1:3)=(/6673,0,6557/)
        data_ped(6674,1:3)=(/6674,6581,2667/)
        data_ped(6675,1:3)=(/6675,6581,2758/)
        data_ped(6676,1:3)=(/6676,6581,2827/)
        data_ped(6677,1:3)=(/6677,1966,6630/)
        data_ped(6678,1:3)=(/6678,6645,2657/)
        data_ped(6679,1:3)=(/6679,6581,2502/)
        data_ped(6680,1:3)=(/6680,6629,2870/)
        data_ped(6681,1:3)=(/6681,2548,6553/)
        data_ped(6682,1:3)=(/6682,6629,2883/)
        data_ped(6683,1:3)=(/6683,6651,2981/)
        data_ped(6684,1:3)=(/6684,2791,6566/)
        data_ped(6685,1:3)=(/6685,6581,0/)
        data_ped(6686,1:3)=(/6686,6581,0/)
        data_ped(6687,1:3)=(/6687,2714,6641/)
        data_ped(6688,1:3)=(/6688,2456,6635/)
        data_ped(6689,1:3)=(/6689,6653,2745/)
        data_ped(6690,1:3)=(/6690,3021,6562/)
        data_ped(6691,1:3)=(/6691,2538,6628/)
        data_ped(6692,1:3)=(/6692,6581,0/)
        data_ped(6693,1:3)=(/6693,6581,0/)
        data_ped(6694,1:3)=(/6694,6651,2981/)
        data_ped(6695,1:3)=(/6695,2445,6659/)
        data_ped(6696,1:3)=(/6696,6651,2634/)
        data_ped(6697,1:3)=(/6697,2391,6650/)
        data_ped(6698,1:3)=(/6698,6651,3073/)
        data_ped(6699,1:3)=(/6699,6581,2275/)
        data_ped(6700,1:3)=(/6700,6581,2332/)
        data_ped(6701,1:3)=(/6701,6581,2830/)
        data_ped(6702,1:3)=(/6702,6581,2924/)
        data_ped(6703,1:3)=(/6703,2521,6666/)
        data_ped(6704,1:3)=(/6704,2456,6644/)
        data_ped(6705,1:3)=(/6705,6654,2680/)
        data_ped(6706,1:3)=(/6706,2625,6663/)
        data_ped(6707,1:3)=(/6707,2791,6560/)
        data_ped(6708,1:3)=(/6708,3109,6669/)
        data_ped(6709,1:3)=(/6709,2631,6632/)
        data_ped(6710,1:3)=(/6710,2625,6675/)
        data_ped(6711,1:3)=(/6711,2053,6668/)
        data_ped(6712,1:3)=(/6712,6586,1220/)
        data_ped(6713,1:3)=(/6713,2308,6637/)
        data_ped(6714,1:3)=(/6714,2443,6672/)
        data_ped(6715,1:3)=(/6715,3109,6660/)
        data_ped(6716,1:3)=(/6716,6581,2589/)
        data_ped(6717,1:3)=(/6717,6629,2891/)
        data_ped(6718,1:3)=(/6718,2224,6667/)
        data_ped(6719,1:3)=(/6719,3197,6491/)
        data_ped(6720,1:3)=(/6720,2456,6635/)
        data_ped(6721,1:3)=(/6721,6680,2741/)
        data_ped(6722,1:3)=(/6722,6645,2806/)
        data_ped(6723,1:3)=(/6723,2053,6671/)
        data_ped(6724,1:3)=(/6724,3283,6643/)
        data_ped(6725,1:3)=(/6725,2375,6655/)
        data_ped(6726,1:3)=(/6726,6646,1201/)
        data_ped(6727,1:3)=(/6727,6590,6508/)
        data_ped(6728,1:3)=(/6728,3021,6638/)
        data_ped(6729,1:3)=(/6729,1138,6662/)
        data_ped(6730,1:3)=(/6730,6651,2750/)
        data_ped(6731,1:3)=(/6731,3177,6634/)
        data_ped(6732,1:3)=(/6732,3187,6574/)
        data_ped(6733,1:3)=(/6733,6629,2869/)
        data_ped(6734,1:3)=(/6734,3177,6636/)
        data_ped(6735,1:3)=(/6735,6629,2181/)
        data_ped(6736,1:3)=(/6736,2053,6671/)
        data_ped(6737,1:3)=(/6737,2053,6702/)
        data_ped(6738,1:3)=(/6738,6651,2415/)
        data_ped(6739,1:3)=(/6739,6645,3067/)
        data_ped(6740,1:3)=(/6740,2933,6656/)
        data_ped(6741,1:3)=(/6741,2053,6676/)
        data_ped(6742,1:3)=(/6742,3020,6683/)
        data_ped(6743,1:3)=(/6743,2795,6570/)
        data_ped(6744,1:3)=(/6744,2053,6699/)
        data_ped(6745,1:3)=(/6745,2053,6668/)
        data_ped(6746,1:3)=(/6746,3391,6481/)
        data_ped(6747,1:3)=(/6747,2024,6467/)
        data_ped(6748,1:3)=(/6748,2521,6637/)
        data_ped(6749,1:3)=(/6749,3106,6698/)
        data_ped(6750,1:3)=(/6750,3650,6573/)
        data_ped(6751,1:3)=(/6751,2053,6674/)
        data_ped(6752,1:3)=(/6752,2551,6668/)
        data_ped(6753,1:3)=(/6753,2053,6685/)
        data_ped(6754,1:3)=(/6754,6581,0/)
        data_ped(6755,1:3)=(/6755,6687,3062/)
        data_ped(6756,1:3)=(/6756,0,6555/)
        data_ped(6757,1:3)=(/6757,6679,3255/)
        data_ped(6758,1:3)=(/6758,2955,6674/)
        data_ped(6759,1:3)=(/6759,6664,0/)
        data_ped(6760,1:3)=(/6760,6581,2589/)
        data_ped(6761,1:3)=(/6761,2523,6690/)
        data_ped(6762,1:3)=(/6762,6703,3586/)
        data_ped(6763,1:3)=(/6763,3482,6649/)
        data_ped(6764,1:3)=(/6764,6646,3512/)
        data_ped(6765,1:3)=(/6765,3565,6710/)
        data_ped(6766,1:3)=(/6766,6679,2670/)
        data_ped(6767,1:3)=(/6767,6723,0/)
        data_ped(6768,1:3)=(/6768,6723,3793/)
        data_ped(6769,1:3)=(/6769,6723,3083/)
        data_ped(6770,1:3)=(/6770,6581,2504/)
        data_ped(6771,1:3)=(/6771,6679,3621/)
        data_ped(6772,1:3)=(/6772,6687,3511/)
        data_ped(6773,1:3)=(/6773,6687,3630/)
        data_ped(6774,1:3)=(/6774,3652,6658/)
        data_ped(6775,1:3)=(/6775,0,6565/)
        data_ped(6776,1:3)=(/6776,6723,6702/)
        data_ped(6777,1:3)=(/6777,6723,3780/)
        data_ped(6778,1:3)=(/6778,6679,3080/)
        data_ped(6779,1:3)=(/6779,2863,6693/)
        data_ped(6780,1:3)=(/6780,6688,3266/)
        data_ped(6781,1:3)=(/6781,3565,6710/)
        data_ped(6782,1:3)=(/6782,6723,0/)
        data_ped(6783,1:3)=(/6783,3754,6501/)
        data_ped(6784,1:3)=(/6784,6730,6707/)
        data_ped(6785,1:3)=(/6785,6723,1228/)
        data_ped(6786,1:3)=(/6786,6723,6699/)
        data_ped(6787,1:3)=(/6787,6723,3083/)
        data_ped(6788,1:3)=(/6788,6586,1496/)
        data_ped(6789,1:3)=(/6789,3759,6689/)
        data_ped(6790,1:3)=(/6790,6696,6661/)
        data_ped(6791,1:3)=(/6791,3558,6682/)
        data_ped(6792,1:3)=(/6792,6694,3515/)
        data_ped(6793,1:3)=(/6793,6695,3624/)
        data_ped(6794,1:3)=(/6794,2863,6692/)
        data_ped(6795,1:3)=(/6795,2538,6719/)
        data_ped(6796,1:3)=(/6796,6723,6685/)
        data_ped(6797,1:3)=(/6797,3574,6504/)
        data_ped(6798,1:3)=(/6798,3950,6701/)
        data_ped(6799,1:3)=(/6799,6723,0/)
        data_ped(6800,1:3)=(/6800,3846,6684/)
        data_ped(6801,1:3)=(/6801,6734,3917/)
        data_ped(6802,1:3)=(/6802,3464,6697/)
        data_ped(6803,1:3)=(/6803,6723,6706/)
        data_ped(6804,1:3)=(/6804,2520,6673/)
        data_ped(6805,1:3)=(/6805,3481,6710/)
        data_ped(6806,1:3)=(/6806,6724,2817/)
        data_ped(6807,1:3)=(/6807,6704,6657/)
        data_ped(6808,1:3)=(/6808,3843,6709/)
        data_ped(6809,1:3)=(/6809,6743,0/)
        data_ped(6810,1:3)=(/6810,6750,6708/)
        data_ped(6811,1:3)=(/6811,6723,2924/)
        data_ped(6812,1:3)=(/6812,6723,3083/)
        data_ped(6813,1:3)=(/6813,4063,6736/)
        data_ped(6814,1:3)=(/6814,6733,3496/)
        data_ped(6815,1:3)=(/6815,3283,6705/)
        data_ped(6816,1:3)=(/6816,6748,3224/)
        data_ped(6817,1:3)=(/6817,3848,6702/)
        data_ped(6818,1:3)=(/6818,6742,3987/)
        data_ped(6819,1:3)=(/6819,3935,6483/)
        data_ped(6820,1:3)=(/6820,4160,6738/)
        data_ped(6821,1:3)=(/6821,6714,4018/)
        data_ped(6822,1:3)=(/6822,0,6747/)
        data_ped(6823,1:3)=(/6823,2955,6700/)
        data_ped(6824,1:3)=(/6824,6575,1247/)
        data_ped(6825,1:3)=(/6825,6723,3789/)
        data_ped(6826,1:3)=(/6826,4063,6736/)
        data_ped(6827,1:3)=(/6827,6723,6758/)
        data_ped(6828,1:3)=(/6828,4177,6722/)
        data_ped(6829,1:3)=(/6829,3481,6765/)
        data_ped(6830,1:3)=(/6830,6716,4285/)
        data_ped(6831,1:3)=(/6831,3838,6691/)
        data_ped(6832,1:3)=(/6832,3848,6702/)
        data_ped(6833,1:3)=(/6833,6723,3714/)
        data_ped(6834,1:3)=(/6834,4272,6772/)
        data_ped(6835,1:3)=(/6835,6704,6726/)
        data_ped(6836,1:3)=(/6836,4178,6681/)
        data_ped(6837,1:3)=(/6837,6723,1266/)
        data_ped(6838,1:3)=(/6838,6751,3786/)
        data_ped(6839,1:3)=(/6839,3303,6500/)
        data_ped(6840,1:3)=(/6840,3848,6674/)
        data_ped(6841,1:3)=(/6841,2955,6782/)
        data_ped(6842,1:3)=(/6842,6716,0/)
        data_ped(6843,1:3)=(/6843,6762,6721/)
        data_ped(6844,1:3)=(/6844,6748,3549/)
        data_ped(6845,1:3)=(/6845,4182,6713/)
        data_ped(6846,1:3)=(/6846,6731,3516/)
        data_ped(6847,1:3)=(/6847,4056,6554/)
        data_ped(6848,1:3)=(/6848,6774,4138/)
        data_ped(6849,1:3)=(/6849,2543,6768/)
        data_ped(6850,1:3)=(/6850,3848,6711/)
        data_ped(6851,1:3)=(/6851,6723,4308/)
        data_ped(6852,1:3)=(/6852,2863,6706/)
        data_ped(6853,1:3)=(/6853,6716,0/)
        data_ped(6854,1:3)=(/6854,6811,6488/)
        data_ped(6855,1:3)=(/6855,6776,3694/)
        data_ped(6856,1:3)=(/6856,6670,3814/)
        data_ped(6857,1:3)=(/6857,6743,3878/)
        data_ped(6858,1:3)=(/6858,4337,6773/)
        data_ped(6859,1:3)=(/6859,3563,6563/)
        data_ped(6860,1:3)=(/6860,3848,6685/)
        data_ped(6861,1:3)=(/6861,4166,6691/)
        data_ped(6862,1:3)=(/6862,3468,6710/)
        data_ped(6863,1:3)=(/6863,6723,6737/)
        data_ped(6864,1:3)=(/6864,3848,6674/)
        data_ped(6865,1:3)=(/6865,6755,3222/)
        data_ped(6866,1:3)=(/6866,6749,3727/)
        data_ped(6867,1:3)=(/6867,6750,6469/)
        data_ped(6868,1:3)=(/6868,6775,3721/)
        data_ped(6869,1:3)=(/6869,6791,0/)
        data_ped(6870,1:3)=(/6870,6813,4293/)
        data_ped(6871,1:3)=(/6871,6779,3997/)
        data_ped(6872,1:3)=(/6872,2543,6767/)
        data_ped(6873,1:3)=(/6873,3848,6702/)
        data_ped(6874,1:3)=(/6874,3852,6685/)
        data_ped(6875,1:3)=(/6875,3576,6759/)
        data_ped(6876,1:3)=(/6876,6723,4465/)
        data_ped(6877,1:3)=(/6877,6813,0/)
        data_ped(6878,1:3)=(/6878,6716,3786/)
        data_ped(6879,1:3)=(/6879,4519,6665/)
        data_ped(6880,1:3)=(/6880,6740,4365/)
        data_ped(6881,1:3)=(/6881,3973,6783/)
        data_ped(6882,1:3)=(/6882,4247,6474/)
        data_ped(6883,1:3)=(/6883,6775,3989/)
        data_ped(6884,1:3)=(/6884,6716,3897/)
        data_ped(6885,1:3)=(/6885,3468,6781/)
        data_ped(6886,1:3)=(/6886,4056,6757/)
        data_ped(6887,1:3)=(/6887,6723,4002/)
        data_ped(6888,1:3)=(/6888,6785,3161/)
        data_ped(6889,1:3)=(/6889,6789,4487/)
        data_ped(6890,1:3)=(/6890,6577,4293/)
        data_ped(6891,1:3)=(/6891,6744,4664/)
        data_ped(6892,1:3)=(/6892,6723,0/)
        data_ped(6893,1:3)=(/6893,3659,6808/)
        data_ped(6894,1:3)=(/6894,4499,6801/)
        data_ped(6895,1:3)=(/6895,4156,6800/)
        data_ped(6896,1:3)=(/6896,6776,4592/)
        data_ped(6897,1:3)=(/6897,4596,6473/)
        data_ped(6898,1:3)=(/6898,6776,3251/)
        data_ped(6899,1:3)=(/6899,6723,4002/)
        data_ped(6900,1:3)=(/6900,6724,4022/)
        data_ped(6901,1:3)=(/6901,4056,6766/)
        data_ped(6902,1:3)=(/6902,3848,6702/)
        data_ped(6903,1:3)=(/6903,6753,4006/)
        data_ped(6904,1:3)=(/6904,3848,6706/)
        data_ped(6905,1:3)=(/6905,4171,6798/)
        data_ped(6906,1:3)=(/6906,6838,6832/)
        data_ped(6907,1:3)=(/6907,6827,4557/)
        data_ped(6908,1:3)=(/6908,6751,4460/)
        data_ped(6909,1:3)=(/6909,3847,6809/)
        data_ped(6910,1:3)=(/6910,6735,6718/)
        data_ped(6911,1:3)=(/6911,3672,6784/)
        data_ped(6912,1:3)=(/6912,6816,4770/)
        data_ped(6913,1:3)=(/6913,6776,4440/)
        data_ped(6914,1:3)=(/6914,4712,6761/)
        data_ped(6915,1:3)=(/6915,6741,6830/)
        data_ped(6916,1:3)=(/6916,4164,6794/)
        data_ped(6917,1:3)=(/6917,3468,6781/)
        data_ped(6918,1:3)=(/6918,6744,4724/)
        data_ped(6919,1:3)=(/6919,4056,6757/)
        data_ped(6920,1:3)=(/6920,2955,6768/)
        data_ped(6921,1:3)=(/6921,6751,2924/)
        data_ped(6922,1:3)=(/6922,3468,6710/)
        data_ped(6923,1:3)=(/6923,6745,0/)
        data_ped(6924,1:3)=(/6924,4595,6756/)
        data_ped(6925,1:3)=(/6925,6810,6477/)
        data_ped(6926,1:3)=(/6926,4268,6821/)
        data_ped(6927,1:3)=(/6927,3652,6807/)
        data_ped(6928,1:3)=(/6928,6776,4592/)
        data_ped(6929,1:3)=(/6929,6741,3616/)
        data_ped(6930,1:3)=(/6930,6576,4084/)
        data_ped(6931,1:3)=(/6931,6744,4104/)
        data_ped(6932,1:3)=(/6932,6695,6780/)
        data_ped(6933,1:3)=(/6933,3468,6781/)
        data_ped(6934,1:3)=(/6934,3468,6829/)
        data_ped(6935,1:3)=(/6935,6805,4742/)
        data_ped(6936,1:3)=(/6936,6838,6550/)
        data_ped(6937,1:3)=(/6937,6786,4324/)
        data_ped(6938,1:3)=(/6938,4530,6856/)
        data_ped(6939,1:3)=(/6939,4503,6728/)
        data_ped(6940,1:3)=(/6940,4185,6820/)
        data_ped(6941,1:3)=(/6941,6775,4188/)
        data_ped(6942,1:3)=(/6942,4507,6818/)
        data_ped(6943,1:3)=(/6943,4712,6802/)
        data_ped(6944,1:3)=(/6944,6815,4139/)
        data_ped(6945,1:3)=(/6945,3468,6765/)
        data_ped(6946,1:3)=(/6946,6823,6905/)
        data_ped(6947,1:3)=(/6947,4879,6712/)
        data_ped(6948,1:3)=(/6948,6843,4774/)
        data_ped(6949,1:3)=(/6949,4620,6720/)
        data_ped(6950,1:3)=(/6950,4414,6817/)
        data_ped(6951,1:3)=(/6951,6575,4327/)
        data_ped(6952,1:3)=(/6952,4054,6489/)
        data_ped(6953,1:3)=(/6953,6751,6702/)
        data_ped(6954,1:3)=(/6954,6786,3447/)
        data_ped(6955,1:3)=(/6955,3468,6484/)
        data_ped(6956,1:3)=(/6956,6837,4297/)
        data_ped(6957,1:3)=(/6957,6838,6832/)
        data_ped(6958,1:3)=(/6958,6779,6769/)
        data_ped(6959,1:3)=(/6959,6811,6877/)
        data_ped(6960,1:3)=(/6960,6779,0/)
        data_ped(6961,1:3)=(/6961,6845,4406/)
        data_ped(6962,1:3)=(/6962,6835,6858/)
        data_ped(6963,1:3)=(/6963,6843,6880/)
        data_ped(6964,1:3)=(/6964,3864,6857/)
        data_ped(6965,1:3)=(/6965,3933,6868/)
        data_ped(6966,1:3)=(/6966,6776,4861/)
        data_ped(6967,1:3)=(/6967,4888,6869/)
        data_ped(6968,1:3)=(/6968,6804,4718/)
        data_ped(6969,1:3)=(/6969,6776,6819/)
        data_ped(6970,1:3)=(/6970,6777,4285/)
        data_ped(6971,1:3)=(/6971,6823,3697/)
        data_ped(6972,1:3)=(/6972,6777,3521/)
        data_ped(6973,1:3)=(/6973,3848,6852/)
        data_ped(6974,1:3)=(/6974,4698,6709/)
        data_ped(6975,1:3)=(/6975,6787,6760/)
        data_ped(6976,1:3)=(/6976,6813,6507/)
        data_ped(6977,1:3)=(/6977,4413,6847/)
        data_ped(6978,1:3)=(/6978,6725,4233/)
        data_ped(6979,1:3)=(/6979,6776,4785/)
        data_ped(6980,1:3)=(/6980,6811,5005/)
        data_ped(6981,1:3)=(/6981,6803,4638/)
        data_ped(6982,1:3)=(/6982,6823,6874/)
        data_ped(6983,1:3)=(/6983,6838,4652/)
        data_ped(6984,1:3)=(/6984,4167,6892/)
        data_ped(6985,1:3)=(/6985,6860,4466/)
        data_ped(6986,1:3)=(/6986,6744,4840/)
        data_ped(6987,1:3)=(/6987,6823,6849/)
        data_ped(6988,1:3)=(/6988,4686,6475/)
        data_ped(6989,1:3)=(/6989,6861,3602/)
        data_ped(6990,1:3)=(/6990,6843,4852/)
        data_ped(6991,1:3)=(/6991,4243,6476/)
        data_ped(6992,1:3)=(/6992,4790,6746/)
        data_ped(6993,1:3)=(/6993,6803,4726/)
        data_ped(6994,1:3)=(/6994,6885,6832/)
        data_ped(6995,1:3)=(/6995,6777,4914/)
        data_ped(6996,1:3)=(/6996,6723,4752/)
        data_ped(6997,1:3)=(/6997,6811,5013/)
        data_ped(6998,1:3)=(/6998,6839,0/)
        data_ped(6999,1:3)=(/6999,6825,4663/)
        data_ped(7000,1:3)=(/7000,6803,4846/)
        data_ped(7001,1:3)=(/7001,6843,4488/)
        data_ped(7002,1:3)=(/7002,4531,6717/)
        data_ped(7003,1:3)=(/7003,4266,6739/)
        data_ped(7004,1:3)=(/7004,3864,6857/)
        data_ped(7005,1:3)=(/7005,6838,6872/)
        data_ped(7006,1:3)=(/7006,4166,6831/)
        data_ped(7007,1:3)=(/7007,6752,5085/)
        data_ped(7008,1:3)=(/7008,6840,4457/)
        data_ped(7009,1:3)=(/7009,6803,4638/)
        data_ped(7010,1:3)=(/7010,6803,4741/)
        data_ped(7011,1:3)=(/7011,3468,6765/)
        data_ped(7012,1:3)=(/7012,6754,5105/)
        data_ped(7013,1:3)=(/7013,6752,4312/)
        data_ped(7014,1:3)=(/7014,5142,6797/)
        data_ped(7015,1:3)=(/7015,6803,4839/)
        data_ped(7016,1:3)=(/7016,6840,6878/)
        data_ped(7017,1:3)=(/7017,6754,0/)
        data_ped(7018,1:3)=(/7018,6885,4849/)
        data_ped(7019,1:3)=(/7019,6839,0/)
        data_ped(7020,1:3)=(/7020,3935,6855/)
        data_ped(7021,1:3)=(/7021,6776,4953/)
        data_ped(7022,1:3)=(/7022,6803,4651/)
        data_ped(7023,1:3)=(/7023,6803,4910/)
        data_ped(7024,1:3)=(/7024,6837,4297/)
        data_ped(7025,1:3)=(/7025,4961,6485/)
        data_ped(7026,1:3)=(/7026,6840,3997/)
        data_ped(7027,1:3)=(/7027,6752,4122/)
        data_ped(7028,1:3)=(/7028,6741,5014/)
        data_ped(7029,1:3)=(/7029,6838,4657/)
        data_ped(7030,1:3)=(/7030,6838,4657/)
        data_ped(7031,1:3)=(/7031,6826,3906/)
        data_ped(7032,1:3)=(/7032,6861,4095/)
        data_ped(7033,1:3)=(/7033,6776,4592/)
        data_ped(7034,1:3)=(/7034,5066,6834/)
        data_ped(7035,1:3)=(/7035,3864,6857/)
        data_ped(7036,1:3)=(/7036,5067,6881/)
        data_ped(7037,1:3)=(/7037,6865,6678/)
        data_ped(7038,1:3)=(/7038,4697,6709/)
        data_ped(7039,1:3)=(/7039,6822,4728/)
        data_ped(7040,1:3)=(/7040,6744,6873/)
        data_ped(7041,1:3)=(/7041,6915,4639/)
        data_ped(7042,1:3)=(/7042,5073,6795/)
        data_ped(7043,1:3)=(/7043,1521,6502/)
        data_ped(7044,1:3)=(/7044,6861,6893/)
        data_ped(7045,1:3)=(/7045,6886,4131/)
        data_ped(7046,1:3)=(/7046,5280,6479/)
        data_ped(7047,1:3)=(/7047,6929,6701/)
        data_ped(7048,1:3)=(/7048,4867,6871/)
        data_ped(7049,1:3)=(/7049,6803,4638/)
        data_ped(7050,1:3)=(/7050,6744,6933/)
        data_ped(7051,1:3)=(/7051,6886,4204/)
        data_ped(7052,1:3)=(/7052,6945,4835/)
        data_ped(7053,1:3)=(/7053,6841,4660/)
        data_ped(7054,1:3)=(/7054,6803,5017/)
        data_ped(7055,1:3)=(/7055,6864,6770/)
        data_ped(7056,1:3)=(/7056,6744,0/)
        data_ped(7057,1:3)=(/7057,3468,6829/)
        data_ped(7058,1:3)=(/7058,6862,5078/)
        data_ped(7059,1:3)=(/7059,4184,6729/)
        data_ped(7060,1:3)=(/7060,6928,4043/)
        data_ped(7061,1:3)=(/7061,6866,6911/)
        data_ped(7062,1:3)=(/7062,6886,4551/)
        data_ped(7063,1:3)=(/7063,6919,1310/)
        data_ped(7064,1:3)=(/7064,6842,5096/)
        data_ped(7065,1:3)=(/7065,6887,4470/)
        data_ped(7066,1:3)=(/7066,6998,4323/)
        data_ped(7067,1:3)=(/7067,6888,5100/)
        data_ped(7068,1:3)=(/7068,6788,5023/)
        data_ped(7069,1:3)=(/7069,6838,3448/)
        data_ped(7070,1:3)=(/7070,3468,6829/)
        data_ped(7071,1:3)=(/7071,6803,4819/)
        data_ped(7072,1:3)=(/7072,6825,6956/)
        data_ped(7073,1:3)=(/7073,6811,6890/)
        data_ped(7074,1:3)=(/7074,6996,4123/)
        data_ped(7075,1:3)=(/7075,6921,4740/)
        data_ped(7076,1:3)=(/7076,6899,6907/)
        data_ped(7077,1:3)=(/7077,6922,4369/)
        data_ped(7078,1:3)=(/7078,4170,6536/)
        data_ped(7079,1:3)=(/7079,6811,6799/)
        data_ped(7080,1:3)=(/7080,6601,5106/)
        data_ped(7081,1:3)=(/7081,6883,4188/)
        data_ped(7082,1:3)=(/7082,6824,4786/)
        data_ped(7083,1:3)=(/7083,6733,4670/)
        data_ped(7084,1:3)=(/7084,6948,4402/)
        data_ped(7085,1:3)=(/7085,4438,6942/)
        data_ped(7086,1:3)=(/7086,4814,6938/)
        data_ped(7087,1:3)=(/7087,5162,6910/)
        data_ped(7088,1:3)=(/7088,6927,6943/)
        data_ped(7089,1:3)=(/7089,6803,4677/)
        data_ped(7090,1:3)=(/7090,6836,4678/)
        data_ped(7091,1:3)=(/7091,3934,6913/)
        data_ped(7092,1:3)=(/7092,6812,6781/)
        data_ped(7093,1:3)=(/7093,6923,1311/)
        data_ped(7094,1:3)=(/7094,6920,4918/)
        data_ped(7095,1:3)=(/7095,6929,6985/)
        data_ped(7096,1:3)=(/7096,7028,6959/)
        data_ped(7097,1:3)=(/7097,4796,6771/)
        data_ped(7098,1:3)=(/7098,6884,0/)
        data_ped(7099,1:3)=(/7099,4797,6517/)
        data_ped(7100,1:3)=(/7100,6954,6936/)
        data_ped(7101,1:3)=(/7101,6811,5211/)
        data_ped(7102,1:3)=(/7102,6786,0/)
        data_ped(7103,1:3)=(/7103,6928,5344/)
        data_ped(7104,1:3)=(/7104,6953,6829/)
        data_ped(7105,1:3)=(/7105,6884,6902/)
        data_ped(7106,1:3)=(/7106,6887,6980/)
        data_ped(7107,1:3)=(/7107,6803,4638/)
        data_ped(7108,1:3)=(/7108,4698,6709/)
        data_ped(7109,1:3)=(/7109,5330,6831/)
        data_ped(7110,1:3)=(/7110,6839,4543/)
        data_ped(7111,1:3)=(/7111,6914,4899/)
        data_ped(7112,1:3)=(/7112,6950,6966/)
        data_ped(7113,1:3)=(/7113,6733,6963/)
        data_ped(7114,1:3)=(/7114,3973,6991/)
        data_ped(7115,1:3)=(/7115,5414,6897/)
        data_ped(7116,1:3)=(/7116,5161,6990/)
        data_ped(7117,1:3)=(/7117,5157,6962/)
        data_ped(7118,1:3)=(/7118,4965,6796/)
        data_ped(7119,1:3)=(/7119,6969,3871/)
        data_ped(7120,1:3)=(/7120,6838,6906/)
        data_ped(7121,1:3)=(/7121,6886,4225/)
        data_ped(7122,1:3)=(/7122,6998,7019/)
        data_ped(7123,1:3)=(/7123,7030,5212/)
        data_ped(7124,1:3)=(/7124,6970,4723/)
        data_ped(7125,1:3)=(/7125,6969,3522/)
        data_ped(7126,1:3)=(/7126,6953,4105/)
        data_ped(7127,1:3)=(/7127,4056,6766/)
        data_ped(7128,1:3)=(/7128,6953,6934/)
        data_ped(7129,1:3)=(/7129,6825,5358/)
        data_ped(7130,1:3)=(/7130,6751,6833/)
        data_ped(7131,1:3)=(/7131,6823,5189/)
        data_ped(7132,1:3)=(/7132,7000,5202/)
        data_ped(7133,1:3)=(/7133,6969,5344/)
        data_ped(7134,1:3)=(/7134,6953,3705/)
        data_ped(7135,1:3)=(/7135,6811,6870/)
        data_ped(7136,1:3)=(/7136,6950,5497/)
        data_ped(7137,1:3)=(/7137,6950,5314/)
        data_ped(7138,1:3)=(/7138,6950,4953/)
        data_ped(7139,1:3)=(/7139,5155,6939/)
        data_ped(7140,1:3)=(/7140,4815,6912/)
        data_ped(7141,1:3)=(/7141,6814,5033/)
        data_ped(7142,1:3)=(/7142,4428,6879/)
        data_ped(7143,1:3)=(/7143,6924,6881/)
        data_ped(7144,1:3)=(/7144,3864,6991/)
        data_ped(7145,1:3)=(/7145,6790,4136/)
        data_ped(7146,1:3)=(/7146,6900,6940/)
        data_ped(7147,1:3)=(/7147,6806,1288/)
        data_ped(7148,1:3)=(/7148,5401,7005/)
        data_ped(7149,1:3)=(/7149,6917,7009/)
        data_ped(7150,1:3)=(/7150,6994,7011/)
        data_ped(7151,1:3)=(/7151,4614,7012/)
        data_ped(7152,1:3)=(/7152,6777,6960/)
        data_ped(7153,1:3)=(/7153,6608,0/)
        data_ped(7154,1:3)=(/7154,6969,5504/)
        data_ped(7155,1:3)=(/7155,6918,6935/)
        data_ped(7156,1:3)=(/7156,7029,6727/)
        data_ped(7157,1:3)=(/7157,4422,6719/)
        data_ped(7158,1:3)=(/7158,6744,5170/)
        data_ped(7159,1:3)=(/7159,6744,6850/)
        data_ped(7160,1:3)=(/7160,6812,6933/)
        data_ped(7161,1:3)=(/7161,6823,5354/)
        data_ped(7162,1:3)=(/7162,7049,5580/)
        data_ped(7163,1:3)=(/7163,6996,6984/)
        data_ped(7164,1:3)=(/7164,6580,6505/)
        data_ped(7165,1:3)=(/7165,7073,0/)
        data_ped(7166,1:3)=(/7166,5250,7038/)
        data_ped(7167,1:3)=(/7167,7025,4370/)
        data_ped(7168,1:3)=(/7168,6886,5281/)
        data_ped(7169,1:3)=(/7169,6889,5214/)
        data_ped(7170,1:3)=(/7170,5270,7036/)
        data_ped(7171,1:3)=(/7171,6950,5497/)
        data_ped(7172,1:3)=(/7172,6950,5544/)
        data_ped(7173,1:3)=(/7173,6950,4944/)
        data_ped(7174,1:3)=(/7174,4180,6913/)
        data_ped(7175,1:3)=(/7175,3864,6964/)
        data_ped(7176,1:3)=(/7176,6918,5239/)
        data_ped(7177,1:3)=(/7177,5600,7063/)
        data_ped(7178,1:3)=(/7178,7026,7064/)
        data_ped(7179,1:3)=(/7179,7049,7018/)
        data_ped(7180,1:3)=(/7180,7077,5020/)
        data_ped(7181,1:3)=(/7181,6884,7067/)
        data_ped(7182,1:3)=(/7182,7000,5021/)
        data_ped(7183,1:3)=(/7183,6609,0/)
        data_ped(7184,1:3)=(/7184,5530,7039/)
        data_ped(7185,1:3)=(/7185,7048,6930/)
        data_ped(7186,1:3)=(/7186,6885,6931/)
        data_ped(7187,1:3)=(/7187,6751,7007/)
        data_ped(7188,1:3)=(/7188,6777,6958/)
        data_ped(7189,1:3)=(/7189,4865,7014/)
        data_ped(7190,1:3)=(/7190,7054,7015/)
        data_ped(7191,1:3)=(/7191,4961,7042/)
        data_ped(7192,1:3)=(/7192,4341,6848/)
        data_ped(7193,1:3)=(/7193,7106,7041/)
        data_ped(7194,1:3)=(/7194,6917,6981/)
        data_ped(7195,1:3)=(/7195,6974,5364/)
        data_ped(7196,1:3)=(/7196,7110,7066/)
        data_ped(7197,1:3)=(/7197,6611,6936/)
        data_ped(7198,1:3)=(/7198,6969,5344/)
        data_ped(7199,1:3)=(/7199,7024,4647/)
        data_ped(7200,1:3)=(/7200,6945,5183/)
        data_ped(7201,1:3)=(/7201,4886,7084/)
        data_ped(7202,1:3)=(/7202,6978,1325/)
        data_ped(7203,1:3)=(/7203,5556,7060/)
        data_ped(7204,1:3)=(/7204,5062,6961/)
        data_ped(7205,1:3)=(/7205,4180,6966/)
        data_ped(7206,1:3)=(/7206,4180,6896/)
        data_ped(7207,1:3)=(/7207,6950,7046/)
        data_ped(7208,1:3)=(/7208,6950,5592/)
        data_ped(7209,1:3)=(/7209,6950,5487/)
        data_ped(7210,1:3)=(/7210,6950,4778/)
        data_ped(7211,1:3)=(/7211,6950,5499/)
        data_ped(7212,1:3)=(/7212,5154,6989/)
        data_ped(7213,1:3)=(/7213,4992,6993/)
        data_ped(7214,1:3)=(/7214,7079,7065/)
        data_ped(7215,1:3)=(/7215,6946,7093/)
        data_ped(7216,1:3)=(/7216,4056,6766/)
        data_ped(7217,1:3)=(/7217,7049,5441/)
        data_ped(7218,1:3)=(/7218,6995,6853/)
        data_ped(7219,1:3)=(/7219,6826,5101/)
        data_ped(7220,1:3)=(/7220,6886,5520/)
        data_ped(7221,1:3)=(/7221,3864,7006/)
        data_ped(7222,1:3)=(/7222,7023,5562/)
        data_ped(7223,1:3)=(/7223,7049,5352/)
        data_ped(7224,1:3)=(/7224,6995,4459/)
        data_ped(7225,1:3)=(/7225,6812,6933/)
        data_ped(7226,1:3)=(/7226,6994,7011/)
        data_ped(7227,1:3)=(/7227,6886,5195/)
        data_ped(7228,1:3)=(/7228,6863,5579/)
        data_ped(7229,1:3)=(/7229,5538,6524/)
        data_ped(7230,1:3)=(/7230,6863,6854/)
        data_ped(7231,1:3)=(/7231,4870,7045/)
        data_ped(7232,1:3)=(/7232,7049,6955/)
        data_ped(7233,1:3)=(/7233,6612,0/)
        data_ped(7234,1:3)=(/7234,5066,7090/)
        data_ped(7235,1:3)=(/7235,6898,5639/)
        data_ped(7236,1:3)=(/7236,6866,6932/)
        data_ped(7237,1:3)=(/7237,6919,5079/)
        data_ped(7238,1:3)=(/7238,6951,4482/)
        data_ped(7239,1:3)=(/7239,6950,5314/)
        data_ped(7240,1:3)=(/7240,6950,5552/)
        data_ped(7241,1:3)=(/7241,6950,5239/)
        data_ped(7242,1:3)=(/7242,3973,6993/)
        data_ped(7243,1:3)=(/7243,7057,7114/)
        data_ped(7244,1:3)=(/7244,7057,5636/)
        data_ped(7245,1:3)=(/7245,7057,6991/)
        data_ped(7246,1:3)=(/7246,5063,6896/)
        data_ped(7247,1:3)=(/7247,5542,6992/)
        data_ped(7248,1:3)=(/7248,7034,5234/)
        data_ped(7249,1:3)=(/7249,6917,7071/)
        data_ped(7250,1:3)=(/7250,6885,7118/)
        data_ped(7251,1:3)=(/7251,6886,5577/)
        data_ped(7252,1:3)=(/7252,6973,6851/)
        data_ped(7253,1:3)=(/7253,7049,5428/)
        data_ped(7254,1:3)=(/7254,6812,6933/)
        data_ped(7255,1:3)=(/7255,6999,4562/)
        data_ped(7256,1:3)=(/7256,7135,5646/)
        data_ped(7257,1:3)=(/7257,6838,6986/)
        data_ped(7258,1:3)=(/7258,7074,0/)
        data_ped(7259,1:3)=(/7259,6915,5708/)
        data_ped(7260,1:3)=(/7260,7108,5658/)
        data_ped(7261,1:3)=(/7261,6917,7010/)
        data_ped(7262,1:3)=(/7262,6811,5292/)
        data_ped(7263,1:3)=(/7263,6997,7096/)
        data_ped(7264,1:3)=(/7264,6929,5462/)
        data_ped(7265,1:3)=(/7265,6884,5713/)
        data_ped(7266,1:3)=(/7266,7026,7064/)
        data_ped(7267,1:3)=(/7267,7081,5588/)
        data_ped(7268,1:3)=(/7268,7072,5353/)
        data_ped(7269,1:3)=(/7269,7050,7040/)
        data_ped(7270,1:3)=(/7270,6813,6957/)
        data_ped(7271,1:3)=(/7271,7105,6850/)
        data_ped(7272,1:3)=(/7272,7116,7083/)
        data_ped(7273,1:3)=(/7273,6951,4573/)
        data_ped(7274,1:3)=(/7274,6979,7103/)
        data_ped(7275,1:3)=(/7275,4529,6895/)
        data_ped(7276,1:3)=(/7276,7057,5699/)
        data_ped(7277,1:3)=(/7277,7057,5636/)
        data_ped(7278,1:3)=(/7278,7057,6991/)
        data_ped(7279,1:3)=(/7279,7057,5769/)
        data_ped(7280,1:3)=(/7280,4988,6944/)
        data_ped(7281,1:3)=(/7281,5063,6896/)
        data_ped(7282,1:3)=(/7282,7111,7117/)
        data_ped(7283,1:3)=(/7283,7004,5630/)
        data_ped(7284,1:3)=(/7284,6968,4783/)
        data_ped(7285,1:3)=(/7285,7135,5654/)
        data_ped(7286,1:3)=(/7286,5467,7124/)
        data_ped(7287,1:3)=(/7287,6886,5370/)
        data_ped(7288,1:3)=(/7288,4614,6908/)
        data_ped(7289,1:3)=(/7289,6921,5096/)
        data_ped(7290,1:3)=(/7290,7072,5353/)
        data_ped(7291,1:3)=(/7291,5138,7008/)
        data_ped(7292,1:3)=(/7292,6915,7095/)
        data_ped(7293,1:3)=(/7293,7073,6957/)
        data_ped(7294,1:3)=(/7294,6827,4119/)
        data_ped(7295,1:3)=(/7295,7049,4638/)
        data_ped(7296,1:3)=(/7296,7194,5640/)
        data_ped(7297,1:3)=(/7297,7163,5437/)
        data_ped(7298,1:3)=(/7298,6995,7098/)
        data_ped(7299,1:3)=(/7299,7057,5225/)
        data_ped(7300,1:3)=(/7300,5531,7062/)
        data_ped(7301,1:3)=(/7301,5628,7115/)
        data_ped(7302,1:3)=(/7302,4992,6988/)
        data_ped(7303,1:3)=(/7303,7057,7144/)
        data_ped(7304,1:3)=(/7304,6979,4861/)
        data_ped(7305,1:3)=(/7305,7061,4769/)
        data_ped(7306,1:3)=(/7306,7091,7021/)
        data_ped(7307,1:3)=(/7307,6950,5524/)
        data_ped(7308,1:3)=(/7308,6950,5777/)
        data_ped(7309,1:3)=(/7309,7087,6967/)
        data_ped(7310,1:3)=(/7310,7139,5750/)
        data_ped(7311,1:3)=(/7311,7049,5239/)
        data_ped(7312,1:3)=(/7312,7229,5099/)
        data_ped(7313,1:3)=(/7313,7196,7122/)
        data_ped(7314,1:3)=(/7314,6826,5371/)
        data_ped(7315,1:3)=(/7315,5539,6909/)
        data_ped(7316,1:3)=(/7316,6946,5299/)
        data_ped(7317,1:3)=(/7317,7165,7153/)
        data_ped(7318,1:3)=(/7318,5251,6516/)
        data_ped(7319,1:3)=(/7319,7129,4220/)
        data_ped(7320,1:3)=(/7320,7195,5788/)
        data_ped(7321,1:3)=(/7321,6975,7020/)
        data_ped(7322,1:3)=(/7322,6915,7151/)
        data_ped(7323,1:3)=(/7323,6812,6987/)
        data_ped(7324,1:3)=(/7324,6838,6903/)
        data_ped(7325,1:3)=(/7325,7133,5241/)
        data_ped(7326,1:3)=(/7326,7133,5819/)
        data_ped(7327,1:3)=(/7327,7133,5389/)
        data_ped(7328,1:3)=(/7328,7167,6989/)
        data_ped(7329,1:3)=(/7329,7167,7212/)
        data_ped(7330,1:3)=(/7330,7035,5799/)
        data_ped(7331,1:3)=(/7331,5545,7141/)
        data_ped(7332,1:3)=(/7332,7057,5231/)
        data_ped(7333,1:3)=(/7333,7057,5699/)
        data_ped(7334,1:3)=(/7334,7088,5696/)
        data_ped(7335,1:3)=(/7335,7091,7112/)
        data_ped(7336,1:3)=(/7336,7091,7138/)
        data_ped(7337,1:3)=(/7337,4977,7170/)
        data_ped(7338,1:3)=(/7338,6950,5845/)
        data_ped(7339,1:3)=(/7339,7002,5485/)
        data_ped(7340,1:3)=(/7340,6951,5309/)
        data_ped(7341,1:3)=(/7341,7003,5826/)
        data_ped(7342,1:3)=(/7342,7171,5491/)
        data_ped(7343,1:3)=(/7343,6886,5515/)
        data_ped(7344,1:3)=(/7344,7026,7017/)
        data_ped(7345,1:3)=(/7345,7161,7215/)
        data_ped(7346,1:3)=(/7346,6916,5087/)
        data_ped(7347,1:3)=(/7347,5537,7152/)
        data_ped(7348,1:3)=(/7348,6883,5207/)
        data_ped(7349,1:3)=(/7349,6608,5650/)
        data_ped(7350,1:3)=(/7350,7080,7183/)
        data_ped(7351,1:3)=(/7351,7176,7021/)
        data_ped(7352,1:3)=(/7352,6979,5729/)
        data_ped(7353,1:3)=(/7353,6846,7086/)
        data_ped(7354,1:3)=(/7354,5614,7240/)
        data_ped(7355,1:3)=(/7355,7057,4861/)
        data_ped(7356,1:3)=(/7356,7057,6913/)
        data_ped(7357,1:3)=(/7357,7134,7126/)
        data_ped(7358,1:3)=(/7358,7134,7242/)
        data_ped(7359,1:3)=(/7359,7134,7249/)
        data_ped(7360,1:3)=(/7360,4616,7137/)
        data_ped(7361,1:3)=(/7361,6950,5726/)
        data_ped(7362,1:3)=(/7362,6950,5887/)
        data_ped(7363,1:3)=(/7363,7142,5385/)
        data_ped(7364,1:3)=(/7364,7171,5753/)
        data_ped(7365,1:3)=(/7365,6889,5751/)
        data_ped(7366,1:3)=(/7366,7157,5322/)
        data_ped(7367,1:3)=(/7367,7157,5931/)
        data_ped(7368,1:3)=(/7368,6918,5239/)
        data_ped(7369,1:3)=(/7369,7128,7211/)
        data_ped(7370,1:3)=(/7370,7052,0/)
        data_ped(7371,1:3)=(/7371,5263,6548/)
        data_ped(7372,1:3)=(/7372,7193,7259/)
        data_ped(7373,1:3)=(/7373,7226,6982/)
        data_ped(7374,1:3)=(/7374,5984,7175/)
        data_ped(7375,1:3)=(/7375,5987,7216/)
        data_ped(7376,1:3)=(/7376,7228,7252/)
        data_ped(7377,1:3)=(/7377,7162,7217/)
        data_ped(7378,1:3)=(/7378,7076,5854/)
        data_ped(7379,1:3)=(/7379,7224,5300/)
        data_ped(7380,1:3)=(/7380,5947,7121/)
        data_ped(7381,1:3)=(/7381,7107,7149/)
        data_ped(7382,1:3)=(/7382,7159,7250/)
        data_ped(7383,1:3)=(/7383,7158,7092/)
        data_ped(7384,1:3)=(/7384,5872,7251/)
        data_ped(7385,1:3)=(/7385,5984,7221/)
        data_ped(7386,1:3)=(/7386,7051,5581/)
        data_ped(7387,1:3)=(/7387,6812,6933/)
        data_ped(7388,1:3)=(/7388,7207,5592/)
        data_ped(7389,1:3)=(/7389,6979,4330/)
        data_ped(7390,1:3)=(/7390,5341,7174/)
        data_ped(7391,1:3)=(/7391,7057,7186/)
        data_ped(7392,1:3)=(/7392,7057,5769/)
        data_ped(7393,1:3)=(/7393,7057,7114/)
        data_ped(7394,1:3)=(/7394,5554,7277/)
        data_ped(7395,1:3)=(/7395,7134,5970/)
        data_ped(7396,1:3)=(/7396,7134,5728/)
        data_ped(7397,1:3)=(/7397,7176,7208/)
        data_ped(7398,1:3)=(/7398,7176,7125/)
        data_ped(7399,1:3)=(/7399,7058,5996/)
        data_ped(7400,1:3)=(/7400,7058,7274/)
        data_ped(7401,1:3)=(/7401,7245,7213/)
        data_ped(7402,1:3)=(/7402,7245,6993/)
        data_ped(7403,1:3)=(/7403,7245,5774/)
        data_ped(7404,1:3)=(/7404,5340,7085/)
        data_ped(7405,1:3)=(/7405,7168,4858/)
        data_ped(7406,1:3)=(/7406,7191,5932/)
        data_ped(7407,1:3)=(/7407,7192,5126/)
        data_ped(7408,1:3)=(/7408,5809,6506/)
        data_ped(7409,1:3)=(/7409,7225,7254/)
        data_ped(7410,1:3)=(/7410,5336,7100/)
        data_ped(7411,1:3)=(/7411,7131,5927/)
        data_ped(7412,1:3)=(/7412,6886,5519/)
        data_ped(7413,1:3)=(/7413,7197,5302/)
        data_ped(7414,1:3)=(/7414,7056,7102/)
        data_ped(7415,1:3)=(/7415,5251,7123/)
        data_ped(7416,1:3)=(/7416,7228,7252/)
        data_ped(7417,1:3)=(/7417,7232,5722/)
        data_ped(7418,1:3)=(/7418,6886,5723/)
        data_ped(7419,1:3)=(/7419,7107,6981/)
        data_ped(7420,1:3)=(/7420,7051,4745/)
        data_ped(7421,1:3)=(/7421,7055,5365/)
        data_ped(7422,1:3)=(/7422,5806,7016/)
        data_ped(7423,1:3)=(/7423,7033,5042/)
        data_ped(7424,1:3)=(/7424,7245,7205/)
        data_ped(7425,1:3)=(/7425,7245,5889/)
        data_ped(7426,1:3)=(/7426,5614,6989/)
        data_ped(7427,1:3)=(/7427,5614,7212/)
        data_ped(7428,1:3)=(/7428,5956,7302/)
        data_ped(7429,1:3)=(/7429,7057,6057/)
        data_ped(7430,1:3)=(/7430,7057,7136/)
        data_ped(7431,1:3)=(/7431,6950,7185/)
        data_ped(7432,1:3)=(/7432,6950,5885/)
        data_ped(7433,1:3)=(/7433,6950,6012/)
        data_ped(7434,1:3)=(/7434,7082,5955/)
        data_ped(7435,1:3)=(/7435,7168,5118/)
        data_ped(7436,1:3)=(/7436,7168,4858/)
        data_ped(7437,1:3)=(/7437,4611,7218/)
        data_ped(7438,1:3)=(/7438,7051,5434/)
        data_ped(7439,1:3)=(/7439,7075,7178/)
        data_ped(7440,1:3)=(/7440,6972,7288/)
        data_ped(7441,1:3)=(/7441,5939,7182/)
        data_ped(7442,1:3)=(/7442,7132,5795/)
        data_ped(7443,1:3)=(/7443,5741,7220/)
        data_ped(7444,1:3)=(/7444,7264,7047/)
        data_ped(7445,1:3)=(/7445,6823,6934/)
        data_ped(7446,1:3)=(/7446,7265,5849/)
        data_ped(7447,1:3)=(/7447,7130,7148/)
        data_ped(7448,1:3)=(/7448,7135,6071/)
        data_ped(7449,1:3)=(/7449,7135,7155/)
        data_ped(7450,1:3)=(/7450,7279,5833/)
        data_ped(7451,1:3)=(/7451,7057,7136/)
        data_ped(7452,1:3)=(/7452,7134,6009/)
        data_ped(7453,1:3)=(/7453,7269,7104/)
        data_ped(7454,1:3)=(/7454,7269,6011/)
        data_ped(7455,1:3)=(/7455,7269,7187/)
        data_ped(7456,1:3)=(/7456,7133,7209/)
        data_ped(7457,1:3)=(/7457,7299,6014/)
        data_ped(7458,1:3)=(/7458,7234,7284/)
        data_ped(7459,1:3)=(/7459,5613,7170/)
        data_ped(7460,1:3)=(/7460,6950,7304/)
        data_ped(7461,1:3)=(/7461,6950,7336/)
        data_ped(7462,1:3)=(/7462,7168,4858/)
        data_ped(7463,1:3)=(/7463,5621,7341/)
        data_ped(7464,1:3)=(/7464,7192,7247/)
        data_ped(7465,1:3)=(/7465,7296,7179/)
        data_ped(7466,1:3)=(/7466,6607,5449/)
        data_ped(7467,1:3)=(/7467,7349,7069/)
        data_ped(7468,1:3)=(/7468,7160,7120/)
        data_ped(7469,1:3)=(/7469,5984,7222/)
        data_ped(7470,1:3)=(/7470,4969,7314/)
        data_ped(7471,1:3)=(/7471,5991,7258/)
        data_ped(7472,1:3)=(/7472,7023,5936/)
        data_ped(7473,1:3)=(/7473,7164,6066/)
        data_ped(7474,1:3)=(/7474,7273,5695/)
        data_ped(7475,1:3)=(/7475,4991,7325/)
        data_ped(7476,1:3)=(/7476,7057,5654/)
        data_ped(7477,1:3)=(/7477,7134,7036/)
        data_ped(7478,1:3)=(/7478,7324,5798/)
        data_ped(7479,1:3)=(/7479,7236,7145/)
        data_ped(7480,1:3)=(/7480,7091,5933/)
        data_ped(7481,1:3)=(/7481,7166,5954/)
        data_ped(7482,1:3)=(/7482,7032,6111/)
        data_ped(7483,1:3)=(/7483,7032,5931/)
        data_ped(7484,1:3)=(/7484,5163,7246/)
        data_ped(7485,1:3)=(/7485,6098,7358/)
        data_ped(7486,1:3)=(/7486,7001,7204/)
        data_ped(7487,1:3)=(/7487,7089,5216/)
        data_ped(7488,1:3)=(/7488,7107,7022/)
        data_ped(7489,1:3)=(/7489,7293,7345/)
        data_ped(7490,1:3)=(/7490,5808,7219/)
        data_ped(7491,1:3)=(/7491,7233,6107/)
        data_ped(7492,1:3)=(/7492,7294,7188/)
        data_ped(7493,1:3)=(/7493,7214,6065/)
        data_ped(7494,1:3)=(/7494,7129,5513/)
        data_ped(7495,1:3)=(/7495,7023,5982/)
        data_ped(7496,1:3)=(/7496,6147,7316/)
        data_ped(7497,1:3)=(/7497,7298,7344/)
        data_ped(7498,1:3)=(/7498,7023,5936/)
        data_ped(7499,1:3)=(/7499,7023,5561/)
        data_ped(7500,1:3)=(/7500,5668,7190/)
        data_ped(7501,1:3)=(/7501,4979,7290/)
        data_ped(7502,1:3)=(/7502,7057,5654/)
        data_ped(7503,1:3)=(/7503,6952,5897/)
        data_ped(7504,1:3)=(/7504,7269,7126/)
        data_ped(7505,1:3)=(/7505,5907,7302/)
        data_ped(7506,1:3)=(/7506,7279,6042/)
        data_ped(7507,1:3)=(/7507,7311,5524/)
        data_ped(7508,1:3)=(/7508,7357,7205/)
        data_ped(7509,1:3)=(/7509,5154,7328/)
        data_ped(7510,1:3)=(/7510,7033,7125/)
        data_ped(7511,1:3)=(/7511,6979,5522/)
        data_ped(7512,1:3)=(/7512,7133,7336/)
        data_ped(7513,1:3)=(/7513,6953,7291/)
        data_ped(7514,1:3)=(/7514,7300,5458/)
        data_ped(7515,1:3)=(/7515,7300,5526/)
        data_ped(7516,1:3)=(/7516,7300,6131/)
        data_ped(7517,1:3)=(/7517,6950,7373/)
        data_ped(7518,1:3)=(/7518,7299,5968/)
        data_ped(7519,1:3)=(/7519,7299,5861/)
        data_ped(7520,1:3)=(/7520,7299,7352/)
        data_ped(7521,1:3)=(/7521,7113,5551/)
        data_ped(7522,1:3)=(/7522,7331,5382/)
        data_ped(7523,1:3)=(/7523,7272,1330/)
        data_ped(7524,1:3)=(/7524,7381,7241/)
        data_ped(7525,1:3)=(/7525,7379,7257/)
        data_ped(7526,1:3)=(/7526,5264,6937/)
        data_ped(7527,1:3)=(/7527,7295,7189/)
        data_ped(7528,1:3)=(/7528,7266,7013/)
        data_ped(7529,1:3)=(/7529,7110,7313/)
        data_ped(7530,1:3)=(/7530,5537,7439/)
        data_ped(7531,1:3)=(/7531,6077,7043/)
        data_ped(7532,1:3)=(/7532,7378,7255/)
        data_ped(7533,1:3)=(/7533,6928,5823/)
        data_ped(7534,1:3)=(/7534,6950,7373/)
        data_ped(7535,1:3)=(/7535,7300,5526/)
        data_ped(7536,1:3)=(/7536,7089,7369/)
        data_ped(7537,1:3)=(/7537,7176,7401/)
        data_ped(7538,1:3)=(/7538,7286,5934/)
        data_ped(7539,1:3)=(/7539,7286,6125/)
        data_ped(7540,1:3)=(/7540,7356,6123/)
        data_ped(7541,1:3)=(/7541,7366,7260/)
        data_ped(7542,1:3)=(/7542,7033,7185/)
        data_ped(7543,1:3)=(/7543,7133,7336/)
        data_ped(7544,1:3)=(/7544,7133,7389/)
        data_ped(7545,1:3)=(/7545,5279,7307/)
        data_ped(7546,1:3)=(/7546,6121,7424/)
        data_ped(7547,1:3)=(/7547,6977,5617/)
        data_ped(7548,1:3)=(/7548,7382,7345/)
        data_ped(7549,1:3)=(/7549,7322,7292/)
        data_ped(7550,1:3)=(/7550,5665,7375/)
        data_ped(7551,1:3)=(/7551,7380,5978/)
        data_ped(7552,1:3)=(/7552,6844,7169/)
        data_ped(7553,1:3)=(/7553,7299,6057/)
        data_ped(7554,1:3)=(/7554,7436,1350/)
        data_ped(7555,1:3)=(/7555,7436,1341/)
        data_ped(7556,1:3)=(/7556,7436,1337/)
        data_ped(7557,1:3)=(/7557,7271,5563/)
        data_ped(7558,1:3)=(/7558,7381,7368/)
        data_ped(7559,1:3)=(/7559,6979,5820/)
        data_ped(7560,1:3)=(/7560,7300,6234/)
        data_ped(7561,1:3)=(/7561,7300,6126/)
        data_ped(7562,1:3)=(/7562,7300,6233/)
        data_ped(7563,1:3)=(/7563,7300,6165/)
        data_ped(7564,1:3)=(/7564,7300,5837/)
        data_ped(7565,1:3)=(/7565,7329,6228/)
        data_ped(7566,1:3)=(/7566,7462,6160/)
        data_ped(7567,1:3)=(/7567,7434,6163/)
        data_ped(7568,1:3)=(/7568,7385,5982/)
        data_ped(7569,1:3)=(/7569,6034,7412/)
        data_ped(7570,1:3)=(/7570,7443,6245/)
        data_ped(7571,1:3)=(/7571,7109,7044/)
        data_ped(7572,1:3)=(/7572,7377,7465/)
        data_ped(7573,1:3)=(/7573,7296,7256/)
        data_ped(7574,1:3)=(/7574,7471,7411/)
        data_ped(7575,1:3)=(/7575,6082,7413/)
        data_ped(7576,1:3)=(/7576,5665,7312/)
        data_ped(7577,1:3)=(/7577,7470,5922/)
        data_ped(7578,1:3)=(/7578,7299,5380/)
        data_ped(7579,1:3)=(/7579,7235,6158/)
        data_ped(7580,1:3)=(/7580,6120,7435/)
        data_ped(7581,1:3)=(/7581,7444,7112/)
        data_ped(7582,1:3)=(/7582,7300,6200/)
        data_ped(7583,1:3)=(/7583,7300,6165/)
        data_ped(7584,1:3)=(/7584,7271,7376/)
        data_ped(7585,1:3)=(/7585,7305,7334/)
        data_ped(7586,1:3)=(/7586,7462,7429/)
        data_ped(7587,1:3)=(/7587,7462,6267/)
        data_ped(7588,1:3)=(/7588,7420,6045/)
        data_ped(7589,1:3)=(/7589,7472,5980/)
        data_ped(7590,1:3)=(/7590,7348,5724/)
        data_ped(7591,1:3)=(/7591,7440,7437/)
        data_ped(7592,1:3)=(/7592,7441,6241/)
        data_ped(7593,1:3)=(/7593,7027,7181/)
        data_ped(7594,1:3)=(/7594,7350,6108/)
        data_ped(7595,1:3)=(/7595,7381,7397/)
        data_ped(7596,1:3)=(/7596,7300,7438/)
        data_ped(7597,1:3)=(/7597,7300,6201/)
        data_ped(7598,1:3)=(/7598,7446,5733/)
        data_ped(7599,1:3)=(/7599,7461,6220/)
        data_ped(7600,1:3)=(/7600,7461,7510/)
        data_ped(7601,1:3)=(/7601,7355,7433/)
        data_ped(7602,1:3)=(/7602,5847,6965/)
        data_ped(7603,1:3)=(/7603,7497,6033/)
        data_ped(7604,1:3)=(/7604,6080,7068/)
        data_ped(7605,1:3)=(/7605,7231,6242/)
        data_ped(7606,1:3)=(/7606,5609,7466/)
        data_ped(7607,1:3)=(/7607,7134,7188/)
        data_ped(7608,1:3)=(/7608,7446,7504/)
        data_ped(7609,1:3)=(/7609,7301,5767/)
        data_ped(7610,1:3)=(/7610,7329,7406/)
        data_ped(7611,1:3)=(/7611,6106,7515/)
        data_ped(7612,1:3)=(/7612,7455,7513/)
        data_ped(7613,1:3)=(/7613,7461,7336/)
        data_ped(7614,1:3)=(/7614,7200,7527/)
        data_ped(7615,1:3)=(/7615,7200,7372/)
        data_ped(7616,1:3)=(/7616,7499,7374/)
        data_ped(7617,1:3)=(/7617,6926,6061/)
        data_ped(7618,1:3)=(/7618,7287,6244/)
        data_ped(7619,1:3)=(/7619,7522,6303/)
        data_ped(7620,1:3)=(/7620,5880,7388/)
        data_ped(7621,1:3)=(/7621,7461,6261/)
        data_ped(7622,1:3)=(/7622,7472,6011/)
        data_ped(7623,1:3)=(/7623,7386,7517/)
        data_ped(7624,1:3)=(/7624,5543,7533/)
        data_ped(7625,1:3)=(/7625,7416,7557/)
        data_ped(7626,1:3)=(/7626,7416,7549/)
        data_ped(7627,1:3)=(/7627,7416,6096/)
        data_ped(7628,1:3)=(/7628,7419,7518/)
        data_ped(7629,1:3)=(/7629,7454,7564/)
        data_ped(7630,1:3)=(/7630,7420,6210/)
        data_ped(7631,1:3)=(/7631,7490,7529/)
        data_ped(7632,1:3)=(/7632,6329,7415/)
        data_ped(7633,1:3)=(/7633,7349,7491/)
        data_ped(7634,1:3)=(/7634,7347,7289/)
        data_ped(7635,1:3)=(/7635,7442,6144/)
        data_ped(7636,1:3)=(/7636,7447,5764/)
        data_ped(7637,1:3)=(/7637,6047,7505/)
        data_ped(7638,1:3)=(/7638,5993,7460/)
        data_ped(7639,1:3)=(/7639,7340,6223/)
        data_ped(7640,1:3)=(/7640,7584,6096/)
        data_ped(7641,1:3)=(/7641,7238,6133/)
        data_ped(7642,1:3)=(/7642,5952,7560/)
        data_ped(7643,1:3)=(/7643,5894,7563/)
        data_ped(7644,1:3)=(/7644,7574,7415/)
        data_ped(7645,1:3)=(/7645,6082,7526/)
        data_ped(7646,1:3)=(/7646,7323,7414/)
        data_ped(7647,1:3)=(/7647,6332,7525/)
        data_ped(7648,1:3)=(/7648,7281,1389/)
        data_ped(7649,1:3)=(/7649,7446,7591/)
        data_ped(7650,1:3)=(/7650,7422,7492/)
        data_ped(7651,1:3)=(/7651,7408,6287/)
        data_ped(7652,1:3)=(/7652,7419,6360/)
        data_ped(7653,1:3)=(/7653,7542,7543/)
        data_ped(7654,1:3)=(/7654,7484,7431/)
        data_ped(7655,1:3)=(/7655,6314,7580/)
        data_ped(7656,1:3)=(/7656,6084,7464/)
        data_ped(7657,1:3)=(/7657,5984,7541/)
        data_ped(7658,1:3)=(/7658,7618,6243/)
        data_ped(7659,1:3)=(/7659,6323,7483/)
        data_ped(7660,1:3)=(/7660,7310,7609/)
        data_ped(7661,1:3)=(/7661,6150,7571/)
        data_ped(7662,1:3)=(/7662,7338,6232/)
        data_ped(7663,1:3)=(/7663,7446,6291/)
        data_ped(7664,1:3)=(/7664,7612,7608/)
        data_ped(7665,1:3)=(/7665,7445,7402/)
        data_ped(7666,1:3)=(/7666,7576,7562/)
        data_ped(7667,1:3)=(/7667,7576,6370/)
        data_ped(7668,1:3)=(/7668,6037,7583/)
        data_ped(7669,1:3)=(/7669,7343,6368/)
        data_ped(7670,1:3)=(/7670,7565,5998/)
        data_ped(7671,1:3)=(/7671,7430,7280/)
        data_ped(7672,1:3)=(/7672,7498,1372/)
        data_ped(7673,1:3)=(/7673,6106,7535/)
        data_ped(7674,1:3)=(/7674,7198,7511/)
        data_ped(7675,1:3)=(/7675,7548,7321/)
        data_ped(7676,1:3)=(/7676,7530,7528/)
        data_ped(7677,1:3)=(/7677,6279,7285/)
        data_ped(7678,1:3)=(/7678,6162,7474/)
        data_ped(7679,1:3)=(/7679,7576,6375/)
        data_ped(7680,1:3)=(/7680,7381,7561/)
        data_ped(7681,1:3)=(/7681,6345,7621/)
        data_ped(7682,1:3)=(/7682,7416,7468/)
        data_ped(7683,1:3)=(/7683,7383,7603/)
        data_ped(7684,1:3)=(/7684,6365,7315/)
        data_ped(7685,1:3)=(/7685,7632,7496/)
        data_ped(7686,1:3)=(/7686,7267,7590/)
        data_ped(7687,1:3)=(/7687,7356,6273/)
        data_ped(7688,1:3)=(/7688,6279,7285/)
        data_ped(7689,1:3)=(/7689,7199,7154/)
        data_ped(7690,1:3)=(/7690,7419,6171/)
        data_ped(7691,1:3)=(/7691,7512,7185/)
        data_ped(7692,1:3)=(/7692,7268,6316/)
        data_ped(7693,1:3)=(/7693,7531,6028/)
        data_ped(7694,1:3)=(/7694,6001,7585/)
        data_ped(7695,1:3)=(/7695,7393,7362/)
        data_ped(7696,1:3)=(/7696,7612,7608/)
        data_ped(7697,1:3)=(/7697,7206,7319/)
        data_ped(7698,1:3)=(/7698,7631,7572/)
        data_ped(7699,1:3)=(/7699,7588,6393/)
        data_ped(7700,1:3)=(/7700,7644,7496/)
        data_ped(7701,1:3)=(/7701,7545,1359/)
        data_ped(7702,1:3)=(/7702,7430,7521/)
        data_ped(7703,1:3)=(/7703,7387,7627/)
        data_ped(7704,1:3)=(/7704,7436,5660/)
        data_ped(7705,1:3)=(/7705,7286,7475/)
        data_ped(7706,1:3)=(/7706,7419,7583/)
        data_ped(7707,1:3)=(/7707,6150,7303/)
        data_ped(7708,1:3)=(/7708,7329,7626/)
        data_ped(7709,1:3)=(/7709,6049,7602/)
        data_ped(7710,1:3)=(/7710,7385,7374/)
        data_ped(7711,1:3)=(/7711,5950,7407/)
        data_ped(7712,1:3)=(/7712,7279,1374/)
        data_ped(7713,1:3)=(/7713,7286,1377/)
        data_ped(7714,1:3)=(/7714,7606,6366/)
        data_ped(7715,1:3)=(/7715,7461,7653/)
        data_ped(7716,1:3)=(/7716,7461,7306/)
        data_ped(7717,1:3)=(/7717,7489,7389/)
        data_ped(7718,1:3)=(/7718,7382,7544/)
        data_ped(7719,1:3)=(/7719,6341,7601/)
        data_ped(7720,1:3)=(/7720,7534,6184/)
        data_ped(7721,1:3)=(/7721,7391,6197/)
        data_ped(7722,1:3)=(/7722,7286,7513/)
        data_ped(7723,1:3)=(/7723,7387,7625/)
        data_ped(7724,1:3)=(/7724,7594,6306/)
        data_ped(7725,1:3)=(/7725,7633,7645/)
        data_ped(7726,1:3)=(/7726,7499,6207/)
        data_ped(7727,1:3)=(/7727,7661,7451/)
        data_ped(7728,1:3)=(/7728,7286,7540/)
        data_ped(7729,1:3)=(/7729,7297,7555/)
        data_ped(7730,1:3)=(/7730,5830,7651/)
        data_ped(7731,1:3)=(/7731,7419,7243/)
        data_ped(7732,1:3)=(/7732,7237,6206/)
        data_ped(7733,1:3)=(/7733,6396,7642/)
        data_ped(7734,1:3)=(/7734,7532,6105/)
        data_ped(7735,1:3)=(/7735,7604,7410/)
        data_ped(7736,1:3)=(/7736,7634,7593/)
        data_ped(7737,1:3)=(/7737,7569,7570/)
        data_ped(7738,1:3)=(/7738,7338,6156/)
        data_ped(7739,1:3)=(/7739,7461,6261/)
        data_ped(7740,1:3)=(/7740,6427,7607/)
        data_ped(7741,1:3)=(/7741,7649,1394/)
        data_ped(7742,1:3)=(/7742,6361,7353/)
        data_ped(7743,1:3)=(/7743,7500,6046/)
        data_ped(7744,1:3)=(/7744,7516,7678/)
        data_ped(7745,1:3)=(/7745,7507,6425/)
        data_ped(7746,1:3)=(/7746,7507,6404/)
        data_ped(7747,1:3)=(/7747,7696,1368/)
        data_ped(7748,1:3)=(/7748,7428,6400/)
        data_ped(7749,1:3)=(/7749,6219,7659/)
        data_ped(7750,1:3)=(/7750,7417,6398/)
        data_ped(7751,1:3)=(/7751,7356,1395/)
        data_ped(7752,1:3)=(/7752,6219,7659/)
        data_ped(7753,1:3)=(/7753,6190,7611/)
        data_ped(7754,1:3)=(/7754,7457,6343/)
        data_ped(7755,1:3)=(/7755,7582,6284/)
        data_ped(7756,1:3)=(/7756,7405,7623/)
        data_ped(7757,1:3)=(/7757,7484,6358/)
        data_ped(7758,1:3)=(/7758,6334,7320/)
        data_ped(7759,1:3)=(/7759,7605,1412/)
        data_ped(7760,1:3)=(/7760,7237,1375/)
        data_ped(7761,1:3)=(/7761,7237,7546/)
        data_ped(7762,1:3)=(/7762,2942,6471/)
        data_ped(7763,1:3)=(/7763,6715,0/)
        data_ped(7764,1:3)=(/7764,3026,6470/)
        data_ped(7765,1:3)=(/7765,0,6478/)
        data_ped(7766,1:3)=(/7766,2054,6480/)
        data_ped(7767,1:3)=(/7767,6581,0/)
        data_ped(7768,1:3)=(/7768,3956,6686/)
        data_ped(7769,1:3)=(/7769,6582,1434/)
        data_ped(7770,1:3)=(/7770,6619,87/)
        data_ped(7771,1:3)=(/7771,6587,4394/)
        data_ped(7772,1:3)=(/7772,6586,0/)
        data_ped(7773,1:3)=(/7773,6579,114/)
        data_ped(7774,1:3)=(/7774,6591,1456/)
        data_ped(7775,1:3)=(/7775,6841,6511/)
        data_ped(7776,1:3)=(/7776,340,6513/)
        data_ped(7777,1:3)=(/7777,6811,4222/)
        data_ped(7778,1:3)=(/7778,366,6521/)
        data_ped(7779,1:3)=(/7779,364,7778/)
        data_ped(7780,1:3)=(/7780,384,6523/)
        data_ped(7781,1:3)=(/7781,352,6525/)
        data_ped(7782,1:3)=(/7782,358,6527/)
        data_ped(7783,1:3)=(/7783,355,6526/)
        data_ped(7784,1:3)=(/7784,364,7778/)
        data_ped(7785,1:3)=(/7785,1554,6530/)
        data_ped(7786,1:3)=(/7786,6593,0/)
        data_ped(7787,1:3)=(/7787,6599,1487/)
        data_ped(7788,1:3)=(/7788,1554,6538/)
        data_ped(7789,1:3)=(/7789,1554,6540/)
        data_ped(7790,1:3)=(/7790,6597,7789/)
        data_ped(7791,1:3)=(/7791,1557,6544/)
        data_ped(7792,1:3)=(/7792,6947,6545/)
        data_ped(7793,1:3)=(/7793,6602,7788/)
        data_ped(7794,1:3)=(/7794,3853,6546/)
        data_ped(7795,1:3)=(/7795,1560,6547/)
        data_ped(7796,1:3)=(/7796,6606,6549/)
        data_ped(7797,1:3)=(/7797,6599,1504/)
        data_ped(7798,1:3)=(/7798,6595,4126/)
        data_ped(7799,1:3)=(/7799,5263,6552/)
        data_ped(7800,1:3)=(/7800,2243,7770/)
        data_ped(7801,1:3)=(/7801,6584,6499/)
        data_ped(7802,1:3)=(/7802,2543,6671/)
        data_ped(7803,1:3)=(/7803,6588,6875/)
        data_ped(7804,1:3)=(/7804,6723,6736/)
        data_ped(7805,1:3)=(/7805,6589,7774/)
        data_ped(7806,1:3)=(/7806,6591,4010/)
        data_ped(7807,1:3)=(/7807,6813,6517/)
        data_ped(7808,1:3)=(/7808,4063,7777/)
        data_ped(7809,1:3)=(/7809,6594,6531/)
        data_ped(7810,1:3)=(/7810,6596,7786/)
        data_ped(7811,1:3)=(/7811,383,7780/)
        data_ped(7812,1:3)=(/7812,6599,7785/)
        data_ped(7813,1:3)=(/7813,7810,6543/)
        data_ped(7814,1:3)=(/7814,6603,7794/)
        data_ped(7815,1:3)=(/7815,7809,1497/)
        data_ped(7816,1:3)=(/7816,6605,7797/)
        data_ped(7817,1:3)=(/7817,1809,7769/)
        data_ped(7818,1:3)=(/7818,2122,7817/)
        data_ped(7819,1:3)=(/7819,2304,7763/)
        data_ped(7820,1:3)=(/7820,2857,7800/)
        data_ped(7821,1:3)=(/7821,7818,6652/)
        data_ped(7822,1:3)=(/7822,6578,7765/)
        data_ped(7823,1:3)=(/7823,7818,3245/)
        data_ped(7824,1:3)=(/7824,2538,7766/)
        data_ped(7825,1:3)=(/7825,2456,7820/)
        data_ped(7826,1:3)=(/7826,7821,2984/)
        data_ped(7827,1:3)=(/7827,2954,7765/)
        data_ped(7828,1:3)=(/7828,2955,7767/)
        data_ped(7829,1:3)=(/7829,3034,7823/)
        data_ped(7830,1:3)=(/7830,6585,7772/)
        data_ped(7831,1:3)=(/7831,3197,7827/)
        data_ped(7832,1:3)=(/7832,7826,3811/)
        data_ped(7833,1:3)=(/7833,3565,7824/)
        data_ped(7834,1:3)=(/7834,7818,6764/)
        data_ped(7835,1:3)=(/7835,4504,7829/)
        data_ped(7836,1:3)=(/7836,7825,6793/)
        data_ped(7837,1:3)=(/7837,7804,3893/)
        data_ped(7838,1:3)=(/7838,4414,7828/)
        data_ped(7839,1:3)=(/7839,6859,7762/)
        data_ped(7840,1:3)=(/7840,4607,7822/)
        data_ped(7841,1:3)=(/7841,7804,0/)
        data_ped(7842,1:3)=(/7842,7801,4011/)
        data_ped(7843,1:3)=(/7843,4506,7832/)
        data_ped(7844,1:3)=(/7844,7831,4382/)
        data_ped(7845,1:3)=(/7845,6803,7768/)
        data_ped(7846,1:3)=(/7846,7830,3166/)
        data_ped(7847,1:3)=(/7847,7806,1500/)
        data_ped(7848,1:3)=(/7848,6826,7841/)
        data_ped(7849,1:3)=(/7849,5404,7771/)
        data_ped(7850,1:3)=(/7850,7838,4498/)
        data_ped(7851,1:3)=(/7851,4060,7840/)
        data_ped(7852,1:3)=(/7852,6971,7833/)
        data_ped(7853,1:3)=(/7853,6949,7843/)
        data_ped(7854,1:3)=(/7854,7806,6514/)
        data_ped(7855,1:3)=(/7855,4063,7837/)
        data_ped(7856,1:3)=(/7856,7810,6543/)
        data_ped(7857,1:3)=(/7857,7836,5032/)
        data_ped(7858,1:3)=(/7858,7845,6891/)
        data_ped(7859,1:3)=(/7859,7807,7099/)
        data_ped(7860,1:3)=(/7860,6882,7773/)
        data_ped(7861,1:3)=(/7861,7838,4498/)
        data_ped(7862,1:3)=(/7862,7808,5297/)
        data_ped(7863,1:3)=(/7863,7802,6904/)
        data_ped(7864,1:3)=(/7864,4063,7837/)
        data_ped(7865,1:3)=(/7865,5057,7792/)
        data_ped(7866,1:3)=(/7866,5253,7849/)
        data_ped(7867,1:3)=(/7867,6598,7791/)
        data_ped(7868,1:3)=(/7868,7031,7841/)
        data_ped(7869,1:3)=(/7869,7848,6876/)
        data_ped(7870,1:3)=(/7870,7848,5586/)
        data_ped(7871,1:3)=(/7871,7834,4481/)
        data_ped(7872,1:3)=(/7872,7851,7844/)
        data_ped(7873,1:3)=(/7873,7814,7793/)
        data_ped(7874,1:3)=(/7874,7078,7789/)
        data_ped(7875,1:3)=(/7875,7806,7775/)
        data_ped(7876,1:3)=(/7876,7855,5856/)
        data_ped(7877,1:3)=(/7877,7194,7858/)
        data_ped(7878,1:3)=(/7878,7866,5178/)
        data_ped(7879,1:3)=(/7879,7864,7150/)
        data_ped(7880,1:3)=(/7880,7864,7120/)
        data_ped(7881,1:3)=(/7881,5257,7792/)
        data_ped(7882,1:3)=(/7882,5747,7861/)
        data_ped(7883,1:3)=(/7883,7866,5786/)
        data_ped(7884,1:3)=(/7884,7224,7863/)
        data_ped(7885,1:3)=(/7885,7846,6983/)
        data_ped(7886,1:3)=(/7886,7805,7156/)
        data_ped(7887,1:3)=(/7887,7803,7842/)
        data_ped(7888,1:3)=(/7888,7857,7871/)
        data_ped(7889,1:3)=(/7889,7864,7868/)
        data_ped(7890,1:3)=(/7890,7805,7156/)
        data_ped(7891,1:3)=(/7891,7866,4905/)
        data_ped(7892,1:3)=(/7892,7813,7790/)
        data_ped(7893,1:3)=(/7893,7245,7850/)
        data_ped(7894,1:3)=(/7894,6950,7861/)
        data_ped(7895,1:3)=(/7895,6591,7854/)
        data_ped(7896,1:3)=(/7896,7870,7119/)
        data_ped(7897,1:3)=(/7897,6606,7799/)
        data_ped(7898,1:3)=(/7898,7853,7282/)
        data_ped(7899,1:3)=(/7899,6610,7798/)
        data_ped(7900,1:3)=(/7900,7864,7370/)
        data_ped(7901,1:3)=(/7901,7881,6067/)
        data_ped(7902,1:3)=(/7902,5572,7860/)
        data_ped(7903,1:3)=(/7903,5635,7861/)
        data_ped(7904,1:3)=(/7904,6006,7882/)
        data_ped(7905,1:3)=(/7905,7852,1348/)
        data_ped(7906,1:3)=(/7906,7230,7889/)
        data_ped(7907,1:3)=(/7907,7847,7317/)
        data_ped(7908,1:3)=(/7908,7880,7409/)
        data_ped(7909,1:3)=(/7909,7879,7373/)
        data_ped(7910,1:3)=(/7910,7869,7262/)
        data_ped(7911,1:3)=(/7911,7230,7889/)
        data_ped(7912,1:3)=(/7912,6976,7859/)
        data_ped(7913,1:3)=(/7913,6005,7883/)
        data_ped(7914,1:3)=(/7914,7884,7468/)
        data_ped(7915,1:3)=(/7915,7885,7477/)
        data_ped(7916,1:3)=(/7916,6239,7894/)
        data_ped(7917,1:3)=(/7917,7459,7886/)
        data_ped(7918,1:3)=(/7918,7874,7873/)
        data_ped(7919,1:3)=(/7919,5609,7897/)
        data_ped(7920,1:3)=(/7920,7885,7318/)
        data_ped(7921,1:3)=(/7921,7497,7862/)
        data_ped(7922,1:3)=(/7922,7869,7101/)
        data_ped(7923,1:3)=(/7923,7427,7878/)
        data_ped(7924,1:3)=(/7924,7885,7337/)
        data_ped(7925,1:3)=(/7925,7896,7321/)
        data_ped(7926,1:3)=(/7926,7911,7900/)
        data_ped(7927,1:3)=(/7927,7856,7899/)
        data_ped(7928,1:3)=(/7928,7887,7365/)
        data_ped(7929,1:3)=(/7929,7875,7890/)
        data_ped(7930,1:3)=(/7930,5993,7898/)
        data_ped(7931,1:3)=(/7931,7875,7920/)
        data_ped(7932,1:3)=(/7932,7875,7915/)
        data_ped(7933,1:3)=(/7933,7885,7318/)
        data_ped(7934,1:3)=(/7934,7885,7890/)
        data_ped(7935,1:3)=(/7935,7910,7909/)
        data_ped(7936,1:3)=(/7936,7542,7925/)
        data_ped(7937,1:3)=(/7937,7922,7551/)
        data_ped(7938,1:3)=(/7938,6230,7888/)
        data_ped(7939,1:3)=(/7939,7166,7916/)
        data_ped(7940,1:3)=(/7940,7908,7519/)
        data_ped(7941,1:3)=(/7941,7885,7929/)
        data_ped(7942,1:3)=(/7942,7902,6401/)
        data_ped(7943,1:3)=(/7943,7881,7467/)
        data_ped(7944,1:3)=(/7944,7907,6307/)
        data_ped(7945,1:3)=(/7945,6219,7904/)
        data_ped(7946,1:3)=(/7946,7912,6156/)
        data_ped(7947,1:3)=(/7947,7875,7933/)
        data_ped(7948,1:3)=(/7948,7885,7895/)
        data_ped(7949,1:3)=(/7949,7908,7508/)
        data_ped(7950,1:3)=(/7950,7931,7924/)
        data_ped(7951,1:3)=(/7951,7934,7917/)
        data_ped(7952,1:3)=(/7952,7921,7682/)
        data_ped(7953,1:3)=(/7953,7926,7935/)
        data_ped(7954,1:3)=(/7954,7819,7764/)
        data_ped(7955,1:3)=(/7955,5259,7776/)
        data_ped(7956,1:3)=(/7956,7806,6515/)
        data_ped(7957,1:3)=(/7957,7811,1472/)
        data_ped(7958,1:3)=(/7958,390,7781/)
        data_ped(7959,1:3)=(/7959,375,7782/)
        data_ped(7960,1:3)=(/7960,351,7957/)
        data_ped(7961,1:3)=(/7961,353,7783/)
        data_ped(7962,1:3)=(/7962,7810,6534/)
        data_ped(7963,1:3)=(/7963,6597,7787/)
        data_ped(7964,1:3)=(/7964,7810,6539/)
        data_ped(7965,1:3)=(/7965,7806,6542/)
        data_ped(7966,1:3)=(/7966,7867,7795/)
        data_ped(7967,1:3)=(/7967,7816,7796/)
        data_ped(7968,1:3)=(/7968,5540,7964/)
        data_ped(7969,1:3)=(/7969,362,7961/)
        data_ped(7970,1:3)=(/7970,362,7961/)
        data_ped(7971,1:3)=(/7971,7970,7779/)
        data_ped(7972,1:3)=(/7972,387,7961/)
        data_ped(7973,1:3)=(/7973,7971,1467/)
        data_ped(7974,1:3)=(/7974,451,7973/)
        data_ped(7975,1:3)=(/7975,391,7959/)
        data_ped(7976,1:3)=(/7976,6732,7954/)
        data_ped(7977,1:3)=(/7977,6867,7976/)
        data_ped(7978,1:3)=(/7978,7053,7956/)
        data_ped(7979,1:3)=(/7979,7839,7977/)
        data_ped(7980,1:3)=(/7980,6592,7965/)
        data_ped(7981,1:3)=(/7981,5668,7955/)
        data_ped(7982,1:3)=(/7982,7812,7963/)
        data_ped(7983,1:3)=(/7983,6925,7979/)
        data_ped(7984,1:3)=(/7984,7815,7962/)
        data_ped(7985,1:3)=(/7985,6148,7968/)
        data_ped(7986,1:3)=(/7986,7892,7966/)
        data_ped(7987,1:3)=(/7987,5609,7984/)
        data_ped(7988,1:3)=(/7988,7550,7981/)
        data_ped(7989,1:3)=(/7989,7885,7978/)
        data_ped(7990,1:3)=(/7990,7885,7980/)
        data_ped(7991,1:3)=(/7991,7865,7985/)
        data_ped(7992,1:3)=(/7992,7575,7987/)
        data_ped(7993,1:3)=(/7993,7988,7372/)
        data_ped(7994,1:3)=(/7994,7875,7990/)
        data_ped(7995,1:3)=(/7995,6195,7983/)
        data_ped(7996,1:3)=(/7996,7986,7932/)
        data_ped(7997,1:3)=(/7997,7986,6271/)
        data_ped(7998,1:3)=(/7998,7986,7942/)
        data_ped(7999,1:3)=(/7999,7989,7928/)
        data_ped(8000,1:3)=(/8000,382,7957/)
        data_ped(8001,1:3)=(/8001,389,7958/)
        data_ped(8002,1:3)=(/8002,7972,1474/)
        data_ped(8003,1:3)=(/8003,350,7961/)
        data_ped(8004,1:3)=(/8004,7970,7784/)
        data_ped(8005,1:3)=(/8005,7969,1474/)
        data_ped(8006,1:3)=(/8006,7971,8002/)
        data_ped(8007,1:3)=(/8007,375,8003/)
        data_ped(8008,1:3)=(/8008,8007,1475/)
        data_ped(8009,1:3)=(/8009,8006,6522/)
        data_ped(8010,1:3)=(/8010,8006,6519/)
        data_ped(8011,1:3)=(/8011,8006,8008/)
        data_ped(8012,1:3)=(/8012,370,8009/)
        data_ped(8013,1:3)=(/8013,7975,8011/)
        data_ped(8014,1:3)=(/8014,451,8012/)
        data_ped(8015,1:3)=(/8015,8014,0/)
        data_ped(8016,1:3)=(/8016,8014,0/)
        data_ped(8017,1:3)=(/8017,8014,1759/)
        data_ped(8018,1:3)=(/8018,8016,736/)
        data_ped(8019,1:3)=(/8019,8015,0/)
        data_ped(8020,1:3)=(/8020,8019,0/)
        data_ped(8021,1:3)=(/8021,0,8018/)
        data_ped(8022,1:3)=(/8022,8016,2414/)
        data_ped(8023,1:3)=(/8023,8019,0/)
        data_ped(8024,1:3)=(/8024,2530,8017/)
        data_ped(8025,1:3)=(/8025,8022,2288/)
        data_ped(8026,1:3)=(/8026,8021,2398/)
        data_ped(8027,1:3)=(/8027,8022,2809/)
        data_ped(8028,1:3)=(/8028,8019,2292/)
        data_ped(8029,1:3)=(/8029,8019,1185/)
        data_ped(8030,1:3)=(/8030,8022,2925/)
        data_ped(8031,1:3)=(/8031,2845,8023/)
        data_ped(8032,1:3)=(/8032,8019,2592/)
        data_ped(8033,1:3)=(/8033,2790,8020/)
        data_ped(8034,1:3)=(/8034,8022,2652/)
        data_ped(8035,1:3)=(/8035,8028,2766/)
        data_ped(8036,1:3)=(/8036,2444,8024/)
        data_ped(8037,1:3)=(/8037,3482,8033/)
        data_ped(8038,1:3)=(/8038,8029,3088/)
        data_ped(8039,1:3)=(/8039,8029,3620/)
        data_ped(8040,1:3)=(/8040,8029,3444/)
        data_ped(8041,1:3)=(/8041,8029,6649/)
        data_ped(8042,1:3)=(/8042,8029,3785/)
        data_ped(8043,1:3)=(/8043,8029,3356/)
        data_ped(8044,1:3)=(/8044,8029,3803/)
        data_ped(8045,1:3)=(/8045,8029,6757/)
        data_ped(8046,1:3)=(/8046,3097,8034/)
        data_ped(8047,1:3)=(/8047,4056,8038/)
        data_ped(8048,1:3)=(/8048,4177,8026/)
        data_ped(8049,1:3)=(/8049,3097,8034/)
        data_ped(8050,1:3)=(/8050,3482,8039/)
        data_ped(8051,1:3)=(/8051,8035,6649/)
        data_ped(8052,1:3)=(/8052,4159,8041/)
        data_ped(8053,1:3)=(/8053,3954,8040/)
        data_ped(8054,1:3)=(/8054,3954,8037/)
        data_ped(8055,1:3)=(/8055,8028,3904/)
        data_ped(8056,1:3)=(/8056,8036,3069/)
        data_ped(8057,1:3)=(/8057,3482,8042/)
        data_ped(8058,1:3)=(/8058,8028,4090/)
        data_ped(8059,1:3)=(/8059,4159,8044/)
        data_ped(8060,1:3)=(/8060,4056,8037/)
        data_ped(8061,1:3)=(/8061,8028,6763/)
        data_ped(8062,1:3)=(/8062,4159,8037/)
        data_ped(8063,1:3)=(/8063,8035,3978/)
        data_ped(8064,1:3)=(/8064,4419,8047/)
        data_ped(8065,1:3)=(/8065,4056,8045/)
        data_ped(8066,1:3)=(/8066,8028,3524/)
        data_ped(8067,1:3)=(/8067,4056,8038/)
        data_ped(8068,1:3)=(/8068,8049,3827/)
        data_ped(8069,1:3)=(/8069,3475,8051/)
        data_ped(8070,1:3)=(/8070,8060,4748/)
        data_ped(8071,1:3)=(/8071,8065,5080/)
        data_ped(8072,1:3)=(/8072,4787,8058/)
        data_ped(8073,1:3)=(/8073,4626,8025/)
        data_ped(8074,1:3)=(/8074,4787,8057/)
        data_ped(8075,1:3)=(/8075,8032,8037/)
        data_ped(8076,1:3)=(/8076,4056,8059/)
        data_ped(8077,1:3)=(/8077,4626,8025/)
        data_ped(8078,1:3)=(/8078,8032,8033/)
        data_ped(8079,1:3)=(/8079,3947,8031/)
        data_ped(8080,1:3)=(/8080,8060,6901/)
        data_ped(8081,1:3)=(/8081,4683,8047/)
        data_ped(8082,1:3)=(/8082,4056,8054/)
        data_ped(8083,1:3)=(/8083,8064,4100/)
        data_ped(8084,1:3)=(/8084,4800,8050/)
        data_ped(8085,1:3)=(/8085,6778,8062/)
        data_ped(8086,1:3)=(/8086,4056,8062/)
        data_ped(8087,1:3)=(/8087,4064,8040/)
        data_ped(8088,1:3)=(/8088,8060,6901/)
        data_ped(8089,1:3)=(/8089,3680,8048/)
        data_ped(8090,1:3)=(/8090,8061,1295/)
        data_ped(8091,1:3)=(/8091,8065,4923/)
        data_ped(8092,1:3)=(/8092,4056,8075/)
        data_ped(8093,1:3)=(/8093,4056,8059/)
        data_ped(8094,1:3)=(/8094,6977,8072/)
        data_ped(8095,1:3)=(/8095,4064,8044/)
        data_ped(8096,1:3)=(/8096,8043,5176/)
        data_ped(8097,1:3)=(/8097,8060,5188/)
        data_ped(8098,1:3)=(/8098,8065,4911/)
        data_ped(8099,1:3)=(/8099,8064,5301/)
        data_ped(8100,1:3)=(/8100,6950,8077/)
        data_ped(8101,1:3)=(/8101,6977,8072/)
        data_ped(8102,1:3)=(/8102,8065,5428/)
        data_ped(8103,1:3)=(/8103,8064,4658/)
        data_ped(8104,1:3)=(/8104,5255,8069/)
        data_ped(8105,1:3)=(/8105,8078,8076/)
        data_ped(8106,1:3)=(/8106,8064,4128/)
        data_ped(8107,1:3)=(/8107,6977,8072/)
        data_ped(8108,1:3)=(/8108,5532,8052/)
        data_ped(8109,1:3)=(/8109,8084,5286/)
        data_ped(8110,1:3)=(/8110,8065,5283/)
        data_ped(8111,1:3)=(/8111,6950,8073/)
        data_ped(8112,1:3)=(/8112,7057,8074/)
        data_ped(8113,1:3)=(/8113,8061,1326/)
        data_ped(8114,1:3)=(/8114,8061,6881/)
        data_ped(8115,1:3)=(/8115,8060,7070/)
        data_ped(8116,1:3)=(/8116,8060,7094/)
        data_ped(8117,1:3)=(/8117,8065,5283/)
        data_ped(8118,1:3)=(/8118,8083,5393/)
        data_ped(8119,1:3)=(/8119,6950,8073/)
        data_ped(8120,1:3)=(/8120,4983,8090/)
        data_ped(8121,1:3)=(/8121,4875,8070/)
        data_ped(8122,1:3)=(/8122,8065,5587/)
        data_ped(8123,1:3)=(/8123,5053,8082/)
        data_ped(8124,1:3)=(/8124,8084,5435/)
        data_ped(8125,1:3)=(/8125,8099,5201/)
        data_ped(8126,1:3)=(/8126,8060,5188/)
        data_ped(8127,1:3)=(/8127,8060,5009/)
        data_ped(8128,1:3)=(/8128,8078,5511/)
        data_ped(8129,1:3)=(/8129,8068,5224/)
        data_ped(8130,1:3)=(/8130,8064,5836/)
        data_ped(8131,1:3)=(/8131,8060,5797/)
        data_ped(8132,1:3)=(/8132,8092,7177/)
        data_ped(8133,1:3)=(/8133,8065,5283/)
        data_ped(8134,1:3)=(/8134,8093,7097/)
        data_ped(8135,1:3)=(/8135,5531,8047/)
        data_ped(8136,1:3)=(/8136,8081,5778/)
        data_ped(8137,1:3)=(/8137,8109,4734/)
        data_ped(8138,1:3)=(/8138,8060,5865/)
        data_ped(8139,1:3)=(/8139,8061,7143/)
        data_ped(8140,1:3)=(/8140,8061,1299/)
        data_ped(8141,1:3)=(/8141,8061,4766/)
        data_ped(8142,1:3)=(/8142,8061,5233/)
        data_ped(8143,1:3)=(/8143,8088,5764/)
        data_ped(8144,1:3)=(/8144,8088,5842/)
        data_ped(8145,1:3)=(/8145,8088,5629/)
        data_ped(8146,1:3)=(/8146,8088,7136/)
        data_ped(8147,1:3)=(/8147,8060,5865/)
        data_ped(8148,1:3)=(/8148,5400,8080/)
        data_ped(8149,1:3)=(/8149,8105,8108/)
        data_ped(8150,1:3)=(/8150,8060,5783/)
        data_ped(8151,1:3)=(/8151,8079,8086/)
        data_ped(8152,1:3)=(/8152,8106,5789/)
        data_ped(8153,1:3)=(/8153,8066,8053/)
        data_ped(8154,1:3)=(/8154,8104,5648/)
        data_ped(8155,1:3)=(/8155,8115,7144/)
        data_ped(8156,1:3)=(/8156,8115,5831/)
        data_ped(8157,1:3)=(/8157,8115,7244/)
        data_ped(8158,1:3)=(/8158,6977,8112/)
        data_ped(8159,1:3)=(/8159,8061,4942/)
        data_ped(8160,1:3)=(/8160,8060,8063/)
        data_ped(8161,1:3)=(/8161,8060,5866/)
        data_ped(8162,1:3)=(/8162,8078,5582/)
        data_ped(8163,1:3)=(/8163,8109,5718/)
        data_ped(8164,1:3)=(/8164,8109,5447/)
        data_ped(8165,1:3)=(/8165,8117,5804/)
        data_ped(8166,1:3)=(/8166,8117,7261/)
        data_ped(8167,1:3)=(/8167,8110,7253/)
        data_ped(8168,1:3)=(/8168,8085,5857/)
        data_ped(8169,1:3)=(/8169,5805,8071/)
        data_ped(8170,1:3)=(/8170,8065,5283/)
        data_ped(8171,1:3)=(/8171,8117,5781/)
        data_ped(8172,1:3)=(/8172,8117,5286/)
        data_ped(8173,1:3)=(/8173,8061,5733/)
        data_ped(8174,1:3)=(/8174,8061,7184/)
        data_ped(8175,1:3)=(/8175,8115,7144/)
        data_ped(8176,1:3)=(/8176,8065,6002/)
        data_ped(8177,1:3)=(/8177,5400,8063/)
        data_ped(8178,1:3)=(/8178,7194,8102/)
        data_ped(8179,1:3)=(/8179,5664,8055/)
        data_ped(8180,1:3)=(/8180,8079,8124/)
        data_ped(8181,1:3)=(/8181,8060,5983/)
        data_ped(8182,1:3)=(/8182,5400,8080/)
        data_ped(8183,1:3)=(/8183,8127,8116/)
        data_ped(8184,1:3)=(/8184,5738,8067/)
        data_ped(8185,1:3)=(/8185,8061,5521/)
        data_ped(8186,1:3)=(/8186,8115,7243/)
        data_ped(8187,1:3)=(/8187,8115,5616/)
        data_ped(8188,1:3)=(/8188,8065,5884/)
        data_ped(8189,1:3)=(/8189,5665,8126/)
        data_ped(8190,1:3)=(/8190,5814,8095/)
        data_ped(8191,1:3)=(/8191,5668,8121/)
        data_ped(8192,1:3)=(/8192,8138,5936/)
        data_ped(8193,1:3)=(/8193,5676,8107/)
        data_ped(8194,1:3)=(/8194,5676,8101/)
        data_ped(8195,1:3)=(/8195,7089,8100/)
        data_ped(8196,1:3)=(/8196,8115,7303/)
        data_ped(8197,1:3)=(/8197,8115,5831/)
        data_ped(8198,1:3)=(/8198,8115,7276/)
        data_ped(8199,1:3)=(/8199,8118,5972/)
        data_ped(8200,1:3)=(/8200,8061,5834/)
        data_ped(8201,1:3)=(/8201,8061,5967/)
        data_ped(8202,1:3)=(/8202,8061,5774/)
        data_ped(8203,1:3)=(/8203,8129,5313/)
        data_ped(8204,1:3)=(/8204,5870,8096/)
        data_ped(8205,1:3)=(/8205,8110,5463/)
        data_ped(8206,1:3)=(/8206,8170,7877/)
        data_ped(8207,1:3)=(/8207,5987,8091/)
        data_ped(8208,1:3)=(/8208,8098,7180/)
        data_ped(8209,1:3)=(/8209,5984,8147/)
        data_ped(8210,1:3)=(/8210,8061,4672/)
        data_ped(8211,1:3)=(/8211,8061,5834/)
        data_ped(8212,1:3)=(/8212,8061,7359/)
        data_ped(8213,1:3)=(/8213,7269,8146/)
        data_ped(8214,1:3)=(/8214,8115,5831/)
        data_ped(8215,1:3)=(/8215,7270,8155/)
        data_ped(8216,1:3)=(/8216,8172,6143/)
        data_ped(8217,1:3)=(/8217,5067,8159/)
        data_ped(8218,1:3)=(/8218,8115,7332/)
        data_ped(8219,1:3)=(/8219,8115,6991/)
        data_ped(8220,1:3)=(/8220,8115,6180/)
        data_ped(8221,1:3)=(/8221,6010,8145/)
        data_ped(8222,1:3)=(/8222,5676,8107/)
        data_ped(8223,1:3)=(/8223,8117,7278/)
        data_ped(8224,1:3)=(/8224,8128,8177/)
        data_ped(8225,1:3)=(/8225,8134,5517/)
        data_ped(8226,1:3)=(/8226,5400,8080/)
        data_ped(8227,1:3)=(/8227,5667,8163/)
        data_ped(8228,1:3)=(/8228,8183,5731/)
        data_ped(8229,1:3)=(/8229,8182,5919/)
        data_ped(8230,1:3)=(/8230,8171,5557/)
        data_ped(8231,1:3)=(/8231,8171,6141/)
        data_ped(8232,1:3)=(/8232,8131,5381/)
        data_ped(8233,1:3)=(/8233,5676,8158/)
        data_ped(8234,1:3)=(/8234,8110,6166/)
        data_ped(8235,1:3)=(/8235,8115,7333/)
        data_ped(8236,1:3)=(/8236,8115,5763/)
        data_ped(8237,1:3)=(/8237,8176,7138/)
        data_ped(8238,1:3)=(/8238,8156,7346/)
        data_ped(8239,1:3)=(/8239,7194,8122/)
        data_ped(8240,1:3)=(/8240,7127,8148/)
        data_ped(8241,1:3)=(/8241,5990,8189/)
        data_ped(8242,1:3)=(/8242,7361,8132/)
        data_ped(8243,1:3)=(/8243,8161,7278/)
        data_ped(8244,1:3)=(/8244,8061,5970/)
        data_ped(8245,1:3)=(/8245,8061,6159/)
        data_ped(8246,1:3)=(/8246,6106,8166/)
        data_ped(8247,1:3)=(/8247,8156,7402/)
        data_ped(8248,1:3)=(/8248,8176,7138/)
        data_ped(8249,1:3)=(/8249,8130,5559/)
        data_ped(8250,1:3)=(/8250,8171,5463/)
        data_ped(8251,1:3)=(/8251,8171,6053/)
        data_ped(8252,1:3)=(/8252,8182,7981/)
        data_ped(8253,1:3)=(/8253,8182,8097/)
        data_ped(8254,1:3)=(/8254,8182,6042/)
        data_ped(8255,1:3)=(/8255,8135,6208/)
        data_ped(8256,1:3)=(/8256,7227,8125/)
        data_ped(8257,1:3)=(/8257,8206,8178/)
        data_ped(8258,1:3)=(/8258,8111,7872/)
        data_ped(8259,1:3)=(/8259,8061,5616/)
        data_ped(8260,1:3)=(/8260,8061,7448/)
        data_ped(8261,1:3)=(/8261,8118,5800/)
        data_ped(8262,1:3)=(/8262,8137,5591/)
        data_ped(8263,1:3)=(/8263,8171,7449/)
        data_ped(8264,1:3)=(/8264,8094,6087/)
        data_ped(8265,1:3)=(/8265,8144,5727/)
        data_ped(8266,1:3)=(/8266,8156,6040/)
        data_ped(8267,1:3)=(/8267,8156,6015/)
        data_ped(8268,1:3)=(/8268,8192,6213/)
        data_ped(8269,1:3)=(/8269,8209,5323/)
        data_ped(8270,1:3)=(/8270,5988,8179/)
        data_ped(8271,1:3)=(/8271,8190,6021/)
        data_ped(8272,1:3)=(/8272,5988,8123/)
        data_ped(8273,1:3)=(/8273,5400,8103/)
        data_ped(8274,1:3)=(/8274,8191,7876/)
        data_ped(8275,1:3)=(/8275,8114,1352/)
        data_ped(8276,1:3)=(/8276,7454,8215/)
        data_ped(8277,1:3)=(/8277,7452,8150/)
        data_ped(8278,1:3)=(/8278,8081,5591/)
        data_ped(8279,1:3)=(/8279,8081,5690/)
        data_ped(8280,1:3)=(/8280,7497,8165/)
        data_ped(8281,1:3)=(/8281,8152,6246/)
        data_ped(8282,1:3)=(/8282,8190,8207/)
        data_ped(8283,1:3)=(/8283,8190,8151/)
        data_ped(8284,1:3)=(/8284,8180,8169/)
        data_ped(8285,1:3)=(/8285,7194,8133/)
        data_ped(8286,1:3)=(/8286,8173,7403/)
        data_ped(8287,1:3)=(/8287,8224,8166/)
        data_ped(8288,1:3)=(/8288,8224,6266/)
        data_ped(8289,1:3)=(/8289,7906,8133/)
        data_ped(8290,1:3)=(/8290,5873,8216/)
        data_ped(8291,1:3)=(/8291,8153,5860/)
        data_ped(8292,1:3)=(/8292,8111,1382/)
        data_ped(8293,1:3)=(/8293,8231,7501/)
        data_ped(8294,1:3)=(/8294,8231,6294/)
        data_ped(8295,1:3)=(/8295,8229,7547/)
        data_ped(8296,1:3)=(/8296,8241,6256/)
        data_ped(8297,1:3)=(/8297,8061,6263/)
        data_ped(8298,1:3)=(/8298,6116,8227/)
        data_ped(8299,1:3)=(/8299,8241,7558/)
        data_ped(8300,1:3)=(/8300,8241,7524/)
        data_ped(8301,1:3)=(/8301,8234,7453/)
        data_ped(8302,1:3)=(/8302,8231,6292/)
        data_ped(8303,1:3)=(/8303,8224,8230/)
        data_ped(8304,1:3)=(/8304,8061,7485/)
        data_ped(8305,1:3)=(/8305,8061,5834/)
        data_ped(8306,1:3)=(/8306,8061,8187/)
        data_ped(8307,1:3)=(/8307,5942,8208/)
        data_ped(8308,1:3)=(/8308,8285,8239/)
        data_ped(8309,1:3)=(/8309,8237,6322/)
        data_ped(8310,1:3)=(/8310,8160,6285/)
        data_ped(8311,1:3)=(/8311,8254,7568/)
        data_ped(8312,1:3)=(/8312,8194,5754/)
        data_ped(8313,1:3)=(/8313,8231,8233/)
        data_ped(8314,1:3)=(/8314,8247,7903/)
        data_ped(8315,1:3)=(/8315,7982,8139/)
        data_ped(8316,1:3)=(/8316,8181,7502/)
        data_ped(8317,1:3)=(/8317,7550,8252/)
        data_ped(8318,1:3)=(/8318,8172,5892/)
        data_ped(8319,1:3)=(/8319,8061,7485/)
        data_ped(8320,1:3)=(/8320,6314,8196/)
        data_ped(8321,1:3)=(/8321,6377,8149/)
        data_ped(8322,1:3)=(/8322,8257,8164/)
        data_ped(8323,1:3)=(/8323,8234,7363/)
        data_ped(8324,1:3)=(/8324,8222,7463/)
        data_ped(8325,1:3)=(/8325,8246,6381/)
        data_ped(8326,1:3)=(/8326,8246,8167/)
        data_ped(8327,1:3)=(/8327,8246,6179/)
        data_ped(8328,1:3)=(/8328,7532,8250/)
        data_ped(8329,1:3)=(/8329,8061,7330/)
        data_ped(8330,1:3)=(/8330,6314,8214/)
        data_ped(8331,1:3)=(/8331,8231,8221/)
        data_ped(8332,1:3)=(/8332,8247,7903/)
        data_ped(8333,1:3)=(/8333,8286,6247/)
        data_ped(8334,1:3)=(/8334,8286,8280/)
        data_ped(8335,1:3)=(/8335,5843,8217/)
        data_ped(8336,1:3)=(/8336,8273,7524/)
        data_ped(8337,1:3)=(/8337,7425,8277/)
        data_ped(8338,1:3)=(/8338,7577,8281/)
        data_ped(8339,1:3)=(/8339,8225,7597/)
        data_ped(8340,1:3)=(/8340,8205,6382/)
        data_ped(8341,1:3)=(/8341,8194,7502/)
        data_ped(8342,1:3)=(/8342,8194,7476/)
        data_ped(8343,1:3)=(/8343,8247,5999/)
        data_ped(8344,1:3)=(/8344,7885,8139/)
        data_ped(8345,1:3)=(/8345,8231,6206/)
        data_ped(8346,1:3)=(/8346,5549,8233/)
        data_ped(8347,1:3)=(/8347,8260,6164/)
        data_ped(8348,1:3)=(/8348,6391,8283/)
        data_ped(8349,1:3)=(/8349,8115,7392/)
        data_ped(8350,1:3)=(/8350,8240,8193/)
        data_ped(8351,1:3)=(/8351,6298,8253/)
        data_ped(8352,1:3)=(/8352,8286,7629/)
        data_ped(8353,1:3)=(/8353,8286,8276/)
        data_ped(8354,1:3)=(/8354,8226,6277/)
        data_ped(8355,1:3)=(/8355,8256,7967/)
        data_ped(8356,1:3)=(/8356,8272,7371/)
        data_ped(8357,1:3)=(/8357,8205,7538/)
        data_ped(8358,1:3)=(/8358,8244,7918/)
        data_ped(8359,1:3)=(/8359,8247,7539/)
        data_ped(8360,1:3)=(/8360,8200,6325/)
        data_ped(8361,1:3)=(/8361,8200,7638/)
        data_ped(8362,1:3)=(/8362,7382,8113/)
        data_ped(8363,1:3)=(/8363,8274,8282/)
        data_ped(8364,1:3)=(/8364,8173,6112/)
        data_ped(8365,1:3)=(/8365,7576,8265/)
        data_ped(8366,1:3)=(/8366,8260,8142/)
        data_ped(8367,1:3)=(/8367,7616,8269/)
        data_ped(8368,1:3)=(/8368,8197,6318/)
        data_ped(8369,1:3)=(/8369,8205,7351/)
        data_ped(8370,1:3)=(/8370,8205,7641/)
        data_ped(8371,1:3)=(/8371,8244,7643/)
        data_ped(8372,1:3)=(/8372,7478,8213/)
        data_ped(8373,1:3)=(/8373,6317,8232/)
        data_ped(8374,1:3)=(/8374,8200,6278/)
        data_ped(8375,1:3)=(/8375,8200,6394/)
        data_ped(8376,1:3)=(/8376,7382,8140/)
        data_ped(8377,1:3)=(/8377,8186,7624/)
        data_ped(8378,1:3)=(/8378,7286,8141/)
        data_ped(8379,1:3)=(/8379,8286,6115/)
        data_ped(8380,1:3)=(/8380,8286,1384/)
        data_ped(8381,1:3)=(/8381,7329,8223/)
        data_ped(8382,1:3)=(/8382,6363,8220/)
        data_ped(8383,1:3)=(/8383,8310,6259/)
        data_ped(8384,1:3)=(/8384,6251,8143/)
        data_ped(8385,1:3)=(/8385,8260,1376/)
        data_ped(8386,1:3)=(/8386,7223,8255/)
        data_ped(8387,1:3)=(/8387,8249,7390/)
        data_ped(8388,1:3)=(/8388,7532,8199/)
        data_ped(8389,1:3)=(/8389,8273,8300/)
        data_ped(8390,1:3)=(/8390,8246,8263/)
        data_ped(8391,1:3)=(/8391,8274,8214/)
        data_ped(8392,1:3)=(/8392,8200,6353/)
        data_ped(8393,1:3)=(/8393,8200,6395/)
        data_ped(8394,1:3)=(/8394,8200,6392/)
        data_ped(8395,1:3)=(/8395,8200,6308/)
        data_ped(8396,1:3)=(/8396,8200,7901/)
        data_ped(8397,1:3)=(/8397,8200,7585/)
        data_ped(8398,1:3)=(/8398,8304,7394/)
        data_ped(8399,1:3)=(/8399,8304,5997/)
        data_ped(8400,1:3)=(/8400,8264,6373/)
        data_ped(8401,1:3)=(/8401,7654,8335/)
        data_ped(8402,1:3)=(/8402,8197,7567/)
        data_ped(8403,1:3)=(/8403,6219,8202/)
        data_ped(8404,1:3)=(/8404,8205,6359/)
        data_ped(8405,1:3)=(/8405,8317,7595/)
        data_ped(8406,1:3)=(/8406,8329,6344/)
        data_ped(8407,1:3)=(/8407,8286,1390/)
        data_ped(8408,1:3)=(/8408,6237,8120/)
        data_ped(8409,1:3)=(/8409,7532,8199/)
        data_ped(8410,1:3)=(/8410,8266,7263/)
        data_ped(8411,1:3)=(/8411,8236,7573/)
        data_ped(8412,1:3)=(/8412,8289,8287/)
        data_ped(8413,1:3)=(/8413,8271,6385/)
        data_ped(8414,1:3)=(/8414,8211,6270/)
        data_ped(8415,1:3)=(/8415,8205,7619/)
        data_ped(8416,1:3)=(/8416,8205,7450/)
        data_ped(8417,1:3)=(/8417,8160,8212/)
        data_ped(8418,1:3)=(/8418,8244,8203/)
        data_ped(8419,1:3)=(/8419,8304,8349/)
        data_ped(8420,1:3)=(/8420,7614,8261/)
        data_ped(8421,1:3)=(/8421,8200,7432/)
        data_ped(8422,1:3)=(/8422,8200,6311/)
        data_ped(8423,1:3)=(/8423,7480,8292/)
        data_ped(8424,1:3)=(/8424,6227,8235/)
        data_ped(8425,1:3)=(/8425,8329,7639/)
        data_ped(8426,1:3)=(/8426,8329,1378/)
        data_ped(8427,1:3)=(/8427,8329,8303/)
        data_ped(8428,1:3)=(/8428,8301,8251/)
        data_ped(8429,1:3)=(/8429,8274,6390/)
        data_ped(8430,1:3)=(/8430,8173,8321/)
        data_ped(8431,1:3)=(/8431,8173,7482/)
        data_ped(8432,1:3)=(/8432,8231,8277/)
        data_ped(8433,1:3)=(/8433,8211,1370/)
        data_ped(8434,1:3)=(/8434,8274,7589/)
        data_ped(8435,1:3)=(/8435,8274,7469/)
        data_ped(8436,1:3)=(/8436,5952,8298/)
        data_ped(8437,1:3)=(/8437,8248,7891/)
        data_ped(8438,1:3)=(/8438,8211,6211/)
        data_ped(8439,1:3)=(/8439,8211,7433/)
        data_ped(8440,1:3)=(/8440,8222,7664/)
        data_ped(8441,1:3)=(/8441,8160,7537/)
        data_ped(8442,1:3)=(/8442,8205,7914/)
        data_ped(8443,1:3)=(/8443,8197,1348/)
        data_ped(8444,1:3)=(/8444,8306,7540/)
        data_ped(8445,1:3)=(/8445,8278,6348/)
        data_ped(8446,1:3)=(/8446,8200,7579/)
        data_ped(8447,1:3)=(/8447,8200,6311/)
        data_ped(8448,1:3)=(/8448,8329,7923/)
        data_ped(8449,1:3)=(/8449,8173,6168/)
        data_ped(8450,1:3)=(/8450,8173,8238/)
        data_ped(8451,1:3)=(/8451,8173,7448/)
        data_ped(8452,1:3)=(/8452,7885,8315/)
        data_ped(8453,1:3)=(/8453,8327,7580/)
        data_ped(8454,1:3)=(/8454,8363,8326/)
        data_ped(8455,1:3)=(/8455,8291,7646/)
        data_ped(8456,1:3)=(/8456,8244,7523/)
        data_ped(8457,1:3)=(/8457,8244,1392/)
        data_ped(8458,1:3)=(/8458,8197,7914/)
        data_ped(8459,1:3)=(/8459,8211,6238/)
        data_ped(8460,1:3)=(/8460,8200,6311/)
        data_ped(8461,1:3)=(/8461,8200,8228/)
        data_ped(8462,1:3)=(/8462,8173,6240/)
        data_ped(8463,1:3)=(/8463,8308,7706/)
        data_ped(8464,1:3)=(/8464,8286,7636/)
        data_ped(8465,1:3)=(/8465,8136,7404/)
        data_ped(8466,1:3)=(/8466,7686,8296/)
        data_ped(8467,1:3)=(/8467,8137,8279/)
        data_ped(8468,1:3)=(/8468,8157,7339/)
        data_ped(8469,1:3)=(/8469,6326,8198/)
        data_ped(8470,1:3)=(/8470,8205,7578/)
        data_ped(8471,1:3)=(/8471,8231,7560/)
        data_ped(8472,1:3)=(/8472,8372,7723/)
        data_ped(8473,1:3)=(/8473,8205,8408/)
        data_ped(8474,1:3)=(/8474,8244,1379/)
        data_ped(8475,1:3)=(/8475,8322,8390/)
        data_ped(8476,1:3)=(/8476,7545,8401/)
        data_ped(8477,1:3)=(/8477,8391,8412/)
        data_ped(8478,1:3)=(/8478,8356,7994/)
        data_ped(8479,1:3)=(/8479,8211,6193/)
        data_ped(8480,1:3)=(/8480,8205,8425/)
        data_ped(8481,1:3)=(/8481,381,8000/)
        data_ped(8482,1:3)=(/8482,388,8001/)
        data_ped(8483,1:3)=(/8483,8005,7960/)
        data_ped(8484,1:3)=(/8484,363,8483/)
        data_ped(8485,1:3)=(/8485,1551,8482/)
        data_ped(8486,1:3)=(/8486,8087,4469/)
        data_ped(8487,1:3)=(/8487,8006,8481/)
        data_ped(8488,1:3)=(/8488,452,8487/)
        data_ped(8489,1:3)=(/8489,377,8488/)
        data_ped(8490,1:3)=(/8490,8014,8489/)
        data_ped(8491,1:3)=(/8491,0,8489/)
        data_ped(8492,1:3)=(/8492,0,8490/)
        data_ped(8493,1:3)=(/8493,0,8491/)
        data_ped(8494,1:3)=(/8494,2444,8492/)
        data_ped(8495,1:3)=(/8495,2530,8494/)
        data_ped(8496,1:3)=(/8496,2444,8492/)
        data_ped(8497,1:3)=(/8497,8495,3826/)
        data_ped(8498,1:3)=(/8498,8495,3926/)
        data_ped(8499,1:3)=(/8499,8498,4583/)
        data_ped(8500,1:3)=(/8500,8495,4034/)
        data_ped(8501,1:3)=(/8501,4176,8496/)
        data_ped(8502,1:3)=(/8502,8498,4036/)
        data_ped(8503,1:3)=(/8503,8495,4034/)
        data_ped(8504,1:3)=(/8504,8495,4142/)
        data_ped(8505,1:3)=(/8505,8500,5128/)
        data_ped(8506,1:3)=(/8506,4520,8497/)
        data_ped(8507,1:3)=(/8507,5492,8503/)
        data_ped(8508,1:3)=(/8508,8499,5686/)
        data_ped(8509,1:3)=(/8509,8504,5822/)
        data_ped(8510,1:3)=(/8510,5825,8505/)
        data_ped(8511,1:3)=(/8511,8509,8508/)
        data_ped(8512,1:3)=(/8512,8162,8486/)
        data_ped(8513,1:3)=(/8513,8507,6337/)
        data_ped(8514,1:3)=(/8514,7486,8510/)
        data_ped(8515,1:3)=(/8515,8244,8511/)
        data_ped(8516,1:3)=(/8516,8154,8513/)
        data_ped(8517,1:3)=(/8517,8268,8514/)
        data_ped(8518,1:3)=(/8518,373,8481/)
        data_ped(8519,1:3)=(/8519,8485,8004/)
        data_ped(8520,1:3)=(/8520,8484,8518/)
        data_ped(8521,1:3)=(/8521,8519,6520/)
        data_ped(8522,1:3)=(/8522,374,8518/)
        data_ped(8523,1:3)=(/8523,8520,7974/)
        data_ped(8524,1:3)=(/8524,8520,6615/)
        data_ped(8525,1:3)=(/8525,8522,8010/)
        data_ped(8526,1:3)=(/8526,8524,6618/)
        data_ped(8527,1:3)=(/8527,8014,8521/)
        data_ped(8528,1:3)=(/8528,8014,8523/)
        data_ped(8529,1:3)=(/8529,8525,0/)
        data_ped(8530,1:3)=(/8530,8525,8013/)
        data_ped(8531,1:3)=(/8531,8530,8526/)
        data_ped(8532,1:3)=(/8532,8525,8527/)
        data_ped(8533,1:3)=(/8533,8525,8490/)
        data_ped(8534,1:3)=(/8534,8533,1089/)
        data_ped(8535,1:3)=(/8535,2119,8532/)
        data_ped(8536,1:3)=(/8536,8533,8493/)
        data_ped(8537,1:3)=(/8537,8528,2595/)
        data_ped(8538,1:3)=(/8538,8528,8532/)
        data_ped(8539,1:3)=(/8539,8014,8531/)
        data_ped(8540,1:3)=(/8540,8529,2806/)
        data_ped(8541,1:3)=(/8541,0,8534/)
        data_ped(8542,1:3)=(/8542,0,8532/)
        data_ped(8543,1:3)=(/8543,8533,2925/)
        data_ped(8544,1:3)=(/8544,8533,1023/)
        data_ped(8545,1:3)=(/8545,8536,1213/)
        data_ped(8546,1:3)=(/8546,8536,8532/)
        data_ped(8547,1:3)=(/8547,2444,8531/)
        data_ped(8548,1:3)=(/8548,8540,8018/)
        data_ped(8549,1:3)=(/8549,8533,8539/)
        data_ped(8550,1:3)=(/8550,8536,2866/)
        data_ped(8551,1:3)=(/8551,8536,8492/)
        data_ped(8552,1:3)=(/8552,8536,8535/)
        data_ped(8553,1:3)=(/8553,8533,8539/)
        data_ped(8554,1:3)=(/8554,8538,2986/)
        data_ped(8555,1:3)=(/8555,8533,8017/)
        data_ped(8556,1:3)=(/8556,8540,2652/)
        data_ped(8557,1:3)=(/8557,3115,8541/)
        data_ped(8558,1:3)=(/8558,8538,3458/)
        data_ped(8559,1:3)=(/8559,3018,8543/)
        data_ped(8560,1:3)=(/8560,2444,8534/)
        data_ped(8561,1:3)=(/8561,8538,3306/)
        data_ped(8562,1:3)=(/8562,8538,2881/)
        data_ped(8563,1:3)=(/8563,8536,8535/)
        data_ped(8564,1:3)=(/8564,3101,8545/)
        data_ped(8565,1:3)=(/8565,8536,3176/)
        data_ped(8566,1:3)=(/8566,8546,2575/)
        data_ped(8567,1:3)=(/8567,8546,8030/)
        data_ped(8568,1:3)=(/8568,8550,2488/)
        data_ped(8569,1:3)=(/8569,3018,8543/)
        data_ped(8570,1:3)=(/8570,3101,8544/)
        data_ped(8571,1:3)=(/8571,2444,8534/)
        data_ped(8572,1:3)=(/8572,8550,3641/)
        data_ped(8573,1:3)=(/8573,3018,8548/)
        data_ped(8574,1:3)=(/8574,8546,8027/)
        data_ped(8575,1:3)=(/8575,8550,6677/)
        data_ped(8576,1:3)=(/8576,8547,3592/)
        data_ped(8577,1:3)=(/8577,8561,4236/)
        data_ped(8578,1:3)=(/8578,8560,3815/)
        data_ped(8579,1:3)=(/8579,6755,8557/)
        data_ped(8580,1:3)=(/8580,8536,3176/)
        data_ped(8581,1:3)=(/8581,8565,4134/)
        data_ped(8582,1:3)=(/8582,8561,3144/)
        data_ped(8583,1:3)=(/8583,3101,8542/)
        data_ped(8584,1:3)=(/8584,8546,3985/)
        data_ped(8585,1:3)=(/8585,3018,8563/)
        data_ped(8586,1:3)=(/8586,8550,3984/)
        data_ped(8587,1:3)=(/8587,4176,8559/)
        data_ped(8588,1:3)=(/8588,3018,8548/)
        data_ped(8589,1:3)=(/8589,8538,3644/)
        data_ped(8590,1:3)=(/8590,8536,3314/)
        data_ped(8591,1:3)=(/8591,8536,8535/)
        data_ped(8592,1:3)=(/8592,3379,8566/)
        data_ped(8593,1:3)=(/8593,8558,3767/)
        data_ped(8594,1:3)=(/8594,8538,3644/)
        data_ped(8595,1:3)=(/8595,8561,2804/)
        data_ped(8596,1:3)=(/8596,8565,8537/)
        data_ped(8597,1:3)=(/8597,8561,4236/)
        data_ped(8598,1:3)=(/8598,4412,8562/)
        data_ped(8599,1:3)=(/8599,3018,8563/)
        data_ped(8600,1:3)=(/8600,8569,1278/)
        data_ped(8601,1:3)=(/8601,8546,8030/)
        data_ped(8602,1:3)=(/8602,8577,4674/)
        data_ped(8603,1:3)=(/8603,8579,3831/)
        data_ped(8604,1:3)=(/8604,8550,4330/)
        data_ped(8605,1:3)=(/8605,3018,8563/)
        data_ped(8606,1:3)=(/8606,8573,3984/)
        data_ped(8607,1:3)=(/8607,3097,8556/)
        data_ped(8608,1:3)=(/8608,8580,3735/)
        data_ped(8609,1:3)=(/8609,8546,4576/)
        data_ped(8610,1:3)=(/8610,6828,8554/)
        data_ped(8611,1:3)=(/8611,8546,8583/)
        data_ped(8612,1:3)=(/8612,8550,4778/)
        data_ped(8613,1:3)=(/8613,3018,8548/)
        data_ped(8614,1:3)=(/8614,4151,8585/)
        data_ped(8615,1:3)=(/8615,4336,8571/)
        data_ped(8616,1:3)=(/8616,8579,4491/)
        data_ped(8617,1:3)=(/8617,4184,8570/)
        data_ped(8618,1:3)=(/8618,8538,3543/)
        data_ped(8619,1:3)=(/8619,8538,8025/)
        data_ped(8620,1:3)=(/8620,4180,8563/)
        data_ped(8621,1:3)=(/8621,4180,8548/)
        data_ped(8622,1:3)=(/8622,3864,8586/)
        data_ped(8623,1:3)=(/8623,8536,3925/)
        data_ped(8624,1:3)=(/8624,3018,8572/)
        data_ped(8625,1:3)=(/8625,4184,8542/)
        data_ped(8626,1:3)=(/8626,8590,4492/)
        data_ped(8627,1:3)=(/8627,4184,8570/)
        data_ped(8628,1:3)=(/8628,8546,8549/)
        data_ped(8629,1:3)=(/8629,8538,8548/)
        data_ped(8630,1:3)=(/8630,4264,8551/)
        data_ped(8631,1:3)=(/8631,8546,8559/)
        data_ped(8632,1:3)=(/8632,8546,8030/)
        data_ped(8633,1:3)=(/8633,4626,8586/)
        data_ped(8634,1:3)=(/8634,4264,8585/)
        data_ped(8635,1:3)=(/8635,4264,8599/)
        data_ped(8636,1:3)=(/8636,8582,1263/)
        data_ped(8637,1:3)=(/8637,8538,8549/)
        data_ped(8638,1:3)=(/8638,8578,3827/)
        data_ped(8639,1:3)=(/8639,4807,8581/)
        data_ped(8640,1:3)=(/8640,8546,8583/)
        data_ped(8641,1:3)=(/8641,4626,8564/)
        data_ped(8642,1:3)=(/8642,8538,8548/)
        data_ped(8643,1:3)=(/8643,8538,8549/)
        data_ped(8644,1:3)=(/8644,5064,8593/)
        data_ped(8645,1:3)=(/8645,8606,3777/)
        data_ped(8646,1:3)=(/8646,8561,8576/)
        data_ped(8647,1:3)=(/8647,4626,8605/)
        data_ped(8648,1:3)=(/8648,8536,4590/)
        data_ped(8649,1:3)=(/8649,4713,8572/)
        data_ped(8650,1:3)=(/8650,8538,8555/)
        data_ped(8651,1:3)=(/8651,8578,4857/)
        data_ped(8652,1:3)=(/8652,4264,8563/)
        data_ped(8653,1:3)=(/8653,4180,8583/)
        data_ped(8654,1:3)=(/8654,3973,8604/)
        data_ped(8655,1:3)=(/8655,5069,8548/)
        data_ped(8656,1:3)=(/8656,5069,8567/)
        data_ped(8657,1:3)=(/8657,4626,8605/)
        data_ped(8658,1:3)=(/8658,4713,8572/)
        data_ped(8659,1:3)=(/8659,8538,4238/)
        data_ped(8660,1:3)=(/8660,4713,8611/)
        data_ped(8661,1:3)=(/8661,8608,4490/)
        data_ped(8662,1:3)=(/8662,8578,4780/)
        data_ped(8663,1:3)=(/8663,8578,8056/)
        data_ped(8664,1:3)=(/8664,4713,8572/)
        data_ped(8665,1:3)=(/8665,4713,8619/)
        data_ped(8666,1:3)=(/8666,4626,8564/)
        data_ped(8667,1:3)=(/8667,4626,8586/)
        data_ped(8668,1:3)=(/8668,6755,8584/)
        data_ped(8669,1:3)=(/8669,5069,8629/)
        data_ped(8670,1:3)=(/8670,4523,8592/)
        data_ped(8671,1:3)=(/8671,5069,8548/)
        data_ped(8672,1:3)=(/8672,4713,8621/)
        data_ped(8673,1:3)=(/8673,4713,8612/)
        data_ped(8674,1:3)=(/8674,8565,4579/)
        data_ped(8675,1:3)=(/8675,4626,8586/)
        data_ped(8676,1:3)=(/8676,5153,8635/)
        data_ped(8677,1:3)=(/8677,8595,5303/)
        data_ped(8678,1:3)=(/8678,8565,4856/)
        data_ped(8679,1:3)=(/8679,4713,8619/)
        data_ped(8680,1:3)=(/8680,4713,8612/)
        data_ped(8681,1:3)=(/8681,4713,8609/)
        data_ped(8682,1:3)=(/8682,8616,5037/)
        data_ped(8683,1:3)=(/8683,8628,4238/)
        data_ped(8684,1:3)=(/8684,5268,8631/)
        data_ped(8685,1:3)=(/8685,8565,8610/)
        data_ped(8686,1:3)=(/8686,8582,1306/)
        data_ped(8687,1:3)=(/8687,5069,8613/)
        data_ped(8688,1:3)=(/8688,5270,8605/)
        data_ped(8689,1:3)=(/8689,4890,8618/)
        data_ped(8690,1:3)=(/8690,4623,8553/)
        data_ped(8691,1:3)=(/8691,4982,8625/)
        data_ped(8692,1:3)=(/8692,4431,8575/)
        data_ped(8693,1:3)=(/8693,8637,8625/)
        data_ped(8694,1:3)=(/8694,4713,8624/)
        data_ped(8695,1:3)=(/8695,8615,5312/)
        data_ped(8696,1:3)=(/8696,5069,8642/)
        data_ped(8697,1:3)=(/8697,5270,8647/)
        data_ped(8698,1:3)=(/8698,4184,8598/)
        data_ped(8699,1:3)=(/8699,4978,8623/)
        data_ped(8700,1:3)=(/8700,8637,8632/)
        data_ped(8701,1:3)=(/8701,8601,6941/)
        data_ped(8702,1:3)=(/8702,8607,1312/)
        data_ped(8703,1:3)=(/8703,6950,8658/)
        data_ped(8704,1:3)=(/8704,6950,8620/)
        data_ped(8705,1:3)=(/8705,8615,5318/)
        data_ped(8706,1:3)=(/8706,3973,8611/)
        data_ped(8707,1:3)=(/8707,4983,8657/)
        data_ped(8708,1:3)=(/8708,8637,8625/)
        data_ped(8709,1:3)=(/8709,8601,5304/)
        data_ped(8710,1:3)=(/8710,8601,8643/)
        data_ped(8711,1:3)=(/8711,8589,5499/)
        data_ped(8712,1:3)=(/8712,8500,8603/)
        data_ped(8713,1:3)=(/8713,8579,4779/)
        data_ped(8714,1:3)=(/8714,5268,8631/)
        data_ped(8715,1:3)=(/8715,8640,1297/)
        data_ped(8716,1:3)=(/8716,5341,8594/)
        data_ped(8717,1:3)=(/8717,6924,8604/)
        data_ped(8718,1:3)=(/8718,3973,8611/)
        data_ped(8719,1:3)=(/8719,3973,8673/)
        data_ped(8720,1:3)=(/8720,5072,8644/)
        data_ped(8721,1:3)=(/8721,5160,8652/)
        data_ped(8722,1:3)=(/8722,5482,8584/)
        data_ped(8723,1:3)=(/8723,8628,8659/)
        data_ped(8724,1:3)=(/8724,8608,7835/)
        data_ped(8725,1:3)=(/8725,6950,8604/)
        data_ped(8726,1:3)=(/8726,4626,8609/)
        data_ped(8727,1:3)=(/8727,4626,8564/)
        data_ped(8728,1:3)=(/8728,8597,8089/)
        data_ped(8729,1:3)=(/8729,8671,8588/)
        data_ped(8730,1:3)=(/8730,5270,8647/)
        data_ped(8731,1:3)=(/8731,5272,8645/)
        data_ped(8732,1:3)=(/8732,5069,8642/)
        data_ped(8733,1:3)=(/8733,5069,8629/)
        data_ped(8734,1:3)=(/8734,3973,8552/)
        data_ped(8735,1:3)=(/8735,5072,8644/)
        data_ped(8736,1:3)=(/8736,4890,8586/)
        data_ped(8737,1:3)=(/8737,7057,8660/)
        data_ped(8738,1:3)=(/8738,7057,8680/)
        data_ped(8739,1:3)=(/8739,5270,8658/)
        data_ped(8740,1:3)=(/8740,5556,8620/)
        data_ped(8741,1:3)=(/8741,8685,5490/)
        data_ped(8742,1:3)=(/8742,8668,8046/)
        data_ped(8743,1:3)=(/8743,8596,5306/)
        data_ped(8744,1:3)=(/8744,8640,8694/)
        data_ped(8745,1:3)=(/8745,8640,5460/)
        data_ped(8746,1:3)=(/8746,8637,8630/)
        data_ped(8747,1:3)=(/8747,8637,8665/)
        data_ped(8748,1:3)=(/8748,6950,8609/)
        data_ped(8749,1:3)=(/8749,6950,8654/)
        data_ped(8750,1:3)=(/8750,4626,8564/)
        data_ped(8751,1:3)=(/8751,4626,8688/)
        data_ped(8752,1:3)=(/8752,7057,8611/)
        data_ped(8753,1:3)=(/8753,4989,8667/)
        data_ped(8754,1:3)=(/8754,5063,8614/)
        data_ped(8755,1:3)=(/8755,5069,8613/)
        data_ped(8756,1:3)=(/8756,5069,8629/)
        data_ped(8757,1:3)=(/8757,6792,8587/)
        data_ped(8758,1:3)=(/8758,5069,8687/)
        data_ped(8759,1:3)=(/8759,8671,8588/)
        data_ped(8760,1:3)=(/8760,8671,1331/)
        data_ped(8761,1:3)=(/8761,8637,5115/)
        data_ped(8762,1:3)=(/8762,8637,8690/)
        data_ped(8763,1:3)=(/8763,8682,5310/)
        data_ped(8764,1:3)=(/8764,8649,1314/)
        data_ped(8765,1:3)=(/8765,4626,8669/)
        data_ped(8766,1:3)=(/8766,4626,8634/)
        data_ped(8767,1:3)=(/8767,8670,5484/)
        data_ped(8768,1:3)=(/8768,8638,6894/)
        data_ped(8769,1:3)=(/8769,8626,8677/)
        data_ped(8770,1:3)=(/8770,8662,5114/)
        data_ped(8771,1:3)=(/8771,8695,1319/)
        data_ped(8772,1:3)=(/8772,3973,8653/)
        data_ped(8773,1:3)=(/8773,7057,8612/)
        data_ped(8774,1:3)=(/8774,7057,8680/)
        data_ped(8775,1:3)=(/8775,7057,8633/)
        data_ped(8776,1:3)=(/8776,8606,5634/)
        data_ped(8777,1:3)=(/8777,5072,8644/)
        data_ped(8778,1:3)=(/8778,8568,4775/)
        data_ped(8779,1:3)=(/8779,5069,8656/)
        data_ped(8780,1:3)=(/8780,4989,8697/)
        data_ped(8781,1:3)=(/8781,5067,8667/)
        data_ped(8782,1:3)=(/8782,4891,8593/)
        data_ped(8783,1:3)=(/8783,5338,8720/)
        data_ped(8784,1:3)=(/8784,5338,8644/)
        data_ped(8785,1:3)=(/8785,5338,8676/)
        data_ped(8786,1:3)=(/8786,4989,8622/)
        data_ped(8787,1:3)=(/8787,8617,7172/)
        data_ped(8788,1:3)=(/8788,8606,8654/)
        data_ped(8789,1:3)=(/8789,8655,1323/)
        data_ped(8790,1:3)=(/8790,5153,8631/)
        data_ped(8791,1:3)=(/8791,8589,4778/)
        data_ped(8792,1:3)=(/8792,4623,8652/)
        data_ped(8793,1:3)=(/8793,8591,8696/)
        data_ped(8794,1:3)=(/8794,8637,8684/)
        data_ped(8795,1:3)=(/8795,5556,8704/)
        data_ped(8796,1:3)=(/8796,8674,8501/)
        data_ped(8797,1:3)=(/8797,8678,5384/)
        data_ped(8798,1:3)=(/8798,5069,8613/)
        data_ped(8799,1:3)=(/8799,5676,8679/)
        data_ped(8800,1:3)=(/8800,7057,8611/)
        data_ped(8801,1:3)=(/8801,8597,5484/)
        data_ped(8802,1:3)=(/8802,8637,8690/)
        data_ped(8803,1:3)=(/8803,8637,8630/)
        data_ped(8804,1:3)=(/8804,6950,8609/)
        data_ped(8805,1:3)=(/8805,6950,8604/)
        data_ped(8806,1:3)=(/8806,6950,8735/)
        data_ped(8807,1:3)=(/8807,8681,5305/)
        data_ped(8808,1:3)=(/8808,8506,8663/)
        data_ped(8809,1:3)=(/8809,8710,8717/)
        data_ped(8810,1:3)=(/8810,8710,8665/)
        data_ped(8811,1:3)=(/8811,5835,8622/)
        data_ped(8812,1:3)=(/8812,8115,8752/)
        data_ped(8813,1:3)=(/8813,7057,8598/)
        data_ped(8814,1:3)=(/8814,5069,8564/)
        data_ped(8815,1:3)=(/8815,5762,8696/)
        data_ped(8816,1:3)=(/8816,5762,8755/)
        data_ped(8817,1:3)=(/8817,7134,8749/)
        data_ped(8818,1:3)=(/8818,5678,8753/)
        data_ped(8819,1:3)=(/8819,5747,8691/)
        data_ped(8820,1:3)=(/8820,8061,8624/)
        data_ped(8821,1:3)=(/8821,8061,8600/)
        data_ped(8822,1:3)=(/8822,4616,8619/)
        data_ped(8823,1:3)=(/8823,8589,4778/)
        data_ped(8824,1:3)=(/8824,8589,5706/)
        data_ped(8825,1:3)=(/8825,8608,4409/)
        data_ped(8826,1:3)=(/8826,4989,8666/)
        data_ped(8827,1:3)=(/8827,5270,8676/)
        data_ped(8828,1:3)=(/8828,6950,8740/)
        data_ped(8829,1:3)=(/8829,6950,8609/)
        data_ped(8830,1:3)=(/8830,6950,8654/)
        data_ped(8831,1:3)=(/8831,8651,5627/)
        data_ped(8832,1:3)=(/8832,8675,5828/)
        data_ped(8833,1:3)=(/8833,8648,5130/)
        data_ped(8834,1:3)=(/8834,8706,8644/)
        data_ped(8835,1:3)=(/8835,8574,0/)
        data_ped(8836,1:3)=(/8836,5337,8736/)
        data_ped(8837,1:3)=(/8837,8061,8743/)
        data_ped(8838,1:3)=(/8838,8639,1342/)
        data_ped(8839,1:3)=(/8839,7057,8611/)
        data_ped(8840,1:3)=(/8840,7134,8749/)
        data_ped(8841,1:3)=(/8841,5067,8667/)
        data_ped(8842,1:3)=(/8842,5675,8658/)
        data_ped(8843,1:3)=(/8843,8689,7202/)
        data_ped(8844,1:3)=(/8844,8627,7239/)
        data_ped(8845,1:3)=(/8845,5956,8780/)
        data_ped(8846,1:3)=(/8846,5956,8753/)
        data_ped(8847,1:3)=(/8847,4987,8636/)
        data_ped(8848,1:3)=(/8848,6950,8735/)
        data_ped(8849,1:3)=(/8849,8650,4943/)
        data_ped(8850,1:3)=(/8850,8671,8684/)
        data_ped(8851,1:3)=(/8851,8671,8654/)
        data_ped(8852,1:3)=(/8852,8671,8665/)
        data_ped(8853,1:3)=(/8853,8683,8711/)
        data_ped(8854,1:3)=(/8854,7140,8763/)
        data_ped(8855,1:3)=(/8855,8706,7203/)
        data_ped(8856,1:3)=(/8856,8706,8708/)
        data_ped(8857,1:3)=(/8857,8710,8745/)
        data_ped(8858,1:3)=(/8858,8723,8721/)
        data_ped(8859,1:3)=(/8859,7168,8659/)
        data_ped(8860,1:3)=(/8860,7168,8609/)
        data_ped(8861,1:3)=(/8861,7168,8726/)
        data_ped(8862,1:3)=(/8862,7168,8604/)
        data_ped(8863,1:3)=(/8863,7037,8702/)
        data_ped(8864,1:3)=(/8864,5847,8700/)
        data_ped(8865,1:3)=(/8865,5067,8644/)
        data_ped(8866,1:3)=(/8866,5067,8783/)
        data_ped(8867,1:3)=(/8867,7269,8749/)
        data_ped(8868,1:3)=(/8868,5675,8786/)
        data_ped(8869,1:3)=(/8869,7057,8673/)
        data_ped(8870,1:3)=(/8870,8115,8738/)
        data_ped(8871,1:3)=(/8871,6950,8777/)
        data_ped(8872,1:3)=(/8872,6950,8754/)
        data_ped(8873,1:3)=(/8873,7168,8622/)
        data_ped(8874,1:3)=(/8874,7168,8788/)
        data_ped(8875,1:3)=(/8875,8671,8792/)
        data_ped(8876,1:3)=(/8876,8675,7342/)
        data_ped(8877,1:3)=(/8877,8678,5566/)
        data_ped(8878,1:3)=(/8878,8683,8618/)
        data_ped(8879,1:3)=(/8879,8709,8619/)
        data_ped(8880,1:3)=(/8880,8712,8770/)
        data_ped(8881,1:3)=(/8881,8728,8502/)
        data_ped(8882,1:3)=(/8882,5067,8806/)
        data_ped(8883,1:3)=(/8883,7057,8633/)
        data_ped(8884,1:3)=(/8884,5633,8785/)
        data_ped(8885,1:3)=(/8885,8061,8753/)
        data_ped(8886,1:3)=(/8886,7134,8784/)
        data_ped(8887,1:3)=(/8887,8130,8705/)
        data_ped(8888,1:3)=(/8888,8591,8758/)
        data_ped(8889,1:3)=(/8889,4989,8766/)
        data_ped(8890,1:3)=(/8890,8637,6041/)
        data_ped(8891,1:3)=(/8891,6950,8730/)
        data_ped(8892,1:3)=(/8892,8650,8733/)
        data_ped(8893,1:3)=(/8893,8692,5683/)
        data_ped(8894,1:3)=(/8894,8701,8707/)
        data_ped(8895,1:3)=(/8895,8709,7210/)
        data_ped(8896,1:3)=(/8896,8709,8711/)
        data_ped(8897,1:3)=(/8897,8716,7137/)
        data_ped(8898,1:3)=(/8898,8729,5775/)
        data_ped(8899,1:3)=(/8899,8729,1346/)
        data_ped(8900,1:3)=(/8900,8797,8724/)
        data_ped(8901,1:3)=(/8901,8706,8672/)
        data_ped(8902,1:3)=(/8902,4979,8731/)
        data_ped(8903,1:3)=(/8903,8061,8776/)
        data_ped(8904,1:3)=(/8904,7134,8806/)
        data_ped(8905,1:3)=(/8905,7058,8726/)
        data_ped(8906,1:3)=(/8906,8744,6023/)
        data_ped(8907,1:3)=(/8907,5817,8739/)
        data_ped(8908,1:3)=(/8908,5817,8721/)
        data_ped(8909,1:3)=(/8909,6977,8799/)
        data_ped(8910,1:3)=(/8910,8606,8815/)
        data_ped(8911,1:3)=(/8911,6008,8751/)
        data_ped(8912,1:3)=(/8912,8591,7173/)
        data_ped(8913,1:3)=(/8913,8591,8795/)
        data_ped(8914,1:3)=(/8914,5481,8654/)
        data_ped(8915,1:3)=(/8915,8671,8687/)
        data_ped(8916,1:3)=(/8916,8671,8746/)
        data_ped(8917,1:3)=(/8917,8722,5029/)
        data_ped(8918,1:3)=(/8918,8709,8119/)
        data_ped(8919,1:3)=(/8919,8709,5862/)
        data_ped(8920,1:3)=(/8920,8061,8753/)
        data_ped(8921,1:3)=(/8921,7057,8673/)
        data_ped(8922,1:3)=(/8922,7058,8747/)
        data_ped(8923,1:3)=(/8923,5554,8812/)
        data_ped(8924,1:3)=(/8924,5554,8718/)
        data_ped(8925,1:3)=(/8925,5818,8732/)
        data_ped(8926,1:3)=(/8926,8131,8826/)
        data_ped(8927,1:3)=(/8927,6001,8677/)
        data_ped(8928,1:3)=(/8928,5749,8842/)
        data_ped(8929,1:3)=(/8929,7299,8654/)
        data_ped(8930,1:3)=(/8930,8661,5995/)
        data_ped(8931,1:3)=(/8931,8709,8174/)
        data_ped(8932,1:3)=(/8932,8734,8663/)
        data_ped(8933,1:3)=(/8933,8767,5313/)
        data_ped(8934,1:3)=(/8934,8803,7275/)
        data_ped(8935,1:3)=(/8935,8803,8632/)
        data_ped(8936,1:3)=(/8936,8802,5828/)
        data_ped(8937,1:3)=(/8937,1358,8835/)
        data_ped(8938,1:3)=(/8938,4989,8827/)
        data_ped(8939,1:3)=(/8939,6928,8863/)
        data_ped(8940,1:3)=(/8940,8671,8794/)
        data_ped(8941,1:3)=(/8941,8671,5499/)
        data_ped(8942,1:3)=(/8942,8692,8646/)
        data_ped(8943,1:3)=(/8943,8723,8652/)
        data_ped(8944,1:3)=(/8944,8791,7308/)
        data_ped(8945,1:3)=(/8945,8802,5959/)
        data_ped(8946,1:3)=(/8946,8802,5962/)
        data_ped(8947,1:3)=(/8947,8803,8185/)
        data_ped(8948,1:3)=(/8948,8803,8759/)
        data_ped(8949,1:3)=(/8949,8803,8822/)
        data_ped(8950,1:3)=(/8950,8802,6135/)
        data_ped(8951,1:3)=(/8951,8802,8807/)
        data_ped(8952,1:3)=(/8952,5747,8824/)
        data_ped(8953,1:3)=(/8953,5069,8613/)
        data_ped(8954,1:3)=(/8954,8061,8856/)
        data_ped(8955,1:3)=(/8955,7134,8848/)
        data_ped(8956,1:3)=(/8956,5818,8732/)
        data_ped(8957,1:3)=(/8957,8719,5453/)
        data_ped(8958,1:3)=(/8958,8773,8844/)
        data_ped(8959,1:3)=(/8959,8637,8755/)
        data_ped(8960,1:3)=(/8960,8650,5971/)
        data_ped(8961,1:3)=(/8961,8650,8888/)
        data_ped(8962,1:3)=(/8962,5747,8731/)
        data_ped(8963,1:3)=(/8963,5883,8771/)
        data_ped(8964,1:3)=(/8964,8741,1338/)
        data_ped(8965,1:3)=(/8965,8748,7354/)
        data_ped(8966,1:3)=(/8966,8802,7398/)
        data_ped(8967,1:3)=(/8967,8803,8849/)
        data_ped(8968,1:3)=(/8968,7057,8612/)
        data_ped(8969,1:3)=(/8969,8719,7396/)
        data_ped(8970,1:3)=(/8970,8773,5218/)
        data_ped(8971,1:3)=(/8971,8793,8834/)
        data_ped(8972,1:3)=(/8972,8061,8776/)
        data_ped(8973,1:3)=(/8973,6161,8870/)
        data_ped(8974,1:3)=(/8974,4893,8698/)
        data_ped(8975,1:3)=(/8975,8655,8714/)
        data_ped(8976,1:3)=(/8976,8868,8858/)
        data_ped(8977,1:3)=(/8977,8868,8851/)
        data_ped(8978,1:3)=(/8978,8637,8630/)
        data_ped(8979,1:3)=(/8979,5498,8602/)
        data_ped(8980,1:3)=(/8980,5560,8782/)
        data_ped(8981,1:3)=(/8981,8713,8686/)
        data_ped(8982,1:3)=(/8982,8761,8838/)
        data_ped(8983,1:3)=(/8983,8762,7399/)
        data_ped(8984,1:3)=(/8984,8796,8769/)
        data_ped(8985,1:3)=(/8985,8802,8798/)
        data_ped(8986,1:3)=(/8986,8205,8901/)
        data_ped(8987,1:3)=(/8987,8727,8809/)
        data_ped(8988,1:3)=(/8988,8793,5705/)
        data_ped(8989,1:3)=(/8989,8793,8687/)
        data_ped(8990,1:3)=(/8990,6161,8870/)
        data_ped(8991,1:3)=(/8991,8664,5820/)
        data_ped(8992,1:3)=(/8992,8868,8851/)
        data_ped(8993,1:3)=(/8993,8868,8908/)
        data_ped(8994,1:3)=(/8994,8868,8858/)
        data_ped(8995,1:3)=(/8995,8874,6172/)
        data_ped(8996,1:3)=(/8996,7488,8725/)
        data_ped(8997,1:3)=(/8997,8637,8765/)
        data_ped(8998,1:3)=(/8998,8637,8896/)
        data_ped(8999,1:3)=(/8999,7299,8873/)
        data_ped(9000,1:3)=(/9000,8802,5727/)
        data_ped(9001,1:3)=(/9001,8802,7398/)
        data_ped(9002,1:3)=(/9002,8878,7429/)
        data_ped(9003,1:3)=(/9003,8641,1307/)
        data_ped(9004,1:3)=(/9004,8655,8853/)
        data_ped(9005,1:3)=(/9005,5554,8775/)
        data_ped(9006,1:3)=(/9006,8737,6199/)
        data_ped(9007,1:3)=(/9007,8800,5821/)
        data_ped(9008,1:3)=(/9008,7356,8867/)
        data_ped(9009,1:3)=(/9009,8173,8753/)
        data_ped(9010,1:3)=(/9010,8868,8851/)
        data_ped(9011,1:3)=(/9011,6298,8889/)
        data_ped(9012,1:3)=(/9012,8094,8836/)
        data_ped(9013,1:3)=(/9013,8061,8776/)
        data_ped(9014,1:3)=(/9014,8892,8750/)
        data_ped(9015,1:3)=(/9015,8637,8659/)
        data_ped(9016,1:3)=(/9016,8637,7494/)
        data_ped(9017,1:3)=(/9017,8650,8943/)
        data_ped(9018,1:3)=(/9018,7447,8938/)
        data_ped(9019,1:3)=(/9019,8710,6073/)
        data_ped(9020,1:3)=(/9020,8710,8916/)
        data_ped(9021,1:3)=(/9021,8723,8619/)
        data_ped(9022,1:3)=(/9022,8723,6057/)
        data_ped(9023,1:3)=(/9023,8729,5730/)
        data_ped(9024,1:3)=(/9024,8729,6137/)
        data_ped(9025,1:3)=(/9025,8748,8857/)
        data_ped(9026,1:3)=(/9026,8803,7137/)
        data_ped(9027,1:3)=(/9027,8803,6248/)
        data_ped(9028,1:3)=(/9028,8803,8145/)
        data_ped(9029,1:3)=(/9029,8825,8927/)
        data_ped(9030,1:3)=(/9030,5502,8881/)
        data_ped(9031,1:3)=(/9031,8710,5981/)
        data_ped(9032,1:3)=(/9032,8723,8944/)
        data_ped(9033,1:3)=(/9033,8762,5758/)
        data_ped(9034,1:3)=(/9034,8664,6325/)
        data_ped(9035,1:3)=(/9035,8664,6220/)
        data_ped(9036,1:3)=(/9036,8773,7277/)
        data_ped(9037,1:3)=(/9037,8811,1327/)
        data_ped(9038,1:3)=(/9038,6181,8898/)
        data_ped(9039,1:3)=(/9039,5891,8865/)
        data_ped(9040,1:3)=(/9040,8173,8753/)
        data_ped(9041,1:3)=(/9041,8173,8920/)
        data_ped(9042,1:3)=(/9042,8061,8749/)
        data_ped(9043,1:3)=(/9043,8820,8917/)
        data_ped(9044,1:3)=(/9044,8814,8974/)
        data_ped(9045,1:3)=(/9045,8911,1366/)
        data_ped(9046,1:3)=(/9046,8768,7201/)
        data_ped(9047,1:3)=(/9047,7248,8934/)
        data_ped(9048,1:3)=(/9048,8723,8861/)
        data_ped(9049,1:3)=(/9049,8802,5959/)
        data_ped(9050,1:3)=(/9050,8802,8788/)
        data_ped(9051,1:3)=(/9051,8879,8693/)
        data_ped(9052,1:3)=(/9052,8803,8624/)
        data_ped(9053,1:3)=(/9053,8709,5799/)
        data_ped(9054,1:3)=(/9054,8650,6248/)
        data_ped(9055,1:3)=(/9055,8734,1329/)
        data_ped(9056,1:3)=(/9056,8734,8808/)
        data_ped(9057,1:3)=(/9057,8710,6330/)
        data_ped(9058,1:3)=(/9058,8710,8790/)
        data_ped(9059,1:3)=(/9059,8723,6157/)
        data_ped(9060,1:3)=(/9060,8803,8897/)
        data_ped(9061,1:3)=(/9061,8710,8751/)
        data_ped(9062,1:3)=(/9062,8811,1340/)
        data_ped(9063,1:3)=(/9063,8664,7423/)
        data_ped(9064,1:3)=(/9064,8922,8908/)
        data_ped(9065,1:3)=(/9065,8977,8926/)
        data_ped(9066,1:3)=(/9066,8800,7395/)
        data_ped(9067,1:3)=(/9067,8800,6225/)
        data_ped(9068,1:3)=(/9068,8773,6191/)
        data_ped(9069,1:3)=(/9069,8879,5758/)
        data_ped(9070,1:3)=(/9070,8800,8840/)
        data_ped(9071,1:3)=(/9071,8929,8090/)
        data_ped(9072,1:3)=(/9072,6286,8955/)
        data_ped(9073,1:3)=(/9073,8061,8957/)
        data_ped(9074,1:3)=(/9074,8061,8776/)
        data_ped(9075,1:3)=(/9075,8860,8996/)
        data_ped(9076,1:3)=(/9076,8829,7351/)
        data_ped(9077,1:3)=(/9077,8683,8828/)
        data_ped(9078,1:3)=(/9078,8699,8900/)
        data_ped(9079,1:3)=(/9079,8803,7553/)
        data_ped(9080,1:3)=(/9080,8803,7360/)
        data_ped(9081,1:3)=(/9081,8911,6061/)
        data_ped(9082,1:3)=(/9082,8710,6030/)
        data_ped(9083,1:3)=(/9083,8710,5913/)
        data_ped(9084,1:3)=(/9084,8814,6282/)
        data_ped(9085,1:3)=(/9085,8814,7398/)
        data_ped(9086,1:3)=(/9086,8710,5981/)
        data_ped(9087,1:3)=(/9087,8793,8774/)
        data_ped(9088,1:3)=(/9088,8903,8753/)
        data_ped(9089,1:3)=(/9089,8886,8217/)
        data_ped(9090,1:3)=(/9090,8977,8926/)
        data_ped(9091,1:3)=(/9091,5948,8904/)
        data_ped(9092,1:3)=(/9092,8875,8855/)
        data_ped(9093,1:3)=(/9093,8061,8749/)
        data_ped(9094,1:3)=(/9094,8061,8725/)
        data_ped(9095,1:3)=(/9095,8719,7327/)
        data_ped(9096,1:3)=(/9096,5994,8890/)
        data_ped(9097,1:3)=(/9097,8923,8218/)
        data_ped(9098,1:3)=(/9098,7418,8937/)
        data_ped(9099,1:3)=(/9099,8801,7487/)
        data_ped(9100,1:3)=(/9100,8756,1385/)
        data_ped(9101,1:3)=(/9101,8823,8884/)
        data_ped(9102,1:3)=(/9102,8930,8963/)
        data_ped(9103,1:3)=(/9103,8675,8979/)
        data_ped(9104,1:3)=(/9104,8883,1344/)
        data_ped(9105,1:3)=(/9105,8883,7479/)
        data_ped(9106,1:3)=(/9106,8802,8145/)
        data_ped(9107,1:3)=(/9107,8878,8845/)
        data_ped(9108,1:3)=(/9108,8814,6362/)
        data_ped(9109,1:3)=(/9109,8814,8259/)
        data_ped(9110,1:3)=(/9110,8683,8834/)
        data_ped(9111,1:3)=(/9111,8803,6304/)
        data_ped(9112,1:3)=(/9112,8710,6128/)
        data_ped(9113,1:3)=(/9113,8710,6090/)
        data_ped(9114,1:3)=(/9114,8710,5802/)
        data_ped(9115,1:3)=(/9115,8805,1343/)
        data_ped(9116,1:3)=(/9116,8862,1345/)
        data_ped(9117,1:3)=(/9117,8803,7360/)
        data_ped(9118,1:3)=(/9118,7146,8757/)
        data_ped(9119,1:3)=(/9119,8719,8876/)
        data_ped(9120,1:3)=(/9120,8922,8994/)
        data_ped(9121,1:3)=(/9121,8977,8993/)
        data_ped(9122,1:3)=(/9122,8874,7536/)
        data_ped(9123,1:3)=(/9123,8929,7523/)
        data_ped(9124,1:3)=(/9124,8864,6072/)
        data_ped(9125,1:3)=(/9125,8231,8931/)
        data_ped(9126,1:3)=(/9126,8903,5774/)
        data_ped(9127,1:3)=(/9127,8800,5896/)
        data_ped(9128,1:3)=(/9128,8617,1351/)
        data_ped(9129,1:3)=(/9129,8915,6169/)
        data_ped(9130,1:3)=(/9130,8885,8866/)
        data_ped(9131,1:3)=(/9131,8949,6249/)
        data_ped(9132,1:3)=(/9132,8921,5800/)
        data_ped(9133,1:3)=(/9133,8061,8780/)
        data_ped(9134,1:3)=(/9134,7893,8819/)
        data_ped(9135,1:3)=(/9135,8791,8964/)
        data_ped(9136,1:3)=(/9136,8928,7147/)
        data_ped(9137,1:3)=(/9137,8804,8778/)
        data_ped(9138,1:3)=(/9138,6060,9018/)
        data_ped(9139,1:3)=(/9139,8734,6248/)
        data_ped(9140,1:3)=(/9140,8710,6032/)
        data_ped(9141,1:3)=(/9141,8805,1381/)
        data_ped(9142,1:3)=(/9142,8883,1380/)
        data_ped(9143,1:3)=(/9143,8883,8258/)
        data_ped(9144,1:3)=(/9144,8814,6272/)
        data_ped(9145,1:3)=(/9145,8862,7364/)
        data_ped(9146,1:3)=(/9146,8850,5764/)
        data_ped(9147,1:3)=(/9147,8950,9026/)
        data_ped(9148,1:3)=(/9148,8950,8962/)
        data_ped(9149,1:3)=(/9149,8734,6356/)
        data_ped(9150,1:3)=(/9150,8802,8902/)
        data_ped(9151,1:3)=(/9151,8802,8810/)
        data_ped(9152,1:3)=(/9152,8803,8966/)
        data_ped(9153,1:3)=(/9153,8650,9000/)
        data_ped(9154,1:3)=(/9154,8929,6175/)
        data_ped(9155,1:3)=(/9155,8929,1357/)
        data_ped(9156,1:3)=(/9156,8875,6070/)
        data_ped(9157,1:3)=(/9157,8818,7283/)
        data_ped(9158,1:3)=(/9158,8818,7335/)
        data_ped(9159,1:3)=(/9159,8703,8195/)
        data_ped(9160,1:3)=(/9160,8664,6278/)
        data_ped(9161,1:3)=(/9161,8664,7309/)
        data_ped(9162,1:3)=(/9162,8664,1354/)
        data_ped(9163,1:3)=(/9163,7382,8991/)
        data_ped(9164,1:3)=(/9164,8972,7210/)
        data_ped(9165,1:3)=(/9165,6230,9012/)
        data_ped(9166,1:3)=(/9166,8903,8201/)
        data_ped(9167,1:3)=(/9167,8800,9003/)
        data_ped(9168,1:3)=(/9168,8719,8262/)
        data_ped(9169,1:3)=(/9169,8719,7559/)
        data_ped(9170,1:3)=(/9170,8719,6164/)
        data_ped(9171,1:3)=(/9171,8949,9001/)
        data_ped(9172,1:3)=(/9172,8922,9010/)
        data_ped(9173,1:3)=(/9173,8617,7913/)
        data_ped(9174,1:3)=(/9174,8992,8926/)
        data_ped(9175,1:3)=(/9175,8989,6229/)
        data_ped(9176,1:3)=(/9176,6136,8906/)
        data_ped(9177,1:3)=(/9177,8804,8869/)
        data_ped(9178,1:3)=(/9178,8709,7212/)
        data_ped(9179,1:3)=(/9179,8729,5730/)
        data_ped(9180,1:3)=(/9180,8877,6099/)
        data_ped(9181,1:3)=(/9181,8650,8914/)
        data_ped(9182,1:3)=(/9182,8710,8859/)
        data_ped(9183,1:3)=(/9183,8710,8906/)
        data_ped(9184,1:3)=(/9184,8710,6282/)
        data_ped(9185,1:3)=(/9185,8710,8781/)
        data_ped(9186,1:3)=(/9186,8802,5962/)
        data_ped(9187,1:3)=(/9187,8802,9057/)
        data_ped(9188,1:3)=(/9188,8802,8816/)
        data_ped(9189,1:3)=(/9189,8802,8895/)
        data_ped(9190,1:3)=(/9190,8961,7587/)
        data_ped(9191,1:3)=(/9191,8814,8259/)
        data_ped(9192,1:3)=(/9192,8950,6309/)
        data_ped(9193,1:3)=(/9193,8950,8988/)
        data_ped(9194,1:3)=(/9194,8950,8751/)
        data_ped(9195,1:3)=(/9195,8945,7326/)
        data_ped(9196,1:3)=(/9196,8929,7523/)
        data_ped(9197,1:3)=(/9197,8952,6276/)
        data_ped(9198,1:3)=(/9198,8184,8970/)
        data_ped(9199,1:3)=(/9199,8664,5834/)
        data_ped(9200,1:3)=(/9200,8664,8204/)
        data_ped(9201,1:3)=(/9201,6121,8899/)
        data_ped(9202,1:3)=(/9202,5993,9063/)
        data_ped(9203,1:3)=(/9203,8997,8726/)
        data_ped(9204,1:3)=(/9204,8779,5692/)
        data_ped(9205,1:3)=(/9205,8989,5935/)
        data_ped(9206,1:3)=(/9206,8832,8939/)
        data_ped(9207,1:3)=(/9207,8832,6102/)
        data_ped(9208,1:3)=(/9208,9021,6129/)
        data_ped(9209,1:3)=(/9209,8802,6224/)
        data_ped(9210,1:3)=(/9210,8650,8956/)
        data_ped(9211,1:3)=(/9211,7545,8786/)
        data_ped(9212,1:3)=(/9212,8949,8958/)
        data_ped(9213,1:3)=(/9213,8949,8894/)
        data_ped(9214,1:3)=(/9214,8802,6257/)
        data_ped(9215,1:3)=(/9215,8950,8512/)
        data_ped(9216,1:3)=(/9216,8950,8978/)
        data_ped(9217,1:3)=(/9217,8862,6138/)
        data_ped(9218,1:3)=(/9218,8802,6263/)
        data_ped(9219,1:3)=(/9219,8802,9002/)
        data_ped(9220,1:3)=(/9220,8802,6094/)
        data_ped(9221,1:3)=(/9221,8734,6355/)
        data_ped(9222,1:3)=(/9222,8683,9033/)
        data_ped(9223,1:3)=(/9223,8710,8975/)
        data_ped(9224,1:3)=(/9224,8650,8953/)
        data_ped(9225,1:3)=(/9225,8650,8916/)
        data_ped(9226,1:3)=(/9226,8650,8910/)
        data_ped(9227,1:3)=(/9227,8650,9022/)
        data_ped(9228,1:3)=(/9228,8205,8843/)
        data_ped(9229,1:3)=(/9229,8883,7905/)
        data_ped(9230,1:3)=(/9230,8883,1362/)
        data_ped(9231,1:3)=(/9231,8791,7367/)
        data_ped(9232,1:3)=(/9232,8852,8924/)
        data_ped(9233,1:3)=(/9233,8222,9038/)
        data_ped(9234,1:3)=(/9234,8247,8901/)
        data_ped(9235,1:3)=(/9235,8922,8976/)
        data_ped(9236,1:3)=(/9236,8922,9065/)
        data_ped(9237,1:3)=(/9237,8959,8242/)
        data_ped(9238,1:3)=(/9238,8929,1356/)
        data_ped(9239,1:3)=(/9239,8664,9018/)
        data_ped(9240,1:3)=(/9240,8200,9034/)
        data_ped(9241,1:3)=(/9241,7382,8991/)
        data_ped(9242,1:3)=(/9242,7384,8909/)
        data_ped(9243,1:3)=(/9243,8800,8817/)
        data_ped(9244,1:3)=(/9244,6262,9096/)
        data_ped(9245,1:3)=(/9245,7408,8985/)
        data_ped(9246,1:3)=(/9246,8874,8987/)
        data_ped(9247,1:3)=(/9247,8874,8830/)
        data_ped(9248,1:3)=(/9248,8874,7506/)
        data_ped(9249,1:3)=(/9249,8832,7495/)
        data_ped(9250,1:3)=(/9250,8995,6308/)
        data_ped(9251,1:3)=(/9251,8995,6321/)
        data_ped(9252,1:3)=(/9252,7421,8870/)
        data_ped(9253,1:3)=(/9253,8841,8715/)
        data_ped(9254,1:3)=(/9254,9072,8217/)
        data_ped(9255,1:3)=(/9255,9072,8871/)
        data_ped(9256,1:3)=(/9256,6001,8984/)
        data_ped(9257,1:3)=(/9257,7473,9037/)
        data_ped(9258,1:3)=(/9258,8875,6551/)
        data_ped(9259,1:3)=(/9259,8729,8951/)
        data_ped(9260,1:3)=(/9260,8945,9051/)
        data_ped(9261,1:3)=(/9261,8734,8831/)
        data_ped(9262,1:3)=(/9262,8734,7327/)
        data_ped(9263,1:3)=(/9263,8734,7456/)
        data_ped(9264,1:3)=(/9264,8710,6330/)
        data_ped(9265,1:3)=(/9265,8862,6226/)
        data_ped(9266,1:3)=(/9266,8907,8982/)
        data_ped(9267,1:3)=(/9267,8950,6222/)
        data_ped(9268,1:3)=(/9268,8950,8751/)
        data_ped(9269,1:3)=(/9269,8802,8971/)
        data_ped(9270,1:3)=(/9270,8802,9032/)
        data_ped(9271,1:3)=(/9271,8802,9113/)
        data_ped(9272,1:3)=(/9272,8802,8858/)
        data_ped(9273,1:3)=(/9273,8935,7566/)
        data_ped(9274,1:3)=(/9274,6372,8764/)
        data_ped(9275,1:3)=(/9275,8205,9089/)
        data_ped(9276,1:3)=(/9276,8205,9076/)
        data_ped(9277,1:3)=(/9277,8829,1363/)
        data_ped(9278,1:3)=(/9278,8244,9008/)
        data_ped(9279,1:3)=(/9279,8709,8880/)
        data_ped(9280,1:3)=(/9280,8650,8284/)
        data_ped(9281,1:3)=(/9281,8650,6289/)
        data_ped(9282,1:3)=(/9282,8197,8999/)
        data_ped(9283,1:3)=(/9283,6001,8893/)
        data_ped(9284,1:3)=(/9284,8952,6276/)
        data_ped(9285,1:3)=(/9285,8959,6052/)
        data_ped(9286,1:3)=(/9286,8959,9048/)
        data_ped(9287,1:3)=(/9287,7586,8976/)
        data_ped(9288,1:3)=(/9288,7586,8926/)
        data_ped(9289,1:3)=(/9289,6151,9036/)
        data_ped(9290,1:3)=(/9290,7286,9097/)
        data_ped(9291,1:3)=(/9291,6396,9125/)
        data_ped(9292,1:3)=(/9292,9109,8299/)
        data_ped(9293,1:3)=(/9293,8972,8846/)
        data_ped(9294,1:3)=(/9294,8941,9069/)
        data_ped(9295,1:3)=(/9295,8997,8968/)
        data_ped(9296,1:3)=(/9296,8800,8789/)
        data_ped(9297,1:3)=(/9297,8929,9030/)
        data_ped(9298,1:3)=(/9298,8986,6134/)
        data_ped(9299,1:3)=(/9299,6049,9054/)
        data_ped(9300,1:3)=(/9300,6049,9027/)
        data_ped(9301,1:3)=(/9301,6049,9047/)
        data_ped(9302,1:3)=(/9302,7059,8871/)
        data_ped(9303,1:3)=(/9303,8923,7276/)
        data_ped(9304,1:3)=(/9304,8923,8175/)
        data_ped(9305,1:3)=(/9305,8940,7493/)
        data_ped(9306,1:3)=(/9306,9042,6402/)
        data_ped(9307,1:3)=(/9307,9015,8910/)
        data_ped(9308,1:3)=(/9308,8200,8882/)
        data_ped(9309,1:3)=(/9309,8200,9034/)
        data_ped(9310,1:3)=(/9310,7286,9126/)
        data_ped(9311,1:3)=(/9311,7520,8932/)
        data_ped(9312,1:3)=(/9312,7322,9036/)
        data_ped(9313,1:3)=(/9313,8820,6134/)
        data_ped(9314,1:3)=(/9314,8820,9068/)
        data_ped(9315,1:3)=(/9315,8967,6309/)
        data_ped(9316,1:3)=(/9316,8200,9160/)
        data_ped(9317,1:3)=(/9317,8913,9077/)
        data_ped(9318,1:3)=(/9318,8959,9048/)
        data_ped(9319,1:3)=(/9319,9070,1364/)
        data_ped(9320,1:3)=(/9320,8923,8813/)
        data_ped(9321,1:3)=(/9321,8304,8839/)
        data_ped(9322,1:3)=(/9322,8832,7659/)
        data_ped(9323,1:3)=(/9323,8829,8319/)
        data_ped(9324,1:3)=(/9324,8244,8855/)
        data_ped(9325,1:3)=(/9325,8805,9116/)
        data_ped(9326,1:3)=(/9326,8950,8270/)
        data_ped(9327,1:3)=(/9327,8650,8925/)
        data_ped(9328,1:3)=(/9328,8802,8887/)
        data_ped(9329,1:3)=(/9329,8802,9059/)
        data_ped(9330,1:3)=(/9330,8802,7630/)
        data_ped(9331,1:3)=(/9331,8710,6312/)
        data_ped(9332,1:3)=(/9332,8710,9049/)
        data_ped(9333,1:3)=(/9333,8710,8975/)
        data_ped(9334,1:3)=(/9334,8907,8982/)
        data_ped(9335,1:3)=(/9335,8960,1404/)
        data_ped(9336,1:3)=(/9336,8935,9050/)
        data_ped(9337,1:3)=(/9337,8935,5828/)
        data_ped(9338,1:3)=(/9338,9016,9064/)
        data_ped(9339,1:3)=(/9339,6049,9139/)
        data_ped(9340,1:3)=(/9340,8297,9005/)
        data_ped(9341,1:3)=(/9341,9166,7610/)
        data_ped(9342,1:3)=(/9342,6253,9044/)
        data_ped(9343,1:3)=(/9343,9031,6269/)
        data_ped(9344,1:3)=(/9344,8912,9170/)
        data_ped(9345,1:3)=(/9345,8912,1349/)
        data_ped(9346,1:3)=(/9346,8922,9121/)
        data_ped(9347,1:3)=(/9347,8874,8210/)
        data_ped(9348,1:3)=(/9348,8860,8220/)
        data_ped(9349,1:3)=(/9349,8929,8891/)
        data_ped(9350,1:3)=(/9350,9017,9014/)
        data_ped(9351,1:3)=(/9351,8852,7660/)
        data_ped(9352,1:3)=(/9352,8211,9024/)
        data_ped(9353,1:3)=(/9353,8950,8270/)
        data_ped(9354,1:3)=(/9354,8950,8998/)
        data_ped(9355,1:3)=(/9355,8160,9146/)
        data_ped(9356,1:3)=(/9356,9042,8983/)
        data_ped(9357,1:3)=(/9357,9042,7598/)
        data_ped(9358,1:3)=(/9358,8961,9061/)
        data_ped(9359,1:3)=(/9359,8197,9053/)
        data_ped(9360,1:3)=(/9360,8243,9046/)
        data_ped(9361,1:3)=(/9361,8804,8332/)
        data_ped(9362,1:3)=(/9362,8923,8219/)
        data_ped(9363,1:3)=(/9363,8860,8847/)
        data_ped(9364,1:3)=(/9364,9091,1391/)
        data_ped(9365,1:3)=(/9365,8297,9212/)
        data_ped(9366,1:3)=(/9366,6253,8973/)
        data_ped(9367,1:3)=(/9367,9011,7552/)
        data_ped(9368,1:3)=(/9368,8200,9006/)
        data_ped(9369,1:3)=(/9369,8200,9200/)
        data_ped(9370,1:3)=(/9370,8954,1367/)
        data_ped(9371,1:3)=(/9371,8929,8854/)
        data_ped(9372,1:3)=(/9372,8929,6268/)
        data_ped(9373,1:3)=(/9373,8929,6320/)
        data_ped(9374,1:3)=(/9374,9133,9155/)
        data_ped(9375,1:3)=(/9375,9133,9154/)
        data_ped(9376,1:3)=(/9376,9171,7400/)
        data_ped(9377,1:3)=(/9377,8260,8831/)
        data_ped(9378,1:3)=(/9378,9021,8948/)
        data_ped(9379,1:3)=(/9379,8967,6250/)
        data_ped(9380,1:3)=(/9380,8772,8821/)
        data_ped(9381,1:3)=(/9381,9138,9073/)
        data_ped(9382,1:3)=(/9382,9016,9121/)
        data_ped(9383,1:3)=(/9383,8905,6176/)
        data_ped(9384,1:3)=(/9384,8997,8288/)
        data_ped(9385,1:3)=(/9385,8912,7554/)
        data_ped(9386,1:3)=(/9386,8329,8742/)
        data_ped(9387,1:3)=(/9387,8918,8865/)
        data_ped(9388,1:3)=(/9388,9109,7556/)
        data_ped(9389,1:3)=(/9389,8947,7456/)
        data_ped(9390,1:3)=(/9390,8874,8760/)
        data_ped(9391,1:3)=(/9391,8211,8833/)
        data_ped(9392,1:3)=(/9392,7581,9144/)
        data_ped(9393,1:3)=(/9393,8921,9019/)
        data_ped(9394,1:3)=(/9394,8921,145/)
        data_ped(9395,1:3)=(/9395,6350,9004/)
        data_ped(9396,1:3)=(/9396,8960,9106/)
        data_ped(9397,1:3)=(/9397,9042,6380/)
        data_ped(9398,1:3)=(/9398,6372,9135/)
        data_ped(9399,1:3)=(/9399,8244,9007/)
        data_ped(9400,1:3)=(/9400,8160,8893/)
        data_ped(9401,1:3)=(/9401,8205,9104/)
        data_ped(9402,1:3)=(/9402,9020,9194/)
        data_ped(9403,1:3)=(/9403,8946,9137/)
        data_ped(9404,1:3)=(/9404,8946,7481/)
        data_ped(9405,1:3)=(/9405,8946,8212/)
        data_ped(9406,1:3)=(/9406,8946,8203/)
        data_ped(9407,1:3)=(/9407,9016,9174/)
        data_ped(9408,1:3)=(/9408,8860,8357/)
        data_ped(9409,1:3)=(/9409,9138,9073/)
        data_ped(9410,1:3)=(/9410,9138,8305/)
        data_ped(9411,1:3)=(/9411,8200,9035/)
        data_ped(9412,1:3)=(/9412,8200,8882/)
        data_ped(9413,1:3)=(/9413,9031,7426/)
        data_ped(9414,1:3)=(/9414,8928,7509/)
        data_ped(9415,1:3)=(/9415,8173,9181/)
        data_ped(9416,1:3)=(/9416,9092,7367/)
        data_ped(9417,1:3)=(/9417,8832,8344/)
        data_ped(9418,1:3)=(/9418,9052,8267/)
        data_ped(9419,1:3)=(/9419,8297,9005/)
        data_ped(9420,1:3)=(/9420,8748,1400/)
        data_ped(9421,1:3)=(/9421,6388,9029/)
        data_ped(9422,1:3)=(/9422,9109,6349/)
        data_ped(9423,1:3)=(/9423,6346,8902/)
        data_ped(9424,1:3)=(/9424,8136,9129/)
        data_ped(9425,1:3)=(/9425,8136,8981/)
        data_ped(9426,1:3)=(/9426,8912,9170/)
        data_ped(9427,1:3)=(/9427,9168,1403/)
        data_ped(9428,1:3)=(/9428,9168,8270/)
        data_ped(9429,1:3)=(/9429,9168,7559/)
        data_ped(9430,1:3)=(/9430,6297,8927/)
        data_ped(9431,1:3)=(/9431,6297,9256/)
        data_ped(9432,1:3)=(/9432,8980,9231/)
        data_ped(9433,1:3)=(/9433,8772,8837/)
        data_ped(9434,1:3)=(/9434,8954,6336/)
        data_ped(9435,1:3)=(/9435,8188,8933/)
        data_ped(9436,1:3)=(/9436,6237,8936/)
        data_ped(9437,1:3)=(/9437,8947,8290/)
        data_ped(9438,1:3)=(/9438,8907,7458/)
        data_ped(9439,1:3)=(/9439,8946,6059/)
        data_ped(9440,1:3)=(/9440,8946,9043/)
        data_ped(9441,1:3)=(/9441,8911,6379/)
        data_ped(9442,1:3)=(/9442,8802,9058/)
        data_ped(9443,1:3)=(/9443,6258,9110/)
        data_ped(9444,1:3)=(/9444,8160,8942/)
        data_ped(9445,1:3)=(/9445,8160,9136/)
        data_ped(9446,1:3)=(/9446,8929,9004/)
        data_ped(9447,1:3)=(/9447,7620,8872/)
        data_ped(9448,1:3)=(/9448,7680,9295/)
        data_ped(9449,1:3)=(/9449,9042,9286/)
        data_ped(9450,1:3)=(/9450,8211,9099/)
        data_ped(9451,1:3)=(/9451,6436,8965/)
        data_ped(9452,1:3)=(/9452,8950,8185/)
        data_ped(9453,1:3)=(/9453,7615,8990/)
        data_ped(9454,1:3)=(/9454,8997,9087/)
        data_ped(9455,1:3)=(/9455,8200,8882/)
        data_ped(9456,1:3)=(/9456,8820,9039/)
        data_ped(9457,1:3)=(/9457,9168,8168/)
        data_ped(9458,1:3)=(/9458,9168,9170/)
        data_ped(9459,1:3)=(/9459,9168,6331/)
        data_ped(9460,1:3)=(/9460,8136,9118/)
        data_ped(9461,1:3)=(/9461,6194,9102/)
        data_ped(9462,1:3)=(/9462,8919,7592/)
        data_ped(9463,1:3)=(/9463,8862,6411/)
        data_ped(9464,1:3)=(/9464,8929,6153/)
        data_ped(9465,1:3)=(/9465,9132,7556/)
        data_ped(9466,1:3)=(/9466,8329,8787/)
        data_ped(9467,1:3)=(/9467,9023,6408/)
        data_ped(9468,1:3)=(/9468,8954,6110/)
        data_ped(9469,1:3)=(/9469,9078,9106/)
        data_ped(9470,1:3)=(/9470,8820,9039/)
        data_ped(9471,1:3)=(/9471,9092,7367/)
        data_ped(9472,1:3)=(/9472,8928,7509/)
        data_ped(9473,1:3)=(/9473,7937,9204/)
        data_ped(9474,1:3)=(/9474,9282,9296/)
        data_ped(9475,1:3)=(/9475,0,0/)
        data_ped(9476,1:3)=(/9476,0,0/)
        data_ped(9477,1:3)=(/9477,0,0/)
        data_ped(9478,1:3)=(/9478,0,0/)
        data_ped(9479,1:3)=(/9479,0,0/)
        data_ped(9480,1:3)=(/9480,0,0/)
        data_ped(9481,1:3)=(/9481,0,0/)


    end subroutine set_data

    subroutine set_solution

        nlevtotref=15
        nhaploref=1
        naniref=347
        testref=3.1363378
        pvalueref=7.74428397E-02
        solmuref=0
        stdmuref=0
        solref(1)=1.0169750
        stdref(1)=0.57424718

        Aref(1,:)=(/43.394993,20.353405,4.3683014,18.673288,6.5284505,&
            3.2162218,5.8564043,2.7361889,5.4243741,0.91206294,&
            5.3763709,1.1040761,4.5123115,0.91206294,6.8164701/)

        Bref(1,1) = 0.32975984

        Cref(1,:)=(/ 0.0000000,-0.0000000,1.88885592E-02,4.44671558E-03,-0.83706260,&
            -0.89402080,-0.81141675,-0.84151697,-0.86200690,-0.87918329,-0.83641231,&
            -0.91889989,-0.85015082,-0.88332295,-0.86247855/)

        Daniref(:)=(/1397,1401,1403,1405,1406,1407,1409,1410,6351,6378,6384,6386,6406,6407,&
            6408,6409,6410,6412,6413,6414,6415,6419,6421,6423,6426,6429,6430,6431,6434,6435,6437,6439,&
            6440,6441,6442,6443,6445,6446,6447,6448,6449,6450,6452,7559,7596,7617,7628,7635,7637,7640,&
            7655,7656,7657,7658,7662,7663,7665,7666,7668,7672,7675,7677,7679,7684,7687,7688,7690,7691,&
            7693,7698,7702,7704,7707,7711,7714,7716,7717,7719,7721,7724,7725,7726,7727,7728,7730,7732,&
            7734,7735,7736,7737,7930,7938,7939,7943,7947,7948,7949,7950,7991,7992,8200,8205,8211,8244,&
            8275,8286,8295,8307,8311,8312,8313,8314,8324,8328,8330,8331,8334,8338,8342,8343,8347,8351,&
            8355,8360,8361,8362,8364,8365,8366,8367,8368,8369,8374,8375,8376,8377,8379,8380,8383,8384,&
            8385,8386,8387,8388,8389,8392,8393,8394,8395,8396,8398,8399,8403,8404,8405,8406,8407,8408,&
            8409,8413,8415,8416,8417,8418,8419,8421,8422,8423,8424,8426,8427,8428,8429,8430,8431,8434,&
            8435,8515,8516,8946,9025,9045,9056,9075,9080,9081,9086,9090,9101,9103,9105,9115,9119,9120,&
            9122,9123,9128,9130,9131,9134,9141,9143,9145,9151,9152,9153,9155,9157,9158,9159,9160,9161,&
            9162,9163,9164,9165,9167,9172,9177,9178,9183,9189,9195,9196,9197,9201,9202,9206,9213,9217,&
            9218,9225,9228,9229,9230,9234,9235,9236,9237,9240,9244,9245,9246,9249,9250,9252,9253,9254,&
            9255,9256,9258,9259,9261,9264,9265,9266,9267,9269,9270,9272,9274,9276,9277,9282,9283,9284,&
            9287,9288,9290,9291,9292,9294,9297,9300,9301,9302,9303,9304,9305,9306,9307,9309,9311,9314,&
            9315,9316,9319,9320,9321,9322,9323,9324,9326,9327,9328,9329,9330,9333,9334,9335,9338,9339,&
            9340,9341,9342,9343,9344,9345,9346,9347,9348,9350,9351,9352,9353,9354,9356,9357,9358,9359,&
            9360,9361,9362,9364,9368,9369,9370,9371,9372,9374,9376,9377,9378,9381,9382,9383,9384,9385,&
            9386,9387,9388,9390,9392,9393,9395,9475,9480/)

        Dref(1,:)=(/6.56908900E-02,6.35447577E-02,-4.07355987E-02,7.49202669E-02,6.40456080E-02,-5.65608591E-02,&
            -5.31228967E-02,-4.85616177E-02,-3.61309834E-02,-3.88808474E-02,-4.53371927E-02,-6.27859905E-02,-5.97652607E-02,&
            -4.98821102E-02,8.05565864E-02,8.67901295E-02,-1.54887876E-02,-7.22115934E-02,-2.08238717E-02,7.80307502E-02,&
            -4.69163321E-02,-3.24934162E-02,8.34544450E-02,-7.68644139E-02,-7.31711760E-02,-4.69501019E-02,-4.00474928E-02,&
            -1.80141311E-02,0.11850835,-3.66700329E-02,-6.70048967E-02,-6.20004833E-02,9.63365287E-02,-6.61190078E-02,&
            0.17198642,6.64786473E-02,-6.01829998E-02,7.28709400E-02,7.23136216E-02,0.19642791,5.88510595E-02,-3.83410528E-02,&
            -5.64776137E-02,6.20517433E-02,7.88844600E-02,4.52171862E-02,-8.27148408E-02,-5.61172478E-02,-4.72673327E-02,&
            -5.41785620E-02,9.39353108E-02,6.86917976E-02,0.19151963,7.63911232E-02,4.10304107E-02,0.20489283,-6.45048842E-02,&
            7.92446285E-02,-1.68051589E-02,-5.88224940E-02,-5.03945313E-02,7.17981309E-02,8.54093283E-02,-6.59677833E-02,&
            -6.17754087E-02,7.64625892E-02,-4.04262990E-02,-7.37133548E-02,7.50947967E-02,-5.16560562E-02,-8.14900994E-02,&
            -4.99940701E-02,-4.59909588E-02,0.18276423,-3.01259942E-02,-7.76449591E-02,5.03317565E-02,-7.86104426E-02,&
            -7.20259845E-02,-4.67119403E-02,-4.85884771E-02,-4.01078761E-02,-7.83411488E-02,7.05823675E-02,-6.73083663E-02,&
            9.14994851E-02,5.69372810E-02,-6.51051924E-02,6.79373071E-02,-3.80446352E-02,8.98458883E-02,-3.49117368E-02,&
            -6.61722124E-02,-4.93702628E-02,7.43450671E-02,9.15167108E-02,-7.42000639E-02,8.98320824E-02,6.39843643E-02,&
            7.09712729E-02,-7.75745958E-02,-7.89530128E-02,-6.31527230E-02,-0.11363846,3.48658971E-02,-9.01499689E-02,&
            0.20810644,5.93932047E-02,0.10015085,-3.08669098E-02,-1.49429459E-02,-1.82671584E-02,-4.66800481E-02,-3.84164900E-02,&
            9.37058181E-02,7.05748126E-02,-7.27221668E-02,0.20139915,-3.73607390E-02,6.43368512E-02,-7.38366470E-02,0.10694800,&
            -5.81091754E-02,-6.83395714E-02,-6.43001497E-02,4.02212590E-02,-7.77109489E-02,-2.19403263E-02,-0.10905322,&
            6.60915077E-02,7.77544305E-02,-9.13098231E-02,-6.37816563E-02,5.45601547E-02,-5.46283834E-02,-5.64331114E-02,&
            2.04801671E-02,&
            -7.80362412E-02,8.88958722E-02,8.57332200E-02,-8.08570608E-02,6.84257373E-02,-5.73606417E-02,6.35932684E-02,&
            -2.70070191E-02,6.01415075E-02,-5.63942492E-02,-3.66525166E-02,-4.72838394E-02,1.41493101E-02,-0.10169994,&
            -8.46255273E-02,4.82743643E-02,-7.52661526E-02,6.60761148E-02,-7.33691975E-02,-7.14252964E-02,-8.93226415E-02,&
            -2.86245495E-02,-2.84410194E-02,2.48992182E-02,-6.66612610E-02,5.27014919E-02,-9.06658769E-02,-9.25673246E-02,&
            -8.85014534E-02,-4.98206951E-02,3.48541252E-02,6.47397041E-02,2.74215117E-02,4.36341986E-02,-5.93026169E-02,&
            -4.35178988E-02,-7.32052699E-02,-9.54541191E-02,-6.08883053E-02,5.61123118E-02,-9.05165672E-02,-6.21881150E-02,&
            0.17003782,-6.17244616E-02,-7.27358386E-02,8.99087563E-02,-3.93738635E-02,-9.14451405E-02,-6.35482818E-02,&
            4.20221128E-02,&
            -5.31339496E-02,5.84222451E-02,-8.10220242E-02,-6.72706589E-02,-7.79562071E-02,-6.79641217E-02,3.90786417E-02,&
            5.46050705E-02,-6.00680970E-02,6.01581782E-02,-8.38623196E-02,-8.18051398E-02,-5.55592515E-02,-7.01843873E-02,&
            -6.79621100E-02,-4.48933505E-02,-4.96553630E-02,-7.34519586E-02,-7.53914043E-02,-6.17901646E-02,-6.75383285E-02,&
            -5.10065332E-02,-9.52039659E-02,-5.63194826E-02,5.32267615E-02,-4.71231304E-02,-4.47143391E-02,-8.83694291E-02,&
            6.17422312E-02,-7.78070614E-02,-4.92786355E-02,-7.75005445E-02,7.86104500E-02,-5.89572750E-02,4.79543731E-02,&
            7.14402422E-02,-7.29824603E-02,0.10670533,-3.56310941E-02,-1.23858703E-02,-6.52170256E-02,-9.37346816E-02,0.16728604,&
            0.17701925,-9.30134133E-02,-7.89277926E-02,3.44902426E-02,3.46127413E-02,6.32096678E-02,-4.95486334E-02,&
            -4.72677015E-02,&
            -6.51124269E-02,-9.17380527E-02,-5.43749221E-02,-5.11201806E-02,-2.76503321E-02,3.03113218E-02,7.50659257E-02,&
            -5.51347211E-02,-8.40347558E-02,-7.92118534E-02,-6.98831230E-02,-5.84580787E-02,7.73197934E-02,6.75886422E-02,&
            7.26965517E-02,4.56163473E-02,-4.99894060E-02,-8.02619681E-02,-5.21121509E-02,-3.80428061E-02,-5.19797951E-02,&
            -4.31218818E-02,7.69329146E-02,-8.72950107E-02,3.04528996E-02,-2.02725902E-02,-6.75219297E-02,0.18848504,&
            -6.84959367E-02,&
            -6.57649636E-02,8.87271836E-02,7.28277937E-02,-7.17043281E-02,-5.35983592E-02,-8.21628347E-02,-7.59699047E-02,&
            -7.63154104E-02,2.94271223E-02,-5.64136691E-02,4.95815165E-02,-5.92096858E-02,-6.52191266E-02,-6.07770830E-02,&
            -0.10184709,7.00257421E-02,-8.18268210E-02,4.92037572E-02,-7.33736083E-02,-6.58877045E-02,2.72713471E-02,-0.10942860,&
            -6.65087253E-02,-8.48259628E-02,-7.97905549E-02,-5.06411195E-02,-9.05567780E-02,-3.65123376E-02,4.32831384E-02,&
            -2.66364086E-02,4.07284647E-02,-8.57761502E-02,-8.93729627E-02,6.16331361E-02,2.72264145E-02,-7.36248940E-02,&
            -7.59818852E-02,-6.51214272E-02,-6.05536290E-02,-4.95377369E-02,4.61579971E-02,5.22228591E-02,5.75874932E-02,&
            6.23211153E-02,-9.73965675E-02,5.48889562E-02,2.73899604E-02,-5.06411195E-02,4.18086387E-02,-7.79087692E-02,&
            2.06688233E-02,-8.44797269E-02,8.51689056E-02,-6.54129833E-02,3.20045538E-02,-2.65659802E-02,-5.60312420E-02,&
            1.49065498E-02,&
            -6.01791553E-02,-7.12926239E-02,-8.75041857E-02,-6.38593361E-02,-9.63964239E-02,-6.25250041E-02,-6.93021715E-02,&
            -8.18693265E-02,-6.17308281E-02,6.51991740E-02,6.42785653E-02,4.19470146E-02,-4.70415540E-02,-7.86530077E-02,&
            -6.31558001E-02,-7.25537911E-02,6.53914660E-02,3.94378193E-02,-7.36589208E-02,-5.97289689E-02,-4.70908619E-02,&
            6.40456080E-02/)

    end subroutine set_solution

end program PROG_TESTMODBLUPF90
